(self["webpackChunkActiveHTMLWidget"] = self["webpackChunkActiveHTMLWidget"] || []).push([["css_index_css"],{

/***/ "./node_modules/css-loader/dist/cjs.js!./css/bootstrap.css":
/*!*****************************************************************!*\
  !*** ./node_modules/css-loader/dist/cjs.js!./css/bootstrap.css ***!
  \*****************************************************************/
/***/ ((module, exports, __webpack_require__) => {

// Imports
var ___CSS_LOADER_API_IMPORT___ = __webpack_require__(/*! ../node_modules/css-loader/dist/runtime/api.js */ "./node_modules/css-loader/dist/runtime/api.js");
exports = ___CSS_LOADER_API_IMPORT___(false);
// Module
exports.push([module.id, "@charset \"UTF-8\";\n:root {\n  --bs-blue: #0d6efd;\n  --bs-indigo: #6610f2;\n  --bs-purple: #6f42c1;\n  --bs-pink: #d63384;\n  --bs-red: #dc3545;\n  --bs-orange: #fd7e14;\n  --bs-yellow: #ffc107;\n  --bs-green: #198754;\n  --bs-teal: #20c997;\n  --bs-cyan: #0dcaf0;\n  --bs-white: #fff;\n  --bs-gray: #6c757d;\n  --bs-gray-dark: #343a40;\n  --bs-gray-100: #f8f9fa;\n  --bs-gray-200: #e9ecef;\n  --bs-gray-300: #dee2e6;\n  --bs-gray-400: #ced4da;\n  --bs-gray-500: #adb5bd;\n  --bs-gray-600: #6c757d;\n  --bs-gray-700: #495057;\n  --bs-gray-800: #343a40;\n  --bs-gray-900: #212529;\n  --bs-primary: #0d6efd;\n  --bs-secondary: #6c757d;\n  --bs-success: #198754;\n  --bs-info: #0dcaf0;\n  --bs-warning: #ffc107;\n  --bs-danger: #dc3545;\n  --bs-light: #f8f9fa;\n  --bs-dark: #212529;\n  --bs-primary-rgb: 13, 110, 253;\n  --bs-secondary-rgb: 108, 117, 125;\n  --bs-success-rgb: 25, 135, 84;\n  --bs-info-rgb: 13, 202, 240;\n  --bs-warning-rgb: 255, 193, 7;\n  --bs-danger-rgb: 220, 53, 69;\n  --bs-light-rgb: 248, 249, 250;\n  --bs-dark-rgb: 33, 37, 41;\n  --bs-white-rgb: 255, 255, 255;\n  --bs-black-rgb: 0, 0, 0;\n  --bs-body-color-rgb: 33, 37, 41;\n  --bs-body-bg-rgb: 255, 255, 255;\n  --bs-font-sans-serif: system-ui, -apple-system, \"Segoe UI\", Roboto, \"Helvetica Neue\", Arial, \"Noto Sans\", \"Liberation Sans\", sans-serif, \"Apple Color Emoji\", \"Segoe UI Emoji\", \"Segoe UI Symbol\", \"Noto Color Emoji\";\n  --bs-font-monospace: SFMono-Regular, Menlo, Monaco, Consolas, \"Liberation Mono\", \"Courier New\", monospace;\n  --bs-gradient: linear-gradient(180deg, rgba(255, 255, 255, 0.15), rgba(255, 255, 255, 0));\n  --bs-body-font-family: var(--bs-font-sans-serif);\n  --bs-body-font-size: 1rem;\n  --bs-body-font-weight: 400;\n  --bs-body-line-height: 1.5;\n  --bs-body-color: #212529;\n  --bs-body-bg: #fff;\n}\n\n.jhtml {\n  /*!\n   * Bootstrap v5.1.3 (https://getbootstrap.com/)\n   * Copyright 2011-2021 The Bootstrap Authors\n   * Copyright 2011-2021 Twitter, Inc.\n   * Licensed under MIT (https://github.com/twbs/bootstrap/blob/main/LICENSE)\n   */\n  /* rtl:raw:\n  [type=\"tel\"],\n  [type=\"url\"],\n  [type=\"email\"],\n  [type=\"number\"] {\n    direction: ltr;\n  }\n  */\n  /* rtl:begin:ignore */\n  /* rtl:end:ignore */\n  /* rtl:options: {\n    \"autoRename\": true,\n    \"stringMap\":[ {\n      \"name\"    : \"prev-next\",\n      \"search\"  : \"prev\",\n      \"replace\" : \"next\"\n    } ]\n  } */\n  /* rtl:begin:remove */\n  /* rtl:end:remove */\n}\n.jhtml :root {\n  --bs-blue: #0d6efd;\n  --bs-indigo: #6610f2;\n  --bs-purple: #6f42c1;\n  --bs-pink: #d63384;\n  --bs-red: #dc3545;\n  --bs-orange: #fd7e14;\n  --bs-yellow: #ffc107;\n  --bs-green: #198754;\n  --bs-teal: #20c997;\n  --bs-cyan: #0dcaf0;\n  --bs-white: #fff;\n  --bs-gray: #6c757d;\n  --bs-gray-dark: #343a40;\n  --bs-gray-100: #f8f9fa;\n  --bs-gray-200: #e9ecef;\n  --bs-gray-300: #dee2e6;\n  --bs-gray-400: #ced4da;\n  --bs-gray-500: #adb5bd;\n  --bs-gray-600: #6c757d;\n  --bs-gray-700: #495057;\n  --bs-gray-800: #343a40;\n  --bs-gray-900: #212529;\n  --bs-primary: #0d6efd;\n  --bs-secondary: #6c757d;\n  --bs-success: #198754;\n  --bs-info: #0dcaf0;\n  --bs-warning: #ffc107;\n  --bs-danger: #dc3545;\n  --bs-light: #f8f9fa;\n  --bs-dark: #212529;\n  --bs-primary-rgb: 13, 110, 253;\n  --bs-secondary-rgb: 108, 117, 125;\n  --bs-success-rgb: 25, 135, 84;\n  --bs-info-rgb: 13, 202, 240;\n  --bs-warning-rgb: 255, 193, 7;\n  --bs-danger-rgb: 220, 53, 69;\n  --bs-light-rgb: 248, 249, 250;\n  --bs-dark-rgb: 33, 37, 41;\n  --bs-white-rgb: 255, 255, 255;\n  --bs-black-rgb: 0, 0, 0;\n  --bs-body-color-rgb: 33, 37, 41;\n  --bs-body-bg-rgb: 255, 255, 255;\n  --bs-font-sans-serif: system-ui, -apple-system, \"Segoe UI\", Roboto, \"Helvetica Neue\", Arial, \"Noto Sans\", \"Liberation Sans\", sans-serif, \"Apple Color Emoji\", \"Segoe UI Emoji\", \"Segoe UI Symbol\", \"Noto Color Emoji\";\n  --bs-font-monospace: SFMono-Regular, Menlo, Monaco, Consolas, \"Liberation Mono\", \"Courier New\", monospace;\n  --bs-gradient: linear-gradient(180deg, rgba(255, 255, 255, 0.15), rgba(255, 255, 255, 0));\n  --bs-body-font-family: var(--bs-font-sans-serif);\n  --bs-body-font-size: 1rem;\n  --bs-body-font-weight: 400;\n  --bs-body-line-height: 1.5;\n  --bs-body-color: #212529;\n  --bs-body-bg: #fff;\n}\n.jhtml *,\n.jhtml *::before,\n.jhtml *::after {\n  box-sizing: border-box;\n}\n@media (prefers-reduced-motion: no-preference) {\n  .jhtml :root {\n    scroll-behavior: smooth;\n  }\n}\n.jhtml body , .jhtml {\n  margin: 0;\n  font-family: var(--bs-body-font-family);\n  font-size: var(--bs-body-font-size);\n  font-weight: var(--bs-body-font-weight);\n  line-height: var(--bs-body-line-height);\n  color: var(--bs-body-color);\n  text-align: var(--bs-body-text-align);\n  background-color: var(--bs-body-bg);\n  -webkit-text-size-adjust: 100%;\n  -webkit-tap-highlight-color: rgba(0, 0, 0, 0);\n}\n.jhtml hr {\n  margin: 1rem 0;\n  color: inherit;\n  background-color: currentColor;\n  border: 0;\n  opacity: 0.25;\n}\n.jhtml hr:not([size]) {\n  height: 1px;\n}\n.jhtml h6,  .jhtml .h6,  .jhtml h5,  .jhtml .h5,  .jhtml h4,  .jhtml .h4,  .jhtml h3,  .jhtml .h3,  .jhtml h2,  .jhtml .h2,  .jhtml h1,  .jhtml .h1 ,  .jhtml.h6,  .jhtml.h5,  .jhtml.h4,  .jhtml.h3,  .jhtml.h2,  .jhtml.h1 {\n  margin-top: 0;\n  margin-bottom: 0.5rem;\n  font-weight: 500;\n  line-height: 1.2;\n}\n.jhtml h1,  .jhtml .h1 ,  .jhtml.h1 {\n  font-size: calc(1.375rem + 1.5vw);\n}\n@media (min-width: 1200px) {\n  .jhtml h1,  .jhtml .h1 ,  .jhtml.h1 {\n    font-size: 2.5rem;\n  }\n}\n.jhtml h2,  .jhtml .h2 ,  .jhtml.h2 {\n  font-size: calc(1.325rem + 0.9vw);\n}\n@media (min-width: 1200px) {\n  .jhtml h2,  .jhtml .h2 ,  .jhtml.h2 {\n    font-size: 2rem;\n  }\n}\n.jhtml h3,  .jhtml .h3 ,  .jhtml.h3 {\n  font-size: calc(1.3rem + 0.6vw);\n}\n@media (min-width: 1200px) {\n  .jhtml h3,  .jhtml .h3 ,  .jhtml.h3 {\n    font-size: 1.75rem;\n  }\n}\n.jhtml h4,  .jhtml .h4 ,  .jhtml.h4 {\n  font-size: calc(1.275rem + 0.3vw);\n}\n@media (min-width: 1200px) {\n  .jhtml h4,  .jhtml .h4 ,  .jhtml.h4 {\n    font-size: 1.5rem;\n  }\n}\n.jhtml h5,  .jhtml .h5 ,  .jhtml.h5 {\n  font-size: 1.25rem;\n}\n.jhtml h6,  .jhtml .h6 ,  .jhtml.h6 {\n  font-size: 1rem;\n}\n.jhtml p {\n  margin-top: 0;\n  margin-bottom: 1rem;\n}\n.jhtml abbr[title],\n.jhtml abbr[data-bs-original-title] {\n  text-decoration: underline dotted;\n  cursor: help;\n  text-decoration-skip-ink: none;\n}\n.jhtml address {\n  margin-bottom: 1rem;\n  font-style: normal;\n  line-height: inherit;\n}\n.jhtml ol,\n.jhtml ul {\n  padding-left: 2rem;\n}\n.jhtml ol,\n.jhtml ul,\n.jhtml dl {\n  margin-top: 0;\n  margin-bottom: 1rem;\n}\n.jhtml ol ol,\n.jhtml ul ul,\n.jhtml ol ul,\n.jhtml ul ol {\n  margin-bottom: 0;\n}\n.jhtml dt {\n  font-weight: 700;\n}\n.jhtml dd {\n  margin-bottom: 0.5rem;\n  margin-left: 0;\n}\n.jhtml blockquote {\n  margin: 0 0 1rem;\n}\n.jhtml b,\n.jhtml strong {\n  font-weight: bolder;\n}\n.jhtml small,  .jhtml .small ,  .jhtml.small {\n  font-size: 0.875em;\n}\n.jhtml mark,  .jhtml .mark ,  .jhtml.mark {\n  padding: 0.2em;\n  background-color: #fcf8e3;\n}\n.jhtml sub,\n.jhtml sup {\n  position: relative;\n  font-size: 0.75em;\n  line-height: 0;\n  vertical-align: baseline;\n}\n.jhtml sub {\n  bottom: -0.25em;\n}\n.jhtml sup {\n  top: -0.5em;\n}\n.jhtml a {\n  color: #0d6efd;\n  text-decoration: underline;\n}\n.jhtml a:hover {\n  color: #0a58ca;\n}\n.jhtml a:not([href]):not([class]), .jhtml a:not([href]):not([class]):hover {\n  color: inherit;\n  text-decoration: none;\n}\n.jhtml pre,\n.jhtml code,\n.jhtml kbd,\n.jhtml samp {\n  font-family: var(--bs-font-monospace);\n  font-size: 1em;\n  direction: ltr /* rtl:ignore */;\n  unicode-bidi: bidi-override;\n}\n.jhtml pre {\n  display: block;\n  margin-top: 0;\n  margin-bottom: 1rem;\n  overflow: auto;\n  font-size: 0.875em;\n}\n.jhtml pre code {\n  font-size: inherit;\n  color: inherit;\n  word-break: normal;\n}\n.jhtml code {\n  font-size: 0.875em;\n  color: #d63384;\n  word-wrap: break-word;\n}\n.jhtml a > code {\n  color: inherit;\n}\n.jhtml kbd {\n  padding: 0.2rem 0.4rem;\n  font-size: 0.875em;\n  color: #fff;\n  background-color: #212529;\n  border-radius: 0.2rem;\n}\n.jhtml kbd kbd {\n  padding: 0;\n  font-size: 1em;\n  font-weight: 700;\n}\n.jhtml figure {\n  margin: 0 0 1rem;\n}\n.jhtml img,\n.jhtml svg {\n  vertical-align: middle;\n}\n.jhtml table {\n  caption-side: bottom;\n  border-collapse: collapse;\n}\n.jhtml caption {\n  padding-top: 0.5rem;\n  padding-bottom: 0.5rem;\n  color: #6c757d;\n  text-align: left;\n}\n.jhtml th {\n  text-align: inherit;\n  text-align: -webkit-match-parent;\n}\n.jhtml thead,\n.jhtml tbody,\n.jhtml tfoot,\n.jhtml tr,\n.jhtml td,\n.jhtml th {\n  border-color: inherit;\n  border-style: solid;\n  border-width: 0;\n}\n.jhtml label {\n  display: inline-block;\n}\n.jhtml button {\n  border-radius: 0;\n}\n.jhtml button:focus:not(:focus-visible) {\n  outline: 0;\n}\n.jhtml input,\n.jhtml button,\n.jhtml select,\n.jhtml optgroup,\n.jhtml textarea {\n  margin: 0;\n  font-family: inherit;\n  font-size: inherit;\n  line-height: inherit;\n}\n.jhtml button,\n.jhtml select {\n  text-transform: none;\n}\n.jhtml [role=button] {\n  cursor: pointer;\n}\n.jhtml select {\n  word-wrap: normal;\n}\n.jhtml select:disabled {\n  opacity: 1;\n}\n.jhtml [list]::-webkit-calendar-picker-indicator {\n  display: none;\n}\n.jhtml button,\n.jhtml [type=button],\n.jhtml [type=reset],\n.jhtml [type=submit] {\n  -webkit-appearance: button;\n}\n.jhtml button:not(:disabled),\n.jhtml [type=button]:not(:disabled),\n.jhtml [type=reset]:not(:disabled),\n.jhtml [type=submit]:not(:disabled) {\n  cursor: pointer;\n}\n.jhtml ::-moz-focus-inner {\n  padding: 0;\n  border-style: none;\n}\n.jhtml textarea {\n  resize: vertical;\n}\n.jhtml fieldset {\n  min-width: 0;\n  padding: 0;\n  margin: 0;\n  border: 0;\n}\n.jhtml legend {\n  float: left;\n  width: 100%;\n  padding: 0;\n  margin-bottom: 0.5rem;\n  font-size: calc(1.275rem + 0.3vw);\n  line-height: inherit;\n}\n@media (min-width: 1200px) {\n  .jhtml legend {\n    font-size: 1.5rem;\n  }\n}\n.jhtml legend + * {\n  clear: left;\n}\n.jhtml ::-webkit-datetime-edit-fields-wrapper,\n.jhtml ::-webkit-datetime-edit-text,\n.jhtml ::-webkit-datetime-edit-minute,\n.jhtml ::-webkit-datetime-edit-hour-field,\n.jhtml ::-webkit-datetime-edit-day-field,\n.jhtml ::-webkit-datetime-edit-month-field,\n.jhtml ::-webkit-datetime-edit-year-field {\n  padding: 0;\n}\n.jhtml ::-webkit-inner-spin-button {\n  height: auto;\n}\n.jhtml [type=search] {\n  outline-offset: -2px;\n  -webkit-appearance: textfield;\n}\n.jhtml ::-webkit-search-decoration {\n  -webkit-appearance: none;\n}\n.jhtml ::-webkit-color-swatch-wrapper {\n  padding: 0;\n}\n.jhtml ::file-selector-button {\n  font: inherit;\n}\n.jhtml ::-webkit-file-upload-button {\n  font: inherit;\n  -webkit-appearance: button;\n}\n.jhtml output {\n  display: inline-block;\n}\n.jhtml iframe {\n  border: 0;\n}\n.jhtml summary {\n  display: list-item;\n  cursor: pointer;\n}\n.jhtml progress {\n  vertical-align: baseline;\n}\n.jhtml [hidden] {\n  display: none !important;\n}\n.jhtml .lead , .jhtml.lead {\n  font-size: 1.25rem;\n  font-weight: 300;\n}\n.jhtml .display-1 , .jhtml.display-1 {\n  font-size: calc(1.625rem + 4.5vw);\n  font-weight: 300;\n  line-height: 1.2;\n}\n@media (min-width: 1200px) {\n  .jhtml .display-1 , .jhtml.display-1 {\n    font-size: 5rem;\n  }\n}\n.jhtml .display-2 , .jhtml.display-2 {\n  font-size: calc(1.575rem + 3.9vw);\n  font-weight: 300;\n  line-height: 1.2;\n}\n@media (min-width: 1200px) {\n  .jhtml .display-2 , .jhtml.display-2 {\n    font-size: 4.5rem;\n  }\n}\n.jhtml .display-3 , .jhtml.display-3 {\n  font-size: calc(1.525rem + 3.3vw);\n  font-weight: 300;\n  line-height: 1.2;\n}\n@media (min-width: 1200px) {\n  .jhtml .display-3 , .jhtml.display-3 {\n    font-size: 4rem;\n  }\n}\n.jhtml .display-4 , .jhtml.display-4 {\n  font-size: calc(1.475rem + 2.7vw);\n  font-weight: 300;\n  line-height: 1.2;\n}\n@media (min-width: 1200px) {\n  .jhtml .display-4 , .jhtml.display-4 {\n    font-size: 3.5rem;\n  }\n}\n.jhtml .display-5 , .jhtml.display-5 {\n  font-size: calc(1.425rem + 2.1vw);\n  font-weight: 300;\n  line-height: 1.2;\n}\n@media (min-width: 1200px) {\n  .jhtml .display-5 , .jhtml.display-5 {\n    font-size: 3rem;\n  }\n}\n.jhtml .display-6 , .jhtml.display-6 {\n  font-size: calc(1.375rem + 1.5vw);\n  font-weight: 300;\n  line-height: 1.2;\n}\n@media (min-width: 1200px) {\n  .jhtml .display-6 , .jhtml.display-6 {\n    font-size: 2.5rem;\n  }\n}\n.jhtml .list-unstyled , .jhtml.list-unstyled {\n  padding-left: 0;\n  list-style: none;\n}\n.jhtml .list-inline , .jhtml.list-inline {\n  padding-left: 0;\n  list-style: none;\n}\n.jhtml .list-inline-item , .jhtml.list-inline-item {\n  display: inline-block;\n}\n.jhtml .list-inline-item:not(:last-child) , .jhtml.list-inline-item:not(:last-child) {\n  margin-right: 0.5rem;\n}\n.jhtml .initialism , .jhtml.initialism {\n  font-size: 0.875em;\n  text-transform: uppercase;\n}\n.jhtml .blockquote , .jhtml.blockquote {\n  margin-bottom: 1rem;\n  font-size: 1.25rem;\n}\n.jhtml .blockquote > :last-child , .jhtml.blockquote > :last-child {\n  margin-bottom: 0;\n}\n.jhtml .blockquote-footer , .jhtml.blockquote-footer {\n  margin-top: -1rem;\n  margin-bottom: 1rem;\n  font-size: 0.875em;\n  color: #6c757d;\n}\n.jhtml .blockquote-footer::before , .jhtml.blockquote-footer::before {\n  content: \"— \";\n}\n.jhtml .img-fluid , .jhtml.img-fluid {\n  max-width: 100%;\n  height: auto;\n}\n.jhtml .img-thumbnail , .jhtml.img-thumbnail {\n  padding: 0.25rem;\n  background-color: #fff;\n  border: 1px solid #dee2e6;\n  border-radius: 0.25rem;\n  max-width: 100%;\n  height: auto;\n}\n.jhtml .figure , .jhtml.figure {\n  display: inline-block;\n}\n.jhtml .figure-img , .jhtml.figure-img {\n  margin-bottom: 0.5rem;\n  line-height: 1;\n}\n.jhtml .figure-caption , .jhtml.figure-caption {\n  font-size: 0.875em;\n  color: #6c757d;\n}\n.jhtml .container, \n.jhtml .container-fluid, \n.jhtml .container-xxl, \n.jhtml .container-xl, \n.jhtml .container-lg, \n.jhtml .container-md, \n.jhtml .container-sm , .jhtml.container, \n.jhtml.container-fluid, \n.jhtml.container-xxl, \n.jhtml.container-xl, \n.jhtml.container-lg, \n.jhtml.container-md, \n.jhtml.container-sm {\n  width: 100%;\n  padding-right: var(--bs-gutter-x, 0.75rem);\n  padding-left: var(--bs-gutter-x, 0.75rem);\n  margin-right: auto;\n  margin-left: auto;\n}\n@media (min-width: 576px) {\n  .jhtml .container-sm,  .jhtml .container , .jhtml.container-sm,  .jhtml.container {\n    max-width: 540px;\n  }\n}\n@media (min-width: 768px) {\n  .jhtml .container-md,  .jhtml .container-sm,  .jhtml .container , .jhtml.container-md,  .jhtml.container-sm,  .jhtml.container {\n    max-width: 720px;\n  }\n}\n@media (min-width: 992px) {\n  .jhtml .container-lg,  .jhtml .container-md,  .jhtml .container-sm,  .jhtml .container , .jhtml.container-lg,  .jhtml.container-md,  .jhtml.container-sm,  .jhtml.container {\n    max-width: 960px;\n  }\n}\n@media (min-width: 1200px) {\n  .jhtml .container-xl,  .jhtml .container-lg,  .jhtml .container-md,  .jhtml .container-sm,  .jhtml .container , .jhtml.container-xl,  .jhtml.container-lg,  .jhtml.container-md,  .jhtml.container-sm,  .jhtml.container {\n    max-width: 1140px;\n  }\n}\n@media (min-width: 1400px) {\n  .jhtml .container-xxl,  .jhtml .container-xl,  .jhtml .container-lg,  .jhtml .container-md,  .jhtml .container-sm,  .jhtml .container , .jhtml.container-xxl,  .jhtml.container-xl,  .jhtml.container-lg,  .jhtml.container-md,  .jhtml.container-sm,  .jhtml.container {\n    max-width: 1320px;\n  }\n}\n.jhtml .row , .jhtml.row {\n  --bs-gutter-x: 1.5rem;\n  --bs-gutter-y: 0;\n  display: flex;\n  flex-wrap: wrap;\n  margin-top: calc(-1 * var(--bs-gutter-y));\n  margin-right: calc(-0.5 * var(--bs-gutter-x));\n  margin-left: calc(-0.5 * var(--bs-gutter-x));\n}\n.jhtml .row > * , .jhtml.row > * {\n  flex-shrink: 0;\n  width: 100%;\n  max-width: 100%;\n  padding-right: calc(var(--bs-gutter-x) * 0.5);\n  padding-left: calc(var(--bs-gutter-x) * 0.5);\n  margin-top: var(--bs-gutter-y);\n}\n.jhtml .col , .jhtml.col {\n  flex: 1 0 0%;\n}\n.jhtml .row-cols-auto > * , .jhtml.row-cols-auto > * {\n  flex: 0 0 auto;\n  width: auto;\n}\n.jhtml .row-cols-1 > * , .jhtml.row-cols-1 > * {\n  flex: 0 0 auto;\n  width: 100%;\n}\n.jhtml .row-cols-2 > * , .jhtml.row-cols-2 > * {\n  flex: 0 0 auto;\n  width: 50%;\n}\n.jhtml .row-cols-3 > * , .jhtml.row-cols-3 > * {\n  flex: 0 0 auto;\n  width: 33.3333333333%;\n}\n.jhtml .row-cols-4 > * , .jhtml.row-cols-4 > * {\n  flex: 0 0 auto;\n  width: 25%;\n}\n.jhtml .row-cols-5 > * , .jhtml.row-cols-5 > * {\n  flex: 0 0 auto;\n  width: 20%;\n}\n.jhtml .row-cols-6 > * , .jhtml.row-cols-6 > * {\n  flex: 0 0 auto;\n  width: 16.6666666667%;\n}\n.jhtml .col-auto , .jhtml.col-auto {\n  flex: 0 0 auto;\n  width: auto;\n}\n.jhtml .col-1 , .jhtml.col-1 {\n  flex: 0 0 auto;\n  width: 8.33333333%;\n}\n.jhtml .col-2 , .jhtml.col-2 {\n  flex: 0 0 auto;\n  width: 16.66666667%;\n}\n.jhtml .col-3 , .jhtml.col-3 {\n  flex: 0 0 auto;\n  width: 25%;\n}\n.jhtml .col-4 , .jhtml.col-4 {\n  flex: 0 0 auto;\n  width: 33.33333333%;\n}\n.jhtml .col-5 , .jhtml.col-5 {\n  flex: 0 0 auto;\n  width: 41.66666667%;\n}\n.jhtml .col-6 , .jhtml.col-6 {\n  flex: 0 0 auto;\n  width: 50%;\n}\n.jhtml .col-7 , .jhtml.col-7 {\n  flex: 0 0 auto;\n  width: 58.33333333%;\n}\n.jhtml .col-8 , .jhtml.col-8 {\n  flex: 0 0 auto;\n  width: 66.66666667%;\n}\n.jhtml .col-9 , .jhtml.col-9 {\n  flex: 0 0 auto;\n  width: 75%;\n}\n.jhtml .col-10 , .jhtml.col-10 {\n  flex: 0 0 auto;\n  width: 83.33333333%;\n}\n.jhtml .col-11 , .jhtml.col-11 {\n  flex: 0 0 auto;\n  width: 91.66666667%;\n}\n.jhtml .col-12 , .jhtml.col-12 {\n  flex: 0 0 auto;\n  width: 100%;\n}\n.jhtml .offset-1 , .jhtml.offset-1 {\n  margin-left: 8.33333333%;\n}\n.jhtml .offset-2 , .jhtml.offset-2 {\n  margin-left: 16.66666667%;\n}\n.jhtml .offset-3 , .jhtml.offset-3 {\n  margin-left: 25%;\n}\n.jhtml .offset-4 , .jhtml.offset-4 {\n  margin-left: 33.33333333%;\n}\n.jhtml .offset-5 , .jhtml.offset-5 {\n  margin-left: 41.66666667%;\n}\n.jhtml .offset-6 , .jhtml.offset-6 {\n  margin-left: 50%;\n}\n.jhtml .offset-7 , .jhtml.offset-7 {\n  margin-left: 58.33333333%;\n}\n.jhtml .offset-8 , .jhtml.offset-8 {\n  margin-left: 66.66666667%;\n}\n.jhtml .offset-9 , .jhtml.offset-9 {\n  margin-left: 75%;\n}\n.jhtml .offset-10 , .jhtml.offset-10 {\n  margin-left: 83.33333333%;\n}\n.jhtml .offset-11 , .jhtml.offset-11 {\n  margin-left: 91.66666667%;\n}\n.jhtml .g-0, \n.jhtml .gx-0 , .jhtml.g-0, \n.jhtml.gx-0 {\n  --bs-gutter-x: 0;\n}\n.jhtml .g-0, \n.jhtml .gy-0 , .jhtml.g-0, \n.jhtml.gy-0 {\n  --bs-gutter-y: 0;\n}\n.jhtml .g-1, \n.jhtml .gx-1 , .jhtml.g-1, \n.jhtml.gx-1 {\n  --bs-gutter-x: 0.25rem;\n}\n.jhtml .g-1, \n.jhtml .gy-1 , .jhtml.g-1, \n.jhtml.gy-1 {\n  --bs-gutter-y: 0.25rem;\n}\n.jhtml .g-2, \n.jhtml .gx-2 , .jhtml.g-2, \n.jhtml.gx-2 {\n  --bs-gutter-x: 0.5rem;\n}\n.jhtml .g-2, \n.jhtml .gy-2 , .jhtml.g-2, \n.jhtml.gy-2 {\n  --bs-gutter-y: 0.5rem;\n}\n.jhtml .g-3, \n.jhtml .gx-3 , .jhtml.g-3, \n.jhtml.gx-3 {\n  --bs-gutter-x: 1rem;\n}\n.jhtml .g-3, \n.jhtml .gy-3 , .jhtml.g-3, \n.jhtml.gy-3 {\n  --bs-gutter-y: 1rem;\n}\n.jhtml .g-4, \n.jhtml .gx-4 , .jhtml.g-4, \n.jhtml.gx-4 {\n  --bs-gutter-x: 1.5rem;\n}\n.jhtml .g-4, \n.jhtml .gy-4 , .jhtml.g-4, \n.jhtml.gy-4 {\n  --bs-gutter-y: 1.5rem;\n}\n.jhtml .g-5, \n.jhtml .gx-5 , .jhtml.g-5, \n.jhtml.gx-5 {\n  --bs-gutter-x: 3rem;\n}\n.jhtml .g-5, \n.jhtml .gy-5 , .jhtml.g-5, \n.jhtml.gy-5 {\n  --bs-gutter-y: 3rem;\n}\n@media (min-width: 576px) {\n  .jhtml .col-sm , .jhtml.col-sm {\n    flex: 1 0 0%;\n  }\n  .jhtml .row-cols-sm-auto > * , .jhtml.row-cols-sm-auto > * {\n    flex: 0 0 auto;\n    width: auto;\n  }\n  .jhtml .row-cols-sm-1 > * , .jhtml.row-cols-sm-1 > * {\n    flex: 0 0 auto;\n    width: 100%;\n  }\n  .jhtml .row-cols-sm-2 > * , .jhtml.row-cols-sm-2 > * {\n    flex: 0 0 auto;\n    width: 50%;\n  }\n  .jhtml .row-cols-sm-3 > * , .jhtml.row-cols-sm-3 > * {\n    flex: 0 0 auto;\n    width: 33.3333333333%;\n  }\n  .jhtml .row-cols-sm-4 > * , .jhtml.row-cols-sm-4 > * {\n    flex: 0 0 auto;\n    width: 25%;\n  }\n  .jhtml .row-cols-sm-5 > * , .jhtml.row-cols-sm-5 > * {\n    flex: 0 0 auto;\n    width: 20%;\n  }\n  .jhtml .row-cols-sm-6 > * , .jhtml.row-cols-sm-6 > * {\n    flex: 0 0 auto;\n    width: 16.6666666667%;\n  }\n  .jhtml .col-sm-auto , .jhtml.col-sm-auto {\n    flex: 0 0 auto;\n    width: auto;\n  }\n  .jhtml .col-sm-1 , .jhtml.col-sm-1 {\n    flex: 0 0 auto;\n    width: 8.33333333%;\n  }\n  .jhtml .col-sm-2 , .jhtml.col-sm-2 {\n    flex: 0 0 auto;\n    width: 16.66666667%;\n  }\n  .jhtml .col-sm-3 , .jhtml.col-sm-3 {\n    flex: 0 0 auto;\n    width: 25%;\n  }\n  .jhtml .col-sm-4 , .jhtml.col-sm-4 {\n    flex: 0 0 auto;\n    width: 33.33333333%;\n  }\n  .jhtml .col-sm-5 , .jhtml.col-sm-5 {\n    flex: 0 0 auto;\n    width: 41.66666667%;\n  }\n  .jhtml .col-sm-6 , .jhtml.col-sm-6 {\n    flex: 0 0 auto;\n    width: 50%;\n  }\n  .jhtml .col-sm-7 , .jhtml.col-sm-7 {\n    flex: 0 0 auto;\n    width: 58.33333333%;\n  }\n  .jhtml .col-sm-8 , .jhtml.col-sm-8 {\n    flex: 0 0 auto;\n    width: 66.66666667%;\n  }\n  .jhtml .col-sm-9 , .jhtml.col-sm-9 {\n    flex: 0 0 auto;\n    width: 75%;\n  }\n  .jhtml .col-sm-10 , .jhtml.col-sm-10 {\n    flex: 0 0 auto;\n    width: 83.33333333%;\n  }\n  .jhtml .col-sm-11 , .jhtml.col-sm-11 {\n    flex: 0 0 auto;\n    width: 91.66666667%;\n  }\n  .jhtml .col-sm-12 , .jhtml.col-sm-12 {\n    flex: 0 0 auto;\n    width: 100%;\n  }\n  .jhtml .offset-sm-0 , .jhtml.offset-sm-0 {\n    margin-left: 0;\n  }\n  .jhtml .offset-sm-1 , .jhtml.offset-sm-1 {\n    margin-left: 8.33333333%;\n  }\n  .jhtml .offset-sm-2 , .jhtml.offset-sm-2 {\n    margin-left: 16.66666667%;\n  }\n  .jhtml .offset-sm-3 , .jhtml.offset-sm-3 {\n    margin-left: 25%;\n  }\n  .jhtml .offset-sm-4 , .jhtml.offset-sm-4 {\n    margin-left: 33.33333333%;\n  }\n  .jhtml .offset-sm-5 , .jhtml.offset-sm-5 {\n    margin-left: 41.66666667%;\n  }\n  .jhtml .offset-sm-6 , .jhtml.offset-sm-6 {\n    margin-left: 50%;\n  }\n  .jhtml .offset-sm-7 , .jhtml.offset-sm-7 {\n    margin-left: 58.33333333%;\n  }\n  .jhtml .offset-sm-8 , .jhtml.offset-sm-8 {\n    margin-left: 66.66666667%;\n  }\n  .jhtml .offset-sm-9 , .jhtml.offset-sm-9 {\n    margin-left: 75%;\n  }\n  .jhtml .offset-sm-10 , .jhtml.offset-sm-10 {\n    margin-left: 83.33333333%;\n  }\n  .jhtml .offset-sm-11 , .jhtml.offset-sm-11 {\n    margin-left: 91.66666667%;\n  }\n  .jhtml .g-sm-0, \n.jhtml .gx-sm-0 , .jhtml.g-sm-0, \n.jhtml.gx-sm-0 {\n    --bs-gutter-x: 0;\n  }\n  .jhtml .g-sm-0, \n.jhtml .gy-sm-0 , .jhtml.g-sm-0, \n.jhtml.gy-sm-0 {\n    --bs-gutter-y: 0;\n  }\n  .jhtml .g-sm-1, \n.jhtml .gx-sm-1 , .jhtml.g-sm-1, \n.jhtml.gx-sm-1 {\n    --bs-gutter-x: 0.25rem;\n  }\n  .jhtml .g-sm-1, \n.jhtml .gy-sm-1 , .jhtml.g-sm-1, \n.jhtml.gy-sm-1 {\n    --bs-gutter-y: 0.25rem;\n  }\n  .jhtml .g-sm-2, \n.jhtml .gx-sm-2 , .jhtml.g-sm-2, \n.jhtml.gx-sm-2 {\n    --bs-gutter-x: 0.5rem;\n  }\n  .jhtml .g-sm-2, \n.jhtml .gy-sm-2 , .jhtml.g-sm-2, \n.jhtml.gy-sm-2 {\n    --bs-gutter-y: 0.5rem;\n  }\n  .jhtml .g-sm-3, \n.jhtml .gx-sm-3 , .jhtml.g-sm-3, \n.jhtml.gx-sm-3 {\n    --bs-gutter-x: 1rem;\n  }\n  .jhtml .g-sm-3, \n.jhtml .gy-sm-3 , .jhtml.g-sm-3, \n.jhtml.gy-sm-3 {\n    --bs-gutter-y: 1rem;\n  }\n  .jhtml .g-sm-4, \n.jhtml .gx-sm-4 , .jhtml.g-sm-4, \n.jhtml.gx-sm-4 {\n    --bs-gutter-x: 1.5rem;\n  }\n  .jhtml .g-sm-4, \n.jhtml .gy-sm-4 , .jhtml.g-sm-4, \n.jhtml.gy-sm-4 {\n    --bs-gutter-y: 1.5rem;\n  }\n  .jhtml .g-sm-5, \n.jhtml .gx-sm-5 , .jhtml.g-sm-5, \n.jhtml.gx-sm-5 {\n    --bs-gutter-x: 3rem;\n  }\n  .jhtml .g-sm-5, \n.jhtml .gy-sm-5 , .jhtml.g-sm-5, \n.jhtml.gy-sm-5 {\n    --bs-gutter-y: 3rem;\n  }\n}\n@media (min-width: 768px) {\n  .jhtml .col-md , .jhtml.col-md {\n    flex: 1 0 0%;\n  }\n  .jhtml .row-cols-md-auto > * , .jhtml.row-cols-md-auto > * {\n    flex: 0 0 auto;\n    width: auto;\n  }\n  .jhtml .row-cols-md-1 > * , .jhtml.row-cols-md-1 > * {\n    flex: 0 0 auto;\n    width: 100%;\n  }\n  .jhtml .row-cols-md-2 > * , .jhtml.row-cols-md-2 > * {\n    flex: 0 0 auto;\n    width: 50%;\n  }\n  .jhtml .row-cols-md-3 > * , .jhtml.row-cols-md-3 > * {\n    flex: 0 0 auto;\n    width: 33.3333333333%;\n  }\n  .jhtml .row-cols-md-4 > * , .jhtml.row-cols-md-4 > * {\n    flex: 0 0 auto;\n    width: 25%;\n  }\n  .jhtml .row-cols-md-5 > * , .jhtml.row-cols-md-5 > * {\n    flex: 0 0 auto;\n    width: 20%;\n  }\n  .jhtml .row-cols-md-6 > * , .jhtml.row-cols-md-6 > * {\n    flex: 0 0 auto;\n    width: 16.6666666667%;\n  }\n  .jhtml .col-md-auto , .jhtml.col-md-auto {\n    flex: 0 0 auto;\n    width: auto;\n  }\n  .jhtml .col-md-1 , .jhtml.col-md-1 {\n    flex: 0 0 auto;\n    width: 8.33333333%;\n  }\n  .jhtml .col-md-2 , .jhtml.col-md-2 {\n    flex: 0 0 auto;\n    width: 16.66666667%;\n  }\n  .jhtml .col-md-3 , .jhtml.col-md-3 {\n    flex: 0 0 auto;\n    width: 25%;\n  }\n  .jhtml .col-md-4 , .jhtml.col-md-4 {\n    flex: 0 0 auto;\n    width: 33.33333333%;\n  }\n  .jhtml .col-md-5 , .jhtml.col-md-5 {\n    flex: 0 0 auto;\n    width: 41.66666667%;\n  }\n  .jhtml .col-md-6 , .jhtml.col-md-6 {\n    flex: 0 0 auto;\n    width: 50%;\n  }\n  .jhtml .col-md-7 , .jhtml.col-md-7 {\n    flex: 0 0 auto;\n    width: 58.33333333%;\n  }\n  .jhtml .col-md-8 , .jhtml.col-md-8 {\n    flex: 0 0 auto;\n    width: 66.66666667%;\n  }\n  .jhtml .col-md-9 , .jhtml.col-md-9 {\n    flex: 0 0 auto;\n    width: 75%;\n  }\n  .jhtml .col-md-10 , .jhtml.col-md-10 {\n    flex: 0 0 auto;\n    width: 83.33333333%;\n  }\n  .jhtml .col-md-11 , .jhtml.col-md-11 {\n    flex: 0 0 auto;\n    width: 91.66666667%;\n  }\n  .jhtml .col-md-12 , .jhtml.col-md-12 {\n    flex: 0 0 auto;\n    width: 100%;\n  }\n  .jhtml .offset-md-0 , .jhtml.offset-md-0 {\n    margin-left: 0;\n  }\n  .jhtml .offset-md-1 , .jhtml.offset-md-1 {\n    margin-left: 8.33333333%;\n  }\n  .jhtml .offset-md-2 , .jhtml.offset-md-2 {\n    margin-left: 16.66666667%;\n  }\n  .jhtml .offset-md-3 , .jhtml.offset-md-3 {\n    margin-left: 25%;\n  }\n  .jhtml .offset-md-4 , .jhtml.offset-md-4 {\n    margin-left: 33.33333333%;\n  }\n  .jhtml .offset-md-5 , .jhtml.offset-md-5 {\n    margin-left: 41.66666667%;\n  }\n  .jhtml .offset-md-6 , .jhtml.offset-md-6 {\n    margin-left: 50%;\n  }\n  .jhtml .offset-md-7 , .jhtml.offset-md-7 {\n    margin-left: 58.33333333%;\n  }\n  .jhtml .offset-md-8 , .jhtml.offset-md-8 {\n    margin-left: 66.66666667%;\n  }\n  .jhtml .offset-md-9 , .jhtml.offset-md-9 {\n    margin-left: 75%;\n  }\n  .jhtml .offset-md-10 , .jhtml.offset-md-10 {\n    margin-left: 83.33333333%;\n  }\n  .jhtml .offset-md-11 , .jhtml.offset-md-11 {\n    margin-left: 91.66666667%;\n  }\n  .jhtml .g-md-0, \n.jhtml .gx-md-0 , .jhtml.g-md-0, \n.jhtml.gx-md-0 {\n    --bs-gutter-x: 0;\n  }\n  .jhtml .g-md-0, \n.jhtml .gy-md-0 , .jhtml.g-md-0, \n.jhtml.gy-md-0 {\n    --bs-gutter-y: 0;\n  }\n  .jhtml .g-md-1, \n.jhtml .gx-md-1 , .jhtml.g-md-1, \n.jhtml.gx-md-1 {\n    --bs-gutter-x: 0.25rem;\n  }\n  .jhtml .g-md-1, \n.jhtml .gy-md-1 , .jhtml.g-md-1, \n.jhtml.gy-md-1 {\n    --bs-gutter-y: 0.25rem;\n  }\n  .jhtml .g-md-2, \n.jhtml .gx-md-2 , .jhtml.g-md-2, \n.jhtml.gx-md-2 {\n    --bs-gutter-x: 0.5rem;\n  }\n  .jhtml .g-md-2, \n.jhtml .gy-md-2 , .jhtml.g-md-2, \n.jhtml.gy-md-2 {\n    --bs-gutter-y: 0.5rem;\n  }\n  .jhtml .g-md-3, \n.jhtml .gx-md-3 , .jhtml.g-md-3, \n.jhtml.gx-md-3 {\n    --bs-gutter-x: 1rem;\n  }\n  .jhtml .g-md-3, \n.jhtml .gy-md-3 , .jhtml.g-md-3, \n.jhtml.gy-md-3 {\n    --bs-gutter-y: 1rem;\n  }\n  .jhtml .g-md-4, \n.jhtml .gx-md-4 , .jhtml.g-md-4, \n.jhtml.gx-md-4 {\n    --bs-gutter-x: 1.5rem;\n  }\n  .jhtml .g-md-4, \n.jhtml .gy-md-4 , .jhtml.g-md-4, \n.jhtml.gy-md-4 {\n    --bs-gutter-y: 1.5rem;\n  }\n  .jhtml .g-md-5, \n.jhtml .gx-md-5 , .jhtml.g-md-5, \n.jhtml.gx-md-5 {\n    --bs-gutter-x: 3rem;\n  }\n  .jhtml .g-md-5, \n.jhtml .gy-md-5 , .jhtml.g-md-5, \n.jhtml.gy-md-5 {\n    --bs-gutter-y: 3rem;\n  }\n}\n@media (min-width: 992px) {\n  .jhtml .col-lg , .jhtml.col-lg {\n    flex: 1 0 0%;\n  }\n  .jhtml .row-cols-lg-auto > * , .jhtml.row-cols-lg-auto > * {\n    flex: 0 0 auto;\n    width: auto;\n  }\n  .jhtml .row-cols-lg-1 > * , .jhtml.row-cols-lg-1 > * {\n    flex: 0 0 auto;\n    width: 100%;\n  }\n  .jhtml .row-cols-lg-2 > * , .jhtml.row-cols-lg-2 > * {\n    flex: 0 0 auto;\n    width: 50%;\n  }\n  .jhtml .row-cols-lg-3 > * , .jhtml.row-cols-lg-3 > * {\n    flex: 0 0 auto;\n    width: 33.3333333333%;\n  }\n  .jhtml .row-cols-lg-4 > * , .jhtml.row-cols-lg-4 > * {\n    flex: 0 0 auto;\n    width: 25%;\n  }\n  .jhtml .row-cols-lg-5 > * , .jhtml.row-cols-lg-5 > * {\n    flex: 0 0 auto;\n    width: 20%;\n  }\n  .jhtml .row-cols-lg-6 > * , .jhtml.row-cols-lg-6 > * {\n    flex: 0 0 auto;\n    width: 16.6666666667%;\n  }\n  .jhtml .col-lg-auto , .jhtml.col-lg-auto {\n    flex: 0 0 auto;\n    width: auto;\n  }\n  .jhtml .col-lg-1 , .jhtml.col-lg-1 {\n    flex: 0 0 auto;\n    width: 8.33333333%;\n  }\n  .jhtml .col-lg-2 , .jhtml.col-lg-2 {\n    flex: 0 0 auto;\n    width: 16.66666667%;\n  }\n  .jhtml .col-lg-3 , .jhtml.col-lg-3 {\n    flex: 0 0 auto;\n    width: 25%;\n  }\n  .jhtml .col-lg-4 , .jhtml.col-lg-4 {\n    flex: 0 0 auto;\n    width: 33.33333333%;\n  }\n  .jhtml .col-lg-5 , .jhtml.col-lg-5 {\n    flex: 0 0 auto;\n    width: 41.66666667%;\n  }\n  .jhtml .col-lg-6 , .jhtml.col-lg-6 {\n    flex: 0 0 auto;\n    width: 50%;\n  }\n  .jhtml .col-lg-7 , .jhtml.col-lg-7 {\n    flex: 0 0 auto;\n    width: 58.33333333%;\n  }\n  .jhtml .col-lg-8 , .jhtml.col-lg-8 {\n    flex: 0 0 auto;\n    width: 66.66666667%;\n  }\n  .jhtml .col-lg-9 , .jhtml.col-lg-9 {\n    flex: 0 0 auto;\n    width: 75%;\n  }\n  .jhtml .col-lg-10 , .jhtml.col-lg-10 {\n    flex: 0 0 auto;\n    width: 83.33333333%;\n  }\n  .jhtml .col-lg-11 , .jhtml.col-lg-11 {\n    flex: 0 0 auto;\n    width: 91.66666667%;\n  }\n  .jhtml .col-lg-12 , .jhtml.col-lg-12 {\n    flex: 0 0 auto;\n    width: 100%;\n  }\n  .jhtml .offset-lg-0 , .jhtml.offset-lg-0 {\n    margin-left: 0;\n  }\n  .jhtml .offset-lg-1 , .jhtml.offset-lg-1 {\n    margin-left: 8.33333333%;\n  }\n  .jhtml .offset-lg-2 , .jhtml.offset-lg-2 {\n    margin-left: 16.66666667%;\n  }\n  .jhtml .offset-lg-3 , .jhtml.offset-lg-3 {\n    margin-left: 25%;\n  }\n  .jhtml .offset-lg-4 , .jhtml.offset-lg-4 {\n    margin-left: 33.33333333%;\n  }\n  .jhtml .offset-lg-5 , .jhtml.offset-lg-5 {\n    margin-left: 41.66666667%;\n  }\n  .jhtml .offset-lg-6 , .jhtml.offset-lg-6 {\n    margin-left: 50%;\n  }\n  .jhtml .offset-lg-7 , .jhtml.offset-lg-7 {\n    margin-left: 58.33333333%;\n  }\n  .jhtml .offset-lg-8 , .jhtml.offset-lg-8 {\n    margin-left: 66.66666667%;\n  }\n  .jhtml .offset-lg-9 , .jhtml.offset-lg-9 {\n    margin-left: 75%;\n  }\n  .jhtml .offset-lg-10 , .jhtml.offset-lg-10 {\n    margin-left: 83.33333333%;\n  }\n  .jhtml .offset-lg-11 , .jhtml.offset-lg-11 {\n    margin-left: 91.66666667%;\n  }\n  .jhtml .g-lg-0, \n.jhtml .gx-lg-0 , .jhtml.g-lg-0, \n.jhtml.gx-lg-0 {\n    --bs-gutter-x: 0;\n  }\n  .jhtml .g-lg-0, \n.jhtml .gy-lg-0 , .jhtml.g-lg-0, \n.jhtml.gy-lg-0 {\n    --bs-gutter-y: 0;\n  }\n  .jhtml .g-lg-1, \n.jhtml .gx-lg-1 , .jhtml.g-lg-1, \n.jhtml.gx-lg-1 {\n    --bs-gutter-x: 0.25rem;\n  }\n  .jhtml .g-lg-1, \n.jhtml .gy-lg-1 , .jhtml.g-lg-1, \n.jhtml.gy-lg-1 {\n    --bs-gutter-y: 0.25rem;\n  }\n  .jhtml .g-lg-2, \n.jhtml .gx-lg-2 , .jhtml.g-lg-2, \n.jhtml.gx-lg-2 {\n    --bs-gutter-x: 0.5rem;\n  }\n  .jhtml .g-lg-2, \n.jhtml .gy-lg-2 , .jhtml.g-lg-2, \n.jhtml.gy-lg-2 {\n    --bs-gutter-y: 0.5rem;\n  }\n  .jhtml .g-lg-3, \n.jhtml .gx-lg-3 , .jhtml.g-lg-3, \n.jhtml.gx-lg-3 {\n    --bs-gutter-x: 1rem;\n  }\n  .jhtml .g-lg-3, \n.jhtml .gy-lg-3 , .jhtml.g-lg-3, \n.jhtml.gy-lg-3 {\n    --bs-gutter-y: 1rem;\n  }\n  .jhtml .g-lg-4, \n.jhtml .gx-lg-4 , .jhtml.g-lg-4, \n.jhtml.gx-lg-4 {\n    --bs-gutter-x: 1.5rem;\n  }\n  .jhtml .g-lg-4, \n.jhtml .gy-lg-4 , .jhtml.g-lg-4, \n.jhtml.gy-lg-4 {\n    --bs-gutter-y: 1.5rem;\n  }\n  .jhtml .g-lg-5, \n.jhtml .gx-lg-5 , .jhtml.g-lg-5, \n.jhtml.gx-lg-5 {\n    --bs-gutter-x: 3rem;\n  }\n  .jhtml .g-lg-5, \n.jhtml .gy-lg-5 , .jhtml.g-lg-5, \n.jhtml.gy-lg-5 {\n    --bs-gutter-y: 3rem;\n  }\n}\n@media (min-width: 1200px) {\n  .jhtml .col-xl , .jhtml.col-xl {\n    flex: 1 0 0%;\n  }\n  .jhtml .row-cols-xl-auto > * , .jhtml.row-cols-xl-auto > * {\n    flex: 0 0 auto;\n    width: auto;\n  }\n  .jhtml .row-cols-xl-1 > * , .jhtml.row-cols-xl-1 > * {\n    flex: 0 0 auto;\n    width: 100%;\n  }\n  .jhtml .row-cols-xl-2 > * , .jhtml.row-cols-xl-2 > * {\n    flex: 0 0 auto;\n    width: 50%;\n  }\n  .jhtml .row-cols-xl-3 > * , .jhtml.row-cols-xl-3 > * {\n    flex: 0 0 auto;\n    width: 33.3333333333%;\n  }\n  .jhtml .row-cols-xl-4 > * , .jhtml.row-cols-xl-4 > * {\n    flex: 0 0 auto;\n    width: 25%;\n  }\n  .jhtml .row-cols-xl-5 > * , .jhtml.row-cols-xl-5 > * {\n    flex: 0 0 auto;\n    width: 20%;\n  }\n  .jhtml .row-cols-xl-6 > * , .jhtml.row-cols-xl-6 > * {\n    flex: 0 0 auto;\n    width: 16.6666666667%;\n  }\n  .jhtml .col-xl-auto , .jhtml.col-xl-auto {\n    flex: 0 0 auto;\n    width: auto;\n  }\n  .jhtml .col-xl-1 , .jhtml.col-xl-1 {\n    flex: 0 0 auto;\n    width: 8.33333333%;\n  }\n  .jhtml .col-xl-2 , .jhtml.col-xl-2 {\n    flex: 0 0 auto;\n    width: 16.66666667%;\n  }\n  .jhtml .col-xl-3 , .jhtml.col-xl-3 {\n    flex: 0 0 auto;\n    width: 25%;\n  }\n  .jhtml .col-xl-4 , .jhtml.col-xl-4 {\n    flex: 0 0 auto;\n    width: 33.33333333%;\n  }\n  .jhtml .col-xl-5 , .jhtml.col-xl-5 {\n    flex: 0 0 auto;\n    width: 41.66666667%;\n  }\n  .jhtml .col-xl-6 , .jhtml.col-xl-6 {\n    flex: 0 0 auto;\n    width: 50%;\n  }\n  .jhtml .col-xl-7 , .jhtml.col-xl-7 {\n    flex: 0 0 auto;\n    width: 58.33333333%;\n  }\n  .jhtml .col-xl-8 , .jhtml.col-xl-8 {\n    flex: 0 0 auto;\n    width: 66.66666667%;\n  }\n  .jhtml .col-xl-9 , .jhtml.col-xl-9 {\n    flex: 0 0 auto;\n    width: 75%;\n  }\n  .jhtml .col-xl-10 , .jhtml.col-xl-10 {\n    flex: 0 0 auto;\n    width: 83.33333333%;\n  }\n  .jhtml .col-xl-11 , .jhtml.col-xl-11 {\n    flex: 0 0 auto;\n    width: 91.66666667%;\n  }\n  .jhtml .col-xl-12 , .jhtml.col-xl-12 {\n    flex: 0 0 auto;\n    width: 100%;\n  }\n  .jhtml .offset-xl-0 , .jhtml.offset-xl-0 {\n    margin-left: 0;\n  }\n  .jhtml .offset-xl-1 , .jhtml.offset-xl-1 {\n    margin-left: 8.33333333%;\n  }\n  .jhtml .offset-xl-2 , .jhtml.offset-xl-2 {\n    margin-left: 16.66666667%;\n  }\n  .jhtml .offset-xl-3 , .jhtml.offset-xl-3 {\n    margin-left: 25%;\n  }\n  .jhtml .offset-xl-4 , .jhtml.offset-xl-4 {\n    margin-left: 33.33333333%;\n  }\n  .jhtml .offset-xl-5 , .jhtml.offset-xl-5 {\n    margin-left: 41.66666667%;\n  }\n  .jhtml .offset-xl-6 , .jhtml.offset-xl-6 {\n    margin-left: 50%;\n  }\n  .jhtml .offset-xl-7 , .jhtml.offset-xl-7 {\n    margin-left: 58.33333333%;\n  }\n  .jhtml .offset-xl-8 , .jhtml.offset-xl-8 {\n    margin-left: 66.66666667%;\n  }\n  .jhtml .offset-xl-9 , .jhtml.offset-xl-9 {\n    margin-left: 75%;\n  }\n  .jhtml .offset-xl-10 , .jhtml.offset-xl-10 {\n    margin-left: 83.33333333%;\n  }\n  .jhtml .offset-xl-11 , .jhtml.offset-xl-11 {\n    margin-left: 91.66666667%;\n  }\n  .jhtml .g-xl-0, \n.jhtml .gx-xl-0 , .jhtml.g-xl-0, \n.jhtml.gx-xl-0 {\n    --bs-gutter-x: 0;\n  }\n  .jhtml .g-xl-0, \n.jhtml .gy-xl-0 , .jhtml.g-xl-0, \n.jhtml.gy-xl-0 {\n    --bs-gutter-y: 0;\n  }\n  .jhtml .g-xl-1, \n.jhtml .gx-xl-1 , .jhtml.g-xl-1, \n.jhtml.gx-xl-1 {\n    --bs-gutter-x: 0.25rem;\n  }\n  .jhtml .g-xl-1, \n.jhtml .gy-xl-1 , .jhtml.g-xl-1, \n.jhtml.gy-xl-1 {\n    --bs-gutter-y: 0.25rem;\n  }\n  .jhtml .g-xl-2, \n.jhtml .gx-xl-2 , .jhtml.g-xl-2, \n.jhtml.gx-xl-2 {\n    --bs-gutter-x: 0.5rem;\n  }\n  .jhtml .g-xl-2, \n.jhtml .gy-xl-2 , .jhtml.g-xl-2, \n.jhtml.gy-xl-2 {\n    --bs-gutter-y: 0.5rem;\n  }\n  .jhtml .g-xl-3, \n.jhtml .gx-xl-3 , .jhtml.g-xl-3, \n.jhtml.gx-xl-3 {\n    --bs-gutter-x: 1rem;\n  }\n  .jhtml .g-xl-3, \n.jhtml .gy-xl-3 , .jhtml.g-xl-3, \n.jhtml.gy-xl-3 {\n    --bs-gutter-y: 1rem;\n  }\n  .jhtml .g-xl-4, \n.jhtml .gx-xl-4 , .jhtml.g-xl-4, \n.jhtml.gx-xl-4 {\n    --bs-gutter-x: 1.5rem;\n  }\n  .jhtml .g-xl-4, \n.jhtml .gy-xl-4 , .jhtml.g-xl-4, \n.jhtml.gy-xl-4 {\n    --bs-gutter-y: 1.5rem;\n  }\n  .jhtml .g-xl-5, \n.jhtml .gx-xl-5 , .jhtml.g-xl-5, \n.jhtml.gx-xl-5 {\n    --bs-gutter-x: 3rem;\n  }\n  .jhtml .g-xl-5, \n.jhtml .gy-xl-5 , .jhtml.g-xl-5, \n.jhtml.gy-xl-5 {\n    --bs-gutter-y: 3rem;\n  }\n}\n@media (min-width: 1400px) {\n  .jhtml .col-xxl , .jhtml.col-xxl {\n    flex: 1 0 0%;\n  }\n  .jhtml .row-cols-xxl-auto > * , .jhtml.row-cols-xxl-auto > * {\n    flex: 0 0 auto;\n    width: auto;\n  }\n  .jhtml .row-cols-xxl-1 > * , .jhtml.row-cols-xxl-1 > * {\n    flex: 0 0 auto;\n    width: 100%;\n  }\n  .jhtml .row-cols-xxl-2 > * , .jhtml.row-cols-xxl-2 > * {\n    flex: 0 0 auto;\n    width: 50%;\n  }\n  .jhtml .row-cols-xxl-3 > * , .jhtml.row-cols-xxl-3 > * {\n    flex: 0 0 auto;\n    width: 33.3333333333%;\n  }\n  .jhtml .row-cols-xxl-4 > * , .jhtml.row-cols-xxl-4 > * {\n    flex: 0 0 auto;\n    width: 25%;\n  }\n  .jhtml .row-cols-xxl-5 > * , .jhtml.row-cols-xxl-5 > * {\n    flex: 0 0 auto;\n    width: 20%;\n  }\n  .jhtml .row-cols-xxl-6 > * , .jhtml.row-cols-xxl-6 > * {\n    flex: 0 0 auto;\n    width: 16.6666666667%;\n  }\n  .jhtml .col-xxl-auto , .jhtml.col-xxl-auto {\n    flex: 0 0 auto;\n    width: auto;\n  }\n  .jhtml .col-xxl-1 , .jhtml.col-xxl-1 {\n    flex: 0 0 auto;\n    width: 8.33333333%;\n  }\n  .jhtml .col-xxl-2 , .jhtml.col-xxl-2 {\n    flex: 0 0 auto;\n    width: 16.66666667%;\n  }\n  .jhtml .col-xxl-3 , .jhtml.col-xxl-3 {\n    flex: 0 0 auto;\n    width: 25%;\n  }\n  .jhtml .col-xxl-4 , .jhtml.col-xxl-4 {\n    flex: 0 0 auto;\n    width: 33.33333333%;\n  }\n  .jhtml .col-xxl-5 , .jhtml.col-xxl-5 {\n    flex: 0 0 auto;\n    width: 41.66666667%;\n  }\n  .jhtml .col-xxl-6 , .jhtml.col-xxl-6 {\n    flex: 0 0 auto;\n    width: 50%;\n  }\n  .jhtml .col-xxl-7 , .jhtml.col-xxl-7 {\n    flex: 0 0 auto;\n    width: 58.33333333%;\n  }\n  .jhtml .col-xxl-8 , .jhtml.col-xxl-8 {\n    flex: 0 0 auto;\n    width: 66.66666667%;\n  }\n  .jhtml .col-xxl-9 , .jhtml.col-xxl-9 {\n    flex: 0 0 auto;\n    width: 75%;\n  }\n  .jhtml .col-xxl-10 , .jhtml.col-xxl-10 {\n    flex: 0 0 auto;\n    width: 83.33333333%;\n  }\n  .jhtml .col-xxl-11 , .jhtml.col-xxl-11 {\n    flex: 0 0 auto;\n    width: 91.66666667%;\n  }\n  .jhtml .col-xxl-12 , .jhtml.col-xxl-12 {\n    flex: 0 0 auto;\n    width: 100%;\n  }\n  .jhtml .offset-xxl-0 , .jhtml.offset-xxl-0 {\n    margin-left: 0;\n  }\n  .jhtml .offset-xxl-1 , .jhtml.offset-xxl-1 {\n    margin-left: 8.33333333%;\n  }\n  .jhtml .offset-xxl-2 , .jhtml.offset-xxl-2 {\n    margin-left: 16.66666667%;\n  }\n  .jhtml .offset-xxl-3 , .jhtml.offset-xxl-3 {\n    margin-left: 25%;\n  }\n  .jhtml .offset-xxl-4 , .jhtml.offset-xxl-4 {\n    margin-left: 33.33333333%;\n  }\n  .jhtml .offset-xxl-5 , .jhtml.offset-xxl-5 {\n    margin-left: 41.66666667%;\n  }\n  .jhtml .offset-xxl-6 , .jhtml.offset-xxl-6 {\n    margin-left: 50%;\n  }\n  .jhtml .offset-xxl-7 , .jhtml.offset-xxl-7 {\n    margin-left: 58.33333333%;\n  }\n  .jhtml .offset-xxl-8 , .jhtml.offset-xxl-8 {\n    margin-left: 66.66666667%;\n  }\n  .jhtml .offset-xxl-9 , .jhtml.offset-xxl-9 {\n    margin-left: 75%;\n  }\n  .jhtml .offset-xxl-10 , .jhtml.offset-xxl-10 {\n    margin-left: 83.33333333%;\n  }\n  .jhtml .offset-xxl-11 , .jhtml.offset-xxl-11 {\n    margin-left: 91.66666667%;\n  }\n  .jhtml .g-xxl-0, \n.jhtml .gx-xxl-0 , .jhtml.g-xxl-0, \n.jhtml.gx-xxl-0 {\n    --bs-gutter-x: 0;\n  }\n  .jhtml .g-xxl-0, \n.jhtml .gy-xxl-0 , .jhtml.g-xxl-0, \n.jhtml.gy-xxl-0 {\n    --bs-gutter-y: 0;\n  }\n  .jhtml .g-xxl-1, \n.jhtml .gx-xxl-1 , .jhtml.g-xxl-1, \n.jhtml.gx-xxl-1 {\n    --bs-gutter-x: 0.25rem;\n  }\n  .jhtml .g-xxl-1, \n.jhtml .gy-xxl-1 , .jhtml.g-xxl-1, \n.jhtml.gy-xxl-1 {\n    --bs-gutter-y: 0.25rem;\n  }\n  .jhtml .g-xxl-2, \n.jhtml .gx-xxl-2 , .jhtml.g-xxl-2, \n.jhtml.gx-xxl-2 {\n    --bs-gutter-x: 0.5rem;\n  }\n  .jhtml .g-xxl-2, \n.jhtml .gy-xxl-2 , .jhtml.g-xxl-2, \n.jhtml.gy-xxl-2 {\n    --bs-gutter-y: 0.5rem;\n  }\n  .jhtml .g-xxl-3, \n.jhtml .gx-xxl-3 , .jhtml.g-xxl-3, \n.jhtml.gx-xxl-3 {\n    --bs-gutter-x: 1rem;\n  }\n  .jhtml .g-xxl-3, \n.jhtml .gy-xxl-3 , .jhtml.g-xxl-3, \n.jhtml.gy-xxl-3 {\n    --bs-gutter-y: 1rem;\n  }\n  .jhtml .g-xxl-4, \n.jhtml .gx-xxl-4 , .jhtml.g-xxl-4, \n.jhtml.gx-xxl-4 {\n    --bs-gutter-x: 1.5rem;\n  }\n  .jhtml .g-xxl-4, \n.jhtml .gy-xxl-4 , .jhtml.g-xxl-4, \n.jhtml.gy-xxl-4 {\n    --bs-gutter-y: 1.5rem;\n  }\n  .jhtml .g-xxl-5, \n.jhtml .gx-xxl-5 , .jhtml.g-xxl-5, \n.jhtml.gx-xxl-5 {\n    --bs-gutter-x: 3rem;\n  }\n  .jhtml .g-xxl-5, \n.jhtml .gy-xxl-5 , .jhtml.g-xxl-5, \n.jhtml.gy-xxl-5 {\n    --bs-gutter-y: 3rem;\n  }\n}\n.jhtml .table , .jhtml.table {\n  --bs-table-bg: transparent;\n  --bs-table-accent-bg: transparent;\n  --bs-table-striped-color: #212529;\n  --bs-table-striped-bg: rgba(0, 0, 0, 0.05);\n  --bs-table-active-color: #212529;\n  --bs-table-active-bg: rgba(0, 0, 0, 0.1);\n  --bs-table-hover-color: #212529;\n  --bs-table-hover-bg: rgba(0, 0, 0, 0.075);\n  width: 100%;\n  margin-bottom: 1rem;\n  color: #212529;\n  vertical-align: top;\n  border-color: #dee2e6;\n}\n.jhtml .table > :not(caption) > * > * , .jhtml.table > :not(caption) > * > * {\n  padding: 0.5rem 0.5rem;\n  background-color: var(--bs-table-bg);\n  border-bottom-width: 1px;\n  box-shadow: inset 0 0 0 9999px var(--bs-table-accent-bg);\n}\n.jhtml .table > tbody , .jhtml.table > tbody {\n  vertical-align: inherit;\n}\n.jhtml .table > thead , .jhtml.table > thead {\n  vertical-align: bottom;\n}\n.jhtml .table > :not(:first-child) , .jhtml.table > :not(:first-child) {\n  border-top: 2px solid currentColor;\n}\n.jhtml .caption-top , .jhtml.caption-top {\n  caption-side: top;\n}\n.jhtml .table-sm > :not(caption) > * > * , .jhtml.table-sm > :not(caption) > * > * {\n  padding: 0.25rem 0.25rem;\n}\n.jhtml .table-bordered > :not(caption) > * , .jhtml.table-bordered > :not(caption) > * {\n  border-width: 1px 0;\n}\n.jhtml .table-bordered > :not(caption) > * > * , .jhtml.table-bordered > :not(caption) > * > * {\n  border-width: 0 1px;\n}\n.jhtml .table-borderless > :not(caption) > * > * , .jhtml.table-borderless > :not(caption) > * > * {\n  border-bottom-width: 0;\n}\n.jhtml .table-borderless > :not(:first-child) , .jhtml.table-borderless > :not(:first-child) {\n  border-top-width: 0;\n}\n.jhtml .table-striped > tbody > tr:nth-of-type(odd) > * , .jhtml.table-striped > tbody > tr:nth-of-type(odd) > * {\n  --bs-table-accent-bg: var(--bs-table-striped-bg);\n  color: var(--bs-table-striped-color);\n}\n.jhtml .table-active , .jhtml.table-active {\n  --bs-table-accent-bg: var(--bs-table-active-bg);\n  color: var(--bs-table-active-color);\n}\n.jhtml .table-hover > tbody > tr:hover > * , .jhtml.table-hover > tbody > tr:hover > * {\n  --bs-table-accent-bg: var(--bs-table-hover-bg);\n  color: var(--bs-table-hover-color);\n}\n.jhtml .table-primary , .jhtml.table-primary {\n  --bs-table-bg: #cfe2ff;\n  --bs-table-striped-bg: #c5d7f2;\n  --bs-table-striped-color: #000;\n  --bs-table-active-bg: #bacbe6;\n  --bs-table-active-color: #000;\n  --bs-table-hover-bg: #bfd1ec;\n  --bs-table-hover-color: #000;\n  color: #000;\n  border-color: #bacbe6;\n}\n.jhtml .table-secondary , .jhtml.table-secondary {\n  --bs-table-bg: #e2e3e5;\n  --bs-table-striped-bg: #d7d8da;\n  --bs-table-striped-color: #000;\n  --bs-table-active-bg: #cbccce;\n  --bs-table-active-color: #000;\n  --bs-table-hover-bg: #d1d2d4;\n  --bs-table-hover-color: #000;\n  color: #000;\n  border-color: #cbccce;\n}\n.jhtml .table-success , .jhtml.table-success {\n  --bs-table-bg: #d1e7dd;\n  --bs-table-striped-bg: #c7dbd2;\n  --bs-table-striped-color: #000;\n  --bs-table-active-bg: #bcd0c7;\n  --bs-table-active-color: #000;\n  --bs-table-hover-bg: #c1d6cc;\n  --bs-table-hover-color: #000;\n  color: #000;\n  border-color: #bcd0c7;\n}\n.jhtml .table-info , .jhtml.table-info {\n  --bs-table-bg: #cff4fc;\n  --bs-table-striped-bg: #c5e8ef;\n  --bs-table-striped-color: #000;\n  --bs-table-active-bg: #badce3;\n  --bs-table-active-color: #000;\n  --bs-table-hover-bg: #bfe2e9;\n  --bs-table-hover-color: #000;\n  color: #000;\n  border-color: #badce3;\n}\n.jhtml .table-warning , .jhtml.table-warning {\n  --bs-table-bg: #fff3cd;\n  --bs-table-striped-bg: #f2e7c3;\n  --bs-table-striped-color: #000;\n  --bs-table-active-bg: #e6dbb9;\n  --bs-table-active-color: #000;\n  --bs-table-hover-bg: #ece1be;\n  --bs-table-hover-color: #000;\n  color: #000;\n  border-color: #e6dbb9;\n}\n.jhtml .table-danger , .jhtml.table-danger {\n  --bs-table-bg: #f8d7da;\n  --bs-table-striped-bg: #eccccf;\n  --bs-table-striped-color: #000;\n  --bs-table-active-bg: #dfc2c4;\n  --bs-table-active-color: #000;\n  --bs-table-hover-bg: #e5c7ca;\n  --bs-table-hover-color: #000;\n  color: #000;\n  border-color: #dfc2c4;\n}\n.jhtml .table-light , .jhtml.table-light {\n  --bs-table-bg: #f8f9fa;\n  --bs-table-striped-bg: #ecedee;\n  --bs-table-striped-color: #000;\n  --bs-table-active-bg: #dfe0e1;\n  --bs-table-active-color: #000;\n  --bs-table-hover-bg: #e5e6e7;\n  --bs-table-hover-color: #000;\n  color: #000;\n  border-color: #dfe0e1;\n}\n.jhtml .table-dark , .jhtml.table-dark {\n  --bs-table-bg: #212529;\n  --bs-table-striped-bg: #2c3034;\n  --bs-table-striped-color: #fff;\n  --bs-table-active-bg: #373b3e;\n  --bs-table-active-color: #fff;\n  --bs-table-hover-bg: #323539;\n  --bs-table-hover-color: #fff;\n  color: #fff;\n  border-color: #373b3e;\n}\n.jhtml .table-responsive , .jhtml.table-responsive {\n  overflow-x: auto;\n  -webkit-overflow-scrolling: touch;\n}\n@media (max-width: 575.98px) {\n  .jhtml .table-responsive-sm , .jhtml.table-responsive-sm {\n    overflow-x: auto;\n    -webkit-overflow-scrolling: touch;\n  }\n}\n@media (max-width: 767.98px) {\n  .jhtml .table-responsive-md , .jhtml.table-responsive-md {\n    overflow-x: auto;\n    -webkit-overflow-scrolling: touch;\n  }\n}\n@media (max-width: 991.98px) {\n  .jhtml .table-responsive-lg , .jhtml.table-responsive-lg {\n    overflow-x: auto;\n    -webkit-overflow-scrolling: touch;\n  }\n}\n@media (max-width: 1199.98px) {\n  .jhtml .table-responsive-xl , .jhtml.table-responsive-xl {\n    overflow-x: auto;\n    -webkit-overflow-scrolling: touch;\n  }\n}\n@media (max-width: 1399.98px) {\n  .jhtml .table-responsive-xxl , .jhtml.table-responsive-xxl {\n    overflow-x: auto;\n    -webkit-overflow-scrolling: touch;\n  }\n}\n.jhtml .form-label , .jhtml.form-label {\n  margin-bottom: 0.5rem;\n}\n.jhtml .col-form-label , .jhtml.col-form-label {\n  padding-top: calc(0.375rem + 1px);\n  padding-bottom: calc(0.375rem + 1px);\n  margin-bottom: 0;\n  font-size: inherit;\n  line-height: 1.5;\n}\n.jhtml .col-form-label-lg , .jhtml.col-form-label-lg {\n  padding-top: calc(0.5rem + 1px);\n  padding-bottom: calc(0.5rem + 1px);\n  font-size: 1.25rem;\n}\n.jhtml .col-form-label-sm , .jhtml.col-form-label-sm {\n  padding-top: calc(0.25rem + 1px);\n  padding-bottom: calc(0.25rem + 1px);\n  font-size: 0.875rem;\n}\n.jhtml .form-text , .jhtml.form-text {\n  margin-top: 0.25rem;\n  font-size: 0.875em;\n  color: #6c757d;\n}\n.jhtml .form-control , .jhtml.form-control {\n  display: block;\n  width: 100%;\n  padding: 0.375rem 0.75rem;\n  font-size: 1rem;\n  font-weight: 400;\n  line-height: 1.5;\n  color: #212529;\n  background-color: #fff;\n  background-clip: padding-box;\n  border: 1px solid #ced4da;\n  appearance: none;\n  border-radius: 0.25rem;\n  transition: border-color 0.15s ease-in-out, box-shadow 0.15s ease-in-out;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .form-control , .jhtml.form-control {\n    transition: none;\n  }\n}\n.jhtml .form-control[type=file] , .jhtml.form-control[type=file] {\n  overflow: hidden;\n}\n.jhtml .form-control[type=file]:not(:disabled):not([readonly]) , .jhtml.form-control[type=file]:not(:disabled):not([readonly]) {\n  cursor: pointer;\n}\n.jhtml .form-control:focus , .jhtml.form-control:focus {\n  color: #212529;\n  background-color: #fff;\n  border-color: #86b7fe;\n  outline: 0;\n  box-shadow: 0 0 0 0.25rem rgba(13, 110, 253, 0.25);\n}\n.jhtml .form-control::-webkit-date-and-time-value , .jhtml.form-control::-webkit-date-and-time-value {\n  height: 1.5em;\n}\n.jhtml .form-control::placeholder , .jhtml.form-control::placeholder {\n  color: #6c757d;\n  opacity: 1;\n}\n.jhtml .form-control:disabled,  .jhtml .form-control[readonly] , .jhtml.form-control:disabled,  .jhtml.form-control[readonly] {\n  background-color: #e9ecef;\n  opacity: 1;\n}\n.jhtml .form-control::file-selector-button , .jhtml.form-control::file-selector-button {\n  padding: 0.375rem 0.75rem;\n  margin: -0.375rem -0.75rem;\n  margin-inline-end: 0.75rem;\n  color: #212529;\n  background-color: #e9ecef;\n  pointer-events: none;\n  border-color: inherit;\n  border-style: solid;\n  border-width: 0;\n  border-inline-end-width: 1px;\n  border-radius: 0;\n  transition: color 0.15s ease-in-out, background-color 0.15s ease-in-out, border-color 0.15s ease-in-out, box-shadow 0.15s ease-in-out;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .form-control::file-selector-button , .jhtml.form-control::file-selector-button {\n    transition: none;\n  }\n}\n.jhtml .form-control:hover:not(:disabled):not([readonly])::file-selector-button , .jhtml.form-control:hover:not(:disabled):not([readonly])::file-selector-button {\n  background-color: #dde0e3;\n}\n.jhtml .form-control::-webkit-file-upload-button , .jhtml.form-control::-webkit-file-upload-button {\n  padding: 0.375rem 0.75rem;\n  margin: -0.375rem -0.75rem;\n  margin-inline-end: 0.75rem;\n  color: #212529;\n  background-color: #e9ecef;\n  pointer-events: none;\n  border-color: inherit;\n  border-style: solid;\n  border-width: 0;\n  border-inline-end-width: 1px;\n  border-radius: 0;\n  transition: color 0.15s ease-in-out, background-color 0.15s ease-in-out, border-color 0.15s ease-in-out, box-shadow 0.15s ease-in-out;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .form-control::-webkit-file-upload-button , .jhtml.form-control::-webkit-file-upload-button {\n    transition: none;\n  }\n}\n.jhtml .form-control:hover:not(:disabled):not([readonly])::-webkit-file-upload-button , .jhtml.form-control:hover:not(:disabled):not([readonly])::-webkit-file-upload-button {\n  background-color: #dde0e3;\n}\n.jhtml .form-control-plaintext , .jhtml.form-control-plaintext {\n  display: block;\n  width: 100%;\n  padding: 0.375rem 0;\n  margin-bottom: 0;\n  line-height: 1.5;\n  color: #212529;\n  background-color: transparent;\n  border: solid transparent;\n  border-width: 1px 0;\n}\n.jhtml .form-control-plaintext.form-control-sm,  .jhtml .form-control-plaintext.form-control-lg , .jhtml.form-control-plaintext.form-control-sm,  .jhtml.form-control-plaintext.form-control-lg {\n  padding-right: 0;\n  padding-left: 0;\n}\n.jhtml .form-control-sm , .jhtml.form-control-sm {\n  min-height: calc(1.5em + 0.5rem + 2px);\n  padding: 0.25rem 0.5rem;\n  font-size: 0.875rem;\n  border-radius: 0.2rem;\n}\n.jhtml .form-control-sm::file-selector-button , .jhtml.form-control-sm::file-selector-button {\n  padding: 0.25rem 0.5rem;\n  margin: -0.25rem -0.5rem;\n  margin-inline-end: 0.5rem;\n}\n.jhtml .form-control-sm::-webkit-file-upload-button , .jhtml.form-control-sm::-webkit-file-upload-button {\n  padding: 0.25rem 0.5rem;\n  margin: -0.25rem -0.5rem;\n  margin-inline-end: 0.5rem;\n}\n.jhtml .form-control-lg , .jhtml.form-control-lg {\n  min-height: calc(1.5em + 1rem + 2px);\n  padding: 0.5rem 1rem;\n  font-size: 1.25rem;\n  border-radius: 0.3rem;\n}\n.jhtml .form-control-lg::file-selector-button , .jhtml.form-control-lg::file-selector-button {\n  padding: 0.5rem 1rem;\n  margin: -0.5rem -1rem;\n  margin-inline-end: 1rem;\n}\n.jhtml .form-control-lg::-webkit-file-upload-button , .jhtml.form-control-lg::-webkit-file-upload-button {\n  padding: 0.5rem 1rem;\n  margin: -0.5rem -1rem;\n  margin-inline-end: 1rem;\n}\n.jhtml textarea.form-control {\n  min-height: calc(1.5em + 0.75rem + 2px);\n}\n.jhtml textarea.form-control-sm {\n  min-height: calc(1.5em + 0.5rem + 2px);\n}\n.jhtml textarea.form-control-lg {\n  min-height: calc(1.5em + 1rem + 2px);\n}\n.jhtml .form-control-color , .jhtml.form-control-color {\n  width: 3rem;\n  height: auto;\n  padding: 0.375rem;\n}\n.jhtml .form-control-color:not(:disabled):not([readonly]) , .jhtml.form-control-color:not(:disabled):not([readonly]) {\n  cursor: pointer;\n}\n.jhtml .form-control-color::-moz-color-swatch , .jhtml.form-control-color::-moz-color-swatch {\n  height: 1.5em;\n  border-radius: 0.25rem;\n}\n.jhtml .form-control-color::-webkit-color-swatch , .jhtml.form-control-color::-webkit-color-swatch {\n  height: 1.5em;\n  border-radius: 0.25rem;\n}\n.jhtml .form-select , .jhtml.form-select {\n  display: block;\n  width: 100%;\n  padding: 0.375rem 2.25rem 0.375rem 0.75rem;\n  -moz-padding-start: calc(0.75rem - 3px);\n  font-size: 1rem;\n  font-weight: 400;\n  line-height: 1.5;\n  color: #212529;\n  background-color: #fff;\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 16 16'%3e%3cpath fill='none' stroke='%23343a40' stroke-linecap='round' stroke-linejoin='round' stroke-width='2' d='M2 5l6 6 6-6'/%3e%3c/svg%3e\");\n  background-repeat: no-repeat;\n  background-position: right 0.75rem center;\n  background-size: 16px 12px;\n  border: 1px solid #ced4da;\n  border-radius: 0.25rem;\n  transition: border-color 0.15s ease-in-out, box-shadow 0.15s ease-in-out;\n  appearance: none;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .form-select , .jhtml.form-select {\n    transition: none;\n  }\n}\n.jhtml .form-select:focus , .jhtml.form-select:focus {\n  border-color: #86b7fe;\n  outline: 0;\n  box-shadow: 0 0 0 0.25rem rgba(13, 110, 253, 0.25);\n}\n.jhtml .form-select[multiple],  .jhtml .form-select[size]:not([size=\"1\"]) , .jhtml.form-select[multiple],  .jhtml.form-select[size]:not([size=\"1\"]) {\n  padding-right: 0.75rem;\n  background-image: none;\n}\n.jhtml .form-select:disabled , .jhtml.form-select:disabled {\n  background-color: #e9ecef;\n}\n.jhtml .form-select:-moz-focusring , .jhtml.form-select:-moz-focusring {\n  color: transparent;\n  text-shadow: 0 0 0 #212529;\n}\n.jhtml .form-select-sm , .jhtml.form-select-sm {\n  padding-top: 0.25rem;\n  padding-bottom: 0.25rem;\n  padding-left: 0.5rem;\n  font-size: 0.875rem;\n  border-radius: 0.2rem;\n}\n.jhtml .form-select-lg , .jhtml.form-select-lg {\n  padding-top: 0.5rem;\n  padding-bottom: 0.5rem;\n  padding-left: 1rem;\n  font-size: 1.25rem;\n  border-radius: 0.3rem;\n}\n.jhtml .form-check , .jhtml.form-check {\n  display: block;\n  min-height: 1.5rem;\n  padding-left: 1.5em;\n  margin-bottom: 0.125rem;\n}\n.jhtml .form-check .form-check-input , .jhtml.form-check .form-check-input {\n  float: left;\n  margin-left: -1.5em;\n}\n.jhtml .form-check-input , .jhtml.form-check-input {\n  width: 1em;\n  height: 1em;\n  margin-top: 0.25em;\n  vertical-align: top;\n  background-color: #fff;\n  background-repeat: no-repeat;\n  background-position: center;\n  background-size: contain;\n  border: 1px solid rgba(0, 0, 0, 0.25);\n  appearance: none;\n  color-adjust: exact;\n}\n.jhtml .form-check-input[type=checkbox] , .jhtml.form-check-input[type=checkbox] {\n  border-radius: 0.25em;\n}\n.jhtml .form-check-input[type=radio] , .jhtml.form-check-input[type=radio] {\n  border-radius: 50%;\n}\n.jhtml .form-check-input:active , .jhtml.form-check-input:active {\n  filter: brightness(90%);\n}\n.jhtml .form-check-input:focus , .jhtml.form-check-input:focus {\n  border-color: #86b7fe;\n  outline: 0;\n  box-shadow: 0 0 0 0.25rem rgba(13, 110, 253, 0.25);\n}\n.jhtml .form-check-input:checked , .jhtml.form-check-input:checked {\n  background-color: #0d6efd;\n  border-color: #0d6efd;\n}\n.jhtml .form-check-input:checked[type=checkbox] , .jhtml.form-check-input:checked[type=checkbox] {\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 20 20'%3e%3cpath fill='none' stroke='%23fff' stroke-linecap='round' stroke-linejoin='round' stroke-width='3' d='M6 10l3 3l6-6'/%3e%3c/svg%3e\");\n}\n.jhtml .form-check-input:checked[type=radio] , .jhtml.form-check-input:checked[type=radio] {\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='-4 -4 8 8'%3e%3ccircle r='2' fill='%23fff'/%3e%3c/svg%3e\");\n}\n.jhtml .form-check-input[type=checkbox]:indeterminate , .jhtml.form-check-input[type=checkbox]:indeterminate {\n  background-color: #0d6efd;\n  border-color: #0d6efd;\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 20 20'%3e%3cpath fill='none' stroke='%23fff' stroke-linecap='round' stroke-linejoin='round' stroke-width='3' d='M6 10h8'/%3e%3c/svg%3e\");\n}\n.jhtml .form-check-input:disabled , .jhtml.form-check-input:disabled {\n  pointer-events: none;\n  filter: none;\n  opacity: 0.5;\n}\n.jhtml .form-check-input[disabled] ~ .form-check-label,  .jhtml .form-check-input:disabled ~ .form-check-label , .jhtml.form-check-input[disabled] ~ .form-check-label,  .jhtml.form-check-input:disabled ~ .form-check-label {\n  opacity: 0.5;\n}\n.jhtml .form-switch , .jhtml.form-switch {\n  padding-left: 2.5em;\n}\n.jhtml .form-switch .form-check-input , .jhtml.form-switch .form-check-input {\n  width: 2em;\n  margin-left: -2.5em;\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='-4 -4 8 8'%3e%3ccircle r='3' fill='rgba%280, 0, 0, 0.25%29'/%3e%3c/svg%3e\");\n  background-position: left center;\n  border-radius: 2em;\n  transition: background-position 0.15s ease-in-out;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .form-switch .form-check-input , .jhtml.form-switch .form-check-input {\n    transition: none;\n  }\n}\n.jhtml .form-switch .form-check-input:focus , .jhtml.form-switch .form-check-input:focus {\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='-4 -4 8 8'%3e%3ccircle r='3' fill='%2386b7fe'/%3e%3c/svg%3e\");\n}\n.jhtml .form-switch .form-check-input:checked , .jhtml.form-switch .form-check-input:checked {\n  background-position: right center;\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='-4 -4 8 8'%3e%3ccircle r='3' fill='%23fff'/%3e%3c/svg%3e\");\n}\n.jhtml .form-check-inline , .jhtml.form-check-inline {\n  display: inline-block;\n  margin-right: 1rem;\n}\n.jhtml .btn-check , .jhtml.btn-check {\n  position: absolute;\n  clip: rect(0, 0, 0, 0);\n  pointer-events: none;\n}\n.jhtml .btn-check[disabled] + .btn,  .jhtml .btn-check:disabled + .btn , .jhtml.btn-check[disabled] + .btn,  .jhtml.btn-check:disabled + .btn {\n  pointer-events: none;\n  filter: none;\n  opacity: 0.65;\n}\n.jhtml .form-range , .jhtml.form-range {\n  width: 100%;\n  height: 1.5rem;\n  padding: 0;\n  background-color: transparent;\n  appearance: none;\n}\n.jhtml .form-range:focus , .jhtml.form-range:focus {\n  outline: 0;\n}\n.jhtml .form-range:focus::-webkit-slider-thumb , .jhtml.form-range:focus::-webkit-slider-thumb {\n  box-shadow: 0 0 0 1px #fff, 0 0 0 0.25rem rgba(13, 110, 253, 0.25);\n}\n.jhtml .form-range:focus::-moz-range-thumb , .jhtml.form-range:focus::-moz-range-thumb {\n  box-shadow: 0 0 0 1px #fff, 0 0 0 0.25rem rgba(13, 110, 253, 0.25);\n}\n.jhtml .form-range::-moz-focus-outer , .jhtml.form-range::-moz-focus-outer {\n  border: 0;\n}\n.jhtml .form-range::-webkit-slider-thumb , .jhtml.form-range::-webkit-slider-thumb {\n  width: 1rem;\n  height: 1rem;\n  margin-top: -0.25rem;\n  background-color: #0d6efd;\n  border: 0;\n  border-radius: 1rem;\n  transition: background-color 0.15s ease-in-out, border-color 0.15s ease-in-out, box-shadow 0.15s ease-in-out;\n  appearance: none;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .form-range::-webkit-slider-thumb , .jhtml.form-range::-webkit-slider-thumb {\n    transition: none;\n  }\n}\n.jhtml .form-range::-webkit-slider-thumb:active , .jhtml.form-range::-webkit-slider-thumb:active {\n  background-color: #b6d4fe;\n}\n.jhtml .form-range::-webkit-slider-runnable-track , .jhtml.form-range::-webkit-slider-runnable-track {\n  width: 100%;\n  height: 0.5rem;\n  color: transparent;\n  cursor: pointer;\n  background-color: #dee2e6;\n  border-color: transparent;\n  border-radius: 1rem;\n}\n.jhtml .form-range::-moz-range-thumb , .jhtml.form-range::-moz-range-thumb {\n  width: 1rem;\n  height: 1rem;\n  background-color: #0d6efd;\n  border: 0;\n  border-radius: 1rem;\n  transition: background-color 0.15s ease-in-out, border-color 0.15s ease-in-out, box-shadow 0.15s ease-in-out;\n  appearance: none;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .form-range::-moz-range-thumb , .jhtml.form-range::-moz-range-thumb {\n    transition: none;\n  }\n}\n.jhtml .form-range::-moz-range-thumb:active , .jhtml.form-range::-moz-range-thumb:active {\n  background-color: #b6d4fe;\n}\n.jhtml .form-range::-moz-range-track , .jhtml.form-range::-moz-range-track {\n  width: 100%;\n  height: 0.5rem;\n  color: transparent;\n  cursor: pointer;\n  background-color: #dee2e6;\n  border-color: transparent;\n  border-radius: 1rem;\n}\n.jhtml .form-range:disabled , .jhtml.form-range:disabled {\n  pointer-events: none;\n}\n.jhtml .form-range:disabled::-webkit-slider-thumb , .jhtml.form-range:disabled::-webkit-slider-thumb {\n  background-color: #adb5bd;\n}\n.jhtml .form-range:disabled::-moz-range-thumb , .jhtml.form-range:disabled::-moz-range-thumb {\n  background-color: #adb5bd;\n}\n.jhtml .form-floating , .jhtml.form-floating {\n  position: relative;\n}\n.jhtml .form-floating > .form-control, \n.jhtml .form-floating > .form-select , .jhtml.form-floating > .form-control, \n.jhtml.form-floating > .form-select {\n  height: calc(3.5rem + 2px);\n  line-height: 1.25;\n}\n.jhtml .form-floating > label , .jhtml.form-floating > label {\n  position: absolute;\n  top: 0;\n  left: 0;\n  height: 100%;\n  padding: 1rem 0.75rem;\n  pointer-events: none;\n  border: 1px solid transparent;\n  transform-origin: 0 0;\n  transition: opacity 0.1s ease-in-out, transform 0.1s ease-in-out;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .form-floating > label , .jhtml.form-floating > label {\n    transition: none;\n  }\n}\n.jhtml .form-floating > .form-control , .jhtml.form-floating > .form-control {\n  padding: 1rem 0.75rem;\n}\n.jhtml .form-floating > .form-control::placeholder , .jhtml.form-floating > .form-control::placeholder {\n  color: transparent;\n}\n.jhtml .form-floating > .form-control:focus,  .jhtml .form-floating > .form-control:not(:placeholder-shown) , .jhtml.form-floating > .form-control:focus,  .jhtml.form-floating > .form-control:not(:placeholder-shown) {\n  padding-top: 1.625rem;\n  padding-bottom: 0.625rem;\n}\n.jhtml .form-floating > .form-control:-webkit-autofill , .jhtml.form-floating > .form-control:-webkit-autofill {\n  padding-top: 1.625rem;\n  padding-bottom: 0.625rem;\n}\n.jhtml .form-floating > .form-select , .jhtml.form-floating > .form-select {\n  padding-top: 1.625rem;\n  padding-bottom: 0.625rem;\n}\n.jhtml .form-floating > .form-control:focus ~ label, \n.jhtml .form-floating > .form-control:not(:placeholder-shown) ~ label, \n.jhtml .form-floating > .form-select ~ label , .jhtml.form-floating > .form-control:focus ~ label, \n.jhtml.form-floating > .form-control:not(:placeholder-shown) ~ label, \n.jhtml.form-floating > .form-select ~ label {\n  opacity: 0.65;\n  transform: scale(0.85) translateY(-0.5rem) translateX(0.15rem);\n}\n.jhtml .form-floating > .form-control:-webkit-autofill ~ label , .jhtml.form-floating > .form-control:-webkit-autofill ~ label {\n  opacity: 0.65;\n  transform: scale(0.85) translateY(-0.5rem) translateX(0.15rem);\n}\n.jhtml .input-group , .jhtml.input-group {\n  position: relative;\n  display: flex;\n  flex-wrap: wrap;\n  align-items: stretch;\n  width: 100%;\n}\n.jhtml .input-group > .form-control, \n.jhtml .input-group > .form-select , .jhtml.input-group > .form-control, \n.jhtml.input-group > .form-select {\n  position: relative;\n  flex: 1 1 auto;\n  width: 1%;\n  min-width: 0;\n}\n.jhtml .input-group > .form-control:focus, \n.jhtml .input-group > .form-select:focus , .jhtml.input-group > .form-control:focus, \n.jhtml.input-group > .form-select:focus {\n  z-index: 3;\n}\n.jhtml .input-group .btn , .jhtml.input-group .btn {\n  position: relative;\n  z-index: 2;\n}\n.jhtml .input-group .btn:focus , .jhtml.input-group .btn:focus {\n  z-index: 3;\n}\n.jhtml .input-group-text , .jhtml.input-group-text {\n  display: flex;\n  align-items: center;\n  padding: 0.375rem 0.75rem;\n  font-size: 1rem;\n  font-weight: 400;\n  line-height: 1.5;\n  color: #212529;\n  text-align: center;\n  white-space: nowrap;\n  background-color: #e9ecef;\n  border: 1px solid #ced4da;\n  border-radius: 0.25rem;\n}\n.jhtml .input-group-lg > .form-control, \n.jhtml .input-group-lg > .form-select, \n.jhtml .input-group-lg > .input-group-text, \n.jhtml .input-group-lg > .btn , .jhtml.input-group-lg > .form-control, \n.jhtml.input-group-lg > .form-select, \n.jhtml.input-group-lg > .input-group-text, \n.jhtml.input-group-lg > .btn {\n  padding: 0.5rem 1rem;\n  font-size: 1.25rem;\n  border-radius: 0.3rem;\n}\n.jhtml .input-group-sm > .form-control, \n.jhtml .input-group-sm > .form-select, \n.jhtml .input-group-sm > .input-group-text, \n.jhtml .input-group-sm > .btn , .jhtml.input-group-sm > .form-control, \n.jhtml.input-group-sm > .form-select, \n.jhtml.input-group-sm > .input-group-text, \n.jhtml.input-group-sm > .btn {\n  padding: 0.25rem 0.5rem;\n  font-size: 0.875rem;\n  border-radius: 0.2rem;\n}\n.jhtml .input-group-lg > .form-select, \n.jhtml .input-group-sm > .form-select , .jhtml.input-group-lg > .form-select, \n.jhtml.input-group-sm > .form-select {\n  padding-right: 3rem;\n}\n.jhtml .input-group:not(.has-validation) > :not(:last-child):not(.dropdown-toggle):not(.dropdown-menu), \n.jhtml .input-group:not(.has-validation) > .dropdown-toggle:nth-last-child(n+3) , .jhtml.input-group:not(.has-validation) > :not(:last-child):not(.dropdown-toggle):not(.dropdown-menu), \n.jhtml.input-group:not(.has-validation) > .dropdown-toggle:nth-last-child(n+3) {\n  border-top-right-radius: 0;\n  border-bottom-right-radius: 0;\n}\n.jhtml .input-group.has-validation > :nth-last-child(n+3):not(.dropdown-toggle):not(.dropdown-menu), \n.jhtml .input-group.has-validation > .dropdown-toggle:nth-last-child(n+4) , .jhtml.input-group.has-validation > :nth-last-child(n+3):not(.dropdown-toggle):not(.dropdown-menu), \n.jhtml.input-group.has-validation > .dropdown-toggle:nth-last-child(n+4) {\n  border-top-right-radius: 0;\n  border-bottom-right-radius: 0;\n}\n.jhtml .input-group > :not(:first-child):not(.dropdown-menu):not(.valid-tooltip):not(.valid-feedback):not(.invalid-tooltip):not(.invalid-feedback) , .jhtml.input-group > :not(:first-child):not(.dropdown-menu):not(.valid-tooltip):not(.valid-feedback):not(.invalid-tooltip):not(.invalid-feedback) {\n  margin-left: -1px;\n  border-top-left-radius: 0;\n  border-bottom-left-radius: 0;\n}\n.jhtml .valid-feedback , .jhtml.valid-feedback {\n  display: none;\n  width: 100%;\n  margin-top: 0.25rem;\n  font-size: 0.875em;\n  color: #198754;\n}\n.jhtml .valid-tooltip , .jhtml.valid-tooltip {\n  position: absolute;\n  top: 100%;\n  z-index: 5;\n  display: none;\n  max-width: 100%;\n  padding: 0.25rem 0.5rem;\n  margin-top: 0.1rem;\n  font-size: 0.875rem;\n  color: #fff;\n  background-color: rgba(25, 135, 84, 0.9);\n  border-radius: 0.25rem;\n}\n.jhtml .was-validated :valid ~ .valid-feedback, \n.jhtml .was-validated :valid ~ .valid-tooltip, \n.jhtml .is-valid ~ .valid-feedback, \n.jhtml .is-valid ~ .valid-tooltip , .jhtml.was-validated :valid ~ .valid-feedback, \n.jhtml.was-validated :valid ~ .valid-tooltip, \n.jhtml.is-valid ~ .valid-feedback, \n.jhtml.is-valid ~ .valid-tooltip {\n  display: block;\n}\n.jhtml .was-validated .form-control:valid,  .jhtml .form-control.is-valid , .jhtml.was-validated .form-control:valid,  .jhtml.form-control.is-valid {\n  border-color: #198754;\n  padding-right: calc(1.5em + 0.75rem);\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 8 8'%3e%3cpath fill='%23198754' d='M2.3 6.73L.6 4.53c-.4-1.04.46-1.4 1.1-.8l1.1 1.4 3.4-3.8c.6-.63 1.6-.27 1.2.7l-4 4.6c-.43.5-.8.4-1.1.1z'/%3e%3c/svg%3e\");\n  background-repeat: no-repeat;\n  background-position: right calc(0.375em + 0.1875rem) center;\n  background-size: calc(0.75em + 0.375rem) calc(0.75em + 0.375rem);\n}\n.jhtml .was-validated .form-control:valid:focus,  .jhtml .form-control.is-valid:focus , .jhtml.was-validated .form-control:valid:focus,  .jhtml.form-control.is-valid:focus {\n  border-color: #198754;\n  box-shadow: 0 0 0 0.25rem rgba(25, 135, 84, 0.25);\n}\n.jhtml .was-validated textarea.form-control:valid,  .jhtml textarea.form-control.is-valid , .jhtml.was-validated textarea.form-control:valid{\n  padding-right: calc(1.5em + 0.75rem);\n  background-position: top calc(0.375em + 0.1875rem) right calc(0.375em + 0.1875rem);\n}\n.jhtml .was-validated .form-select:valid,  .jhtml .form-select.is-valid , .jhtml.was-validated .form-select:valid,  .jhtml.form-select.is-valid {\n  border-color: #198754;\n}\n.jhtml .was-validated .form-select:valid:not([multiple]):not([size]),  .jhtml .was-validated .form-select:valid:not([multiple])[size=\"1\"],  .jhtml .form-select.is-valid:not([multiple]):not([size]),  .jhtml .form-select.is-valid:not([multiple])[size=\"1\"] , .jhtml.was-validated .form-select:valid:not([multiple]):not([size]),  .jhtml.was-validated .form-select:valid:not([multiple])[size=\"1\"],  .jhtml.form-select.is-valid:not([multiple]):not([size]),  .jhtml.form-select.is-valid:not([multiple])[size=\"1\"] {\n  padding-right: 4.125rem;\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 16 16'%3e%3cpath fill='none' stroke='%23343a40' stroke-linecap='round' stroke-linejoin='round' stroke-width='2' d='M2 5l6 6 6-6'/%3e%3c/svg%3e\"), url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 8 8'%3e%3cpath fill='%23198754' d='M2.3 6.73L.6 4.53c-.4-1.04.46-1.4 1.1-.8l1.1 1.4 3.4-3.8c.6-.63 1.6-.27 1.2.7l-4 4.6c-.43.5-.8.4-1.1.1z'/%3e%3c/svg%3e\");\n  background-position: right 0.75rem center, center right 2.25rem;\n  background-size: 16px 12px, calc(0.75em + 0.375rem) calc(0.75em + 0.375rem);\n}\n.jhtml .was-validated .form-select:valid:focus,  .jhtml .form-select.is-valid:focus , .jhtml.was-validated .form-select:valid:focus,  .jhtml.form-select.is-valid:focus {\n  border-color: #198754;\n  box-shadow: 0 0 0 0.25rem rgba(25, 135, 84, 0.25);\n}\n.jhtml .was-validated .form-check-input:valid,  .jhtml .form-check-input.is-valid , .jhtml.was-validated .form-check-input:valid,  .jhtml.form-check-input.is-valid {\n  border-color: #198754;\n}\n.jhtml .was-validated .form-check-input:valid:checked,  .jhtml .form-check-input.is-valid:checked , .jhtml.was-validated .form-check-input:valid:checked,  .jhtml.form-check-input.is-valid:checked {\n  background-color: #198754;\n}\n.jhtml .was-validated .form-check-input:valid:focus,  .jhtml .form-check-input.is-valid:focus , .jhtml.was-validated .form-check-input:valid:focus,  .jhtml.form-check-input.is-valid:focus {\n  box-shadow: 0 0 0 0.25rem rgba(25, 135, 84, 0.25);\n}\n.jhtml .was-validated .form-check-input:valid ~ .form-check-label,  .jhtml .form-check-input.is-valid ~ .form-check-label , .jhtml.was-validated .form-check-input:valid ~ .form-check-label,  .jhtml.form-check-input.is-valid ~ .form-check-label {\n  color: #198754;\n}\n.jhtml .form-check-inline .form-check-input ~ .valid-feedback , .jhtml.form-check-inline .form-check-input ~ .valid-feedback {\n  margin-left: 0.5em;\n}\n.jhtml .was-validated .input-group .form-control:valid,  .jhtml .input-group .form-control.is-valid, \n.jhtml .was-validated .input-group .form-select:valid, \n.jhtml .input-group .form-select.is-valid , .jhtml.was-validated .input-group .form-control:valid,  .jhtml.input-group .form-control.is-valid, \n.jhtml.was-validated .input-group .form-select:valid, \n.jhtml.input-group .form-select.is-valid {\n  z-index: 1;\n}\n.jhtml .was-validated .input-group .form-control:valid:focus,  .jhtml .input-group .form-control.is-valid:focus, \n.jhtml .was-validated .input-group .form-select:valid:focus, \n.jhtml .input-group .form-select.is-valid:focus , .jhtml.was-validated .input-group .form-control:valid:focus,  .jhtml.input-group .form-control.is-valid:focus, \n.jhtml.was-validated .input-group .form-select:valid:focus, \n.jhtml.input-group .form-select.is-valid:focus {\n  z-index: 3;\n}\n.jhtml .invalid-feedback , .jhtml.invalid-feedback {\n  display: none;\n  width: 100%;\n  margin-top: 0.25rem;\n  font-size: 0.875em;\n  color: #dc3545;\n}\n.jhtml .invalid-tooltip , .jhtml.invalid-tooltip {\n  position: absolute;\n  top: 100%;\n  z-index: 5;\n  display: none;\n  max-width: 100%;\n  padding: 0.25rem 0.5rem;\n  margin-top: 0.1rem;\n  font-size: 0.875rem;\n  color: #fff;\n  background-color: rgba(220, 53, 69, 0.9);\n  border-radius: 0.25rem;\n}\n.jhtml .was-validated :invalid ~ .invalid-feedback, \n.jhtml .was-validated :invalid ~ .invalid-tooltip, \n.jhtml .is-invalid ~ .invalid-feedback, \n.jhtml .is-invalid ~ .invalid-tooltip , .jhtml.was-validated :invalid ~ .invalid-feedback, \n.jhtml.was-validated :invalid ~ .invalid-tooltip, \n.jhtml.is-invalid ~ .invalid-feedback, \n.jhtml.is-invalid ~ .invalid-tooltip {\n  display: block;\n}\n.jhtml .was-validated .form-control:invalid,  .jhtml .form-control.is-invalid , .jhtml.was-validated .form-control:invalid,  .jhtml.form-control.is-invalid {\n  border-color: #dc3545;\n  padding-right: calc(1.5em + 0.75rem);\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 12 12' width='12' height='12' fill='none' stroke='%23dc3545'%3e%3ccircle cx='6' cy='6' r='4.5'/%3e%3cpath stroke-linejoin='round' d='M5.8 3.6h.4L6 6.5z'/%3e%3ccircle cx='6' cy='8.2' r='.6' fill='%23dc3545' stroke='none'/%3e%3c/svg%3e\");\n  background-repeat: no-repeat;\n  background-position: right calc(0.375em + 0.1875rem) center;\n  background-size: calc(0.75em + 0.375rem) calc(0.75em + 0.375rem);\n}\n.jhtml .was-validated .form-control:invalid:focus,  .jhtml .form-control.is-invalid:focus , .jhtml.was-validated .form-control:invalid:focus,  .jhtml.form-control.is-invalid:focus {\n  border-color: #dc3545;\n  box-shadow: 0 0 0 0.25rem rgba(220, 53, 69, 0.25);\n}\n.jhtml .was-validated textarea.form-control:invalid,  .jhtml textarea.form-control.is-invalid , .jhtml.was-validated textarea.form-control:invalid{\n  padding-right: calc(1.5em + 0.75rem);\n  background-position: top calc(0.375em + 0.1875rem) right calc(0.375em + 0.1875rem);\n}\n.jhtml .was-validated .form-select:invalid,  .jhtml .form-select.is-invalid , .jhtml.was-validated .form-select:invalid,  .jhtml.form-select.is-invalid {\n  border-color: #dc3545;\n}\n.jhtml .was-validated .form-select:invalid:not([multiple]):not([size]),  .jhtml .was-validated .form-select:invalid:not([multiple])[size=\"1\"],  .jhtml .form-select.is-invalid:not([multiple]):not([size]),  .jhtml .form-select.is-invalid:not([multiple])[size=\"1\"] , .jhtml.was-validated .form-select:invalid:not([multiple]):not([size]),  .jhtml.was-validated .form-select:invalid:not([multiple])[size=\"1\"],  .jhtml.form-select.is-invalid:not([multiple]):not([size]),  .jhtml.form-select.is-invalid:not([multiple])[size=\"1\"] {\n  padding-right: 4.125rem;\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 16 16'%3e%3cpath fill='none' stroke='%23343a40' stroke-linecap='round' stroke-linejoin='round' stroke-width='2' d='M2 5l6 6 6-6'/%3e%3c/svg%3e\"), url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 12 12' width='12' height='12' fill='none' stroke='%23dc3545'%3e%3ccircle cx='6' cy='6' r='4.5'/%3e%3cpath stroke-linejoin='round' d='M5.8 3.6h.4L6 6.5z'/%3e%3ccircle cx='6' cy='8.2' r='.6' fill='%23dc3545' stroke='none'/%3e%3c/svg%3e\");\n  background-position: right 0.75rem center, center right 2.25rem;\n  background-size: 16px 12px, calc(0.75em + 0.375rem) calc(0.75em + 0.375rem);\n}\n.jhtml .was-validated .form-select:invalid:focus,  .jhtml .form-select.is-invalid:focus , .jhtml.was-validated .form-select:invalid:focus,  .jhtml.form-select.is-invalid:focus {\n  border-color: #dc3545;\n  box-shadow: 0 0 0 0.25rem rgba(220, 53, 69, 0.25);\n}\n.jhtml .was-validated .form-check-input:invalid,  .jhtml .form-check-input.is-invalid , .jhtml.was-validated .form-check-input:invalid,  .jhtml.form-check-input.is-invalid {\n  border-color: #dc3545;\n}\n.jhtml .was-validated .form-check-input:invalid:checked,  .jhtml .form-check-input.is-invalid:checked , .jhtml.was-validated .form-check-input:invalid:checked,  .jhtml.form-check-input.is-invalid:checked {\n  background-color: #dc3545;\n}\n.jhtml .was-validated .form-check-input:invalid:focus,  .jhtml .form-check-input.is-invalid:focus , .jhtml.was-validated .form-check-input:invalid:focus,  .jhtml.form-check-input.is-invalid:focus {\n  box-shadow: 0 0 0 0.25rem rgba(220, 53, 69, 0.25);\n}\n.jhtml .was-validated .form-check-input:invalid ~ .form-check-label,  .jhtml .form-check-input.is-invalid ~ .form-check-label , .jhtml.was-validated .form-check-input:invalid ~ .form-check-label,  .jhtml.form-check-input.is-invalid ~ .form-check-label {\n  color: #dc3545;\n}\n.jhtml .form-check-inline .form-check-input ~ .invalid-feedback , .jhtml.form-check-inline .form-check-input ~ .invalid-feedback {\n  margin-left: 0.5em;\n}\n.jhtml .was-validated .input-group .form-control:invalid,  .jhtml .input-group .form-control.is-invalid, \n.jhtml .was-validated .input-group .form-select:invalid, \n.jhtml .input-group .form-select.is-invalid , .jhtml.was-validated .input-group .form-control:invalid,  .jhtml.input-group .form-control.is-invalid, \n.jhtml.was-validated .input-group .form-select:invalid, \n.jhtml.input-group .form-select.is-invalid {\n  z-index: 2;\n}\n.jhtml .was-validated .input-group .form-control:invalid:focus,  .jhtml .input-group .form-control.is-invalid:focus, \n.jhtml .was-validated .input-group .form-select:invalid:focus, \n.jhtml .input-group .form-select.is-invalid:focus , .jhtml.was-validated .input-group .form-control:invalid:focus,  .jhtml.input-group .form-control.is-invalid:focus, \n.jhtml.was-validated .input-group .form-select:invalid:focus, \n.jhtml.input-group .form-select.is-invalid:focus {\n  z-index: 3;\n}\n.jhtml .btn , .jhtml.btn {\n  display: inline-block;\n  font-weight: 400;\n  line-height: 1.5;\n  color: #212529;\n  text-align: center;\n  text-decoration: none;\n  vertical-align: middle;\n  cursor: pointer;\n  user-select: none;\n  background-color: transparent;\n  border: 1px solid transparent;\n  padding: 0.375rem 0.75rem;\n  font-size: 1rem;\n  border-radius: 0.25rem;\n  transition: color 0.15s ease-in-out, background-color 0.15s ease-in-out, border-color 0.15s ease-in-out, box-shadow 0.15s ease-in-out;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .btn , .jhtml.btn {\n    transition: none;\n  }\n}\n.jhtml .btn:hover , .jhtml.btn:hover {\n  color: #212529;\n}\n.jhtml .btn-check:focus + .btn,  .jhtml .btn:focus , .jhtml.btn-check:focus + .btn,  .jhtml.btn:focus {\n  outline: 0;\n  box-shadow: 0 0 0 0.25rem rgba(13, 110, 253, 0.25);\n}\n.jhtml .btn:disabled,  .jhtml .btn.disabled,  .jhtml fieldset:disabled .btn , .jhtml.btn:disabled,  .jhtml.btn.disabled{\n  pointer-events: none;\n  opacity: 0.65;\n}\n.jhtml .btn-primary , .jhtml.btn-primary {\n  color: #fff;\n  background-color: #0d6efd;\n  border-color: #0d6efd;\n}\n.jhtml .btn-primary:hover , .jhtml.btn-primary:hover {\n  color: #fff;\n  background-color: #0b5ed7;\n  border-color: #0a58ca;\n}\n.jhtml .btn-check:focus + .btn-primary,  .jhtml .btn-primary:focus , .jhtml.btn-check:focus + .btn-primary,  .jhtml.btn-primary:focus {\n  color: #fff;\n  background-color: #0b5ed7;\n  border-color: #0a58ca;\n  box-shadow: 0 0 0 0.25rem rgba(49, 132, 253, 0.5);\n}\n.jhtml .btn-check:checked + .btn-primary,  .jhtml .btn-check:active + .btn-primary,  .jhtml .btn-primary:active,  .jhtml .btn-primary.active,  .jhtml .show > .btn-primary.dropdown-toggle , .jhtml.btn-check:checked + .btn-primary,  .jhtml.btn-check:active + .btn-primary,  .jhtml.btn-primary:active,  .jhtml.btn-primary.active,  .jhtml.show > .btn-primary.dropdown-toggle {\n  color: #fff;\n  background-color: #0a58ca;\n  border-color: #0a53be;\n}\n.jhtml .btn-check:checked + .btn-primary:focus,  .jhtml .btn-check:active + .btn-primary:focus,  .jhtml .btn-primary:active:focus,  .jhtml .btn-primary.active:focus,  .jhtml .show > .btn-primary.dropdown-toggle:focus , .jhtml.btn-check:checked + .btn-primary:focus,  .jhtml.btn-check:active + .btn-primary:focus,  .jhtml.btn-primary:active:focus,  .jhtml.btn-primary.active:focus,  .jhtml.show > .btn-primary.dropdown-toggle:focus {\n  box-shadow: 0 0 0 0.25rem rgba(49, 132, 253, 0.5);\n}\n.jhtml .btn-primary:disabled,  .jhtml .btn-primary.disabled , .jhtml.btn-primary:disabled,  .jhtml.btn-primary.disabled {\n  color: #fff;\n  background-color: #0d6efd;\n  border-color: #0d6efd;\n}\n.jhtml .btn-secondary , .jhtml.btn-secondary {\n  color: #fff;\n  background-color: #6c757d;\n  border-color: #6c757d;\n}\n.jhtml .btn-secondary:hover , .jhtml.btn-secondary:hover {\n  color: #fff;\n  background-color: #5c636a;\n  border-color: #565e64;\n}\n.jhtml .btn-check:focus + .btn-secondary,  .jhtml .btn-secondary:focus , .jhtml.btn-check:focus + .btn-secondary,  .jhtml.btn-secondary:focus {\n  color: #fff;\n  background-color: #5c636a;\n  border-color: #565e64;\n  box-shadow: 0 0 0 0.25rem rgba(130, 138, 145, 0.5);\n}\n.jhtml .btn-check:checked + .btn-secondary,  .jhtml .btn-check:active + .btn-secondary,  .jhtml .btn-secondary:active,  .jhtml .btn-secondary.active,  .jhtml .show > .btn-secondary.dropdown-toggle , .jhtml.btn-check:checked + .btn-secondary,  .jhtml.btn-check:active + .btn-secondary,  .jhtml.btn-secondary:active,  .jhtml.btn-secondary.active,  .jhtml.show > .btn-secondary.dropdown-toggle {\n  color: #fff;\n  background-color: #565e64;\n  border-color: #51585e;\n}\n.jhtml .btn-check:checked + .btn-secondary:focus,  .jhtml .btn-check:active + .btn-secondary:focus,  .jhtml .btn-secondary:active:focus,  .jhtml .btn-secondary.active:focus,  .jhtml .show > .btn-secondary.dropdown-toggle:focus , .jhtml.btn-check:checked + .btn-secondary:focus,  .jhtml.btn-check:active + .btn-secondary:focus,  .jhtml.btn-secondary:active:focus,  .jhtml.btn-secondary.active:focus,  .jhtml.show > .btn-secondary.dropdown-toggle:focus {\n  box-shadow: 0 0 0 0.25rem rgba(130, 138, 145, 0.5);\n}\n.jhtml .btn-secondary:disabled,  .jhtml .btn-secondary.disabled , .jhtml.btn-secondary:disabled,  .jhtml.btn-secondary.disabled {\n  color: #fff;\n  background-color: #6c757d;\n  border-color: #6c757d;\n}\n.jhtml .btn-success , .jhtml.btn-success {\n  color: #fff;\n  background-color: #198754;\n  border-color: #198754;\n}\n.jhtml .btn-success:hover , .jhtml.btn-success:hover {\n  color: #fff;\n  background-color: #157347;\n  border-color: #146c43;\n}\n.jhtml .btn-check:focus + .btn-success,  .jhtml .btn-success:focus , .jhtml.btn-check:focus + .btn-success,  .jhtml.btn-success:focus {\n  color: #fff;\n  background-color: #157347;\n  border-color: #146c43;\n  box-shadow: 0 0 0 0.25rem rgba(60, 153, 110, 0.5);\n}\n.jhtml .btn-check:checked + .btn-success,  .jhtml .btn-check:active + .btn-success,  .jhtml .btn-success:active,  .jhtml .btn-success.active,  .jhtml .show > .btn-success.dropdown-toggle , .jhtml.btn-check:checked + .btn-success,  .jhtml.btn-check:active + .btn-success,  .jhtml.btn-success:active,  .jhtml.btn-success.active,  .jhtml.show > .btn-success.dropdown-toggle {\n  color: #fff;\n  background-color: #146c43;\n  border-color: #13653f;\n}\n.jhtml .btn-check:checked + .btn-success:focus,  .jhtml .btn-check:active + .btn-success:focus,  .jhtml .btn-success:active:focus,  .jhtml .btn-success.active:focus,  .jhtml .show > .btn-success.dropdown-toggle:focus , .jhtml.btn-check:checked + .btn-success:focus,  .jhtml.btn-check:active + .btn-success:focus,  .jhtml.btn-success:active:focus,  .jhtml.btn-success.active:focus,  .jhtml.show > .btn-success.dropdown-toggle:focus {\n  box-shadow: 0 0 0 0.25rem rgba(60, 153, 110, 0.5);\n}\n.jhtml .btn-success:disabled,  .jhtml .btn-success.disabled , .jhtml.btn-success:disabled,  .jhtml.btn-success.disabled {\n  color: #fff;\n  background-color: #198754;\n  border-color: #198754;\n}\n.jhtml .btn-info , .jhtml.btn-info {\n  color: #000;\n  background-color: #0dcaf0;\n  border-color: #0dcaf0;\n}\n.jhtml .btn-info:hover , .jhtml.btn-info:hover {\n  color: #000;\n  background-color: #31d2f2;\n  border-color: #25cff2;\n}\n.jhtml .btn-check:focus + .btn-info,  .jhtml .btn-info:focus , .jhtml.btn-check:focus + .btn-info,  .jhtml.btn-info:focus {\n  color: #000;\n  background-color: #31d2f2;\n  border-color: #25cff2;\n  box-shadow: 0 0 0 0.25rem rgba(11, 172, 204, 0.5);\n}\n.jhtml .btn-check:checked + .btn-info,  .jhtml .btn-check:active + .btn-info,  .jhtml .btn-info:active,  .jhtml .btn-info.active,  .jhtml .show > .btn-info.dropdown-toggle , .jhtml.btn-check:checked + .btn-info,  .jhtml.btn-check:active + .btn-info,  .jhtml.btn-info:active,  .jhtml.btn-info.active,  .jhtml.show > .btn-info.dropdown-toggle {\n  color: #000;\n  background-color: #3dd5f3;\n  border-color: #25cff2;\n}\n.jhtml .btn-check:checked + .btn-info:focus,  .jhtml .btn-check:active + .btn-info:focus,  .jhtml .btn-info:active:focus,  .jhtml .btn-info.active:focus,  .jhtml .show > .btn-info.dropdown-toggle:focus , .jhtml.btn-check:checked + .btn-info:focus,  .jhtml.btn-check:active + .btn-info:focus,  .jhtml.btn-info:active:focus,  .jhtml.btn-info.active:focus,  .jhtml.show > .btn-info.dropdown-toggle:focus {\n  box-shadow: 0 0 0 0.25rem rgba(11, 172, 204, 0.5);\n}\n.jhtml .btn-info:disabled,  .jhtml .btn-info.disabled , .jhtml.btn-info:disabled,  .jhtml.btn-info.disabled {\n  color: #000;\n  background-color: #0dcaf0;\n  border-color: #0dcaf0;\n}\n.jhtml .btn-warning , .jhtml.btn-warning {\n  color: #000;\n  background-color: #ffc107;\n  border-color: #ffc107;\n}\n.jhtml .btn-warning:hover , .jhtml.btn-warning:hover {\n  color: #000;\n  background-color: #ffca2c;\n  border-color: #ffc720;\n}\n.jhtml .btn-check:focus + .btn-warning,  .jhtml .btn-warning:focus , .jhtml.btn-check:focus + .btn-warning,  .jhtml.btn-warning:focus {\n  color: #000;\n  background-color: #ffca2c;\n  border-color: #ffc720;\n  box-shadow: 0 0 0 0.25rem rgba(217, 164, 6, 0.5);\n}\n.jhtml .btn-check:checked + .btn-warning,  .jhtml .btn-check:active + .btn-warning,  .jhtml .btn-warning:active,  .jhtml .btn-warning.active,  .jhtml .show > .btn-warning.dropdown-toggle , .jhtml.btn-check:checked + .btn-warning,  .jhtml.btn-check:active + .btn-warning,  .jhtml.btn-warning:active,  .jhtml.btn-warning.active,  .jhtml.show > .btn-warning.dropdown-toggle {\n  color: #000;\n  background-color: #ffcd39;\n  border-color: #ffc720;\n}\n.jhtml .btn-check:checked + .btn-warning:focus,  .jhtml .btn-check:active + .btn-warning:focus,  .jhtml .btn-warning:active:focus,  .jhtml .btn-warning.active:focus,  .jhtml .show > .btn-warning.dropdown-toggle:focus , .jhtml.btn-check:checked + .btn-warning:focus,  .jhtml.btn-check:active + .btn-warning:focus,  .jhtml.btn-warning:active:focus,  .jhtml.btn-warning.active:focus,  .jhtml.show > .btn-warning.dropdown-toggle:focus {\n  box-shadow: 0 0 0 0.25rem rgba(217, 164, 6, 0.5);\n}\n.jhtml .btn-warning:disabled,  .jhtml .btn-warning.disabled , .jhtml.btn-warning:disabled,  .jhtml.btn-warning.disabled {\n  color: #000;\n  background-color: #ffc107;\n  border-color: #ffc107;\n}\n.jhtml .btn-danger , .jhtml.btn-danger {\n  color: #fff;\n  background-color: #dc3545;\n  border-color: #dc3545;\n}\n.jhtml .btn-danger:hover , .jhtml.btn-danger:hover {\n  color: #fff;\n  background-color: #bb2d3b;\n  border-color: #b02a37;\n}\n.jhtml .btn-check:focus + .btn-danger,  .jhtml .btn-danger:focus , .jhtml.btn-check:focus + .btn-danger,  .jhtml.btn-danger:focus {\n  color: #fff;\n  background-color: #bb2d3b;\n  border-color: #b02a37;\n  box-shadow: 0 0 0 0.25rem rgba(225, 83, 97, 0.5);\n}\n.jhtml .btn-check:checked + .btn-danger,  .jhtml .btn-check:active + .btn-danger,  .jhtml .btn-danger:active,  .jhtml .btn-danger.active,  .jhtml .show > .btn-danger.dropdown-toggle , .jhtml.btn-check:checked + .btn-danger,  .jhtml.btn-check:active + .btn-danger,  .jhtml.btn-danger:active,  .jhtml.btn-danger.active,  .jhtml.show > .btn-danger.dropdown-toggle {\n  color: #fff;\n  background-color: #b02a37;\n  border-color: #a52834;\n}\n.jhtml .btn-check:checked + .btn-danger:focus,  .jhtml .btn-check:active + .btn-danger:focus,  .jhtml .btn-danger:active:focus,  .jhtml .btn-danger.active:focus,  .jhtml .show > .btn-danger.dropdown-toggle:focus , .jhtml.btn-check:checked + .btn-danger:focus,  .jhtml.btn-check:active + .btn-danger:focus,  .jhtml.btn-danger:active:focus,  .jhtml.btn-danger.active:focus,  .jhtml.show > .btn-danger.dropdown-toggle:focus {\n  box-shadow: 0 0 0 0.25rem rgba(225, 83, 97, 0.5);\n}\n.jhtml .btn-danger:disabled,  .jhtml .btn-danger.disabled , .jhtml.btn-danger:disabled,  .jhtml.btn-danger.disabled {\n  color: #fff;\n  background-color: #dc3545;\n  border-color: #dc3545;\n}\n.jhtml .btn-light , .jhtml.btn-light {\n  color: #000;\n  background-color: #f8f9fa;\n  border-color: #f8f9fa;\n}\n.jhtml .btn-light:hover , .jhtml.btn-light:hover {\n  color: #000;\n  background-color: #f9fafb;\n  border-color: #f9fafb;\n}\n.jhtml .btn-check:focus + .btn-light,  .jhtml .btn-light:focus , .jhtml.btn-check:focus + .btn-light,  .jhtml.btn-light:focus {\n  color: #000;\n  background-color: #f9fafb;\n  border-color: #f9fafb;\n  box-shadow: 0 0 0 0.25rem rgba(211, 212, 213, 0.5);\n}\n.jhtml .btn-check:checked + .btn-light,  .jhtml .btn-check:active + .btn-light,  .jhtml .btn-light:active,  .jhtml .btn-light.active,  .jhtml .show > .btn-light.dropdown-toggle , .jhtml.btn-check:checked + .btn-light,  .jhtml.btn-check:active + .btn-light,  .jhtml.btn-light:active,  .jhtml.btn-light.active,  .jhtml.show > .btn-light.dropdown-toggle {\n  color: #000;\n  background-color: #f9fafb;\n  border-color: #f9fafb;\n}\n.jhtml .btn-check:checked + .btn-light:focus,  .jhtml .btn-check:active + .btn-light:focus,  .jhtml .btn-light:active:focus,  .jhtml .btn-light.active:focus,  .jhtml .show > .btn-light.dropdown-toggle:focus , .jhtml.btn-check:checked + .btn-light:focus,  .jhtml.btn-check:active + .btn-light:focus,  .jhtml.btn-light:active:focus,  .jhtml.btn-light.active:focus,  .jhtml.show > .btn-light.dropdown-toggle:focus {\n  box-shadow: 0 0 0 0.25rem rgba(211, 212, 213, 0.5);\n}\n.jhtml .btn-light:disabled,  .jhtml .btn-light.disabled , .jhtml.btn-light:disabled,  .jhtml.btn-light.disabled {\n  color: #000;\n  background-color: #f8f9fa;\n  border-color: #f8f9fa;\n}\n.jhtml .btn-dark , .jhtml.btn-dark {\n  color: #fff;\n  background-color: #212529;\n  border-color: #212529;\n}\n.jhtml .btn-dark:hover , .jhtml.btn-dark:hover {\n  color: #fff;\n  background-color: #1c1f23;\n  border-color: #1a1e21;\n}\n.jhtml .btn-check:focus + .btn-dark,  .jhtml .btn-dark:focus , .jhtml.btn-check:focus + .btn-dark,  .jhtml.btn-dark:focus {\n  color: #fff;\n  background-color: #1c1f23;\n  border-color: #1a1e21;\n  box-shadow: 0 0 0 0.25rem rgba(66, 70, 73, 0.5);\n}\n.jhtml .btn-check:checked + .btn-dark,  .jhtml .btn-check:active + .btn-dark,  .jhtml .btn-dark:active,  .jhtml .btn-dark.active,  .jhtml .show > .btn-dark.dropdown-toggle , .jhtml.btn-check:checked + .btn-dark,  .jhtml.btn-check:active + .btn-dark,  .jhtml.btn-dark:active,  .jhtml.btn-dark.active,  .jhtml.show > .btn-dark.dropdown-toggle {\n  color: #fff;\n  background-color: #1a1e21;\n  border-color: #191c1f;\n}\n.jhtml .btn-check:checked + .btn-dark:focus,  .jhtml .btn-check:active + .btn-dark:focus,  .jhtml .btn-dark:active:focus,  .jhtml .btn-dark.active:focus,  .jhtml .show > .btn-dark.dropdown-toggle:focus , .jhtml.btn-check:checked + .btn-dark:focus,  .jhtml.btn-check:active + .btn-dark:focus,  .jhtml.btn-dark:active:focus,  .jhtml.btn-dark.active:focus,  .jhtml.show > .btn-dark.dropdown-toggle:focus {\n  box-shadow: 0 0 0 0.25rem rgba(66, 70, 73, 0.5);\n}\n.jhtml .btn-dark:disabled,  .jhtml .btn-dark.disabled , .jhtml.btn-dark:disabled,  .jhtml.btn-dark.disabled {\n  color: #fff;\n  background-color: #212529;\n  border-color: #212529;\n}\n.jhtml .btn-outline-primary , .jhtml.btn-outline-primary {\n  color: #0d6efd;\n  border-color: #0d6efd;\n}\n.jhtml .btn-outline-primary:hover , .jhtml.btn-outline-primary:hover {\n  color: #fff;\n  background-color: #0d6efd;\n  border-color: #0d6efd;\n}\n.jhtml .btn-check:focus + .btn-outline-primary,  .jhtml .btn-outline-primary:focus , .jhtml.btn-check:focus + .btn-outline-primary,  .jhtml.btn-outline-primary:focus {\n  box-shadow: 0 0 0 0.25rem rgba(13, 110, 253, 0.5);\n}\n.jhtml .btn-check:checked + .btn-outline-primary,  .jhtml .btn-check:active + .btn-outline-primary,  .jhtml .btn-outline-primary:active,  .jhtml .btn-outline-primary.active,  .jhtml .btn-outline-primary.dropdown-toggle.show , .jhtml.btn-check:checked + .btn-outline-primary,  .jhtml.btn-check:active + .btn-outline-primary,  .jhtml.btn-outline-primary:active,  .jhtml.btn-outline-primary.active,  .jhtml.btn-outline-primary.dropdown-toggle.show {\n  color: #fff;\n  background-color: #0d6efd;\n  border-color: #0d6efd;\n}\n.jhtml .btn-check:checked + .btn-outline-primary:focus,  .jhtml .btn-check:active + .btn-outline-primary:focus,  .jhtml .btn-outline-primary:active:focus,  .jhtml .btn-outline-primary.active:focus,  .jhtml .btn-outline-primary.dropdown-toggle.show:focus , .jhtml.btn-check:checked + .btn-outline-primary:focus,  .jhtml.btn-check:active + .btn-outline-primary:focus,  .jhtml.btn-outline-primary:active:focus,  .jhtml.btn-outline-primary.active:focus,  .jhtml.btn-outline-primary.dropdown-toggle.show:focus {\n  box-shadow: 0 0 0 0.25rem rgba(13, 110, 253, 0.5);\n}\n.jhtml .btn-outline-primary:disabled,  .jhtml .btn-outline-primary.disabled , .jhtml.btn-outline-primary:disabled,  .jhtml.btn-outline-primary.disabled {\n  color: #0d6efd;\n  background-color: transparent;\n}\n.jhtml .btn-outline-secondary , .jhtml.btn-outline-secondary {\n  color: #6c757d;\n  border-color: #6c757d;\n}\n.jhtml .btn-outline-secondary:hover , .jhtml.btn-outline-secondary:hover {\n  color: #fff;\n  background-color: #6c757d;\n  border-color: #6c757d;\n}\n.jhtml .btn-check:focus + .btn-outline-secondary,  .jhtml .btn-outline-secondary:focus , .jhtml.btn-check:focus + .btn-outline-secondary,  .jhtml.btn-outline-secondary:focus {\n  box-shadow: 0 0 0 0.25rem rgba(108, 117, 125, 0.5);\n}\n.jhtml .btn-check:checked + .btn-outline-secondary,  .jhtml .btn-check:active + .btn-outline-secondary,  .jhtml .btn-outline-secondary:active,  .jhtml .btn-outline-secondary.active,  .jhtml .btn-outline-secondary.dropdown-toggle.show , .jhtml.btn-check:checked + .btn-outline-secondary,  .jhtml.btn-check:active + .btn-outline-secondary,  .jhtml.btn-outline-secondary:active,  .jhtml.btn-outline-secondary.active,  .jhtml.btn-outline-secondary.dropdown-toggle.show {\n  color: #fff;\n  background-color: #6c757d;\n  border-color: #6c757d;\n}\n.jhtml .btn-check:checked + .btn-outline-secondary:focus,  .jhtml .btn-check:active + .btn-outline-secondary:focus,  .jhtml .btn-outline-secondary:active:focus,  .jhtml .btn-outline-secondary.active:focus,  .jhtml .btn-outline-secondary.dropdown-toggle.show:focus , .jhtml.btn-check:checked + .btn-outline-secondary:focus,  .jhtml.btn-check:active + .btn-outline-secondary:focus,  .jhtml.btn-outline-secondary:active:focus,  .jhtml.btn-outline-secondary.active:focus,  .jhtml.btn-outline-secondary.dropdown-toggle.show:focus {\n  box-shadow: 0 0 0 0.25rem rgba(108, 117, 125, 0.5);\n}\n.jhtml .btn-outline-secondary:disabled,  .jhtml .btn-outline-secondary.disabled , .jhtml.btn-outline-secondary:disabled,  .jhtml.btn-outline-secondary.disabled {\n  color: #6c757d;\n  background-color: transparent;\n}\n.jhtml .btn-outline-success , .jhtml.btn-outline-success {\n  color: #198754;\n  border-color: #198754;\n}\n.jhtml .btn-outline-success:hover , .jhtml.btn-outline-success:hover {\n  color: #fff;\n  background-color: #198754;\n  border-color: #198754;\n}\n.jhtml .btn-check:focus + .btn-outline-success,  .jhtml .btn-outline-success:focus , .jhtml.btn-check:focus + .btn-outline-success,  .jhtml.btn-outline-success:focus {\n  box-shadow: 0 0 0 0.25rem rgba(25, 135, 84, 0.5);\n}\n.jhtml .btn-check:checked + .btn-outline-success,  .jhtml .btn-check:active + .btn-outline-success,  .jhtml .btn-outline-success:active,  .jhtml .btn-outline-success.active,  .jhtml .btn-outline-success.dropdown-toggle.show , .jhtml.btn-check:checked + .btn-outline-success,  .jhtml.btn-check:active + .btn-outline-success,  .jhtml.btn-outline-success:active,  .jhtml.btn-outline-success.active,  .jhtml.btn-outline-success.dropdown-toggle.show {\n  color: #fff;\n  background-color: #198754;\n  border-color: #198754;\n}\n.jhtml .btn-check:checked + .btn-outline-success:focus,  .jhtml .btn-check:active + .btn-outline-success:focus,  .jhtml .btn-outline-success:active:focus,  .jhtml .btn-outline-success.active:focus,  .jhtml .btn-outline-success.dropdown-toggle.show:focus , .jhtml.btn-check:checked + .btn-outline-success:focus,  .jhtml.btn-check:active + .btn-outline-success:focus,  .jhtml.btn-outline-success:active:focus,  .jhtml.btn-outline-success.active:focus,  .jhtml.btn-outline-success.dropdown-toggle.show:focus {\n  box-shadow: 0 0 0 0.25rem rgba(25, 135, 84, 0.5);\n}\n.jhtml .btn-outline-success:disabled,  .jhtml .btn-outline-success.disabled , .jhtml.btn-outline-success:disabled,  .jhtml.btn-outline-success.disabled {\n  color: #198754;\n  background-color: transparent;\n}\n.jhtml .btn-outline-info , .jhtml.btn-outline-info {\n  color: #0dcaf0;\n  border-color: #0dcaf0;\n}\n.jhtml .btn-outline-info:hover , .jhtml.btn-outline-info:hover {\n  color: #000;\n  background-color: #0dcaf0;\n  border-color: #0dcaf0;\n}\n.jhtml .btn-check:focus + .btn-outline-info,  .jhtml .btn-outline-info:focus , .jhtml.btn-check:focus + .btn-outline-info,  .jhtml.btn-outline-info:focus {\n  box-shadow: 0 0 0 0.25rem rgba(13, 202, 240, 0.5);\n}\n.jhtml .btn-check:checked + .btn-outline-info,  .jhtml .btn-check:active + .btn-outline-info,  .jhtml .btn-outline-info:active,  .jhtml .btn-outline-info.active,  .jhtml .btn-outline-info.dropdown-toggle.show , .jhtml.btn-check:checked + .btn-outline-info,  .jhtml.btn-check:active + .btn-outline-info,  .jhtml.btn-outline-info:active,  .jhtml.btn-outline-info.active,  .jhtml.btn-outline-info.dropdown-toggle.show {\n  color: #000;\n  background-color: #0dcaf0;\n  border-color: #0dcaf0;\n}\n.jhtml .btn-check:checked + .btn-outline-info:focus,  .jhtml .btn-check:active + .btn-outline-info:focus,  .jhtml .btn-outline-info:active:focus,  .jhtml .btn-outline-info.active:focus,  .jhtml .btn-outline-info.dropdown-toggle.show:focus , .jhtml.btn-check:checked + .btn-outline-info:focus,  .jhtml.btn-check:active + .btn-outline-info:focus,  .jhtml.btn-outline-info:active:focus,  .jhtml.btn-outline-info.active:focus,  .jhtml.btn-outline-info.dropdown-toggle.show:focus {\n  box-shadow: 0 0 0 0.25rem rgba(13, 202, 240, 0.5);\n}\n.jhtml .btn-outline-info:disabled,  .jhtml .btn-outline-info.disabled , .jhtml.btn-outline-info:disabled,  .jhtml.btn-outline-info.disabled {\n  color: #0dcaf0;\n  background-color: transparent;\n}\n.jhtml .btn-outline-warning , .jhtml.btn-outline-warning {\n  color: #ffc107;\n  border-color: #ffc107;\n}\n.jhtml .btn-outline-warning:hover , .jhtml.btn-outline-warning:hover {\n  color: #000;\n  background-color: #ffc107;\n  border-color: #ffc107;\n}\n.jhtml .btn-check:focus + .btn-outline-warning,  .jhtml .btn-outline-warning:focus , .jhtml.btn-check:focus + .btn-outline-warning,  .jhtml.btn-outline-warning:focus {\n  box-shadow: 0 0 0 0.25rem rgba(255, 193, 7, 0.5);\n}\n.jhtml .btn-check:checked + .btn-outline-warning,  .jhtml .btn-check:active + .btn-outline-warning,  .jhtml .btn-outline-warning:active,  .jhtml .btn-outline-warning.active,  .jhtml .btn-outline-warning.dropdown-toggle.show , .jhtml.btn-check:checked + .btn-outline-warning,  .jhtml.btn-check:active + .btn-outline-warning,  .jhtml.btn-outline-warning:active,  .jhtml.btn-outline-warning.active,  .jhtml.btn-outline-warning.dropdown-toggle.show {\n  color: #000;\n  background-color: #ffc107;\n  border-color: #ffc107;\n}\n.jhtml .btn-check:checked + .btn-outline-warning:focus,  .jhtml .btn-check:active + .btn-outline-warning:focus,  .jhtml .btn-outline-warning:active:focus,  .jhtml .btn-outline-warning.active:focus,  .jhtml .btn-outline-warning.dropdown-toggle.show:focus , .jhtml.btn-check:checked + .btn-outline-warning:focus,  .jhtml.btn-check:active + .btn-outline-warning:focus,  .jhtml.btn-outline-warning:active:focus,  .jhtml.btn-outline-warning.active:focus,  .jhtml.btn-outline-warning.dropdown-toggle.show:focus {\n  box-shadow: 0 0 0 0.25rem rgba(255, 193, 7, 0.5);\n}\n.jhtml .btn-outline-warning:disabled,  .jhtml .btn-outline-warning.disabled , .jhtml.btn-outline-warning:disabled,  .jhtml.btn-outline-warning.disabled {\n  color: #ffc107;\n  background-color: transparent;\n}\n.jhtml .btn-outline-danger , .jhtml.btn-outline-danger {\n  color: #dc3545;\n  border-color: #dc3545;\n}\n.jhtml .btn-outline-danger:hover , .jhtml.btn-outline-danger:hover {\n  color: #fff;\n  background-color: #dc3545;\n  border-color: #dc3545;\n}\n.jhtml .btn-check:focus + .btn-outline-danger,  .jhtml .btn-outline-danger:focus , .jhtml.btn-check:focus + .btn-outline-danger,  .jhtml.btn-outline-danger:focus {\n  box-shadow: 0 0 0 0.25rem rgba(220, 53, 69, 0.5);\n}\n.jhtml .btn-check:checked + .btn-outline-danger,  .jhtml .btn-check:active + .btn-outline-danger,  .jhtml .btn-outline-danger:active,  .jhtml .btn-outline-danger.active,  .jhtml .btn-outline-danger.dropdown-toggle.show , .jhtml.btn-check:checked + .btn-outline-danger,  .jhtml.btn-check:active + .btn-outline-danger,  .jhtml.btn-outline-danger:active,  .jhtml.btn-outline-danger.active,  .jhtml.btn-outline-danger.dropdown-toggle.show {\n  color: #fff;\n  background-color: #dc3545;\n  border-color: #dc3545;\n}\n.jhtml .btn-check:checked + .btn-outline-danger:focus,  .jhtml .btn-check:active + .btn-outline-danger:focus,  .jhtml .btn-outline-danger:active:focus,  .jhtml .btn-outline-danger.active:focus,  .jhtml .btn-outline-danger.dropdown-toggle.show:focus , .jhtml.btn-check:checked + .btn-outline-danger:focus,  .jhtml.btn-check:active + .btn-outline-danger:focus,  .jhtml.btn-outline-danger:active:focus,  .jhtml.btn-outline-danger.active:focus,  .jhtml.btn-outline-danger.dropdown-toggle.show:focus {\n  box-shadow: 0 0 0 0.25rem rgba(220, 53, 69, 0.5);\n}\n.jhtml .btn-outline-danger:disabled,  .jhtml .btn-outline-danger.disabled , .jhtml.btn-outline-danger:disabled,  .jhtml.btn-outline-danger.disabled {\n  color: #dc3545;\n  background-color: transparent;\n}\n.jhtml .btn-outline-light , .jhtml.btn-outline-light {\n  color: #f8f9fa;\n  border-color: #f8f9fa;\n}\n.jhtml .btn-outline-light:hover , .jhtml.btn-outline-light:hover {\n  color: #000;\n  background-color: #f8f9fa;\n  border-color: #f8f9fa;\n}\n.jhtml .btn-check:focus + .btn-outline-light,  .jhtml .btn-outline-light:focus , .jhtml.btn-check:focus + .btn-outline-light,  .jhtml.btn-outline-light:focus {\n  box-shadow: 0 0 0 0.25rem rgba(248, 249, 250, 0.5);\n}\n.jhtml .btn-check:checked + .btn-outline-light,  .jhtml .btn-check:active + .btn-outline-light,  .jhtml .btn-outline-light:active,  .jhtml .btn-outline-light.active,  .jhtml .btn-outline-light.dropdown-toggle.show , .jhtml.btn-check:checked + .btn-outline-light,  .jhtml.btn-check:active + .btn-outline-light,  .jhtml.btn-outline-light:active,  .jhtml.btn-outline-light.active,  .jhtml.btn-outline-light.dropdown-toggle.show {\n  color: #000;\n  background-color: #f8f9fa;\n  border-color: #f8f9fa;\n}\n.jhtml .btn-check:checked + .btn-outline-light:focus,  .jhtml .btn-check:active + .btn-outline-light:focus,  .jhtml .btn-outline-light:active:focus,  .jhtml .btn-outline-light.active:focus,  .jhtml .btn-outline-light.dropdown-toggle.show:focus , .jhtml.btn-check:checked + .btn-outline-light:focus,  .jhtml.btn-check:active + .btn-outline-light:focus,  .jhtml.btn-outline-light:active:focus,  .jhtml.btn-outline-light.active:focus,  .jhtml.btn-outline-light.dropdown-toggle.show:focus {\n  box-shadow: 0 0 0 0.25rem rgba(248, 249, 250, 0.5);\n}\n.jhtml .btn-outline-light:disabled,  .jhtml .btn-outline-light.disabled , .jhtml.btn-outline-light:disabled,  .jhtml.btn-outline-light.disabled {\n  color: #f8f9fa;\n  background-color: transparent;\n}\n.jhtml .btn-outline-dark , .jhtml.btn-outline-dark {\n  color: #212529;\n  border-color: #212529;\n}\n.jhtml .btn-outline-dark:hover , .jhtml.btn-outline-dark:hover {\n  color: #fff;\n  background-color: #212529;\n  border-color: #212529;\n}\n.jhtml .btn-check:focus + .btn-outline-dark,  .jhtml .btn-outline-dark:focus , .jhtml.btn-check:focus + .btn-outline-dark,  .jhtml.btn-outline-dark:focus {\n  box-shadow: 0 0 0 0.25rem rgba(33, 37, 41, 0.5);\n}\n.jhtml .btn-check:checked + .btn-outline-dark,  .jhtml .btn-check:active + .btn-outline-dark,  .jhtml .btn-outline-dark:active,  .jhtml .btn-outline-dark.active,  .jhtml .btn-outline-dark.dropdown-toggle.show , .jhtml.btn-check:checked + .btn-outline-dark,  .jhtml.btn-check:active + .btn-outline-dark,  .jhtml.btn-outline-dark:active,  .jhtml.btn-outline-dark.active,  .jhtml.btn-outline-dark.dropdown-toggle.show {\n  color: #fff;\n  background-color: #212529;\n  border-color: #212529;\n}\n.jhtml .btn-check:checked + .btn-outline-dark:focus,  .jhtml .btn-check:active + .btn-outline-dark:focus,  .jhtml .btn-outline-dark:active:focus,  .jhtml .btn-outline-dark.active:focus,  .jhtml .btn-outline-dark.dropdown-toggle.show:focus , .jhtml.btn-check:checked + .btn-outline-dark:focus,  .jhtml.btn-check:active + .btn-outline-dark:focus,  .jhtml.btn-outline-dark:active:focus,  .jhtml.btn-outline-dark.active:focus,  .jhtml.btn-outline-dark.dropdown-toggle.show:focus {\n  box-shadow: 0 0 0 0.25rem rgba(33, 37, 41, 0.5);\n}\n.jhtml .btn-outline-dark:disabled,  .jhtml .btn-outline-dark.disabled , .jhtml.btn-outline-dark:disabled,  .jhtml.btn-outline-dark.disabled {\n  color: #212529;\n  background-color: transparent;\n}\n.jhtml .btn-link , .jhtml.btn-link {\n  font-weight: 400;\n  color: #0d6efd;\n  text-decoration: underline;\n}\n.jhtml .btn-link:hover , .jhtml.btn-link:hover {\n  color: #0a58ca;\n}\n.jhtml .btn-link:disabled,  .jhtml .btn-link.disabled , .jhtml.btn-link:disabled,  .jhtml.btn-link.disabled {\n  color: #6c757d;\n}\n.jhtml .btn-lg,  .jhtml .btn-group-lg > .btn , .jhtml.btn-lg,  .jhtml.btn-group-lg > .btn {\n  padding: 0.5rem 1rem;\n  font-size: 1.25rem;\n  border-radius: 0.3rem;\n}\n.jhtml .btn-sm,  .jhtml .btn-group-sm > .btn , .jhtml.btn-sm,  .jhtml.btn-group-sm > .btn {\n  padding: 0.25rem 0.5rem;\n  font-size: 0.875rem;\n  border-radius: 0.2rem;\n}\n.jhtml .fade , .jhtml.fade {\n  transition: opacity 0.15s linear;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .fade , .jhtml.fade {\n    transition: none;\n  }\n}\n.jhtml .fade:not(.show) , .jhtml.fade:not(.show) {\n  opacity: 0;\n}\n.jhtml .collapse:not(.show) , .jhtml.collapse:not(.show) {\n  display: none;\n}\n.jhtml .collapsing , .jhtml.collapsing {\n  height: 0;\n  overflow: hidden;\n  transition: height 0.35s ease;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .collapsing , .jhtml.collapsing {\n    transition: none;\n  }\n}\n.jhtml .collapsing.collapse-horizontal , .jhtml.collapsing.collapse-horizontal {\n  width: 0;\n  height: auto;\n  transition: width 0.35s ease;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .collapsing.collapse-horizontal , .jhtml.collapsing.collapse-horizontal {\n    transition: none;\n  }\n}\n.jhtml .dropup, \n.jhtml .dropend, \n.jhtml .dropdown, \n.jhtml .dropstart , .jhtml.dropup, \n.jhtml.dropend, \n.jhtml.dropdown, \n.jhtml.dropstart {\n  position: relative;\n}\n.jhtml .dropdown-toggle , .jhtml.dropdown-toggle {\n  white-space: nowrap;\n}\n.jhtml .dropdown-toggle::after , .jhtml.dropdown-toggle::after {\n  display: inline-block;\n  margin-left: 0.255em;\n  vertical-align: 0.255em;\n  content: \"\";\n  border-top: 0.3em solid;\n  border-right: 0.3em solid transparent;\n  border-bottom: 0;\n  border-left: 0.3em solid transparent;\n}\n.jhtml .dropdown-toggle:empty::after , .jhtml.dropdown-toggle:empty::after {\n  margin-left: 0;\n}\n.jhtml .dropdown-menu , .jhtml.dropdown-menu {\n  position: absolute;\n  z-index: 1000;\n  display: none;\n  min-width: 10rem;\n  padding: 0.5rem 0;\n  margin: 0;\n  font-size: 1rem;\n  color: #212529;\n  text-align: left;\n  list-style: none;\n  background-color: #fff;\n  background-clip: padding-box;\n  border: 1px solid rgba(0, 0, 0, 0.15);\n  border-radius: 0.25rem;\n}\n.jhtml .dropdown-menu[data-bs-popper] , .jhtml.dropdown-menu[data-bs-popper] {\n  top: 100%;\n  left: 0;\n  margin-top: 0.125rem;\n}\n.jhtml .dropdown-menu-start , .jhtml.dropdown-menu-start {\n  --bs-position: start;\n}\n.jhtml .dropdown-menu-start[data-bs-popper] , .jhtml.dropdown-menu-start[data-bs-popper] {\n  right: auto;\n  left: 0;\n}\n.jhtml .dropdown-menu-end , .jhtml.dropdown-menu-end {\n  --bs-position: end;\n}\n.jhtml .dropdown-menu-end[data-bs-popper] , .jhtml.dropdown-menu-end[data-bs-popper] {\n  right: 0;\n  left: auto;\n}\n@media (min-width: 576px) {\n  .jhtml .dropdown-menu-sm-start , .jhtml.dropdown-menu-sm-start {\n    --bs-position: start;\n  }\n  .jhtml .dropdown-menu-sm-start[data-bs-popper] , .jhtml.dropdown-menu-sm-start[data-bs-popper] {\n    right: auto;\n    left: 0;\n  }\n  .jhtml .dropdown-menu-sm-end , .jhtml.dropdown-menu-sm-end {\n    --bs-position: end;\n  }\n  .jhtml .dropdown-menu-sm-end[data-bs-popper] , .jhtml.dropdown-menu-sm-end[data-bs-popper] {\n    right: 0;\n    left: auto;\n  }\n}\n@media (min-width: 768px) {\n  .jhtml .dropdown-menu-md-start , .jhtml.dropdown-menu-md-start {\n    --bs-position: start;\n  }\n  .jhtml .dropdown-menu-md-start[data-bs-popper] , .jhtml.dropdown-menu-md-start[data-bs-popper] {\n    right: auto;\n    left: 0;\n  }\n  .jhtml .dropdown-menu-md-end , .jhtml.dropdown-menu-md-end {\n    --bs-position: end;\n  }\n  .jhtml .dropdown-menu-md-end[data-bs-popper] , .jhtml.dropdown-menu-md-end[data-bs-popper] {\n    right: 0;\n    left: auto;\n  }\n}\n@media (min-width: 992px) {\n  .jhtml .dropdown-menu-lg-start , .jhtml.dropdown-menu-lg-start {\n    --bs-position: start;\n  }\n  .jhtml .dropdown-menu-lg-start[data-bs-popper] , .jhtml.dropdown-menu-lg-start[data-bs-popper] {\n    right: auto;\n    left: 0;\n  }\n  .jhtml .dropdown-menu-lg-end , .jhtml.dropdown-menu-lg-end {\n    --bs-position: end;\n  }\n  .jhtml .dropdown-menu-lg-end[data-bs-popper] , .jhtml.dropdown-menu-lg-end[data-bs-popper] {\n    right: 0;\n    left: auto;\n  }\n}\n@media (min-width: 1200px) {\n  .jhtml .dropdown-menu-xl-start , .jhtml.dropdown-menu-xl-start {\n    --bs-position: start;\n  }\n  .jhtml .dropdown-menu-xl-start[data-bs-popper] , .jhtml.dropdown-menu-xl-start[data-bs-popper] {\n    right: auto;\n    left: 0;\n  }\n  .jhtml .dropdown-menu-xl-end , .jhtml.dropdown-menu-xl-end {\n    --bs-position: end;\n  }\n  .jhtml .dropdown-menu-xl-end[data-bs-popper] , .jhtml.dropdown-menu-xl-end[data-bs-popper] {\n    right: 0;\n    left: auto;\n  }\n}\n@media (min-width: 1400px) {\n  .jhtml .dropdown-menu-xxl-start , .jhtml.dropdown-menu-xxl-start {\n    --bs-position: start;\n  }\n  .jhtml .dropdown-menu-xxl-start[data-bs-popper] , .jhtml.dropdown-menu-xxl-start[data-bs-popper] {\n    right: auto;\n    left: 0;\n  }\n  .jhtml .dropdown-menu-xxl-end , .jhtml.dropdown-menu-xxl-end {\n    --bs-position: end;\n  }\n  .jhtml .dropdown-menu-xxl-end[data-bs-popper] , .jhtml.dropdown-menu-xxl-end[data-bs-popper] {\n    right: 0;\n    left: auto;\n  }\n}\n.jhtml .dropup .dropdown-menu[data-bs-popper] , .jhtml.dropup .dropdown-menu[data-bs-popper] {\n  top: auto;\n  bottom: 100%;\n  margin-top: 0;\n  margin-bottom: 0.125rem;\n}\n.jhtml .dropup .dropdown-toggle::after , .jhtml.dropup .dropdown-toggle::after {\n  display: inline-block;\n  margin-left: 0.255em;\n  vertical-align: 0.255em;\n  content: \"\";\n  border-top: 0;\n  border-right: 0.3em solid transparent;\n  border-bottom: 0.3em solid;\n  border-left: 0.3em solid transparent;\n}\n.jhtml .dropup .dropdown-toggle:empty::after , .jhtml.dropup .dropdown-toggle:empty::after {\n  margin-left: 0;\n}\n.jhtml .dropend .dropdown-menu[data-bs-popper] , .jhtml.dropend .dropdown-menu[data-bs-popper] {\n  top: 0;\n  right: auto;\n  left: 100%;\n  margin-top: 0;\n  margin-left: 0.125rem;\n}\n.jhtml .dropend .dropdown-toggle::after , .jhtml.dropend .dropdown-toggle::after {\n  display: inline-block;\n  margin-left: 0.255em;\n  vertical-align: 0.255em;\n  content: \"\";\n  border-top: 0.3em solid transparent;\n  border-right: 0;\n  border-bottom: 0.3em solid transparent;\n  border-left: 0.3em solid;\n}\n.jhtml .dropend .dropdown-toggle:empty::after , .jhtml.dropend .dropdown-toggle:empty::after {\n  margin-left: 0;\n}\n.jhtml .dropend .dropdown-toggle::after , .jhtml.dropend .dropdown-toggle::after {\n  vertical-align: 0;\n}\n.jhtml .dropstart .dropdown-menu[data-bs-popper] , .jhtml.dropstart .dropdown-menu[data-bs-popper] {\n  top: 0;\n  right: 100%;\n  left: auto;\n  margin-top: 0;\n  margin-right: 0.125rem;\n}\n.jhtml .dropstart .dropdown-toggle::after , .jhtml.dropstart .dropdown-toggle::after {\n  display: inline-block;\n  margin-left: 0.255em;\n  vertical-align: 0.255em;\n  content: \"\";\n}\n.jhtml .dropstart .dropdown-toggle::after , .jhtml.dropstart .dropdown-toggle::after {\n  display: none;\n}\n.jhtml .dropstart .dropdown-toggle::before , .jhtml.dropstart .dropdown-toggle::before {\n  display: inline-block;\n  margin-right: 0.255em;\n  vertical-align: 0.255em;\n  content: \"\";\n  border-top: 0.3em solid transparent;\n  border-right: 0.3em solid;\n  border-bottom: 0.3em solid transparent;\n}\n.jhtml .dropstart .dropdown-toggle:empty::after , .jhtml.dropstart .dropdown-toggle:empty::after {\n  margin-left: 0;\n}\n.jhtml .dropstart .dropdown-toggle::before , .jhtml.dropstart .dropdown-toggle::before {\n  vertical-align: 0;\n}\n.jhtml .dropdown-divider , .jhtml.dropdown-divider {\n  height: 0;\n  margin: 0.5rem 0;\n  overflow: hidden;\n  border-top: 1px solid rgba(0, 0, 0, 0.15);\n}\n.jhtml .dropdown-item , .jhtml.dropdown-item {\n  display: block;\n  width: 100%;\n  padding: 0.25rem 1rem;\n  clear: both;\n  font-weight: 400;\n  color: #212529;\n  text-align: inherit;\n  text-decoration: none;\n  white-space: nowrap;\n  background-color: transparent;\n  border: 0;\n}\n.jhtml .dropdown-item:hover,  .jhtml .dropdown-item:focus , .jhtml.dropdown-item:hover,  .jhtml.dropdown-item:focus {\n  color: #1e2125;\n  background-color: #e9ecef;\n}\n.jhtml .dropdown-item.active,  .jhtml .dropdown-item:active , .jhtml.dropdown-item.active,  .jhtml.dropdown-item:active {\n  color: #fff;\n  text-decoration: none;\n  background-color: #0d6efd;\n}\n.jhtml .dropdown-item.disabled,  .jhtml .dropdown-item:disabled , .jhtml.dropdown-item.disabled,  .jhtml.dropdown-item:disabled {\n  color: #adb5bd;\n  pointer-events: none;\n  background-color: transparent;\n}\n.jhtml .dropdown-menu.show , .jhtml.dropdown-menu.show {\n  display: block;\n}\n.jhtml .dropdown-header , .jhtml.dropdown-header {\n  display: block;\n  padding: 0.5rem 1rem;\n  margin-bottom: 0;\n  font-size: 0.875rem;\n  color: #6c757d;\n  white-space: nowrap;\n}\n.jhtml .dropdown-item-text , .jhtml.dropdown-item-text {\n  display: block;\n  padding: 0.25rem 1rem;\n  color: #212529;\n}\n.jhtml .dropdown-menu-dark , .jhtml.dropdown-menu-dark {\n  color: #dee2e6;\n  background-color: #343a40;\n  border-color: rgba(0, 0, 0, 0.15);\n}\n.jhtml .dropdown-menu-dark .dropdown-item , .jhtml.dropdown-menu-dark .dropdown-item {\n  color: #dee2e6;\n}\n.jhtml .dropdown-menu-dark .dropdown-item:hover,  .jhtml .dropdown-menu-dark .dropdown-item:focus , .jhtml.dropdown-menu-dark .dropdown-item:hover,  .jhtml.dropdown-menu-dark .dropdown-item:focus {\n  color: #fff;\n  background-color: rgba(255, 255, 255, 0.15);\n}\n.jhtml .dropdown-menu-dark .dropdown-item.active,  .jhtml .dropdown-menu-dark .dropdown-item:active , .jhtml.dropdown-menu-dark .dropdown-item.active,  .jhtml.dropdown-menu-dark .dropdown-item:active {\n  color: #fff;\n  background-color: #0d6efd;\n}\n.jhtml .dropdown-menu-dark .dropdown-item.disabled,  .jhtml .dropdown-menu-dark .dropdown-item:disabled , .jhtml.dropdown-menu-dark .dropdown-item.disabled,  .jhtml.dropdown-menu-dark .dropdown-item:disabled {\n  color: #adb5bd;\n}\n.jhtml .dropdown-menu-dark .dropdown-divider , .jhtml.dropdown-menu-dark .dropdown-divider {\n  border-color: rgba(0, 0, 0, 0.15);\n}\n.jhtml .dropdown-menu-dark .dropdown-item-text , .jhtml.dropdown-menu-dark .dropdown-item-text {\n  color: #dee2e6;\n}\n.jhtml .dropdown-menu-dark .dropdown-header , .jhtml.dropdown-menu-dark .dropdown-header {\n  color: #adb5bd;\n}\n.jhtml .btn-group, \n.jhtml .btn-group-vertical , .jhtml.btn-group, \n.jhtml.btn-group-vertical {\n  position: relative;\n  display: inline-flex;\n  vertical-align: middle;\n}\n.jhtml .btn-group > .btn, \n.jhtml .btn-group-vertical > .btn , .jhtml.btn-group > .btn, \n.jhtml.btn-group-vertical > .btn {\n  position: relative;\n  flex: 1 1 auto;\n}\n.jhtml .btn-group > .btn-check:checked + .btn, \n.jhtml .btn-group > .btn-check:focus + .btn, \n.jhtml .btn-group > .btn:hover, \n.jhtml .btn-group > .btn:focus, \n.jhtml .btn-group > .btn:active, \n.jhtml .btn-group > .btn.active, \n.jhtml .btn-group-vertical > .btn-check:checked + .btn, \n.jhtml .btn-group-vertical > .btn-check:focus + .btn, \n.jhtml .btn-group-vertical > .btn:hover, \n.jhtml .btn-group-vertical > .btn:focus, \n.jhtml .btn-group-vertical > .btn:active, \n.jhtml .btn-group-vertical > .btn.active , .jhtml.btn-group > .btn-check:checked + .btn, \n.jhtml.btn-group > .btn-check:focus + .btn, \n.jhtml.btn-group > .btn:hover, \n.jhtml.btn-group > .btn:focus, \n.jhtml.btn-group > .btn:active, \n.jhtml.btn-group > .btn.active, \n.jhtml.btn-group-vertical > .btn-check:checked + .btn, \n.jhtml.btn-group-vertical > .btn-check:focus + .btn, \n.jhtml.btn-group-vertical > .btn:hover, \n.jhtml.btn-group-vertical > .btn:focus, \n.jhtml.btn-group-vertical > .btn:active, \n.jhtml.btn-group-vertical > .btn.active {\n  z-index: 1;\n}\n.jhtml .btn-toolbar , .jhtml.btn-toolbar {\n  display: flex;\n  flex-wrap: wrap;\n  justify-content: flex-start;\n}\n.jhtml .btn-toolbar .input-group , .jhtml.btn-toolbar .input-group {\n  width: auto;\n}\n.jhtml .btn-group > .btn:not(:first-child), \n.jhtml .btn-group > .btn-group:not(:first-child) , .jhtml.btn-group > .btn:not(:first-child), \n.jhtml.btn-group > .btn-group:not(:first-child) {\n  margin-left: -1px;\n}\n.jhtml .btn-group > .btn:not(:last-child):not(.dropdown-toggle), \n.jhtml .btn-group > .btn-group:not(:last-child) > .btn , .jhtml.btn-group > .btn:not(:last-child):not(.dropdown-toggle), \n.jhtml.btn-group > .btn-group:not(:last-child) > .btn {\n  border-top-right-radius: 0;\n  border-bottom-right-radius: 0;\n}\n.jhtml .btn-group > .btn:nth-child(n+3), \n.jhtml .btn-group > :not(.btn-check) + .btn, \n.jhtml .btn-group > .btn-group:not(:first-child) > .btn , .jhtml.btn-group > .btn:nth-child(n+3), \n.jhtml.btn-group > :not(.btn-check) + .btn, \n.jhtml.btn-group > .btn-group:not(:first-child) > .btn {\n  border-top-left-radius: 0;\n  border-bottom-left-radius: 0;\n}\n.jhtml .dropdown-toggle-split , .jhtml.dropdown-toggle-split {\n  padding-right: 0.5625rem;\n  padding-left: 0.5625rem;\n}\n.jhtml .dropdown-toggle-split::after,  .jhtml .dropup .dropdown-toggle-split::after,  .jhtml .dropend .dropdown-toggle-split::after , .jhtml.dropdown-toggle-split::after,  .jhtml.dropup .dropdown-toggle-split::after,  .jhtml.dropend .dropdown-toggle-split::after {\n  margin-left: 0;\n}\n.jhtml .dropstart .dropdown-toggle-split::before , .jhtml.dropstart .dropdown-toggle-split::before {\n  margin-right: 0;\n}\n.jhtml .btn-sm + .dropdown-toggle-split,  .jhtml .btn-group-sm > .btn + .dropdown-toggle-split , .jhtml.btn-sm + .dropdown-toggle-split,  .jhtml.btn-group-sm > .btn + .dropdown-toggle-split {\n  padding-right: 0.375rem;\n  padding-left: 0.375rem;\n}\n.jhtml .btn-lg + .dropdown-toggle-split,  .jhtml .btn-group-lg > .btn + .dropdown-toggle-split , .jhtml.btn-lg + .dropdown-toggle-split,  .jhtml.btn-group-lg > .btn + .dropdown-toggle-split {\n  padding-right: 0.75rem;\n  padding-left: 0.75rem;\n}\n.jhtml .btn-group-vertical , .jhtml.btn-group-vertical {\n  flex-direction: column;\n  align-items: flex-start;\n  justify-content: center;\n}\n.jhtml .btn-group-vertical > .btn, \n.jhtml .btn-group-vertical > .btn-group , .jhtml.btn-group-vertical > .btn, \n.jhtml.btn-group-vertical > .btn-group {\n  width: 100%;\n}\n.jhtml .btn-group-vertical > .btn:not(:first-child), \n.jhtml .btn-group-vertical > .btn-group:not(:first-child) , .jhtml.btn-group-vertical > .btn:not(:first-child), \n.jhtml.btn-group-vertical > .btn-group:not(:first-child) {\n  margin-top: -1px;\n}\n.jhtml .btn-group-vertical > .btn:not(:last-child):not(.dropdown-toggle), \n.jhtml .btn-group-vertical > .btn-group:not(:last-child) > .btn , .jhtml.btn-group-vertical > .btn:not(:last-child):not(.dropdown-toggle), \n.jhtml.btn-group-vertical > .btn-group:not(:last-child) > .btn {\n  border-bottom-right-radius: 0;\n  border-bottom-left-radius: 0;\n}\n.jhtml .btn-group-vertical > .btn ~ .btn, \n.jhtml .btn-group-vertical > .btn-group:not(:first-child) > .btn , .jhtml.btn-group-vertical > .btn ~ .btn, \n.jhtml.btn-group-vertical > .btn-group:not(:first-child) > .btn {\n  border-top-left-radius: 0;\n  border-top-right-radius: 0;\n}\n.jhtml .nav , .jhtml.nav {\n  display: flex;\n  flex-wrap: wrap;\n  padding-left: 0;\n  margin-bottom: 0;\n  list-style: none;\n}\n.jhtml .nav-link , .jhtml.nav-link {\n  display: block;\n  padding: 0.5rem 1rem;\n  color: #0d6efd;\n  text-decoration: none;\n  transition: color 0.15s ease-in-out, background-color 0.15s ease-in-out, border-color 0.15s ease-in-out;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .nav-link , .jhtml.nav-link {\n    transition: none;\n  }\n}\n.jhtml .nav-link:hover,  .jhtml .nav-link:focus , .jhtml.nav-link:hover,  .jhtml.nav-link:focus {\n  color: #0a58ca;\n}\n.jhtml .nav-link.disabled , .jhtml.nav-link.disabled {\n  color: #6c757d;\n  pointer-events: none;\n  cursor: default;\n}\n.jhtml .nav-tabs , .jhtml.nav-tabs {\n  border-bottom: 1px solid #dee2e6;\n}\n.jhtml .nav-tabs .nav-link , .jhtml.nav-tabs .nav-link {\n  margin-bottom: -1px;\n  background: none;\n  border: 1px solid transparent;\n  border-top-left-radius: 0.25rem;\n  border-top-right-radius: 0.25rem;\n}\n.jhtml .nav-tabs .nav-link:hover,  .jhtml .nav-tabs .nav-link:focus , .jhtml.nav-tabs .nav-link:hover,  .jhtml.nav-tabs .nav-link:focus {\n  border-color: #e9ecef #e9ecef #dee2e6;\n  isolation: isolate;\n}\n.jhtml .nav-tabs .nav-link.disabled , .jhtml.nav-tabs .nav-link.disabled {\n  color: #6c757d;\n  background-color: transparent;\n  border-color: transparent;\n}\n.jhtml .nav-tabs .nav-link.active, \n.jhtml .nav-tabs .nav-item.show .nav-link , .jhtml.nav-tabs .nav-link.active, \n.jhtml.nav-tabs .nav-item.show .nav-link {\n  color: #495057;\n  background-color: #fff;\n  border-color: #dee2e6 #dee2e6 #fff;\n}\n.jhtml .nav-tabs .dropdown-menu , .jhtml.nav-tabs .dropdown-menu {\n  margin-top: -1px;\n  border-top-left-radius: 0;\n  border-top-right-radius: 0;\n}\n.jhtml .nav-pills .nav-link , .jhtml.nav-pills .nav-link {\n  background: none;\n  border: 0;\n  border-radius: 0.25rem;\n}\n.jhtml .nav-pills .nav-link.active, \n.jhtml .nav-pills .show > .nav-link , .jhtml.nav-pills .nav-link.active, \n.jhtml.nav-pills .show > .nav-link {\n  color: #fff;\n  background-color: #0d6efd;\n}\n.jhtml .nav-fill > .nav-link, \n.jhtml .nav-fill .nav-item , .jhtml.nav-fill > .nav-link, \n.jhtml.nav-fill .nav-item {\n  flex: 1 1 auto;\n  text-align: center;\n}\n.jhtml .nav-justified > .nav-link, \n.jhtml .nav-justified .nav-item , .jhtml.nav-justified > .nav-link, \n.jhtml.nav-justified .nav-item {\n  flex-basis: 0;\n  flex-grow: 1;\n  text-align: center;\n}\n.jhtml .nav-fill .nav-item .nav-link, \n.jhtml .nav-justified .nav-item .nav-link , .jhtml.nav-fill .nav-item .nav-link, \n.jhtml.nav-justified .nav-item .nav-link {\n  width: 100%;\n}\n.jhtml .tab-content > .tab-pane , .jhtml.tab-content > .tab-pane {\n  display: none;\n}\n.jhtml .tab-content > .active , .jhtml.tab-content > .active {\n  display: block;\n}\n.jhtml .navbar , .jhtml.navbar {\n  position: relative;\n  display: flex;\n  flex-wrap: wrap;\n  align-items: center;\n  justify-content: space-between;\n  padding-top: 0.5rem;\n  padding-bottom: 0.5rem;\n}\n.jhtml .navbar > .container, \n.jhtml .navbar > .container-fluid, \n.jhtml .navbar > .container-sm, \n.jhtml .navbar > .container-md, \n.jhtml .navbar > .container-lg, \n.jhtml .navbar > .container-xl, \n.jhtml .navbar > .container-xxl , .jhtml.navbar > .container, \n.jhtml.navbar > .container-fluid, \n.jhtml.navbar > .container-sm, \n.jhtml.navbar > .container-md, \n.jhtml.navbar > .container-lg, \n.jhtml.navbar > .container-xl, \n.jhtml.navbar > .container-xxl {\n  display: flex;\n  flex-wrap: inherit;\n  align-items: center;\n  justify-content: space-between;\n}\n.jhtml .navbar-brand , .jhtml.navbar-brand {\n  padding-top: 0.3125rem;\n  padding-bottom: 0.3125rem;\n  margin-right: 1rem;\n  font-size: 1.25rem;\n  text-decoration: none;\n  white-space: nowrap;\n}\n.jhtml .navbar-nav , .jhtml.navbar-nav {\n  display: flex;\n  flex-direction: column;\n  padding-left: 0;\n  margin-bottom: 0;\n  list-style: none;\n}\n.jhtml .navbar-nav .nav-link , .jhtml.navbar-nav .nav-link {\n  padding-right: 0;\n  padding-left: 0;\n}\n.jhtml .navbar-nav .dropdown-menu , .jhtml.navbar-nav .dropdown-menu {\n  position: static;\n}\n.jhtml .navbar-text , .jhtml.navbar-text {\n  padding-top: 0.5rem;\n  padding-bottom: 0.5rem;\n}\n.jhtml .navbar-collapse , .jhtml.navbar-collapse {\n  flex-basis: 100%;\n  flex-grow: 1;\n  align-items: center;\n}\n.jhtml .navbar-toggler , .jhtml.navbar-toggler {\n  padding: 0.25rem 0.75rem;\n  font-size: 1.25rem;\n  line-height: 1;\n  background-color: transparent;\n  border: 1px solid transparent;\n  border-radius: 0.25rem;\n  transition: box-shadow 0.15s ease-in-out;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .navbar-toggler , .jhtml.navbar-toggler {\n    transition: none;\n  }\n}\n.jhtml .navbar-toggler:hover , .jhtml.navbar-toggler:hover {\n  text-decoration: none;\n}\n.jhtml .navbar-toggler:focus , .jhtml.navbar-toggler:focus {\n  text-decoration: none;\n  outline: 0;\n  box-shadow: 0 0 0 0.25rem;\n}\n.jhtml .navbar-toggler-icon , .jhtml.navbar-toggler-icon {\n  display: inline-block;\n  width: 1.5em;\n  height: 1.5em;\n  vertical-align: middle;\n  background-repeat: no-repeat;\n  background-position: center;\n  background-size: 100%;\n}\n.jhtml .navbar-nav-scroll , .jhtml.navbar-nav-scroll {\n  max-height: var(--bs-scroll-height, 75vh);\n  overflow-y: auto;\n}\n@media (min-width: 576px) {\n  .jhtml .navbar-expand-sm , .jhtml.navbar-expand-sm {\n    flex-wrap: nowrap;\n    justify-content: flex-start;\n  }\n  .jhtml .navbar-expand-sm .navbar-nav , .jhtml.navbar-expand-sm .navbar-nav {\n    flex-direction: row;\n  }\n  .jhtml .navbar-expand-sm .navbar-nav .dropdown-menu , .jhtml.navbar-expand-sm .navbar-nav .dropdown-menu {\n    position: absolute;\n  }\n  .jhtml .navbar-expand-sm .navbar-nav .nav-link , .jhtml.navbar-expand-sm .navbar-nav .nav-link {\n    padding-right: 0.5rem;\n    padding-left: 0.5rem;\n  }\n  .jhtml .navbar-expand-sm .navbar-nav-scroll , .jhtml.navbar-expand-sm .navbar-nav-scroll {\n    overflow: visible;\n  }\n  .jhtml .navbar-expand-sm .navbar-collapse , .jhtml.navbar-expand-sm .navbar-collapse {\n    display: flex !important;\n    flex-basis: auto;\n  }\n  .jhtml .navbar-expand-sm .navbar-toggler , .jhtml.navbar-expand-sm .navbar-toggler {\n    display: none;\n  }\n  .jhtml .navbar-expand-sm .offcanvas-header , .jhtml.navbar-expand-sm .offcanvas-header {\n    display: none;\n  }\n  .jhtml .navbar-expand-sm .offcanvas , .jhtml.navbar-expand-sm .offcanvas {\n    position: inherit;\n    bottom: 0;\n    z-index: 1000;\n    flex-grow: 1;\n    visibility: visible !important;\n    background-color: transparent;\n    border-right: 0;\n    border-left: 0;\n    transition: none;\n    transform: none;\n  }\n  .jhtml .navbar-expand-sm .offcanvas-top, \n.jhtml .navbar-expand-sm .offcanvas-bottom , .jhtml.navbar-expand-sm .offcanvas-top, \n.jhtml.navbar-expand-sm .offcanvas-bottom {\n    height: auto;\n    border-top: 0;\n    border-bottom: 0;\n  }\n  .jhtml .navbar-expand-sm .offcanvas-body , .jhtml.navbar-expand-sm .offcanvas-body {\n    display: flex;\n    flex-grow: 0;\n    padding: 0;\n    overflow-y: visible;\n  }\n}\n@media (min-width: 768px) {\n  .jhtml .navbar-expand-md , .jhtml.navbar-expand-md {\n    flex-wrap: nowrap;\n    justify-content: flex-start;\n  }\n  .jhtml .navbar-expand-md .navbar-nav , .jhtml.navbar-expand-md .navbar-nav {\n    flex-direction: row;\n  }\n  .jhtml .navbar-expand-md .navbar-nav .dropdown-menu , .jhtml.navbar-expand-md .navbar-nav .dropdown-menu {\n    position: absolute;\n  }\n  .jhtml .navbar-expand-md .navbar-nav .nav-link , .jhtml.navbar-expand-md .navbar-nav .nav-link {\n    padding-right: 0.5rem;\n    padding-left: 0.5rem;\n  }\n  .jhtml .navbar-expand-md .navbar-nav-scroll , .jhtml.navbar-expand-md .navbar-nav-scroll {\n    overflow: visible;\n  }\n  .jhtml .navbar-expand-md .navbar-collapse , .jhtml.navbar-expand-md .navbar-collapse {\n    display: flex !important;\n    flex-basis: auto;\n  }\n  .jhtml .navbar-expand-md .navbar-toggler , .jhtml.navbar-expand-md .navbar-toggler {\n    display: none;\n  }\n  .jhtml .navbar-expand-md .offcanvas-header , .jhtml.navbar-expand-md .offcanvas-header {\n    display: none;\n  }\n  .jhtml .navbar-expand-md .offcanvas , .jhtml.navbar-expand-md .offcanvas {\n    position: inherit;\n    bottom: 0;\n    z-index: 1000;\n    flex-grow: 1;\n    visibility: visible !important;\n    background-color: transparent;\n    border-right: 0;\n    border-left: 0;\n    transition: none;\n    transform: none;\n  }\n  .jhtml .navbar-expand-md .offcanvas-top, \n.jhtml .navbar-expand-md .offcanvas-bottom , .jhtml.navbar-expand-md .offcanvas-top, \n.jhtml.navbar-expand-md .offcanvas-bottom {\n    height: auto;\n    border-top: 0;\n    border-bottom: 0;\n  }\n  .jhtml .navbar-expand-md .offcanvas-body , .jhtml.navbar-expand-md .offcanvas-body {\n    display: flex;\n    flex-grow: 0;\n    padding: 0;\n    overflow-y: visible;\n  }\n}\n@media (min-width: 992px) {\n  .jhtml .navbar-expand-lg , .jhtml.navbar-expand-lg {\n    flex-wrap: nowrap;\n    justify-content: flex-start;\n  }\n  .jhtml .navbar-expand-lg .navbar-nav , .jhtml.navbar-expand-lg .navbar-nav {\n    flex-direction: row;\n  }\n  .jhtml .navbar-expand-lg .navbar-nav .dropdown-menu , .jhtml.navbar-expand-lg .navbar-nav .dropdown-menu {\n    position: absolute;\n  }\n  .jhtml .navbar-expand-lg .navbar-nav .nav-link , .jhtml.navbar-expand-lg .navbar-nav .nav-link {\n    padding-right: 0.5rem;\n    padding-left: 0.5rem;\n  }\n  .jhtml .navbar-expand-lg .navbar-nav-scroll , .jhtml.navbar-expand-lg .navbar-nav-scroll {\n    overflow: visible;\n  }\n  .jhtml .navbar-expand-lg .navbar-collapse , .jhtml.navbar-expand-lg .navbar-collapse {\n    display: flex !important;\n    flex-basis: auto;\n  }\n  .jhtml .navbar-expand-lg .navbar-toggler , .jhtml.navbar-expand-lg .navbar-toggler {\n    display: none;\n  }\n  .jhtml .navbar-expand-lg .offcanvas-header , .jhtml.navbar-expand-lg .offcanvas-header {\n    display: none;\n  }\n  .jhtml .navbar-expand-lg .offcanvas , .jhtml.navbar-expand-lg .offcanvas {\n    position: inherit;\n    bottom: 0;\n    z-index: 1000;\n    flex-grow: 1;\n    visibility: visible !important;\n    background-color: transparent;\n    border-right: 0;\n    border-left: 0;\n    transition: none;\n    transform: none;\n  }\n  .jhtml .navbar-expand-lg .offcanvas-top, \n.jhtml .navbar-expand-lg .offcanvas-bottom , .jhtml.navbar-expand-lg .offcanvas-top, \n.jhtml.navbar-expand-lg .offcanvas-bottom {\n    height: auto;\n    border-top: 0;\n    border-bottom: 0;\n  }\n  .jhtml .navbar-expand-lg .offcanvas-body , .jhtml.navbar-expand-lg .offcanvas-body {\n    display: flex;\n    flex-grow: 0;\n    padding: 0;\n    overflow-y: visible;\n  }\n}\n@media (min-width: 1200px) {\n  .jhtml .navbar-expand-xl , .jhtml.navbar-expand-xl {\n    flex-wrap: nowrap;\n    justify-content: flex-start;\n  }\n  .jhtml .navbar-expand-xl .navbar-nav , .jhtml.navbar-expand-xl .navbar-nav {\n    flex-direction: row;\n  }\n  .jhtml .navbar-expand-xl .navbar-nav .dropdown-menu , .jhtml.navbar-expand-xl .navbar-nav .dropdown-menu {\n    position: absolute;\n  }\n  .jhtml .navbar-expand-xl .navbar-nav .nav-link , .jhtml.navbar-expand-xl .navbar-nav .nav-link {\n    padding-right: 0.5rem;\n    padding-left: 0.5rem;\n  }\n  .jhtml .navbar-expand-xl .navbar-nav-scroll , .jhtml.navbar-expand-xl .navbar-nav-scroll {\n    overflow: visible;\n  }\n  .jhtml .navbar-expand-xl .navbar-collapse , .jhtml.navbar-expand-xl .navbar-collapse {\n    display: flex !important;\n    flex-basis: auto;\n  }\n  .jhtml .navbar-expand-xl .navbar-toggler , .jhtml.navbar-expand-xl .navbar-toggler {\n    display: none;\n  }\n  .jhtml .navbar-expand-xl .offcanvas-header , .jhtml.navbar-expand-xl .offcanvas-header {\n    display: none;\n  }\n  .jhtml .navbar-expand-xl .offcanvas , .jhtml.navbar-expand-xl .offcanvas {\n    position: inherit;\n    bottom: 0;\n    z-index: 1000;\n    flex-grow: 1;\n    visibility: visible !important;\n    background-color: transparent;\n    border-right: 0;\n    border-left: 0;\n    transition: none;\n    transform: none;\n  }\n  .jhtml .navbar-expand-xl .offcanvas-top, \n.jhtml .navbar-expand-xl .offcanvas-bottom , .jhtml.navbar-expand-xl .offcanvas-top, \n.jhtml.navbar-expand-xl .offcanvas-bottom {\n    height: auto;\n    border-top: 0;\n    border-bottom: 0;\n  }\n  .jhtml .navbar-expand-xl .offcanvas-body , .jhtml.navbar-expand-xl .offcanvas-body {\n    display: flex;\n    flex-grow: 0;\n    padding: 0;\n    overflow-y: visible;\n  }\n}\n@media (min-width: 1400px) {\n  .jhtml .navbar-expand-xxl , .jhtml.navbar-expand-xxl {\n    flex-wrap: nowrap;\n    justify-content: flex-start;\n  }\n  .jhtml .navbar-expand-xxl .navbar-nav , .jhtml.navbar-expand-xxl .navbar-nav {\n    flex-direction: row;\n  }\n  .jhtml .navbar-expand-xxl .navbar-nav .dropdown-menu , .jhtml.navbar-expand-xxl .navbar-nav .dropdown-menu {\n    position: absolute;\n  }\n  .jhtml .navbar-expand-xxl .navbar-nav .nav-link , .jhtml.navbar-expand-xxl .navbar-nav .nav-link {\n    padding-right: 0.5rem;\n    padding-left: 0.5rem;\n  }\n  .jhtml .navbar-expand-xxl .navbar-nav-scroll , .jhtml.navbar-expand-xxl .navbar-nav-scroll {\n    overflow: visible;\n  }\n  .jhtml .navbar-expand-xxl .navbar-collapse , .jhtml.navbar-expand-xxl .navbar-collapse {\n    display: flex !important;\n    flex-basis: auto;\n  }\n  .jhtml .navbar-expand-xxl .navbar-toggler , .jhtml.navbar-expand-xxl .navbar-toggler {\n    display: none;\n  }\n  .jhtml .navbar-expand-xxl .offcanvas-header , .jhtml.navbar-expand-xxl .offcanvas-header {\n    display: none;\n  }\n  .jhtml .navbar-expand-xxl .offcanvas , .jhtml.navbar-expand-xxl .offcanvas {\n    position: inherit;\n    bottom: 0;\n    z-index: 1000;\n    flex-grow: 1;\n    visibility: visible !important;\n    background-color: transparent;\n    border-right: 0;\n    border-left: 0;\n    transition: none;\n    transform: none;\n  }\n  .jhtml .navbar-expand-xxl .offcanvas-top, \n.jhtml .navbar-expand-xxl .offcanvas-bottom , .jhtml.navbar-expand-xxl .offcanvas-top, \n.jhtml.navbar-expand-xxl .offcanvas-bottom {\n    height: auto;\n    border-top: 0;\n    border-bottom: 0;\n  }\n  .jhtml .navbar-expand-xxl .offcanvas-body , .jhtml.navbar-expand-xxl .offcanvas-body {\n    display: flex;\n    flex-grow: 0;\n    padding: 0;\n    overflow-y: visible;\n  }\n}\n.jhtml .navbar-expand , .jhtml.navbar-expand {\n  flex-wrap: nowrap;\n  justify-content: flex-start;\n}\n.jhtml .navbar-expand .navbar-nav , .jhtml.navbar-expand .navbar-nav {\n  flex-direction: row;\n}\n.jhtml .navbar-expand .navbar-nav .dropdown-menu , .jhtml.navbar-expand .navbar-nav .dropdown-menu {\n  position: absolute;\n}\n.jhtml .navbar-expand .navbar-nav .nav-link , .jhtml.navbar-expand .navbar-nav .nav-link {\n  padding-right: 0.5rem;\n  padding-left: 0.5rem;\n}\n.jhtml .navbar-expand .navbar-nav-scroll , .jhtml.navbar-expand .navbar-nav-scroll {\n  overflow: visible;\n}\n.jhtml .navbar-expand .navbar-collapse , .jhtml.navbar-expand .navbar-collapse {\n  display: flex !important;\n  flex-basis: auto;\n}\n.jhtml .navbar-expand .navbar-toggler , .jhtml.navbar-expand .navbar-toggler {\n  display: none;\n}\n.jhtml .navbar-expand .offcanvas-header , .jhtml.navbar-expand .offcanvas-header {\n  display: none;\n}\n.jhtml .navbar-expand .offcanvas , .jhtml.navbar-expand .offcanvas {\n  position: inherit;\n  bottom: 0;\n  z-index: 1000;\n  flex-grow: 1;\n  visibility: visible !important;\n  background-color: transparent;\n  border-right: 0;\n  border-left: 0;\n  transition: none;\n  transform: none;\n}\n.jhtml .navbar-expand .offcanvas-top, \n.jhtml .navbar-expand .offcanvas-bottom , .jhtml.navbar-expand .offcanvas-top, \n.jhtml.navbar-expand .offcanvas-bottom {\n  height: auto;\n  border-top: 0;\n  border-bottom: 0;\n}\n.jhtml .navbar-expand .offcanvas-body , .jhtml.navbar-expand .offcanvas-body {\n  display: flex;\n  flex-grow: 0;\n  padding: 0;\n  overflow-y: visible;\n}\n.jhtml .navbar-light .navbar-brand , .jhtml.navbar-light .navbar-brand {\n  color: rgba(0, 0, 0, 0.9);\n}\n.jhtml .navbar-light .navbar-brand:hover,  .jhtml .navbar-light .navbar-brand:focus , .jhtml.navbar-light .navbar-brand:hover,  .jhtml.navbar-light .navbar-brand:focus {\n  color: rgba(0, 0, 0, 0.9);\n}\n.jhtml .navbar-light .navbar-nav .nav-link , .jhtml.navbar-light .navbar-nav .nav-link {\n  color: rgba(0, 0, 0, 0.55);\n}\n.jhtml .navbar-light .navbar-nav .nav-link:hover,  .jhtml .navbar-light .navbar-nav .nav-link:focus , .jhtml.navbar-light .navbar-nav .nav-link:hover,  .jhtml.navbar-light .navbar-nav .nav-link:focus {\n  color: rgba(0, 0, 0, 0.7);\n}\n.jhtml .navbar-light .navbar-nav .nav-link.disabled , .jhtml.navbar-light .navbar-nav .nav-link.disabled {\n  color: rgba(0, 0, 0, 0.3);\n}\n.jhtml .navbar-light .navbar-nav .show > .nav-link, \n.jhtml .navbar-light .navbar-nav .nav-link.active , .jhtml.navbar-light .navbar-nav .show > .nav-link, \n.jhtml.navbar-light .navbar-nav .nav-link.active {\n  color: rgba(0, 0, 0, 0.9);\n}\n.jhtml .navbar-light .navbar-toggler , .jhtml.navbar-light .navbar-toggler {\n  color: rgba(0, 0, 0, 0.55);\n  border-color: rgba(0, 0, 0, 0.1);\n}\n.jhtml .navbar-light .navbar-toggler-icon , .jhtml.navbar-light .navbar-toggler-icon {\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 30 30'%3e%3cpath stroke='rgba%280, 0, 0, 0.55%29' stroke-linecap='round' stroke-miterlimit='10' stroke-width='2' d='M4 7h22M4 15h22M4 23h22'/%3e%3c/svg%3e\");\n}\n.jhtml .navbar-light .navbar-text , .jhtml.navbar-light .navbar-text {\n  color: rgba(0, 0, 0, 0.55);\n}\n.jhtml .navbar-light .navbar-text a, \n.jhtml .navbar-light .navbar-text a:hover, \n.jhtml .navbar-light .navbar-text a:focus , .jhtml.navbar-light .navbar-text a, \n.jhtml.navbar-light .navbar-text a:hover, \n.jhtml.navbar-light .navbar-text a:focus {\n  color: rgba(0, 0, 0, 0.9);\n}\n.jhtml .navbar-dark .navbar-brand , .jhtml.navbar-dark .navbar-brand {\n  color: #fff;\n}\n.jhtml .navbar-dark .navbar-brand:hover,  .jhtml .navbar-dark .navbar-brand:focus , .jhtml.navbar-dark .navbar-brand:hover,  .jhtml.navbar-dark .navbar-brand:focus {\n  color: #fff;\n}\n.jhtml .navbar-dark .navbar-nav .nav-link , .jhtml.navbar-dark .navbar-nav .nav-link {\n  color: rgba(255, 255, 255, 0.55);\n}\n.jhtml .navbar-dark .navbar-nav .nav-link:hover,  .jhtml .navbar-dark .navbar-nav .nav-link:focus , .jhtml.navbar-dark .navbar-nav .nav-link:hover,  .jhtml.navbar-dark .navbar-nav .nav-link:focus {\n  color: rgba(255, 255, 255, 0.75);\n}\n.jhtml .navbar-dark .navbar-nav .nav-link.disabled , .jhtml.navbar-dark .navbar-nav .nav-link.disabled {\n  color: rgba(255, 255, 255, 0.25);\n}\n.jhtml .navbar-dark .navbar-nav .show > .nav-link, \n.jhtml .navbar-dark .navbar-nav .nav-link.active , .jhtml.navbar-dark .navbar-nav .show > .nav-link, \n.jhtml.navbar-dark .navbar-nav .nav-link.active {\n  color: #fff;\n}\n.jhtml .navbar-dark .navbar-toggler , .jhtml.navbar-dark .navbar-toggler {\n  color: rgba(255, 255, 255, 0.55);\n  border-color: rgba(255, 255, 255, 0.1);\n}\n.jhtml .navbar-dark .navbar-toggler-icon , .jhtml.navbar-dark .navbar-toggler-icon {\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 30 30'%3e%3cpath stroke='rgba%28255, 255, 255, 0.55%29' stroke-linecap='round' stroke-miterlimit='10' stroke-width='2' d='M4 7h22M4 15h22M4 23h22'/%3e%3c/svg%3e\");\n}\n.jhtml .navbar-dark .navbar-text , .jhtml.navbar-dark .navbar-text {\n  color: rgba(255, 255, 255, 0.55);\n}\n.jhtml .navbar-dark .navbar-text a, \n.jhtml .navbar-dark .navbar-text a:hover, \n.jhtml .navbar-dark .navbar-text a:focus , .jhtml.navbar-dark .navbar-text a, \n.jhtml.navbar-dark .navbar-text a:hover, \n.jhtml.navbar-dark .navbar-text a:focus {\n  color: #fff;\n}\n.jhtml .card , .jhtml.card {\n  position: relative;\n  display: flex;\n  flex-direction: column;\n  min-width: 0;\n  word-wrap: break-word;\n  background-color: #fff;\n  background-clip: border-box;\n  border: 1px solid rgba(0, 0, 0, 0.125);\n  border-radius: 0.25rem;\n}\n.jhtml .card > hr , .jhtml.card > hr {\n  margin-right: 0;\n  margin-left: 0;\n}\n.jhtml .card > .list-group , .jhtml.card > .list-group {\n  border-top: inherit;\n  border-bottom: inherit;\n}\n.jhtml .card > .list-group:first-child , .jhtml.card > .list-group:first-child {\n  border-top-width: 0;\n  border-top-left-radius: calc(0.25rem - 1px);\n  border-top-right-radius: calc(0.25rem - 1px);\n}\n.jhtml .card > .list-group:last-child , .jhtml.card > .list-group:last-child {\n  border-bottom-width: 0;\n  border-bottom-right-radius: calc(0.25rem - 1px);\n  border-bottom-left-radius: calc(0.25rem - 1px);\n}\n.jhtml .card > .card-header + .list-group, \n.jhtml .card > .list-group + .card-footer , .jhtml.card > .card-header + .list-group, \n.jhtml.card > .list-group + .card-footer {\n  border-top: 0;\n}\n.jhtml .card-body , .jhtml.card-body {\n  flex: 1 1 auto;\n  padding: 1rem 1rem;\n}\n.jhtml .card-title , .jhtml.card-title {\n  margin-bottom: 0.5rem;\n}\n.jhtml .card-subtitle , .jhtml.card-subtitle {\n  margin-top: -0.25rem;\n  margin-bottom: 0;\n}\n.jhtml .card-text:last-child , .jhtml.card-text:last-child {\n  margin-bottom: 0;\n}\n.jhtml .card-link + .card-link , .jhtml.card-link + .card-link {\n  margin-left: 1rem;\n}\n.jhtml .card-header , .jhtml.card-header {\n  padding: 0.5rem 1rem;\n  margin-bottom: 0;\n  background-color: rgba(0, 0, 0, 0.03);\n  border-bottom: 1px solid rgba(0, 0, 0, 0.125);\n}\n.jhtml .card-header:first-child , .jhtml.card-header:first-child {\n  border-radius: calc(0.25rem - 1px) calc(0.25rem - 1px) 0 0;\n}\n.jhtml .card-footer , .jhtml.card-footer {\n  padding: 0.5rem 1rem;\n  background-color: rgba(0, 0, 0, 0.03);\n  border-top: 1px solid rgba(0, 0, 0, 0.125);\n}\n.jhtml .card-footer:last-child , .jhtml.card-footer:last-child {\n  border-radius: 0 0 calc(0.25rem - 1px) calc(0.25rem - 1px);\n}\n.jhtml .card-header-tabs , .jhtml.card-header-tabs {\n  margin-right: -0.5rem;\n  margin-bottom: -0.5rem;\n  margin-left: -0.5rem;\n  border-bottom: 0;\n}\n.jhtml .card-header-pills , .jhtml.card-header-pills {\n  margin-right: -0.5rem;\n  margin-left: -0.5rem;\n}\n.jhtml .card-img-overlay , .jhtml.card-img-overlay {\n  position: absolute;\n  top: 0;\n  right: 0;\n  bottom: 0;\n  left: 0;\n  padding: 1rem;\n  border-radius: calc(0.25rem - 1px);\n}\n.jhtml .card-img, \n.jhtml .card-img-top, \n.jhtml .card-img-bottom , .jhtml.card-img, \n.jhtml.card-img-top, \n.jhtml.card-img-bottom {\n  width: 100%;\n}\n.jhtml .card-img, \n.jhtml .card-img-top , .jhtml.card-img, \n.jhtml.card-img-top {\n  border-top-left-radius: calc(0.25rem - 1px);\n  border-top-right-radius: calc(0.25rem - 1px);\n}\n.jhtml .card-img, \n.jhtml .card-img-bottom , .jhtml.card-img, \n.jhtml.card-img-bottom {\n  border-bottom-right-radius: calc(0.25rem - 1px);\n  border-bottom-left-radius: calc(0.25rem - 1px);\n}\n.jhtml .card-group > .card , .jhtml.card-group > .card {\n  margin-bottom: 0.75rem;\n}\n@media (min-width: 576px) {\n  .jhtml .card-group , .jhtml.card-group {\n    display: flex;\n    flex-flow: row wrap;\n  }\n  .jhtml .card-group > .card , .jhtml.card-group > .card {\n    flex: 1 0 0%;\n    margin-bottom: 0;\n  }\n  .jhtml .card-group > .card + .card , .jhtml.card-group > .card + .card {\n    margin-left: 0;\n    border-left: 0;\n  }\n  .jhtml .card-group > .card:not(:last-child) , .jhtml.card-group > .card:not(:last-child) {\n    border-top-right-radius: 0;\n    border-bottom-right-radius: 0;\n  }\n  .jhtml .card-group > .card:not(:last-child) .card-img-top, \n.jhtml .card-group > .card:not(:last-child) .card-header , .jhtml.card-group > .card:not(:last-child) .card-img-top, \n.jhtml.card-group > .card:not(:last-child) .card-header {\n    border-top-right-radius: 0;\n  }\n  .jhtml .card-group > .card:not(:last-child) .card-img-bottom, \n.jhtml .card-group > .card:not(:last-child) .card-footer , .jhtml.card-group > .card:not(:last-child) .card-img-bottom, \n.jhtml.card-group > .card:not(:last-child) .card-footer {\n    border-bottom-right-radius: 0;\n  }\n  .jhtml .card-group > .card:not(:first-child) , .jhtml.card-group > .card:not(:first-child) {\n    border-top-left-radius: 0;\n    border-bottom-left-radius: 0;\n  }\n  .jhtml .card-group > .card:not(:first-child) .card-img-top, \n.jhtml .card-group > .card:not(:first-child) .card-header , .jhtml.card-group > .card:not(:first-child) .card-img-top, \n.jhtml.card-group > .card:not(:first-child) .card-header {\n    border-top-left-radius: 0;\n  }\n  .jhtml .card-group > .card:not(:first-child) .card-img-bottom, \n.jhtml .card-group > .card:not(:first-child) .card-footer , .jhtml.card-group > .card:not(:first-child) .card-img-bottom, \n.jhtml.card-group > .card:not(:first-child) .card-footer {\n    border-bottom-left-radius: 0;\n  }\n}\n.jhtml .accordion-button , .jhtml.accordion-button {\n  position: relative;\n  display: flex;\n  align-items: center;\n  width: 100%;\n  padding: 1rem 1.25rem;\n  font-size: 1rem;\n  color: #212529;\n  text-align: left;\n  background-color: #fff;\n  border: 0;\n  border-radius: 0;\n  overflow-anchor: none;\n  transition: color 0.15s ease-in-out, background-color 0.15s ease-in-out, border-color 0.15s ease-in-out, box-shadow 0.15s ease-in-out, border-radius 0.15s ease;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .accordion-button , .jhtml.accordion-button {\n    transition: none;\n  }\n}\n.jhtml .accordion-button:not(.collapsed) , .jhtml.accordion-button:not(.collapsed) {\n  color: #0c63e4;\n  background-color: #e7f1ff;\n  box-shadow: inset 0 -1px 0 rgba(0, 0, 0, 0.125);\n}\n.jhtml .accordion-button:not(.collapsed)::after , .jhtml.accordion-button:not(.collapsed)::after {\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 16 16' fill='%230c63e4'%3e%3cpath fill-rule='evenodd' d='M1.646 4.646a.5.5 0 0 1 .708 0L8 10.293l5.646-5.647a.5.5 0 0 1 .708.708l-6 6a.5.5 0 0 1-.708 0l-6-6a.5.5 0 0 1 0-.708z'/%3e%3c/svg%3e\");\n  transform: rotate(-180deg);\n}\n.jhtml .accordion-button::after , .jhtml.accordion-button::after {\n  flex-shrink: 0;\n  width: 1.25rem;\n  height: 1.25rem;\n  margin-left: auto;\n  content: \"\";\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 16 16' fill='%23212529'%3e%3cpath fill-rule='evenodd' d='M1.646 4.646a.5.5 0 0 1 .708 0L8 10.293l5.646-5.647a.5.5 0 0 1 .708.708l-6 6a.5.5 0 0 1-.708 0l-6-6a.5.5 0 0 1 0-.708z'/%3e%3c/svg%3e\");\n  background-repeat: no-repeat;\n  background-size: 1.25rem;\n  transition: transform 0.2s ease-in-out;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .accordion-button::after , .jhtml.accordion-button::after {\n    transition: none;\n  }\n}\n.jhtml .accordion-button:hover , .jhtml.accordion-button:hover {\n  z-index: 2;\n}\n.jhtml .accordion-button:focus , .jhtml.accordion-button:focus {\n  z-index: 3;\n  border-color: #86b7fe;\n  outline: 0;\n  box-shadow: 0 0 0 0.25rem rgba(13, 110, 253, 0.25);\n}\n.jhtml .accordion-header , .jhtml.accordion-header {\n  margin-bottom: 0;\n}\n.jhtml .accordion-item , .jhtml.accordion-item {\n  background-color: #fff;\n  border: 1px solid rgba(0, 0, 0, 0.125);\n}\n.jhtml .accordion-item:first-of-type , .jhtml.accordion-item:first-of-type {\n  border-top-left-radius: 0.25rem;\n  border-top-right-radius: 0.25rem;\n}\n.jhtml .accordion-item:first-of-type .accordion-button , .jhtml.accordion-item:first-of-type .accordion-button {\n  border-top-left-radius: calc(0.25rem - 1px);\n  border-top-right-radius: calc(0.25rem - 1px);\n}\n.jhtml .accordion-item:not(:first-of-type) , .jhtml.accordion-item:not(:first-of-type) {\n  border-top: 0;\n}\n.jhtml .accordion-item:last-of-type , .jhtml.accordion-item:last-of-type {\n  border-bottom-right-radius: 0.25rem;\n  border-bottom-left-radius: 0.25rem;\n}\n.jhtml .accordion-item:last-of-type .accordion-button.collapsed , .jhtml.accordion-item:last-of-type .accordion-button.collapsed {\n  border-bottom-right-radius: calc(0.25rem - 1px);\n  border-bottom-left-radius: calc(0.25rem - 1px);\n}\n.jhtml .accordion-item:last-of-type .accordion-collapse , .jhtml.accordion-item:last-of-type .accordion-collapse {\n  border-bottom-right-radius: 0.25rem;\n  border-bottom-left-radius: 0.25rem;\n}\n.jhtml .accordion-body , .jhtml.accordion-body {\n  padding: 1rem 1.25rem;\n}\n.jhtml .accordion-flush .accordion-collapse , .jhtml.accordion-flush .accordion-collapse {\n  border-width: 0;\n}\n.jhtml .accordion-flush .accordion-item , .jhtml.accordion-flush .accordion-item {\n  border-right: 0;\n  border-left: 0;\n  border-radius: 0;\n}\n.jhtml .accordion-flush .accordion-item:first-child , .jhtml.accordion-flush .accordion-item:first-child {\n  border-top: 0;\n}\n.jhtml .accordion-flush .accordion-item:last-child , .jhtml.accordion-flush .accordion-item:last-child {\n  border-bottom: 0;\n}\n.jhtml .accordion-flush .accordion-item .accordion-button , .jhtml.accordion-flush .accordion-item .accordion-button {\n  border-radius: 0;\n}\n.jhtml .breadcrumb , .jhtml.breadcrumb {\n  display: flex;\n  flex-wrap: wrap;\n  padding: 0 0;\n  margin-bottom: 1rem;\n  list-style: none;\n}\n.jhtml .breadcrumb-item + .breadcrumb-item , .jhtml.breadcrumb-item + .breadcrumb-item {\n  padding-left: 0.5rem;\n}\n.jhtml .breadcrumb-item + .breadcrumb-item::before , .jhtml.breadcrumb-item + .breadcrumb-item::before {\n  float: left;\n  padding-right: 0.5rem;\n  color: #6c757d;\n  content: var(--bs-breadcrumb-divider, \"/\") /* rtl: var(--bs-breadcrumb-divider, \"/\") */;\n}\n.jhtml .breadcrumb-item.active , .jhtml.breadcrumb-item.active {\n  color: #6c757d;\n}\n.jhtml .pagination , .jhtml.pagination {\n  display: flex;\n  padding-left: 0;\n  list-style: none;\n}\n.jhtml .page-link , .jhtml.page-link {\n  position: relative;\n  display: block;\n  color: #0d6efd;\n  text-decoration: none;\n  background-color: #fff;\n  border: 1px solid #dee2e6;\n  transition: color 0.15s ease-in-out, background-color 0.15s ease-in-out, border-color 0.15s ease-in-out, box-shadow 0.15s ease-in-out;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .page-link , .jhtml.page-link {\n    transition: none;\n  }\n}\n.jhtml .page-link:hover , .jhtml.page-link:hover {\n  z-index: 2;\n  color: #0a58ca;\n  background-color: #e9ecef;\n  border-color: #dee2e6;\n}\n.jhtml .page-link:focus , .jhtml.page-link:focus {\n  z-index: 3;\n  color: #0a58ca;\n  background-color: #e9ecef;\n  outline: 0;\n  box-shadow: 0 0 0 0.25rem rgba(13, 110, 253, 0.25);\n}\n.jhtml .page-item:not(:first-child) .page-link , .jhtml.page-item:not(:first-child) .page-link {\n  margin-left: -1px;\n}\n.jhtml .page-item.active .page-link , .jhtml.page-item.active .page-link {\n  z-index: 3;\n  color: #fff;\n  background-color: #0d6efd;\n  border-color: #0d6efd;\n}\n.jhtml .page-item.disabled .page-link , .jhtml.page-item.disabled .page-link {\n  color: #6c757d;\n  pointer-events: none;\n  background-color: #fff;\n  border-color: #dee2e6;\n}\n.jhtml .page-link , .jhtml.page-link {\n  padding: 0.375rem 0.75rem;\n}\n.jhtml .page-item:first-child .page-link , .jhtml.page-item:first-child .page-link {\n  border-top-left-radius: 0.25rem;\n  border-bottom-left-radius: 0.25rem;\n}\n.jhtml .page-item:last-child .page-link , .jhtml.page-item:last-child .page-link {\n  border-top-right-radius: 0.25rem;\n  border-bottom-right-radius: 0.25rem;\n}\n.jhtml .pagination-lg .page-link , .jhtml.pagination-lg .page-link {\n  padding: 0.75rem 1.5rem;\n  font-size: 1.25rem;\n}\n.jhtml .pagination-lg .page-item:first-child .page-link , .jhtml.pagination-lg .page-item:first-child .page-link {\n  border-top-left-radius: 0.3rem;\n  border-bottom-left-radius: 0.3rem;\n}\n.jhtml .pagination-lg .page-item:last-child .page-link , .jhtml.pagination-lg .page-item:last-child .page-link {\n  border-top-right-radius: 0.3rem;\n  border-bottom-right-radius: 0.3rem;\n}\n.jhtml .pagination-sm .page-link , .jhtml.pagination-sm .page-link {\n  padding: 0.25rem 0.5rem;\n  font-size: 0.875rem;\n}\n.jhtml .pagination-sm .page-item:first-child .page-link , .jhtml.pagination-sm .page-item:first-child .page-link {\n  border-top-left-radius: 0.2rem;\n  border-bottom-left-radius: 0.2rem;\n}\n.jhtml .pagination-sm .page-item:last-child .page-link , .jhtml.pagination-sm .page-item:last-child .page-link {\n  border-top-right-radius: 0.2rem;\n  border-bottom-right-radius: 0.2rem;\n}\n.jhtml .badge , .jhtml.badge {\n  display: inline-block;\n  padding: 0.35em 0.65em;\n  font-size: 0.75em;\n  font-weight: 700;\n  line-height: 1;\n  color: #fff;\n  text-align: center;\n  white-space: nowrap;\n  vertical-align: baseline;\n  border-radius: 0.25rem;\n}\n.jhtml .badge:empty , .jhtml.badge:empty {\n  display: none;\n}\n.jhtml .btn .badge , .jhtml.btn .badge {\n  position: relative;\n  top: -1px;\n}\n.jhtml .alert , .jhtml.alert {\n  position: relative;\n  padding: 1rem 1rem;\n  margin-bottom: 1rem;\n  border: 1px solid transparent;\n  border-radius: 0.25rem;\n}\n.jhtml .alert-heading , .jhtml.alert-heading {\n  color: inherit;\n}\n.jhtml .alert-link , .jhtml.alert-link {\n  font-weight: 700;\n}\n.jhtml .alert-dismissible , .jhtml.alert-dismissible {\n  padding-right: 3rem;\n}\n.jhtml .alert-dismissible .btn-close , .jhtml.alert-dismissible .btn-close {\n  position: absolute;\n  top: 0;\n  right: 0;\n  z-index: 2;\n  padding: 1.25rem 1rem;\n}\n.jhtml .alert-primary , .jhtml.alert-primary {\n  color: #084298;\n  background-color: #cfe2ff;\n  border-color: #b6d4fe;\n}\n.jhtml .alert-primary .alert-link , .jhtml.alert-primary .alert-link {\n  color: #06357a;\n}\n.jhtml .alert-secondary , .jhtml.alert-secondary {\n  color: #41464b;\n  background-color: #e2e3e5;\n  border-color: #d3d6d8;\n}\n.jhtml .alert-secondary .alert-link , .jhtml.alert-secondary .alert-link {\n  color: #34383c;\n}\n.jhtml .alert-success , .jhtml.alert-success {\n  color: #0f5132;\n  background-color: #d1e7dd;\n  border-color: #badbcc;\n}\n.jhtml .alert-success .alert-link , .jhtml.alert-success .alert-link {\n  color: #0c4128;\n}\n.jhtml .alert-info , .jhtml.alert-info {\n  color: #055160;\n  background-color: #cff4fc;\n  border-color: #b6effb;\n}\n.jhtml .alert-info .alert-link , .jhtml.alert-info .alert-link {\n  color: #04414d;\n}\n.jhtml .alert-warning , .jhtml.alert-warning {\n  color: #664d03;\n  background-color: #fff3cd;\n  border-color: #ffecb5;\n}\n.jhtml .alert-warning .alert-link , .jhtml.alert-warning .alert-link {\n  color: #523e02;\n}\n.jhtml .alert-danger , .jhtml.alert-danger {\n  color: #842029;\n  background-color: #f8d7da;\n  border-color: #f5c2c7;\n}\n.jhtml .alert-danger .alert-link , .jhtml.alert-danger .alert-link {\n  color: #6a1a21;\n}\n.jhtml .alert-light , .jhtml.alert-light {\n  color: #636464;\n  background-color: #fefefe;\n  border-color: #fdfdfe;\n}\n.jhtml .alert-light .alert-link , .jhtml.alert-light .alert-link {\n  color: #4f5050;\n}\n.jhtml .alert-dark , .jhtml.alert-dark {\n  color: #141619;\n  background-color: #d3d3d4;\n  border-color: #bcbebf;\n}\n.jhtml .alert-dark .alert-link , .jhtml.alert-dark .alert-link {\n  color: #101214;\n}\n@keyframes progress-bar-stripes {\n  0% {\n    background-position-x: 1rem;\n  }\n}\n.jhtml .progress , .jhtml.progress {\n  display: flex;\n  height: 1rem;\n  overflow: hidden;\n  font-size: 0.75rem;\n  background-color: #e9ecef;\n  border-radius: 0.25rem;\n}\n.jhtml .progress-bar , .jhtml.progress-bar {\n  display: flex;\n  flex-direction: column;\n  justify-content: center;\n  overflow: hidden;\n  color: #fff;\n  text-align: center;\n  white-space: nowrap;\n  background-color: #0d6efd;\n  transition: width 0.6s ease;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .progress-bar , .jhtml.progress-bar {\n    transition: none;\n  }\n}\n.jhtml .progress-bar-striped , .jhtml.progress-bar-striped {\n  background-image: linear-gradient(45deg, rgba(255, 255, 255, 0.15) 25%, transparent 25%, transparent 50%, rgba(255, 255, 255, 0.15) 50%, rgba(255, 255, 255, 0.15) 75%, transparent 75%, transparent);\n  background-size: 1rem 1rem;\n}\n.jhtml .progress-bar-animated , .jhtml.progress-bar-animated {\n  animation: 1s linear infinite progress-bar-stripes;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .progress-bar-animated , .jhtml.progress-bar-animated {\n    animation: none;\n  }\n}\n.jhtml .list-group , .jhtml.list-group {\n  display: flex;\n  flex-direction: column;\n  padding-left: 0;\n  margin-bottom: 0;\n  border-radius: 0.25rem;\n}\n.jhtml .list-group-numbered , .jhtml.list-group-numbered {\n  list-style-type: none;\n  counter-reset: section;\n}\n.jhtml .list-group-numbered > li::before , .jhtml.list-group-numbered > li::before {\n  content: counters(section, \".\") \". \";\n  counter-increment: section;\n}\n.jhtml .list-group-item-action , .jhtml.list-group-item-action {\n  width: 100%;\n  color: #495057;\n  text-align: inherit;\n}\n.jhtml .list-group-item-action:hover,  .jhtml .list-group-item-action:focus , .jhtml.list-group-item-action:hover,  .jhtml.list-group-item-action:focus {\n  z-index: 1;\n  color: #495057;\n  text-decoration: none;\n  background-color: #f8f9fa;\n}\n.jhtml .list-group-item-action:active , .jhtml.list-group-item-action:active {\n  color: #212529;\n  background-color: #e9ecef;\n}\n.jhtml .list-group-item , .jhtml.list-group-item {\n  position: relative;\n  display: block;\n  padding: 0.5rem 1rem;\n  color: #212529;\n  text-decoration: none;\n  background-color: #fff;\n  border: 1px solid rgba(0, 0, 0, 0.125);\n}\n.jhtml .list-group-item:first-child , .jhtml.list-group-item:first-child {\n  border-top-left-radius: inherit;\n  border-top-right-radius: inherit;\n}\n.jhtml .list-group-item:last-child , .jhtml.list-group-item:last-child {\n  border-bottom-right-radius: inherit;\n  border-bottom-left-radius: inherit;\n}\n.jhtml .list-group-item.disabled,  .jhtml .list-group-item:disabled , .jhtml.list-group-item.disabled,  .jhtml.list-group-item:disabled {\n  color: #6c757d;\n  pointer-events: none;\n  background-color: #fff;\n}\n.jhtml .list-group-item.active , .jhtml.list-group-item.active {\n  z-index: 2;\n  color: #fff;\n  background-color: #0d6efd;\n  border-color: #0d6efd;\n}\n.jhtml .list-group-item + .list-group-item , .jhtml.list-group-item + .list-group-item {\n  border-top-width: 0;\n}\n.jhtml .list-group-item + .list-group-item.active , .jhtml.list-group-item + .list-group-item.active {\n  margin-top: -1px;\n  border-top-width: 1px;\n}\n.jhtml .list-group-horizontal , .jhtml.list-group-horizontal {\n  flex-direction: row;\n}\n.jhtml .list-group-horizontal > .list-group-item:first-child , .jhtml.list-group-horizontal > .list-group-item:first-child {\n  border-bottom-left-radius: 0.25rem;\n  border-top-right-radius: 0;\n}\n.jhtml .list-group-horizontal > .list-group-item:last-child , .jhtml.list-group-horizontal > .list-group-item:last-child {\n  border-top-right-radius: 0.25rem;\n  border-bottom-left-radius: 0;\n}\n.jhtml .list-group-horizontal > .list-group-item.active , .jhtml.list-group-horizontal > .list-group-item.active {\n  margin-top: 0;\n}\n.jhtml .list-group-horizontal > .list-group-item + .list-group-item , .jhtml.list-group-horizontal > .list-group-item + .list-group-item {\n  border-top-width: 1px;\n  border-left-width: 0;\n}\n.jhtml .list-group-horizontal > .list-group-item + .list-group-item.active , .jhtml.list-group-horizontal > .list-group-item + .list-group-item.active {\n  margin-left: -1px;\n  border-left-width: 1px;\n}\n@media (min-width: 576px) {\n  .jhtml .list-group-horizontal-sm , .jhtml.list-group-horizontal-sm {\n    flex-direction: row;\n  }\n  .jhtml .list-group-horizontal-sm > .list-group-item:first-child , .jhtml.list-group-horizontal-sm > .list-group-item:first-child {\n    border-bottom-left-radius: 0.25rem;\n    border-top-right-radius: 0;\n  }\n  .jhtml .list-group-horizontal-sm > .list-group-item:last-child , .jhtml.list-group-horizontal-sm > .list-group-item:last-child {\n    border-top-right-radius: 0.25rem;\n    border-bottom-left-radius: 0;\n  }\n  .jhtml .list-group-horizontal-sm > .list-group-item.active , .jhtml.list-group-horizontal-sm > .list-group-item.active {\n    margin-top: 0;\n  }\n  .jhtml .list-group-horizontal-sm > .list-group-item + .list-group-item , .jhtml.list-group-horizontal-sm > .list-group-item + .list-group-item {\n    border-top-width: 1px;\n    border-left-width: 0;\n  }\n  .jhtml .list-group-horizontal-sm > .list-group-item + .list-group-item.active , .jhtml.list-group-horizontal-sm > .list-group-item + .list-group-item.active {\n    margin-left: -1px;\n    border-left-width: 1px;\n  }\n}\n@media (min-width: 768px) {\n  .jhtml .list-group-horizontal-md , .jhtml.list-group-horizontal-md {\n    flex-direction: row;\n  }\n  .jhtml .list-group-horizontal-md > .list-group-item:first-child , .jhtml.list-group-horizontal-md > .list-group-item:first-child {\n    border-bottom-left-radius: 0.25rem;\n    border-top-right-radius: 0;\n  }\n  .jhtml .list-group-horizontal-md > .list-group-item:last-child , .jhtml.list-group-horizontal-md > .list-group-item:last-child {\n    border-top-right-radius: 0.25rem;\n    border-bottom-left-radius: 0;\n  }\n  .jhtml .list-group-horizontal-md > .list-group-item.active , .jhtml.list-group-horizontal-md > .list-group-item.active {\n    margin-top: 0;\n  }\n  .jhtml .list-group-horizontal-md > .list-group-item + .list-group-item , .jhtml.list-group-horizontal-md > .list-group-item + .list-group-item {\n    border-top-width: 1px;\n    border-left-width: 0;\n  }\n  .jhtml .list-group-horizontal-md > .list-group-item + .list-group-item.active , .jhtml.list-group-horizontal-md > .list-group-item + .list-group-item.active {\n    margin-left: -1px;\n    border-left-width: 1px;\n  }\n}\n@media (min-width: 992px) {\n  .jhtml .list-group-horizontal-lg , .jhtml.list-group-horizontal-lg {\n    flex-direction: row;\n  }\n  .jhtml .list-group-horizontal-lg > .list-group-item:first-child , .jhtml.list-group-horizontal-lg > .list-group-item:first-child {\n    border-bottom-left-radius: 0.25rem;\n    border-top-right-radius: 0;\n  }\n  .jhtml .list-group-horizontal-lg > .list-group-item:last-child , .jhtml.list-group-horizontal-lg > .list-group-item:last-child {\n    border-top-right-radius: 0.25rem;\n    border-bottom-left-radius: 0;\n  }\n  .jhtml .list-group-horizontal-lg > .list-group-item.active , .jhtml.list-group-horizontal-lg > .list-group-item.active {\n    margin-top: 0;\n  }\n  .jhtml .list-group-horizontal-lg > .list-group-item + .list-group-item , .jhtml.list-group-horizontal-lg > .list-group-item + .list-group-item {\n    border-top-width: 1px;\n    border-left-width: 0;\n  }\n  .jhtml .list-group-horizontal-lg > .list-group-item + .list-group-item.active , .jhtml.list-group-horizontal-lg > .list-group-item + .list-group-item.active {\n    margin-left: -1px;\n    border-left-width: 1px;\n  }\n}\n@media (min-width: 1200px) {\n  .jhtml .list-group-horizontal-xl , .jhtml.list-group-horizontal-xl {\n    flex-direction: row;\n  }\n  .jhtml .list-group-horizontal-xl > .list-group-item:first-child , .jhtml.list-group-horizontal-xl > .list-group-item:first-child {\n    border-bottom-left-radius: 0.25rem;\n    border-top-right-radius: 0;\n  }\n  .jhtml .list-group-horizontal-xl > .list-group-item:last-child , .jhtml.list-group-horizontal-xl > .list-group-item:last-child {\n    border-top-right-radius: 0.25rem;\n    border-bottom-left-radius: 0;\n  }\n  .jhtml .list-group-horizontal-xl > .list-group-item.active , .jhtml.list-group-horizontal-xl > .list-group-item.active {\n    margin-top: 0;\n  }\n  .jhtml .list-group-horizontal-xl > .list-group-item + .list-group-item , .jhtml.list-group-horizontal-xl > .list-group-item + .list-group-item {\n    border-top-width: 1px;\n    border-left-width: 0;\n  }\n  .jhtml .list-group-horizontal-xl > .list-group-item + .list-group-item.active , .jhtml.list-group-horizontal-xl > .list-group-item + .list-group-item.active {\n    margin-left: -1px;\n    border-left-width: 1px;\n  }\n}\n@media (min-width: 1400px) {\n  .jhtml .list-group-horizontal-xxl , .jhtml.list-group-horizontal-xxl {\n    flex-direction: row;\n  }\n  .jhtml .list-group-horizontal-xxl > .list-group-item:first-child , .jhtml.list-group-horizontal-xxl > .list-group-item:first-child {\n    border-bottom-left-radius: 0.25rem;\n    border-top-right-radius: 0;\n  }\n  .jhtml .list-group-horizontal-xxl > .list-group-item:last-child , .jhtml.list-group-horizontal-xxl > .list-group-item:last-child {\n    border-top-right-radius: 0.25rem;\n    border-bottom-left-radius: 0;\n  }\n  .jhtml .list-group-horizontal-xxl > .list-group-item.active , .jhtml.list-group-horizontal-xxl > .list-group-item.active {\n    margin-top: 0;\n  }\n  .jhtml .list-group-horizontal-xxl > .list-group-item + .list-group-item , .jhtml.list-group-horizontal-xxl > .list-group-item + .list-group-item {\n    border-top-width: 1px;\n    border-left-width: 0;\n  }\n  .jhtml .list-group-horizontal-xxl > .list-group-item + .list-group-item.active , .jhtml.list-group-horizontal-xxl > .list-group-item + .list-group-item.active {\n    margin-left: -1px;\n    border-left-width: 1px;\n  }\n}\n.jhtml .list-group-flush , .jhtml.list-group-flush {\n  border-radius: 0;\n}\n.jhtml .list-group-flush > .list-group-item , .jhtml.list-group-flush > .list-group-item {\n  border-width: 0 0 1px;\n}\n.jhtml .list-group-flush > .list-group-item:last-child , .jhtml.list-group-flush > .list-group-item:last-child {\n  border-bottom-width: 0;\n}\n.jhtml .list-group-item-primary , .jhtml.list-group-item-primary {\n  color: #084298;\n  background-color: #cfe2ff;\n}\n.jhtml .list-group-item-primary.list-group-item-action:hover,  .jhtml .list-group-item-primary.list-group-item-action:focus , .jhtml.list-group-item-primary.list-group-item-action:hover,  .jhtml.list-group-item-primary.list-group-item-action:focus {\n  color: #084298;\n  background-color: #bacbe6;\n}\n.jhtml .list-group-item-primary.list-group-item-action.active , .jhtml.list-group-item-primary.list-group-item-action.active {\n  color: #fff;\n  background-color: #084298;\n  border-color: #084298;\n}\n.jhtml .list-group-item-secondary , .jhtml.list-group-item-secondary {\n  color: #41464b;\n  background-color: #e2e3e5;\n}\n.jhtml .list-group-item-secondary.list-group-item-action:hover,  .jhtml .list-group-item-secondary.list-group-item-action:focus , .jhtml.list-group-item-secondary.list-group-item-action:hover,  .jhtml.list-group-item-secondary.list-group-item-action:focus {\n  color: #41464b;\n  background-color: #cbccce;\n}\n.jhtml .list-group-item-secondary.list-group-item-action.active , .jhtml.list-group-item-secondary.list-group-item-action.active {\n  color: #fff;\n  background-color: #41464b;\n  border-color: #41464b;\n}\n.jhtml .list-group-item-success , .jhtml.list-group-item-success {\n  color: #0f5132;\n  background-color: #d1e7dd;\n}\n.jhtml .list-group-item-success.list-group-item-action:hover,  .jhtml .list-group-item-success.list-group-item-action:focus , .jhtml.list-group-item-success.list-group-item-action:hover,  .jhtml.list-group-item-success.list-group-item-action:focus {\n  color: #0f5132;\n  background-color: #bcd0c7;\n}\n.jhtml .list-group-item-success.list-group-item-action.active , .jhtml.list-group-item-success.list-group-item-action.active {\n  color: #fff;\n  background-color: #0f5132;\n  border-color: #0f5132;\n}\n.jhtml .list-group-item-info , .jhtml.list-group-item-info {\n  color: #055160;\n  background-color: #cff4fc;\n}\n.jhtml .list-group-item-info.list-group-item-action:hover,  .jhtml .list-group-item-info.list-group-item-action:focus , .jhtml.list-group-item-info.list-group-item-action:hover,  .jhtml.list-group-item-info.list-group-item-action:focus {\n  color: #055160;\n  background-color: #badce3;\n}\n.jhtml .list-group-item-info.list-group-item-action.active , .jhtml.list-group-item-info.list-group-item-action.active {\n  color: #fff;\n  background-color: #055160;\n  border-color: #055160;\n}\n.jhtml .list-group-item-warning , .jhtml.list-group-item-warning {\n  color: #664d03;\n  background-color: #fff3cd;\n}\n.jhtml .list-group-item-warning.list-group-item-action:hover,  .jhtml .list-group-item-warning.list-group-item-action:focus , .jhtml.list-group-item-warning.list-group-item-action:hover,  .jhtml.list-group-item-warning.list-group-item-action:focus {\n  color: #664d03;\n  background-color: #e6dbb9;\n}\n.jhtml .list-group-item-warning.list-group-item-action.active , .jhtml.list-group-item-warning.list-group-item-action.active {\n  color: #fff;\n  background-color: #664d03;\n  border-color: #664d03;\n}\n.jhtml .list-group-item-danger , .jhtml.list-group-item-danger {\n  color: #842029;\n  background-color: #f8d7da;\n}\n.jhtml .list-group-item-danger.list-group-item-action:hover,  .jhtml .list-group-item-danger.list-group-item-action:focus , .jhtml.list-group-item-danger.list-group-item-action:hover,  .jhtml.list-group-item-danger.list-group-item-action:focus {\n  color: #842029;\n  background-color: #dfc2c4;\n}\n.jhtml .list-group-item-danger.list-group-item-action.active , .jhtml.list-group-item-danger.list-group-item-action.active {\n  color: #fff;\n  background-color: #842029;\n  border-color: #842029;\n}\n.jhtml .list-group-item-light , .jhtml.list-group-item-light {\n  color: #636464;\n  background-color: #fefefe;\n}\n.jhtml .list-group-item-light.list-group-item-action:hover,  .jhtml .list-group-item-light.list-group-item-action:focus , .jhtml.list-group-item-light.list-group-item-action:hover,  .jhtml.list-group-item-light.list-group-item-action:focus {\n  color: #636464;\n  background-color: #e5e5e5;\n}\n.jhtml .list-group-item-light.list-group-item-action.active , .jhtml.list-group-item-light.list-group-item-action.active {\n  color: #fff;\n  background-color: #636464;\n  border-color: #636464;\n}\n.jhtml .list-group-item-dark , .jhtml.list-group-item-dark {\n  color: #141619;\n  background-color: #d3d3d4;\n}\n.jhtml .list-group-item-dark.list-group-item-action:hover,  .jhtml .list-group-item-dark.list-group-item-action:focus , .jhtml.list-group-item-dark.list-group-item-action:hover,  .jhtml.list-group-item-dark.list-group-item-action:focus {\n  color: #141619;\n  background-color: #bebebf;\n}\n.jhtml .list-group-item-dark.list-group-item-action.active , .jhtml.list-group-item-dark.list-group-item-action.active {\n  color: #fff;\n  background-color: #141619;\n  border-color: #141619;\n}\n.jhtml .btn-close , .jhtml.btn-close {\n  box-sizing: content-box;\n  width: 1em;\n  height: 1em;\n  padding: 0.25em 0.25em;\n  color: #000;\n  background: transparent url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 16 16' fill='%23000'%3e%3cpath d='M.293.293a1 1 0 011.414 0L8 6.586 14.293.293a1 1 0 111.414 1.414L9.414 8l6.293 6.293a1 1 0 01-1.414 1.414L8 9.414l-6.293 6.293a1 1 0 01-1.414-1.414L6.586 8 .293 1.707a1 1 0 010-1.414z'/%3e%3c/svg%3e\") center/1em auto no-repeat;\n  border: 0;\n  border-radius: 0.25rem;\n  opacity: 0.5;\n}\n.jhtml .btn-close:hover , .jhtml.btn-close:hover {\n  color: #000;\n  text-decoration: none;\n  opacity: 0.75;\n}\n.jhtml .btn-close:focus , .jhtml.btn-close:focus {\n  outline: 0;\n  box-shadow: 0 0 0 0.25rem rgba(13, 110, 253, 0.25);\n  opacity: 1;\n}\n.jhtml .btn-close:disabled,  .jhtml .btn-close.disabled , .jhtml.btn-close:disabled,  .jhtml.btn-close.disabled {\n  pointer-events: none;\n  user-select: none;\n  opacity: 0.25;\n}\n.jhtml .btn-close-white , .jhtml.btn-close-white {\n  filter: invert(1) grayscale(100%) brightness(200%);\n}\n.jhtml .toast , .jhtml.toast {\n  width: 350px;\n  max-width: 100%;\n  font-size: 0.875rem;\n  pointer-events: auto;\n  background-color: rgba(255, 255, 255, 0.85);\n  background-clip: padding-box;\n  border: 1px solid rgba(0, 0, 0, 0.1);\n  box-shadow: 0 0.5rem 1rem rgba(0, 0, 0, 0.15);\n  border-radius: 0.25rem;\n}\n.jhtml .toast.showing , .jhtml.toast.showing {\n  opacity: 0;\n}\n.jhtml .toast:not(.show) , .jhtml.toast:not(.show) {\n  display: none;\n}\n.jhtml .toast-container , .jhtml.toast-container {\n  width: max-content;\n  max-width: 100%;\n  pointer-events: none;\n}\n.jhtml .toast-container > :not(:last-child) , .jhtml.toast-container > :not(:last-child) {\n  margin-bottom: 0.75rem;\n}\n.jhtml .toast-header , .jhtml.toast-header {\n  display: flex;\n  align-items: center;\n  padding: 0.5rem 0.75rem;\n  color: #6c757d;\n  background-color: rgba(255, 255, 255, 0.85);\n  background-clip: padding-box;\n  border-bottom: 1px solid rgba(0, 0, 0, 0.05);\n  border-top-left-radius: calc(0.25rem - 1px);\n  border-top-right-radius: calc(0.25rem - 1px);\n}\n.jhtml .toast-header .btn-close , .jhtml.toast-header .btn-close {\n  margin-right: -0.375rem;\n  margin-left: 0.75rem;\n}\n.jhtml .toast-body , .jhtml.toast-body {\n  padding: 0.75rem;\n  word-wrap: break-word;\n}\n.jhtml .modal , .jhtml.modal {\n  position: fixed;\n  top: 0;\n  left: 0;\n  z-index: 1055;\n  display: none;\n  width: 100%;\n  height: 100%;\n  overflow-x: hidden;\n  overflow-y: auto;\n  outline: 0;\n}\n.jhtml .modal-dialog , .jhtml.modal-dialog {\n  position: relative;\n  width: auto;\n  margin: 0.5rem;\n  pointer-events: none;\n}\n.jhtml .modal.fade .modal-dialog , .jhtml.modal.fade .modal-dialog {\n  transition: transform 0.3s ease-out;\n  transform: translate(0, -50px);\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .modal.fade .modal-dialog , .jhtml.modal.fade .modal-dialog {\n    transition: none;\n  }\n}\n.jhtml .modal.show .modal-dialog , .jhtml.modal.show .modal-dialog {\n  transform: none;\n}\n.jhtml .modal.modal-static .modal-dialog , .jhtml.modal.modal-static .modal-dialog {\n  transform: scale(1.02);\n}\n.jhtml .modal-dialog-scrollable , .jhtml.modal-dialog-scrollable {\n  height: calc(100% - 1rem);\n}\n.jhtml .modal-dialog-scrollable .modal-content , .jhtml.modal-dialog-scrollable .modal-content {\n  max-height: 100%;\n  overflow: hidden;\n}\n.jhtml .modal-dialog-scrollable .modal-body , .jhtml.modal-dialog-scrollable .modal-body {\n  overflow-y: auto;\n}\n.jhtml .modal-dialog-centered , .jhtml.modal-dialog-centered {\n  display: flex;\n  align-items: center;\n  min-height: calc(100% - 1rem);\n}\n.jhtml .modal-content , .jhtml.modal-content {\n  position: relative;\n  display: flex;\n  flex-direction: column;\n  width: 100%;\n  pointer-events: auto;\n  background-color: #fff;\n  background-clip: padding-box;\n  border: 1px solid rgba(0, 0, 0, 0.2);\n  border-radius: 0.3rem;\n  outline: 0;\n}\n.jhtml .modal-backdrop , .jhtml.modal-backdrop {\n  position: fixed;\n  top: 0;\n  left: 0;\n  z-index: 1050;\n  width: 100vw;\n  height: 100vh;\n  background-color: #000;\n}\n.jhtml .modal-backdrop.fade , .jhtml.modal-backdrop.fade {\n  opacity: 0;\n}\n.jhtml .modal-backdrop.show , .jhtml.modal-backdrop.show {\n  opacity: 0.5;\n}\n.jhtml .modal-header , .jhtml.modal-header {\n  display: flex;\n  flex-shrink: 0;\n  align-items: center;\n  justify-content: space-between;\n  padding: 1rem 1rem;\n  border-bottom: 1px solid #dee2e6;\n  border-top-left-radius: calc(0.3rem - 1px);\n  border-top-right-radius: calc(0.3rem - 1px);\n}\n.jhtml .modal-header .btn-close , .jhtml.modal-header .btn-close {\n  padding: 0.5rem 0.5rem;\n  margin: -0.5rem -0.5rem -0.5rem auto;\n}\n.jhtml .modal-title , .jhtml.modal-title {\n  margin-bottom: 0;\n  line-height: 1.5;\n}\n.jhtml .modal-body , .jhtml.modal-body {\n  position: relative;\n  flex: 1 1 auto;\n  padding: 1rem;\n}\n.jhtml .modal-footer , .jhtml.modal-footer {\n  display: flex;\n  flex-wrap: wrap;\n  flex-shrink: 0;\n  align-items: center;\n  justify-content: flex-end;\n  padding: 0.75rem;\n  border-top: 1px solid #dee2e6;\n  border-bottom-right-radius: calc(0.3rem - 1px);\n  border-bottom-left-radius: calc(0.3rem - 1px);\n}\n.jhtml .modal-footer > * , .jhtml.modal-footer > * {\n  margin: 0.25rem;\n}\n@media (min-width: 576px) {\n  .jhtml .modal-dialog , .jhtml.modal-dialog {\n    max-width: 500px;\n    margin: 1.75rem auto;\n  }\n  .jhtml .modal-dialog-scrollable , .jhtml.modal-dialog-scrollable {\n    height: calc(100% - 3.5rem);\n  }\n  .jhtml .modal-dialog-centered , .jhtml.modal-dialog-centered {\n    min-height: calc(100% - 3.5rem);\n  }\n  .jhtml .modal-sm , .jhtml.modal-sm {\n    max-width: 300px;\n  }\n}\n@media (min-width: 992px) {\n  .jhtml .modal-lg, \n.jhtml .modal-xl , .jhtml.modal-lg, \n.jhtml.modal-xl {\n    max-width: 800px;\n  }\n}\n@media (min-width: 1200px) {\n  .jhtml .modal-xl , .jhtml.modal-xl {\n    max-width: 1140px;\n  }\n}\n.jhtml .modal-fullscreen , .jhtml.modal-fullscreen {\n  width: 100vw;\n  max-width: none;\n  height: 100%;\n  margin: 0;\n}\n.jhtml .modal-fullscreen .modal-content , .jhtml.modal-fullscreen .modal-content {\n  height: 100%;\n  border: 0;\n  border-radius: 0;\n}\n.jhtml .modal-fullscreen .modal-header , .jhtml.modal-fullscreen .modal-header {\n  border-radius: 0;\n}\n.jhtml .modal-fullscreen .modal-body , .jhtml.modal-fullscreen .modal-body {\n  overflow-y: auto;\n}\n.jhtml .modal-fullscreen .modal-footer , .jhtml.modal-fullscreen .modal-footer {\n  border-radius: 0;\n}\n@media (max-width: 575.98px) {\n  .jhtml .modal-fullscreen-sm-down , .jhtml.modal-fullscreen-sm-down {\n    width: 100vw;\n    max-width: none;\n    height: 100%;\n    margin: 0;\n  }\n  .jhtml .modal-fullscreen-sm-down .modal-content , .jhtml.modal-fullscreen-sm-down .modal-content {\n    height: 100%;\n    border: 0;\n    border-radius: 0;\n  }\n  .jhtml .modal-fullscreen-sm-down .modal-header , .jhtml.modal-fullscreen-sm-down .modal-header {\n    border-radius: 0;\n  }\n  .jhtml .modal-fullscreen-sm-down .modal-body , .jhtml.modal-fullscreen-sm-down .modal-body {\n    overflow-y: auto;\n  }\n  .jhtml .modal-fullscreen-sm-down .modal-footer , .jhtml.modal-fullscreen-sm-down .modal-footer {\n    border-radius: 0;\n  }\n}\n@media (max-width: 767.98px) {\n  .jhtml .modal-fullscreen-md-down , .jhtml.modal-fullscreen-md-down {\n    width: 100vw;\n    max-width: none;\n    height: 100%;\n    margin: 0;\n  }\n  .jhtml .modal-fullscreen-md-down .modal-content , .jhtml.modal-fullscreen-md-down .modal-content {\n    height: 100%;\n    border: 0;\n    border-radius: 0;\n  }\n  .jhtml .modal-fullscreen-md-down .modal-header , .jhtml.modal-fullscreen-md-down .modal-header {\n    border-radius: 0;\n  }\n  .jhtml .modal-fullscreen-md-down .modal-body , .jhtml.modal-fullscreen-md-down .modal-body {\n    overflow-y: auto;\n  }\n  .jhtml .modal-fullscreen-md-down .modal-footer , .jhtml.modal-fullscreen-md-down .modal-footer {\n    border-radius: 0;\n  }\n}\n@media (max-width: 991.98px) {\n  .jhtml .modal-fullscreen-lg-down , .jhtml.modal-fullscreen-lg-down {\n    width: 100vw;\n    max-width: none;\n    height: 100%;\n    margin: 0;\n  }\n  .jhtml .modal-fullscreen-lg-down .modal-content , .jhtml.modal-fullscreen-lg-down .modal-content {\n    height: 100%;\n    border: 0;\n    border-radius: 0;\n  }\n  .jhtml .modal-fullscreen-lg-down .modal-header , .jhtml.modal-fullscreen-lg-down .modal-header {\n    border-radius: 0;\n  }\n  .jhtml .modal-fullscreen-lg-down .modal-body , .jhtml.modal-fullscreen-lg-down .modal-body {\n    overflow-y: auto;\n  }\n  .jhtml .modal-fullscreen-lg-down .modal-footer , .jhtml.modal-fullscreen-lg-down .modal-footer {\n    border-radius: 0;\n  }\n}\n@media (max-width: 1199.98px) {\n  .jhtml .modal-fullscreen-xl-down , .jhtml.modal-fullscreen-xl-down {\n    width: 100vw;\n    max-width: none;\n    height: 100%;\n    margin: 0;\n  }\n  .jhtml .modal-fullscreen-xl-down .modal-content , .jhtml.modal-fullscreen-xl-down .modal-content {\n    height: 100%;\n    border: 0;\n    border-radius: 0;\n  }\n  .jhtml .modal-fullscreen-xl-down .modal-header , .jhtml.modal-fullscreen-xl-down .modal-header {\n    border-radius: 0;\n  }\n  .jhtml .modal-fullscreen-xl-down .modal-body , .jhtml.modal-fullscreen-xl-down .modal-body {\n    overflow-y: auto;\n  }\n  .jhtml .modal-fullscreen-xl-down .modal-footer , .jhtml.modal-fullscreen-xl-down .modal-footer {\n    border-radius: 0;\n  }\n}\n@media (max-width: 1399.98px) {\n  .jhtml .modal-fullscreen-xxl-down , .jhtml.modal-fullscreen-xxl-down {\n    width: 100vw;\n    max-width: none;\n    height: 100%;\n    margin: 0;\n  }\n  .jhtml .modal-fullscreen-xxl-down .modal-content , .jhtml.modal-fullscreen-xxl-down .modal-content {\n    height: 100%;\n    border: 0;\n    border-radius: 0;\n  }\n  .jhtml .modal-fullscreen-xxl-down .modal-header , .jhtml.modal-fullscreen-xxl-down .modal-header {\n    border-radius: 0;\n  }\n  .jhtml .modal-fullscreen-xxl-down .modal-body , .jhtml.modal-fullscreen-xxl-down .modal-body {\n    overflow-y: auto;\n  }\n  .jhtml .modal-fullscreen-xxl-down .modal-footer , .jhtml.modal-fullscreen-xxl-down .modal-footer {\n    border-radius: 0;\n  }\n}\n.jhtml .tooltip , .jhtml.tooltip {\n  position: absolute;\n  z-index: 1080;\n  display: block;\n  margin: 0;\n  font-family: var(--bs-font-sans-serif);\n  font-style: normal;\n  font-weight: 400;\n  line-height: 1.5;\n  text-align: left;\n  text-align: start;\n  text-decoration: none;\n  text-shadow: none;\n  text-transform: none;\n  letter-spacing: normal;\n  word-break: normal;\n  word-spacing: normal;\n  white-space: normal;\n  line-break: auto;\n  font-size: 0.875rem;\n  word-wrap: break-word;\n  opacity: 0;\n}\n.jhtml .tooltip.show , .jhtml.tooltip.show {\n  opacity: 0.9;\n}\n.jhtml .tooltip .tooltip-arrow , .jhtml.tooltip .tooltip-arrow {\n  position: absolute;\n  display: block;\n  width: 0.8rem;\n  height: 0.4rem;\n}\n.jhtml .tooltip .tooltip-arrow::before , .jhtml.tooltip .tooltip-arrow::before {\n  position: absolute;\n  content: \"\";\n  border-color: transparent;\n  border-style: solid;\n}\n.jhtml .bs-tooltip-top,  .jhtml .bs-tooltip-auto[data-popper-placement^=top] , .jhtml.bs-tooltip-top,  .jhtml.bs-tooltip-auto[data-popper-placement^=top] {\n  padding: 0.4rem 0;\n}\n.jhtml .bs-tooltip-top .tooltip-arrow,  .jhtml .bs-tooltip-auto[data-popper-placement^=top] .tooltip-arrow , .jhtml.bs-tooltip-top .tooltip-arrow,  .jhtml.bs-tooltip-auto[data-popper-placement^=top] .tooltip-arrow {\n  bottom: 0;\n}\n.jhtml .bs-tooltip-top .tooltip-arrow::before,  .jhtml .bs-tooltip-auto[data-popper-placement^=top] .tooltip-arrow::before , .jhtml.bs-tooltip-top .tooltip-arrow::before,  .jhtml.bs-tooltip-auto[data-popper-placement^=top] .tooltip-arrow::before {\n  top: -1px;\n  border-width: 0.4rem 0.4rem 0;\n  border-top-color: #000;\n}\n.jhtml .bs-tooltip-end,  .jhtml .bs-tooltip-auto[data-popper-placement^=right] , .jhtml.bs-tooltip-end,  .jhtml.bs-tooltip-auto[data-popper-placement^=right] {\n  padding: 0 0.4rem;\n}\n.jhtml .bs-tooltip-end .tooltip-arrow,  .jhtml .bs-tooltip-auto[data-popper-placement^=right] .tooltip-arrow , .jhtml.bs-tooltip-end .tooltip-arrow,  .jhtml.bs-tooltip-auto[data-popper-placement^=right] .tooltip-arrow {\n  left: 0;\n  width: 0.4rem;\n  height: 0.8rem;\n}\n.jhtml .bs-tooltip-end .tooltip-arrow::before,  .jhtml .bs-tooltip-auto[data-popper-placement^=right] .tooltip-arrow::before , .jhtml.bs-tooltip-end .tooltip-arrow::before,  .jhtml.bs-tooltip-auto[data-popper-placement^=right] .tooltip-arrow::before {\n  right: -1px;\n  border-width: 0.4rem 0.4rem 0.4rem 0;\n  border-right-color: #000;\n}\n.jhtml .bs-tooltip-bottom,  .jhtml .bs-tooltip-auto[data-popper-placement^=bottom] , .jhtml.bs-tooltip-bottom,  .jhtml.bs-tooltip-auto[data-popper-placement^=bottom] {\n  padding: 0.4rem 0;\n}\n.jhtml .bs-tooltip-bottom .tooltip-arrow,  .jhtml .bs-tooltip-auto[data-popper-placement^=bottom] .tooltip-arrow , .jhtml.bs-tooltip-bottom .tooltip-arrow,  .jhtml.bs-tooltip-auto[data-popper-placement^=bottom] .tooltip-arrow {\n  top: 0;\n}\n.jhtml .bs-tooltip-bottom .tooltip-arrow::before,  .jhtml .bs-tooltip-auto[data-popper-placement^=bottom] .tooltip-arrow::before , .jhtml.bs-tooltip-bottom .tooltip-arrow::before,  .jhtml.bs-tooltip-auto[data-popper-placement^=bottom] .tooltip-arrow::before {\n  bottom: -1px;\n  border-width: 0 0.4rem 0.4rem;\n  border-bottom-color: #000;\n}\n.jhtml .bs-tooltip-start,  .jhtml .bs-tooltip-auto[data-popper-placement^=left] , .jhtml.bs-tooltip-start,  .jhtml.bs-tooltip-auto[data-popper-placement^=left] {\n  padding: 0 0.4rem;\n}\n.jhtml .bs-tooltip-start .tooltip-arrow,  .jhtml .bs-tooltip-auto[data-popper-placement^=left] .tooltip-arrow , .jhtml.bs-tooltip-start .tooltip-arrow,  .jhtml.bs-tooltip-auto[data-popper-placement^=left] .tooltip-arrow {\n  right: 0;\n  width: 0.4rem;\n  height: 0.8rem;\n}\n.jhtml .bs-tooltip-start .tooltip-arrow::before,  .jhtml .bs-tooltip-auto[data-popper-placement^=left] .tooltip-arrow::before , .jhtml.bs-tooltip-start .tooltip-arrow::before,  .jhtml.bs-tooltip-auto[data-popper-placement^=left] .tooltip-arrow::before {\n  left: -1px;\n  border-width: 0.4rem 0 0.4rem 0.4rem;\n  border-left-color: #000;\n}\n.jhtml .tooltip-inner , .jhtml.tooltip-inner {\n  max-width: 200px;\n  padding: 0.25rem 0.5rem;\n  color: #fff;\n  text-align: center;\n  background-color: #000;\n  border-radius: 0.25rem;\n}\n.jhtml .popover , .jhtml.popover {\n  position: absolute;\n  top: 0;\n  left: 0 /* rtl:ignore */;\n  z-index: 1070;\n  display: block;\n  max-width: 276px;\n  font-family: var(--bs-font-sans-serif);\n  font-style: normal;\n  font-weight: 400;\n  line-height: 1.5;\n  text-align: left;\n  text-align: start;\n  text-decoration: none;\n  text-shadow: none;\n  text-transform: none;\n  letter-spacing: normal;\n  word-break: normal;\n  word-spacing: normal;\n  white-space: normal;\n  line-break: auto;\n  font-size: 0.875rem;\n  word-wrap: break-word;\n  background-color: #fff;\n  background-clip: padding-box;\n  border: 1px solid rgba(0, 0, 0, 0.2);\n  border-radius: 0.3rem;\n}\n.jhtml .popover .popover-arrow , .jhtml.popover .popover-arrow {\n  position: absolute;\n  display: block;\n  width: 1rem;\n  height: 0.5rem;\n}\n.jhtml .popover .popover-arrow::before,  .jhtml .popover .popover-arrow::after , .jhtml.popover .popover-arrow::before,  .jhtml.popover .popover-arrow::after {\n  position: absolute;\n  display: block;\n  content: \"\";\n  border-color: transparent;\n  border-style: solid;\n}\n.jhtml .bs-popover-top > .popover-arrow,  .jhtml .bs-popover-auto[data-popper-placement^=top] > .popover-arrow , .jhtml.bs-popover-top > .popover-arrow,  .jhtml.bs-popover-auto[data-popper-placement^=top] > .popover-arrow {\n  bottom: calc(-0.5rem - 1px);\n}\n.jhtml .bs-popover-top > .popover-arrow::before,  .jhtml .bs-popover-auto[data-popper-placement^=top] > .popover-arrow::before , .jhtml.bs-popover-top > .popover-arrow::before,  .jhtml.bs-popover-auto[data-popper-placement^=top] > .popover-arrow::before {\n  bottom: 0;\n  border-width: 0.5rem 0.5rem 0;\n  border-top-color: rgba(0, 0, 0, 0.25);\n}\n.jhtml .bs-popover-top > .popover-arrow::after,  .jhtml .bs-popover-auto[data-popper-placement^=top] > .popover-arrow::after , .jhtml.bs-popover-top > .popover-arrow::after,  .jhtml.bs-popover-auto[data-popper-placement^=top] > .popover-arrow::after {\n  bottom: 1px;\n  border-width: 0.5rem 0.5rem 0;\n  border-top-color: #fff;\n}\n.jhtml .bs-popover-end > .popover-arrow,  .jhtml .bs-popover-auto[data-popper-placement^=right] > .popover-arrow , .jhtml.bs-popover-end > .popover-arrow,  .jhtml.bs-popover-auto[data-popper-placement^=right] > .popover-arrow {\n  left: calc(-0.5rem - 1px);\n  width: 0.5rem;\n  height: 1rem;\n}\n.jhtml .bs-popover-end > .popover-arrow::before,  .jhtml .bs-popover-auto[data-popper-placement^=right] > .popover-arrow::before , .jhtml.bs-popover-end > .popover-arrow::before,  .jhtml.bs-popover-auto[data-popper-placement^=right] > .popover-arrow::before {\n  left: 0;\n  border-width: 0.5rem 0.5rem 0.5rem 0;\n  border-right-color: rgba(0, 0, 0, 0.25);\n}\n.jhtml .bs-popover-end > .popover-arrow::after,  .jhtml .bs-popover-auto[data-popper-placement^=right] > .popover-arrow::after , .jhtml.bs-popover-end > .popover-arrow::after,  .jhtml.bs-popover-auto[data-popper-placement^=right] > .popover-arrow::after {\n  left: 1px;\n  border-width: 0.5rem 0.5rem 0.5rem 0;\n  border-right-color: #fff;\n}\n.jhtml .bs-popover-bottom > .popover-arrow,  .jhtml .bs-popover-auto[data-popper-placement^=bottom] > .popover-arrow , .jhtml.bs-popover-bottom > .popover-arrow,  .jhtml.bs-popover-auto[data-popper-placement^=bottom] > .popover-arrow {\n  top: calc(-0.5rem - 1px);\n}\n.jhtml .bs-popover-bottom > .popover-arrow::before,  .jhtml .bs-popover-auto[data-popper-placement^=bottom] > .popover-arrow::before , .jhtml.bs-popover-bottom > .popover-arrow::before,  .jhtml.bs-popover-auto[data-popper-placement^=bottom] > .popover-arrow::before {\n  top: 0;\n  border-width: 0 0.5rem 0.5rem 0.5rem;\n  border-bottom-color: rgba(0, 0, 0, 0.25);\n}\n.jhtml .bs-popover-bottom > .popover-arrow::after,  .jhtml .bs-popover-auto[data-popper-placement^=bottom] > .popover-arrow::after , .jhtml.bs-popover-bottom > .popover-arrow::after,  .jhtml.bs-popover-auto[data-popper-placement^=bottom] > .popover-arrow::after {\n  top: 1px;\n  border-width: 0 0.5rem 0.5rem 0.5rem;\n  border-bottom-color: #fff;\n}\n.jhtml .bs-popover-bottom .popover-header::before,  .jhtml .bs-popover-auto[data-popper-placement^=bottom] .popover-header::before , .jhtml.bs-popover-bottom .popover-header::before,  .jhtml.bs-popover-auto[data-popper-placement^=bottom] .popover-header::before {\n  position: absolute;\n  top: 0;\n  left: 50%;\n  display: block;\n  width: 1rem;\n  margin-left: -0.5rem;\n  content: \"\";\n  border-bottom: 1px solid #f0f0f0;\n}\n.jhtml .bs-popover-start > .popover-arrow,  .jhtml .bs-popover-auto[data-popper-placement^=left] > .popover-arrow , .jhtml.bs-popover-start > .popover-arrow,  .jhtml.bs-popover-auto[data-popper-placement^=left] > .popover-arrow {\n  right: calc(-0.5rem - 1px);\n  width: 0.5rem;\n  height: 1rem;\n}\n.jhtml .bs-popover-start > .popover-arrow::before,  .jhtml .bs-popover-auto[data-popper-placement^=left] > .popover-arrow::before , .jhtml.bs-popover-start > .popover-arrow::before,  .jhtml.bs-popover-auto[data-popper-placement^=left] > .popover-arrow::before {\n  right: 0;\n  border-width: 0.5rem 0 0.5rem 0.5rem;\n  border-left-color: rgba(0, 0, 0, 0.25);\n}\n.jhtml .bs-popover-start > .popover-arrow::after,  .jhtml .bs-popover-auto[data-popper-placement^=left] > .popover-arrow::after , .jhtml.bs-popover-start > .popover-arrow::after,  .jhtml.bs-popover-auto[data-popper-placement^=left] > .popover-arrow::after {\n  right: 1px;\n  border-width: 0.5rem 0 0.5rem 0.5rem;\n  border-left-color: #fff;\n}\n.jhtml .popover-header , .jhtml.popover-header {\n  padding: 0.5rem 1rem;\n  margin-bottom: 0;\n  font-size: 1rem;\n  background-color: #f0f0f0;\n  border-bottom: 1px solid rgba(0, 0, 0, 0.2);\n  border-top-left-radius: calc(0.3rem - 1px);\n  border-top-right-radius: calc(0.3rem - 1px);\n}\n.jhtml .popover-header:empty , .jhtml.popover-header:empty {\n  display: none;\n}\n.jhtml .popover-body , .jhtml.popover-body {\n  padding: 1rem 1rem;\n  color: #212529;\n}\n.jhtml .carousel , .jhtml.carousel {\n  position: relative;\n}\n.jhtml .carousel.pointer-event , .jhtml.carousel.pointer-event {\n  touch-action: pan-y;\n}\n.jhtml .carousel-inner , .jhtml.carousel-inner {\n  position: relative;\n  width: 100%;\n  overflow: hidden;\n}\n.jhtml .carousel-inner::after , .jhtml.carousel-inner::after {\n  display: block;\n  clear: both;\n  content: \"\";\n}\n.jhtml .carousel-item , .jhtml.carousel-item {\n  position: relative;\n  display: none;\n  float: left;\n  width: 100%;\n  margin-right: -100%;\n  backface-visibility: hidden;\n  transition: transform 0.6s ease-in-out;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .carousel-item , .jhtml.carousel-item {\n    transition: none;\n  }\n}\n.jhtml .carousel-item.active, \n.jhtml .carousel-item-next, \n.jhtml .carousel-item-prev , .jhtml.carousel-item.active, \n.jhtml.carousel-item-next, \n.jhtml.carousel-item-prev {\n  display: block;\n}\n.jhtml .carousel-item-next:not(.carousel-item-start), \n.jhtml .active.carousel-item-end , .jhtml.carousel-item-next:not(.carousel-item-start), \n.jhtml.active.carousel-item-end {\n  transform: translateX(100%);\n}\n.jhtml .carousel-item-prev:not(.carousel-item-end), \n.jhtml .active.carousel-item-start , .jhtml.carousel-item-prev:not(.carousel-item-end), \n.jhtml.active.carousel-item-start {\n  transform: translateX(-100%);\n}\n.jhtml .carousel-fade .carousel-item , .jhtml.carousel-fade .carousel-item {\n  opacity: 0;\n  transition-property: opacity;\n  transform: none;\n}\n.jhtml .carousel-fade .carousel-item.active, \n.jhtml .carousel-fade .carousel-item-next.carousel-item-start, \n.jhtml .carousel-fade .carousel-item-prev.carousel-item-end , .jhtml.carousel-fade .carousel-item.active, \n.jhtml.carousel-fade .carousel-item-next.carousel-item-start, \n.jhtml.carousel-fade .carousel-item-prev.carousel-item-end {\n  z-index: 1;\n  opacity: 1;\n}\n.jhtml .carousel-fade .active.carousel-item-start, \n.jhtml .carousel-fade .active.carousel-item-end , .jhtml.carousel-fade .active.carousel-item-start, \n.jhtml.carousel-fade .active.carousel-item-end {\n  z-index: 0;\n  opacity: 0;\n  transition: opacity 0s 0.6s;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .carousel-fade .active.carousel-item-start, \n.jhtml .carousel-fade .active.carousel-item-end , .jhtml.carousel-fade .active.carousel-item-start, \n.jhtml.carousel-fade .active.carousel-item-end {\n    transition: none;\n  }\n}\n.jhtml .carousel-control-prev, \n.jhtml .carousel-control-next , .jhtml.carousel-control-prev, \n.jhtml.carousel-control-next {\n  position: absolute;\n  top: 0;\n  bottom: 0;\n  z-index: 1;\n  display: flex;\n  align-items: center;\n  justify-content: center;\n  width: 15%;\n  padding: 0;\n  color: #fff;\n  text-align: center;\n  background: none;\n  border: 0;\n  opacity: 0.5;\n  transition: opacity 0.15s ease;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .carousel-control-prev, \n.jhtml .carousel-control-next , .jhtml.carousel-control-prev, \n.jhtml.carousel-control-next {\n    transition: none;\n  }\n}\n.jhtml .carousel-control-prev:hover,  .jhtml .carousel-control-prev:focus, \n.jhtml .carousel-control-next:hover, \n.jhtml .carousel-control-next:focus , .jhtml.carousel-control-prev:hover,  .jhtml.carousel-control-prev:focus, \n.jhtml.carousel-control-next:hover, \n.jhtml.carousel-control-next:focus {\n  color: #fff;\n  text-decoration: none;\n  outline: 0;\n  opacity: 0.9;\n}\n.jhtml .carousel-control-prev , .jhtml.carousel-control-prev {\n  left: 0;\n}\n.jhtml .carousel-control-next , .jhtml.carousel-control-next {\n  right: 0;\n}\n.jhtml .carousel-control-prev-icon, \n.jhtml .carousel-control-next-icon , .jhtml.carousel-control-prev-icon, \n.jhtml.carousel-control-next-icon {\n  display: inline-block;\n  width: 2rem;\n  height: 2rem;\n  background-repeat: no-repeat;\n  background-position: 50%;\n  background-size: 100% 100%;\n}\n.jhtml .carousel-control-prev-icon , .jhtml.carousel-control-prev-icon {\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 16 16' fill='%23fff'%3e%3cpath d='M11.354 1.646a.5.5 0 0 1 0 .708L5.707 8l5.647 5.646a.5.5 0 0 1-.708.708l-6-6a.5.5 0 0 1 0-.708l6-6a.5.5 0 0 1 .708 0z'/%3e%3c/svg%3e\");\n}\n.jhtml .carousel-control-next-icon , .jhtml.carousel-control-next-icon {\n  background-image: url(\"data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 16 16' fill='%23fff'%3e%3cpath d='M4.646 1.646a.5.5 0 0 1 .708 0l6 6a.5.5 0 0 1 0 .708l-6 6a.5.5 0 0 1-.708-.708L10.293 8 4.646 2.354a.5.5 0 0 1 0-.708z'/%3e%3c/svg%3e\");\n}\n.jhtml .carousel-indicators , .jhtml.carousel-indicators {\n  position: absolute;\n  right: 0;\n  bottom: 0;\n  left: 0;\n  z-index: 2;\n  display: flex;\n  justify-content: center;\n  padding: 0;\n  margin-right: 15%;\n  margin-bottom: 1rem;\n  margin-left: 15%;\n  list-style: none;\n}\n.jhtml .carousel-indicators [data-bs-target] , .jhtml.carousel-indicators [data-bs-target] {\n  box-sizing: content-box;\n  flex: 0 1 auto;\n  width: 30px;\n  height: 3px;\n  padding: 0;\n  margin-right: 3px;\n  margin-left: 3px;\n  text-indent: -999px;\n  cursor: pointer;\n  background-color: #fff;\n  background-clip: padding-box;\n  border: 0;\n  border-top: 10px solid transparent;\n  border-bottom: 10px solid transparent;\n  opacity: 0.5;\n  transition: opacity 0.6s ease;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .carousel-indicators [data-bs-target] , .jhtml.carousel-indicators [data-bs-target] {\n    transition: none;\n  }\n}\n.jhtml .carousel-indicators .active , .jhtml.carousel-indicators .active {\n  opacity: 1;\n}\n.jhtml .carousel-caption , .jhtml.carousel-caption {\n  position: absolute;\n  right: 15%;\n  bottom: 1.25rem;\n  left: 15%;\n  padding-top: 1.25rem;\n  padding-bottom: 1.25rem;\n  color: #fff;\n  text-align: center;\n}\n.jhtml .carousel-dark .carousel-control-prev-icon, \n.jhtml .carousel-dark .carousel-control-next-icon , .jhtml.carousel-dark .carousel-control-prev-icon, \n.jhtml.carousel-dark .carousel-control-next-icon {\n  filter: invert(1) grayscale(100);\n}\n.jhtml .carousel-dark .carousel-indicators [data-bs-target] , .jhtml.carousel-dark .carousel-indicators [data-bs-target] {\n  background-color: #000;\n}\n.jhtml .carousel-dark .carousel-caption , .jhtml.carousel-dark .carousel-caption {\n  color: #000;\n}\n@keyframes spinner-border {\n  to {\n    transform: rotate(360deg) /* rtl:ignore */;\n  }\n}\n.jhtml .spinner-border , .jhtml.spinner-border {\n  display: inline-block;\n  width: 2rem;\n  height: 2rem;\n  vertical-align: -0.125em;\n  border: 0.25em solid currentColor;\n  border-right-color: transparent;\n  border-radius: 50%;\n  animation: 0.75s linear infinite spinner-border;\n}\n.jhtml .spinner-border-sm , .jhtml.spinner-border-sm {\n  width: 1rem;\n  height: 1rem;\n  border-width: 0.2em;\n}\n@keyframes spinner-grow {\n  0% {\n    transform: scale(0);\n  }\n  50% {\n    opacity: 1;\n    transform: none;\n  }\n}\n.jhtml .spinner-grow , .jhtml.spinner-grow {\n  display: inline-block;\n  width: 2rem;\n  height: 2rem;\n  vertical-align: -0.125em;\n  background-color: currentColor;\n  border-radius: 50%;\n  opacity: 0;\n  animation: 0.75s linear infinite spinner-grow;\n}\n.jhtml .spinner-grow-sm , .jhtml.spinner-grow-sm {\n  width: 1rem;\n  height: 1rem;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .spinner-border, \n.jhtml .spinner-grow , .jhtml.spinner-border, \n.jhtml.spinner-grow {\n    animation-duration: 1.5s;\n  }\n}\n.jhtml .offcanvas , .jhtml.offcanvas {\n  position: fixed;\n  bottom: 0;\n  z-index: 1045;\n  display: flex;\n  flex-direction: column;\n  max-width: 100%;\n  visibility: hidden;\n  background-color: #fff;\n  background-clip: padding-box;\n  outline: 0;\n  transition: transform 0.3s ease-in-out;\n}\n@media (prefers-reduced-motion: reduce) {\n  .jhtml .offcanvas , .jhtml.offcanvas {\n    transition: none;\n  }\n}\n.jhtml .offcanvas-backdrop , .jhtml.offcanvas-backdrop {\n  position: fixed;\n  top: 0;\n  left: 0;\n  z-index: 1040;\n  width: 100vw;\n  height: 100vh;\n  background-color: #000;\n}\n.jhtml .offcanvas-backdrop.fade , .jhtml.offcanvas-backdrop.fade {\n  opacity: 0;\n}\n.jhtml .offcanvas-backdrop.show , .jhtml.offcanvas-backdrop.show {\n  opacity: 0.5;\n}\n.jhtml .offcanvas-header , .jhtml.offcanvas-header {\n  display: flex;\n  align-items: center;\n  justify-content: space-between;\n  padding: 1rem 1rem;\n}\n.jhtml .offcanvas-header .btn-close , .jhtml.offcanvas-header .btn-close {\n  padding: 0.5rem 0.5rem;\n  margin-top: -0.5rem;\n  margin-right: -0.5rem;\n  margin-bottom: -0.5rem;\n}\n.jhtml .offcanvas-title , .jhtml.offcanvas-title {\n  margin-bottom: 0;\n  line-height: 1.5;\n}\n.jhtml .offcanvas-body , .jhtml.offcanvas-body {\n  flex-grow: 1;\n  padding: 1rem 1rem;\n  overflow-y: auto;\n}\n.jhtml .offcanvas-start , .jhtml.offcanvas-start {\n  top: 0;\n  left: 0;\n  width: 400px;\n  border-right: 1px solid rgba(0, 0, 0, 0.2);\n  transform: translateX(-100%);\n}\n.jhtml .offcanvas-end , .jhtml.offcanvas-end {\n  top: 0;\n  right: 0;\n  width: 400px;\n  border-left: 1px solid rgba(0, 0, 0, 0.2);\n  transform: translateX(100%);\n}\n.jhtml .offcanvas-top , .jhtml.offcanvas-top {\n  top: 0;\n  right: 0;\n  left: 0;\n  height: 30vh;\n  max-height: 100%;\n  border-bottom: 1px solid rgba(0, 0, 0, 0.2);\n  transform: translateY(-100%);\n}\n.jhtml .offcanvas-bottom , .jhtml.offcanvas-bottom {\n  right: 0;\n  left: 0;\n  height: 30vh;\n  max-height: 100%;\n  border-top: 1px solid rgba(0, 0, 0, 0.2);\n  transform: translateY(100%);\n}\n.jhtml .offcanvas.show , .jhtml.offcanvas.show {\n  transform: none;\n}\n.jhtml .placeholder , .jhtml.placeholder {\n  display: inline-block;\n  min-height: 1em;\n  vertical-align: middle;\n  cursor: wait;\n  background-color: currentColor;\n  opacity: 0.5;\n}\n.jhtml .placeholder.btn::before , .jhtml.placeholder.btn::before {\n  display: inline-block;\n  content: \"\";\n}\n.jhtml .placeholder-xs , .jhtml.placeholder-xs {\n  min-height: 0.6em;\n}\n.jhtml .placeholder-sm , .jhtml.placeholder-sm {\n  min-height: 0.8em;\n}\n.jhtml .placeholder-lg , .jhtml.placeholder-lg {\n  min-height: 1.2em;\n}\n.jhtml .placeholder-glow .placeholder , .jhtml.placeholder-glow .placeholder {\n  animation: placeholder-glow 2s ease-in-out infinite;\n}\n@keyframes placeholder-glow {\n  50% {\n    opacity: 0.2;\n  }\n}\n.jhtml .placeholder-wave , .jhtml.placeholder-wave {\n  mask-image: linear-gradient(130deg, #000 55%, rgba(0, 0, 0, 0.8) 75%, #000 95%);\n  mask-size: 200% 100%;\n  animation: placeholder-wave 2s linear infinite;\n}\n@keyframes placeholder-wave {\n  100% {\n    mask-position: -200% 0%;\n  }\n}\n.jhtml .clearfix::after , .jhtml.clearfix::after {\n  display: block;\n  clear: both;\n  content: \"\";\n}\n.jhtml .link-primary , .jhtml.link-primary {\n  color: #0d6efd;\n}\n.jhtml .link-primary:hover,  .jhtml .link-primary:focus , .jhtml.link-primary:hover,  .jhtml.link-primary:focus {\n  color: #0a58ca;\n}\n.jhtml .link-secondary , .jhtml.link-secondary {\n  color: #6c757d;\n}\n.jhtml .link-secondary:hover,  .jhtml .link-secondary:focus , .jhtml.link-secondary:hover,  .jhtml.link-secondary:focus {\n  color: #565e64;\n}\n.jhtml .link-success , .jhtml.link-success {\n  color: #198754;\n}\n.jhtml .link-success:hover,  .jhtml .link-success:focus , .jhtml.link-success:hover,  .jhtml.link-success:focus {\n  color: #146c43;\n}\n.jhtml .link-info , .jhtml.link-info {\n  color: #0dcaf0;\n}\n.jhtml .link-info:hover,  .jhtml .link-info:focus , .jhtml.link-info:hover,  .jhtml.link-info:focus {\n  color: #3dd5f3;\n}\n.jhtml .link-warning , .jhtml.link-warning {\n  color: #ffc107;\n}\n.jhtml .link-warning:hover,  .jhtml .link-warning:focus , .jhtml.link-warning:hover,  .jhtml.link-warning:focus {\n  color: #ffcd39;\n}\n.jhtml .link-danger , .jhtml.link-danger {\n  color: #dc3545;\n}\n.jhtml .link-danger:hover,  .jhtml .link-danger:focus , .jhtml.link-danger:hover,  .jhtml.link-danger:focus {\n  color: #b02a37;\n}\n.jhtml .link-light , .jhtml.link-light {\n  color: #f8f9fa;\n}\n.jhtml .link-light:hover,  .jhtml .link-light:focus , .jhtml.link-light:hover,  .jhtml.link-light:focus {\n  color: #f9fafb;\n}\n.jhtml .link-dark , .jhtml.link-dark {\n  color: #212529;\n}\n.jhtml .link-dark:hover,  .jhtml .link-dark:focus , .jhtml.link-dark:hover,  .jhtml.link-dark:focus {\n  color: #1a1e21;\n}\n.jhtml .ratio , .jhtml.ratio {\n  position: relative;\n  width: 100%;\n}\n.jhtml .ratio::before , .jhtml.ratio::before {\n  display: block;\n  padding-top: var(--bs-aspect-ratio);\n  content: \"\";\n}\n.jhtml .ratio > * , .jhtml.ratio > * {\n  position: absolute;\n  top: 0;\n  left: 0;\n  width: 100%;\n  height: 100%;\n}\n.jhtml .ratio-1x1 , .jhtml.ratio-1x1 {\n  --bs-aspect-ratio: 100%;\n}\n.jhtml .ratio-4x3 , .jhtml.ratio-4x3 {\n  --bs-aspect-ratio: 75%;\n}\n.jhtml .ratio-16x9 , .jhtml.ratio-16x9 {\n  --bs-aspect-ratio: 56.25%;\n}\n.jhtml .ratio-21x9 , .jhtml.ratio-21x9 {\n  --bs-aspect-ratio: 42.8571428571%;\n}\n.jhtml .fixed-top , .jhtml.fixed-top {\n  position: fixed;\n  top: 0;\n  right: 0;\n  left: 0;\n  z-index: 1030;\n}\n.jhtml .fixed-bottom , .jhtml.fixed-bottom {\n  position: fixed;\n  right: 0;\n  bottom: 0;\n  left: 0;\n  z-index: 1030;\n}\n.jhtml .sticky-top , .jhtml.sticky-top {\n  position: sticky;\n  top: 0;\n  z-index: 1020;\n}\n@media (min-width: 576px) {\n  .jhtml .sticky-sm-top , .jhtml.sticky-sm-top {\n    position: sticky;\n    top: 0;\n    z-index: 1020;\n  }\n}\n@media (min-width: 768px) {\n  .jhtml .sticky-md-top , .jhtml.sticky-md-top {\n    position: sticky;\n    top: 0;\n    z-index: 1020;\n  }\n}\n@media (min-width: 992px) {\n  .jhtml .sticky-lg-top , .jhtml.sticky-lg-top {\n    position: sticky;\n    top: 0;\n    z-index: 1020;\n  }\n}\n@media (min-width: 1200px) {\n  .jhtml .sticky-xl-top , .jhtml.sticky-xl-top {\n    position: sticky;\n    top: 0;\n    z-index: 1020;\n  }\n}\n@media (min-width: 1400px) {\n  .jhtml .sticky-xxl-top , .jhtml.sticky-xxl-top {\n    position: sticky;\n    top: 0;\n    z-index: 1020;\n  }\n}\n.jhtml .hstack , .jhtml.hstack {\n  display: flex;\n  flex-direction: row;\n  align-items: center;\n  align-self: stretch;\n}\n.jhtml .vstack , .jhtml.vstack {\n  display: flex;\n  flex: 1 1 auto;\n  flex-direction: column;\n  align-self: stretch;\n}\n.jhtml .visually-hidden, \n.jhtml .visually-hidden-focusable:not(:focus):not(:focus-within) , .jhtml.visually-hidden, \n.jhtml.visually-hidden-focusable:not(:focus):not(:focus-within) {\n  position: absolute !important;\n  width: 1px !important;\n  height: 1px !important;\n  padding: 0 !important;\n  margin: -1px !important;\n  overflow: hidden !important;\n  clip: rect(0, 0, 0, 0) !important;\n  white-space: nowrap !important;\n  border: 0 !important;\n}\n.jhtml .stretched-link::after , .jhtml.stretched-link::after {\n  position: absolute;\n  top: 0;\n  right: 0;\n  bottom: 0;\n  left: 0;\n  z-index: 1;\n  content: \"\";\n}\n.jhtml .text-truncate , .jhtml.text-truncate {\n  overflow: hidden;\n  text-overflow: ellipsis;\n  white-space: nowrap;\n}\n.jhtml .vr , .jhtml.vr {\n  display: inline-block;\n  align-self: stretch;\n  width: 1px;\n  min-height: 1em;\n  background-color: currentColor;\n  opacity: 0.25;\n}\n.jhtml .align-baseline , .jhtml.align-baseline {\n  vertical-align: baseline !important;\n}\n.jhtml .align-top , .jhtml.align-top {\n  vertical-align: top !important;\n}\n.jhtml .align-middle , .jhtml.align-middle {\n  vertical-align: middle !important;\n}\n.jhtml .align-bottom , .jhtml.align-bottom {\n  vertical-align: bottom !important;\n}\n.jhtml .align-text-bottom , .jhtml.align-text-bottom {\n  vertical-align: text-bottom !important;\n}\n.jhtml .align-text-top , .jhtml.align-text-top {\n  vertical-align: text-top !important;\n}\n.jhtml .float-start , .jhtml.float-start {\n  float: left !important;\n}\n.jhtml .float-end , .jhtml.float-end {\n  float: right !important;\n}\n.jhtml .float-none , .jhtml.float-none {\n  float: none !important;\n}\n.jhtml .opacity-0 , .jhtml.opacity-0 {\n  opacity: 0 !important;\n}\n.jhtml .opacity-25 , .jhtml.opacity-25 {\n  opacity: 0.25 !important;\n}\n.jhtml .opacity-50 , .jhtml.opacity-50 {\n  opacity: 0.5 !important;\n}\n.jhtml .opacity-75 , .jhtml.opacity-75 {\n  opacity: 0.75 !important;\n}\n.jhtml .opacity-100 , .jhtml.opacity-100 {\n  opacity: 1 !important;\n}\n.jhtml .overflow-auto , .jhtml.overflow-auto {\n  overflow: auto !important;\n}\n.jhtml .overflow-hidden , .jhtml.overflow-hidden {\n  overflow: hidden !important;\n}\n.jhtml .overflow-visible , .jhtml.overflow-visible {\n  overflow: visible !important;\n}\n.jhtml .overflow-scroll , .jhtml.overflow-scroll {\n  overflow: scroll !important;\n}\n.jhtml .d-inline , .jhtml.d-inline {\n  display: inline !important;\n}\n.jhtml .d-inline-block , .jhtml.d-inline-block {\n  display: inline-block !important;\n}\n.jhtml .d-block , .jhtml.d-block {\n  display: block !important;\n}\n.jhtml .d-grid , .jhtml.d-grid {\n  display: grid !important;\n}\n.jhtml .d-table , .jhtml.d-table {\n  display: table !important;\n}\n.jhtml .d-table-row , .jhtml.d-table-row {\n  display: table-row !important;\n}\n.jhtml .d-table-cell , .jhtml.d-table-cell {\n  display: table-cell !important;\n}\n.jhtml .d-flex , .jhtml.d-flex {\n  display: flex !important;\n}\n.jhtml .d-inline-flex , .jhtml.d-inline-flex {\n  display: inline-flex !important;\n}\n.jhtml .d-none , .jhtml.d-none {\n  display: none !important;\n}\n.jhtml .shadow , .jhtml.shadow {\n  box-shadow: 0 0.5rem 1rem rgba(0, 0, 0, 0.15) !important;\n}\n.jhtml .shadow-sm , .jhtml.shadow-sm {\n  box-shadow: 0 0.125rem 0.25rem rgba(0, 0, 0, 0.075) !important;\n}\n.jhtml .shadow-lg , .jhtml.shadow-lg {\n  box-shadow: 0 1rem 3rem rgba(0, 0, 0, 0.175) !important;\n}\n.jhtml .shadow-none , .jhtml.shadow-none {\n  box-shadow: none !important;\n}\n.jhtml .position-static , .jhtml.position-static {\n  position: static !important;\n}\n.jhtml .position-relative , .jhtml.position-relative {\n  position: relative !important;\n}\n.jhtml .position-absolute , .jhtml.position-absolute {\n  position: absolute !important;\n}\n.jhtml .position-fixed , .jhtml.position-fixed {\n  position: fixed !important;\n}\n.jhtml .position-sticky , .jhtml.position-sticky {\n  position: sticky !important;\n}\n.jhtml .top-0 , .jhtml.top-0 {\n  top: 0 !important;\n}\n.jhtml .top-50 , .jhtml.top-50 {\n  top: 50% !important;\n}\n.jhtml .top-100 , .jhtml.top-100 {\n  top: 100% !important;\n}\n.jhtml .bottom-0 , .jhtml.bottom-0 {\n  bottom: 0 !important;\n}\n.jhtml .bottom-50 , .jhtml.bottom-50 {\n  bottom: 50% !important;\n}\n.jhtml .bottom-100 , .jhtml.bottom-100 {\n  bottom: 100% !important;\n}\n.jhtml .start-0 , .jhtml.start-0 {\n  left: 0 !important;\n}\n.jhtml .start-50 , .jhtml.start-50 {\n  left: 50% !important;\n}\n.jhtml .start-100 , .jhtml.start-100 {\n  left: 100% !important;\n}\n.jhtml .end-0 , .jhtml.end-0 {\n  right: 0 !important;\n}\n.jhtml .end-50 , .jhtml.end-50 {\n  right: 50% !important;\n}\n.jhtml .end-100 , .jhtml.end-100 {\n  right: 100% !important;\n}\n.jhtml .translate-middle , .jhtml.translate-middle {\n  transform: translate(-50%, -50%) !important;\n}\n.jhtml .translate-middle-x , .jhtml.translate-middle-x {\n  transform: translateX(-50%) !important;\n}\n.jhtml .translate-middle-y , .jhtml.translate-middle-y {\n  transform: translateY(-50%) !important;\n}\n.jhtml .border , .jhtml.border {\n  border: 1px solid #dee2e6 !important;\n}\n.jhtml .border-0 , .jhtml.border-0 {\n  border: 0 !important;\n}\n.jhtml .border-top , .jhtml.border-top {\n  border-top: 1px solid #dee2e6 !important;\n}\n.jhtml .border-top-0 , .jhtml.border-top-0 {\n  border-top: 0 !important;\n}\n.jhtml .border-end , .jhtml.border-end {\n  border-right: 1px solid #dee2e6 !important;\n}\n.jhtml .border-end-0 , .jhtml.border-end-0 {\n  border-right: 0 !important;\n}\n.jhtml .border-bottom , .jhtml.border-bottom {\n  border-bottom: 1px solid #dee2e6 !important;\n}\n.jhtml .border-bottom-0 , .jhtml.border-bottom-0 {\n  border-bottom: 0 !important;\n}\n.jhtml .border-start , .jhtml.border-start {\n  border-left: 1px solid #dee2e6 !important;\n}\n.jhtml .border-start-0 , .jhtml.border-start-0 {\n  border-left: 0 !important;\n}\n.jhtml .border-primary , .jhtml.border-primary {\n  border-color: #0d6efd !important;\n}\n.jhtml .border-secondary , .jhtml.border-secondary {\n  border-color: #6c757d !important;\n}\n.jhtml .border-success , .jhtml.border-success {\n  border-color: #198754 !important;\n}\n.jhtml .border-info , .jhtml.border-info {\n  border-color: #0dcaf0 !important;\n}\n.jhtml .border-warning , .jhtml.border-warning {\n  border-color: #ffc107 !important;\n}\n.jhtml .border-danger , .jhtml.border-danger {\n  border-color: #dc3545 !important;\n}\n.jhtml .border-light , .jhtml.border-light {\n  border-color: #f8f9fa !important;\n}\n.jhtml .border-dark , .jhtml.border-dark {\n  border-color: #212529 !important;\n}\n.jhtml .border-white , .jhtml.border-white {\n  border-color: #fff !important;\n}\n.jhtml .border-1 , .jhtml.border-1 {\n  border-width: 1px !important;\n}\n.jhtml .border-2 , .jhtml.border-2 {\n  border-width: 2px !important;\n}\n.jhtml .border-3 , .jhtml.border-3 {\n  border-width: 3px !important;\n}\n.jhtml .border-4 , .jhtml.border-4 {\n  border-width: 4px !important;\n}\n.jhtml .border-5 , .jhtml.border-5 {\n  border-width: 5px !important;\n}\n.jhtml .w-25 , .jhtml.w-25 {\n  width: 25% !important;\n}\n.jhtml .w-50 , .jhtml.w-50 {\n  width: 50% !important;\n}\n.jhtml .w-75 , .jhtml.w-75 {\n  width: 75% !important;\n}\n.jhtml .w-100 , .jhtml.w-100 {\n  width: 100% !important;\n}\n.jhtml .w-auto , .jhtml.w-auto {\n  width: auto !important;\n}\n.jhtml .mw-100 , .jhtml.mw-100 {\n  max-width: 100% !important;\n}\n.jhtml .vw-100 , .jhtml.vw-100 {\n  width: 100vw !important;\n}\n.jhtml .min-vw-100 , .jhtml.min-vw-100 {\n  min-width: 100vw !important;\n}\n.jhtml .h-25 , .jhtml.h-25 {\n  height: 25% !important;\n}\n.jhtml .h-50 , .jhtml.h-50 {\n  height: 50% !important;\n}\n.jhtml .h-75 , .jhtml.h-75 {\n  height: 75% !important;\n}\n.jhtml .h-100 , .jhtml.h-100 {\n  height: 100% !important;\n}\n.jhtml .h-auto , .jhtml.h-auto {\n  height: auto !important;\n}\n.jhtml .mh-100 , .jhtml.mh-100 {\n  max-height: 100% !important;\n}\n.jhtml .vh-100 , .jhtml.vh-100 {\n  height: 100vh !important;\n}\n.jhtml .min-vh-100 , .jhtml.min-vh-100 {\n  min-height: 100vh !important;\n}\n.jhtml .flex-fill , .jhtml.flex-fill {\n  flex: 1 1 auto !important;\n}\n.jhtml .flex-row , .jhtml.flex-row {\n  flex-direction: row !important;\n}\n.jhtml .flex-column , .jhtml.flex-column {\n  flex-direction: column !important;\n}\n.jhtml .flex-row-reverse , .jhtml.flex-row-reverse {\n  flex-direction: row-reverse !important;\n}\n.jhtml .flex-column-reverse , .jhtml.flex-column-reverse {\n  flex-direction: column-reverse !important;\n}\n.jhtml .flex-grow-0 , .jhtml.flex-grow-0 {\n  flex-grow: 0 !important;\n}\n.jhtml .flex-grow-1 , .jhtml.flex-grow-1 {\n  flex-grow: 1 !important;\n}\n.jhtml .flex-shrink-0 , .jhtml.flex-shrink-0 {\n  flex-shrink: 0 !important;\n}\n.jhtml .flex-shrink-1 , .jhtml.flex-shrink-1 {\n  flex-shrink: 1 !important;\n}\n.jhtml .flex-wrap , .jhtml.flex-wrap {\n  flex-wrap: wrap !important;\n}\n.jhtml .flex-nowrap , .jhtml.flex-nowrap {\n  flex-wrap: nowrap !important;\n}\n.jhtml .flex-wrap-reverse , .jhtml.flex-wrap-reverse {\n  flex-wrap: wrap-reverse !important;\n}\n.jhtml .gap-0 , .jhtml.gap-0 {\n  gap: 0 !important;\n}\n.jhtml .gap-1 , .jhtml.gap-1 {\n  gap: 0.25rem !important;\n}\n.jhtml .gap-2 , .jhtml.gap-2 {\n  gap: 0.5rem !important;\n}\n.jhtml .gap-3 , .jhtml.gap-3 {\n  gap: 1rem !important;\n}\n.jhtml .gap-4 , .jhtml.gap-4 {\n  gap: 1.5rem !important;\n}\n.jhtml .gap-5 , .jhtml.gap-5 {\n  gap: 3rem !important;\n}\n.jhtml .justify-content-start , .jhtml.justify-content-start {\n  justify-content: flex-start !important;\n}\n.jhtml .justify-content-end , .jhtml.justify-content-end {\n  justify-content: flex-end !important;\n}\n.jhtml .justify-content-center , .jhtml.justify-content-center {\n  justify-content: center !important;\n}\n.jhtml .justify-content-between , .jhtml.justify-content-between {\n  justify-content: space-between !important;\n}\n.jhtml .justify-content-around , .jhtml.justify-content-around {\n  justify-content: space-around !important;\n}\n.jhtml .justify-content-evenly , .jhtml.justify-content-evenly {\n  justify-content: space-evenly !important;\n}\n.jhtml .align-items-start , .jhtml.align-items-start {\n  align-items: flex-start !important;\n}\n.jhtml .align-items-end , .jhtml.align-items-end {\n  align-items: flex-end !important;\n}\n.jhtml .align-items-center , .jhtml.align-items-center {\n  align-items: center !important;\n}\n.jhtml .align-items-baseline , .jhtml.align-items-baseline {\n  align-items: baseline !important;\n}\n.jhtml .align-items-stretch , .jhtml.align-items-stretch {\n  align-items: stretch !important;\n}\n.jhtml .align-content-start , .jhtml.align-content-start {\n  align-content: flex-start !important;\n}\n.jhtml .align-content-end , .jhtml.align-content-end {\n  align-content: flex-end !important;\n}\n.jhtml .align-content-center , .jhtml.align-content-center {\n  align-content: center !important;\n}\n.jhtml .align-content-between , .jhtml.align-content-between {\n  align-content: space-between !important;\n}\n.jhtml .align-content-around , .jhtml.align-content-around {\n  align-content: space-around !important;\n}\n.jhtml .align-content-stretch , .jhtml.align-content-stretch {\n  align-content: stretch !important;\n}\n.jhtml .align-self-auto , .jhtml.align-self-auto {\n  align-self: auto !important;\n}\n.jhtml .align-self-start , .jhtml.align-self-start {\n  align-self: flex-start !important;\n}\n.jhtml .align-self-end , .jhtml.align-self-end {\n  align-self: flex-end !important;\n}\n.jhtml .align-self-center , .jhtml.align-self-center {\n  align-self: center !important;\n}\n.jhtml .align-self-baseline , .jhtml.align-self-baseline {\n  align-self: baseline !important;\n}\n.jhtml .align-self-stretch , .jhtml.align-self-stretch {\n  align-self: stretch !important;\n}\n.jhtml .order-first , .jhtml.order-first {\n  order: -1 !important;\n}\n.jhtml .order-0 , .jhtml.order-0 {\n  order: 0 !important;\n}\n.jhtml .order-1 , .jhtml.order-1 {\n  order: 1 !important;\n}\n.jhtml .order-2 , .jhtml.order-2 {\n  order: 2 !important;\n}\n.jhtml .order-3 , .jhtml.order-3 {\n  order: 3 !important;\n}\n.jhtml .order-4 , .jhtml.order-4 {\n  order: 4 !important;\n}\n.jhtml .order-5 , .jhtml.order-5 {\n  order: 5 !important;\n}\n.jhtml .order-last , .jhtml.order-last {\n  order: 6 !important;\n}\n.jhtml .m-0 , .jhtml.m-0 {\n  margin: 0 !important;\n}\n.jhtml .m-1 , .jhtml.m-1 {\n  margin: 0.25rem !important;\n}\n.jhtml .m-2 , .jhtml.m-2 {\n  margin: 0.5rem !important;\n}\n.jhtml .m-3 , .jhtml.m-3 {\n  margin: 1rem !important;\n}\n.jhtml .m-4 , .jhtml.m-4 {\n  margin: 1.5rem !important;\n}\n.jhtml .m-5 , .jhtml.m-5 {\n  margin: 3rem !important;\n}\n.jhtml .m-auto , .jhtml.m-auto {\n  margin: auto !important;\n}\n.jhtml .mx-0 , .jhtml.mx-0 {\n  margin-right: 0 !important;\n  margin-left: 0 !important;\n}\n.jhtml .mx-1 , .jhtml.mx-1 {\n  margin-right: 0.25rem !important;\n  margin-left: 0.25rem !important;\n}\n.jhtml .mx-2 , .jhtml.mx-2 {\n  margin-right: 0.5rem !important;\n  margin-left: 0.5rem !important;\n}\n.jhtml .mx-3 , .jhtml.mx-3 {\n  margin-right: 1rem !important;\n  margin-left: 1rem !important;\n}\n.jhtml .mx-4 , .jhtml.mx-4 {\n  margin-right: 1.5rem !important;\n  margin-left: 1.5rem !important;\n}\n.jhtml .mx-5 , .jhtml.mx-5 {\n  margin-right: 3rem !important;\n  margin-left: 3rem !important;\n}\n.jhtml .mx-auto , .jhtml.mx-auto {\n  margin-right: auto !important;\n  margin-left: auto !important;\n}\n.jhtml .my-0 , .jhtml.my-0 {\n  margin-top: 0 !important;\n  margin-bottom: 0 !important;\n}\n.jhtml .my-1 , .jhtml.my-1 {\n  margin-top: 0.25rem !important;\n  margin-bottom: 0.25rem !important;\n}\n.jhtml .my-2 , .jhtml.my-2 {\n  margin-top: 0.5rem !important;\n  margin-bottom: 0.5rem !important;\n}\n.jhtml .my-3 , .jhtml.my-3 {\n  margin-top: 1rem !important;\n  margin-bottom: 1rem !important;\n}\n.jhtml .my-4 , .jhtml.my-4 {\n  margin-top: 1.5rem !important;\n  margin-bottom: 1.5rem !important;\n}\n.jhtml .my-5 , .jhtml.my-5 {\n  margin-top: 3rem !important;\n  margin-bottom: 3rem !important;\n}\n.jhtml .my-auto , .jhtml.my-auto {\n  margin-top: auto !important;\n  margin-bottom: auto !important;\n}\n.jhtml .mt-0 , .jhtml.mt-0 {\n  margin-top: 0 !important;\n}\n.jhtml .mt-1 , .jhtml.mt-1 {\n  margin-top: 0.25rem !important;\n}\n.jhtml .mt-2 , .jhtml.mt-2 {\n  margin-top: 0.5rem !important;\n}\n.jhtml .mt-3 , .jhtml.mt-3 {\n  margin-top: 1rem !important;\n}\n.jhtml .mt-4 , .jhtml.mt-4 {\n  margin-top: 1.5rem !important;\n}\n.jhtml .mt-5 , .jhtml.mt-5 {\n  margin-top: 3rem !important;\n}\n.jhtml .mt-auto , .jhtml.mt-auto {\n  margin-top: auto !important;\n}\n.jhtml .me-0 , .jhtml.me-0 {\n  margin-right: 0 !important;\n}\n.jhtml .me-1 , .jhtml.me-1 {\n  margin-right: 0.25rem !important;\n}\n.jhtml .me-2 , .jhtml.me-2 {\n  margin-right: 0.5rem !important;\n}\n.jhtml .me-3 , .jhtml.me-3 {\n  margin-right: 1rem !important;\n}\n.jhtml .me-4 , .jhtml.me-4 {\n  margin-right: 1.5rem !important;\n}\n.jhtml .me-5 , .jhtml.me-5 {\n  margin-right: 3rem !important;\n}\n.jhtml .me-auto , .jhtml.me-auto {\n  margin-right: auto !important;\n}\n.jhtml .mb-0 , .jhtml.mb-0 {\n  margin-bottom: 0 !important;\n}\n.jhtml .mb-1 , .jhtml.mb-1 {\n  margin-bottom: 0.25rem !important;\n}\n.jhtml .mb-2 , .jhtml.mb-2 {\n  margin-bottom: 0.5rem !important;\n}\n.jhtml .mb-3 , .jhtml.mb-3 {\n  margin-bottom: 1rem !important;\n}\n.jhtml .mb-4 , .jhtml.mb-4 {\n  margin-bottom: 1.5rem !important;\n}\n.jhtml .mb-5 , .jhtml.mb-5 {\n  margin-bottom: 3rem !important;\n}\n.jhtml .mb-auto , .jhtml.mb-auto {\n  margin-bottom: auto !important;\n}\n.jhtml .ms-0 , .jhtml.ms-0 {\n  margin-left: 0 !important;\n}\n.jhtml .ms-1 , .jhtml.ms-1 {\n  margin-left: 0.25rem !important;\n}\n.jhtml .ms-2 , .jhtml.ms-2 {\n  margin-left: 0.5rem !important;\n}\n.jhtml .ms-3 , .jhtml.ms-3 {\n  margin-left: 1rem !important;\n}\n.jhtml .ms-4 , .jhtml.ms-4 {\n  margin-left: 1.5rem !important;\n}\n.jhtml .ms-5 , .jhtml.ms-5 {\n  margin-left: 3rem !important;\n}\n.jhtml .ms-auto , .jhtml.ms-auto {\n  margin-left: auto !important;\n}\n.jhtml .p-0 , .jhtml.p-0 {\n  padding: 0 !important;\n}\n.jhtml .p-1 , .jhtml.p-1 {\n  padding: 0.25rem !important;\n}\n.jhtml .p-2 , .jhtml.p-2 {\n  padding: 0.5rem !important;\n}\n.jhtml .p-3 , .jhtml.p-3 {\n  padding: 1rem !important;\n}\n.jhtml .p-4 , .jhtml.p-4 {\n  padding: 1.5rem !important;\n}\n.jhtml .p-5 , .jhtml.p-5 {\n  padding: 3rem !important;\n}\n.jhtml .px-0 , .jhtml.px-0 {\n  padding-right: 0 !important;\n  padding-left: 0 !important;\n}\n.jhtml .px-1 , .jhtml.px-1 {\n  padding-right: 0.25rem !important;\n  padding-left: 0.25rem !important;\n}\n.jhtml .px-2 , .jhtml.px-2 {\n  padding-right: 0.5rem !important;\n  padding-left: 0.5rem !important;\n}\n.jhtml .px-3 , .jhtml.px-3 {\n  padding-right: 1rem !important;\n  padding-left: 1rem !important;\n}\n.jhtml .px-4 , .jhtml.px-4 {\n  padding-right: 1.5rem !important;\n  padding-left: 1.5rem !important;\n}\n.jhtml .px-5 , .jhtml.px-5 {\n  padding-right: 3rem !important;\n  padding-left: 3rem !important;\n}\n.jhtml .py-0 , .jhtml.py-0 {\n  padding-top: 0 !important;\n  padding-bottom: 0 !important;\n}\n.jhtml .py-1 , .jhtml.py-1 {\n  padding-top: 0.25rem !important;\n  padding-bottom: 0.25rem !important;\n}\n.jhtml .py-2 , .jhtml.py-2 {\n  padding-top: 0.5rem !important;\n  padding-bottom: 0.5rem !important;\n}\n.jhtml .py-3 , .jhtml.py-3 {\n  padding-top: 1rem !important;\n  padding-bottom: 1rem !important;\n}\n.jhtml .py-4 , .jhtml.py-4 {\n  padding-top: 1.5rem !important;\n  padding-bottom: 1.5rem !important;\n}\n.jhtml .py-5 , .jhtml.py-5 {\n  padding-top: 3rem !important;\n  padding-bottom: 3rem !important;\n}\n.jhtml .pt-0 , .jhtml.pt-0 {\n  padding-top: 0 !important;\n}\n.jhtml .pt-1 , .jhtml.pt-1 {\n  padding-top: 0.25rem !important;\n}\n.jhtml .pt-2 , .jhtml.pt-2 {\n  padding-top: 0.5rem !important;\n}\n.jhtml .pt-3 , .jhtml.pt-3 {\n  padding-top: 1rem !important;\n}\n.jhtml .pt-4 , .jhtml.pt-4 {\n  padding-top: 1.5rem !important;\n}\n.jhtml .pt-5 , .jhtml.pt-5 {\n  padding-top: 3rem !important;\n}\n.jhtml .pe-0 , .jhtml.pe-0 {\n  padding-right: 0 !important;\n}\n.jhtml .pe-1 , .jhtml.pe-1 {\n  padding-right: 0.25rem !important;\n}\n.jhtml .pe-2 , .jhtml.pe-2 {\n  padding-right: 0.5rem !important;\n}\n.jhtml .pe-3 , .jhtml.pe-3 {\n  padding-right: 1rem !important;\n}\n.jhtml .pe-4 , .jhtml.pe-4 {\n  padding-right: 1.5rem !important;\n}\n.jhtml .pe-5 , .jhtml.pe-5 {\n  padding-right: 3rem !important;\n}\n.jhtml .pb-0 , .jhtml.pb-0 {\n  padding-bottom: 0 !important;\n}\n.jhtml .pb-1 , .jhtml.pb-1 {\n  padding-bottom: 0.25rem !important;\n}\n.jhtml .pb-2 , .jhtml.pb-2 {\n  padding-bottom: 0.5rem !important;\n}\n.jhtml .pb-3 , .jhtml.pb-3 {\n  padding-bottom: 1rem !important;\n}\n.jhtml .pb-4 , .jhtml.pb-4 {\n  padding-bottom: 1.5rem !important;\n}\n.jhtml .pb-5 , .jhtml.pb-5 {\n  padding-bottom: 3rem !important;\n}\n.jhtml .ps-0 , .jhtml.ps-0 {\n  padding-left: 0 !important;\n}\n.jhtml .ps-1 , .jhtml.ps-1 {\n  padding-left: 0.25rem !important;\n}\n.jhtml .ps-2 , .jhtml.ps-2 {\n  padding-left: 0.5rem !important;\n}\n.jhtml .ps-3 , .jhtml.ps-3 {\n  padding-left: 1rem !important;\n}\n.jhtml .ps-4 , .jhtml.ps-4 {\n  padding-left: 1.5rem !important;\n}\n.jhtml .ps-5 , .jhtml.ps-5 {\n  padding-left: 3rem !important;\n}\n.jhtml .font-monospace , .jhtml.font-monospace {\n  font-family: var(--bs-font-monospace) !important;\n}\n.jhtml .fs-1 , .jhtml.fs-1 {\n  font-size: calc(1.375rem + 1.5vw) !important;\n}\n.jhtml .fs-2 , .jhtml.fs-2 {\n  font-size: calc(1.325rem + 0.9vw) !important;\n}\n.jhtml .fs-3 , .jhtml.fs-3 {\n  font-size: calc(1.3rem + 0.6vw) !important;\n}\n.jhtml .fs-4 , .jhtml.fs-4 {\n  font-size: calc(1.275rem + 0.3vw) !important;\n}\n.jhtml .fs-5 , .jhtml.fs-5 {\n  font-size: 1.25rem !important;\n}\n.jhtml .fs-6 , .jhtml.fs-6 {\n  font-size: 1rem !important;\n}\n.jhtml .fst-italic , .jhtml.fst-italic {\n  font-style: italic !important;\n}\n.jhtml .fst-normal , .jhtml.fst-normal {\n  font-style: normal !important;\n}\n.jhtml .fw-light , .jhtml.fw-light {\n  font-weight: 300 !important;\n}\n.jhtml .fw-lighter , .jhtml.fw-lighter {\n  font-weight: lighter !important;\n}\n.jhtml .fw-normal , .jhtml.fw-normal {\n  font-weight: 400 !important;\n}\n.jhtml .fw-bold , .jhtml.fw-bold {\n  font-weight: 700 !important;\n}\n.jhtml .fw-bolder , .jhtml.fw-bolder {\n  font-weight: bolder !important;\n}\n.jhtml .lh-1 , .jhtml.lh-1 {\n  line-height: 1 !important;\n}\n.jhtml .lh-sm , .jhtml.lh-sm {\n  line-height: 1.25 !important;\n}\n.jhtml .lh-base , .jhtml.lh-base {\n  line-height: 1.5 !important;\n}\n.jhtml .lh-lg , .jhtml.lh-lg {\n  line-height: 2 !important;\n}\n.jhtml .text-start , .jhtml.text-start {\n  text-align: left !important;\n}\n.jhtml .text-end , .jhtml.text-end {\n  text-align: right !important;\n}\n.jhtml .text-center , .jhtml.text-center {\n  text-align: center !important;\n}\n.jhtml .text-decoration-none , .jhtml.text-decoration-none {\n  text-decoration: none !important;\n}\n.jhtml .text-decoration-underline , .jhtml.text-decoration-underline {\n  text-decoration: underline !important;\n}\n.jhtml .text-decoration-line-through , .jhtml.text-decoration-line-through {\n  text-decoration: line-through !important;\n}\n.jhtml .text-lowercase , .jhtml.text-lowercase {\n  text-transform: lowercase !important;\n}\n.jhtml .text-uppercase , .jhtml.text-uppercase {\n  text-transform: uppercase !important;\n}\n.jhtml .text-capitalize , .jhtml.text-capitalize {\n  text-transform: capitalize !important;\n}\n.jhtml .text-wrap , .jhtml.text-wrap {\n  white-space: normal !important;\n}\n.jhtml .text-nowrap , .jhtml.text-nowrap {\n  white-space: nowrap !important;\n}\n.jhtml .text-break , .jhtml.text-break {\n  word-wrap: break-word !important;\n  word-break: break-word !important;\n}\n.jhtml .text-primary , .jhtml.text-primary {\n  --bs-text-opacity: 1;\n  color: rgba(var(--bs-primary-rgb), var(--bs-text-opacity)) !important;\n}\n.jhtml .text-secondary , .jhtml.text-secondary {\n  --bs-text-opacity: 1;\n  color: rgba(var(--bs-secondary-rgb), var(--bs-text-opacity)) !important;\n}\n.jhtml .text-success , .jhtml.text-success {\n  --bs-text-opacity: 1;\n  color: rgba(var(--bs-success-rgb), var(--bs-text-opacity)) !important;\n}\n.jhtml .text-info , .jhtml.text-info {\n  --bs-text-opacity: 1;\n  color: rgba(var(--bs-info-rgb), var(--bs-text-opacity)) !important;\n}\n.jhtml .text-warning , .jhtml.text-warning {\n  --bs-text-opacity: 1;\n  color: rgba(var(--bs-warning-rgb), var(--bs-text-opacity)) !important;\n}\n.jhtml .text-danger , .jhtml.text-danger {\n  --bs-text-opacity: 1;\n  color: rgba(var(--bs-danger-rgb), var(--bs-text-opacity)) !important;\n}\n.jhtml .text-light , .jhtml.text-light {\n  --bs-text-opacity: 1;\n  color: rgba(var(--bs-light-rgb), var(--bs-text-opacity)) !important;\n}\n.jhtml .text-dark , .jhtml.text-dark {\n  --bs-text-opacity: 1;\n  color: rgba(var(--bs-dark-rgb), var(--bs-text-opacity)) !important;\n}\n.jhtml .text-black , .jhtml.text-black {\n  --bs-text-opacity: 1;\n  color: rgba(var(--bs-black-rgb), var(--bs-text-opacity)) !important;\n}\n.jhtml .text-white , .jhtml.text-white {\n  --bs-text-opacity: 1;\n  color: rgba(var(--bs-white-rgb), var(--bs-text-opacity)) !important;\n}\n.jhtml .text-body , .jhtml.text-body {\n  --bs-text-opacity: 1;\n  color: rgba(var(--bs-body-color-rgb), var(--bs-text-opacity)) !important;\n}\n.jhtml .text-muted , .jhtml.text-muted {\n  --bs-text-opacity: 1;\n  color: #6c757d !important;\n}\n.jhtml .text-black-50 , .jhtml.text-black-50 {\n  --bs-text-opacity: 1;\n  color: rgba(0, 0, 0, 0.5) !important;\n}\n.jhtml .text-white-50 , .jhtml.text-white-50 {\n  --bs-text-opacity: 1;\n  color: rgba(255, 255, 255, 0.5) !important;\n}\n.jhtml .text-reset , .jhtml.text-reset {\n  --bs-text-opacity: 1;\n  color: inherit !important;\n}\n.jhtml .text-opacity-25 , .jhtml.text-opacity-25 {\n  --bs-text-opacity: 0.25;\n}\n.jhtml .text-opacity-50 , .jhtml.text-opacity-50 {\n  --bs-text-opacity: 0.5;\n}\n.jhtml .text-opacity-75 , .jhtml.text-opacity-75 {\n  --bs-text-opacity: 0.75;\n}\n.jhtml .text-opacity-100 , .jhtml.text-opacity-100 {\n  --bs-text-opacity: 1;\n}\n.jhtml .bg-primary , .jhtml.bg-primary {\n  --bs-bg-opacity: 1;\n  background-color: rgba(var(--bs-primary-rgb), var(--bs-bg-opacity)) !important;\n}\n.jhtml .bg-secondary , .jhtml.bg-secondary {\n  --bs-bg-opacity: 1;\n  background-color: rgba(var(--bs-secondary-rgb), var(--bs-bg-opacity)) !important;\n}\n.jhtml .bg-success , .jhtml.bg-success {\n  --bs-bg-opacity: 1;\n  background-color: rgba(var(--bs-success-rgb), var(--bs-bg-opacity)) !important;\n}\n.jhtml .bg-info , .jhtml.bg-info {\n  --bs-bg-opacity: 1;\n  background-color: rgba(var(--bs-info-rgb), var(--bs-bg-opacity)) !important;\n}\n.jhtml .bg-warning , .jhtml.bg-warning {\n  --bs-bg-opacity: 1;\n  background-color: rgba(var(--bs-warning-rgb), var(--bs-bg-opacity)) !important;\n}\n.jhtml .bg-danger , .jhtml.bg-danger {\n  --bs-bg-opacity: 1;\n  background-color: rgba(var(--bs-danger-rgb), var(--bs-bg-opacity)) !important;\n}\n.jhtml .bg-light , .jhtml.bg-light {\n  --bs-bg-opacity: 1;\n  background-color: rgba(var(--bs-light-rgb), var(--bs-bg-opacity)) !important;\n}\n.jhtml .bg-dark , .jhtml.bg-dark {\n  --bs-bg-opacity: 1;\n  background-color: rgba(var(--bs-dark-rgb), var(--bs-bg-opacity)) !important;\n}\n.jhtml .bg-black , .jhtml.bg-black {\n  --bs-bg-opacity: 1;\n  background-color: rgba(var(--bs-black-rgb), var(--bs-bg-opacity)) !important;\n}\n.jhtml .bg-white , .jhtml.bg-white {\n  --bs-bg-opacity: 1;\n  background-color: rgba(var(--bs-white-rgb), var(--bs-bg-opacity)) !important;\n}\n.jhtml .bg-body , .jhtml.bg-body {\n  --bs-bg-opacity: 1;\n  background-color: rgba(var(--bs-body-bg-rgb), var(--bs-bg-opacity)) !important;\n}\n.jhtml .bg-transparent , .jhtml.bg-transparent {\n  --bs-bg-opacity: 1;\n  background-color: transparent !important;\n}\n.jhtml .bg-opacity-10 , .jhtml.bg-opacity-10 {\n  --bs-bg-opacity: 0.1;\n}\n.jhtml .bg-opacity-25 , .jhtml.bg-opacity-25 {\n  --bs-bg-opacity: 0.25;\n}\n.jhtml .bg-opacity-50 , .jhtml.bg-opacity-50 {\n  --bs-bg-opacity: 0.5;\n}\n.jhtml .bg-opacity-75 , .jhtml.bg-opacity-75 {\n  --bs-bg-opacity: 0.75;\n}\n.jhtml .bg-opacity-100 , .jhtml.bg-opacity-100 {\n  --bs-bg-opacity: 1;\n}\n.jhtml .bg-gradient , .jhtml.bg-gradient {\n  background-image: var(--bs-gradient) !important;\n}\n.jhtml .user-select-all , .jhtml.user-select-all {\n  user-select: all !important;\n}\n.jhtml .user-select-auto , .jhtml.user-select-auto {\n  user-select: auto !important;\n}\n.jhtml .user-select-none , .jhtml.user-select-none {\n  user-select: none !important;\n}\n.jhtml .pe-none , .jhtml.pe-none {\n  pointer-events: none !important;\n}\n.jhtml .pe-auto , .jhtml.pe-auto {\n  pointer-events: auto !important;\n}\n.jhtml .rounded , .jhtml.rounded {\n  border-radius: 0.25rem !important;\n}\n.jhtml .rounded-0 , .jhtml.rounded-0 {\n  border-radius: 0 !important;\n}\n.jhtml .rounded-1 , .jhtml.rounded-1 {\n  border-radius: 0.2rem !important;\n}\n.jhtml .rounded-2 , .jhtml.rounded-2 {\n  border-radius: 0.25rem !important;\n}\n.jhtml .rounded-3 , .jhtml.rounded-3 {\n  border-radius: 0.3rem !important;\n}\n.jhtml .rounded-circle , .jhtml.rounded-circle {\n  border-radius: 50% !important;\n}\n.jhtml .rounded-pill , .jhtml.rounded-pill {\n  border-radius: 50rem !important;\n}\n.jhtml .rounded-top , .jhtml.rounded-top {\n  border-top-left-radius: 0.25rem !important;\n  border-top-right-radius: 0.25rem !important;\n}\n.jhtml .rounded-end , .jhtml.rounded-end {\n  border-top-right-radius: 0.25rem !important;\n  border-bottom-right-radius: 0.25rem !important;\n}\n.jhtml .rounded-bottom , .jhtml.rounded-bottom {\n  border-bottom-right-radius: 0.25rem !important;\n  border-bottom-left-radius: 0.25rem !important;\n}\n.jhtml .rounded-start , .jhtml.rounded-start {\n  border-bottom-left-radius: 0.25rem !important;\n  border-top-left-radius: 0.25rem !important;\n}\n.jhtml .visible , .jhtml.visible {\n  visibility: visible !important;\n}\n.jhtml .invisible , .jhtml.invisible {\n  visibility: hidden !important;\n}\n@media (min-width: 576px) {\n  .jhtml .float-sm-start , .jhtml.float-sm-start {\n    float: left !important;\n  }\n  .jhtml .float-sm-end , .jhtml.float-sm-end {\n    float: right !important;\n  }\n  .jhtml .float-sm-none , .jhtml.float-sm-none {\n    float: none !important;\n  }\n  .jhtml .d-sm-inline , .jhtml.d-sm-inline {\n    display: inline !important;\n  }\n  .jhtml .d-sm-inline-block , .jhtml.d-sm-inline-block {\n    display: inline-block !important;\n  }\n  .jhtml .d-sm-block , .jhtml.d-sm-block {\n    display: block !important;\n  }\n  .jhtml .d-sm-grid , .jhtml.d-sm-grid {\n    display: grid !important;\n  }\n  .jhtml .d-sm-table , .jhtml.d-sm-table {\n    display: table !important;\n  }\n  .jhtml .d-sm-table-row , .jhtml.d-sm-table-row {\n    display: table-row !important;\n  }\n  .jhtml .d-sm-table-cell , .jhtml.d-sm-table-cell {\n    display: table-cell !important;\n  }\n  .jhtml .d-sm-flex , .jhtml.d-sm-flex {\n    display: flex !important;\n  }\n  .jhtml .d-sm-inline-flex , .jhtml.d-sm-inline-flex {\n    display: inline-flex !important;\n  }\n  .jhtml .d-sm-none , .jhtml.d-sm-none {\n    display: none !important;\n  }\n  .jhtml .flex-sm-fill , .jhtml.flex-sm-fill {\n    flex: 1 1 auto !important;\n  }\n  .jhtml .flex-sm-row , .jhtml.flex-sm-row {\n    flex-direction: row !important;\n  }\n  .jhtml .flex-sm-column , .jhtml.flex-sm-column {\n    flex-direction: column !important;\n  }\n  .jhtml .flex-sm-row-reverse , .jhtml.flex-sm-row-reverse {\n    flex-direction: row-reverse !important;\n  }\n  .jhtml .flex-sm-column-reverse , .jhtml.flex-sm-column-reverse {\n    flex-direction: column-reverse !important;\n  }\n  .jhtml .flex-sm-grow-0 , .jhtml.flex-sm-grow-0 {\n    flex-grow: 0 !important;\n  }\n  .jhtml .flex-sm-grow-1 , .jhtml.flex-sm-grow-1 {\n    flex-grow: 1 !important;\n  }\n  .jhtml .flex-sm-shrink-0 , .jhtml.flex-sm-shrink-0 {\n    flex-shrink: 0 !important;\n  }\n  .jhtml .flex-sm-shrink-1 , .jhtml.flex-sm-shrink-1 {\n    flex-shrink: 1 !important;\n  }\n  .jhtml .flex-sm-wrap , .jhtml.flex-sm-wrap {\n    flex-wrap: wrap !important;\n  }\n  .jhtml .flex-sm-nowrap , .jhtml.flex-sm-nowrap {\n    flex-wrap: nowrap !important;\n  }\n  .jhtml .flex-sm-wrap-reverse , .jhtml.flex-sm-wrap-reverse {\n    flex-wrap: wrap-reverse !important;\n  }\n  .jhtml .gap-sm-0 , .jhtml.gap-sm-0 {\n    gap: 0 !important;\n  }\n  .jhtml .gap-sm-1 , .jhtml.gap-sm-1 {\n    gap: 0.25rem !important;\n  }\n  .jhtml .gap-sm-2 , .jhtml.gap-sm-2 {\n    gap: 0.5rem !important;\n  }\n  .jhtml .gap-sm-3 , .jhtml.gap-sm-3 {\n    gap: 1rem !important;\n  }\n  .jhtml .gap-sm-4 , .jhtml.gap-sm-4 {\n    gap: 1.5rem !important;\n  }\n  .jhtml .gap-sm-5 , .jhtml.gap-sm-5 {\n    gap: 3rem !important;\n  }\n  .jhtml .justify-content-sm-start , .jhtml.justify-content-sm-start {\n    justify-content: flex-start !important;\n  }\n  .jhtml .justify-content-sm-end , .jhtml.justify-content-sm-end {\n    justify-content: flex-end !important;\n  }\n  .jhtml .justify-content-sm-center , .jhtml.justify-content-sm-center {\n    justify-content: center !important;\n  }\n  .jhtml .justify-content-sm-between , .jhtml.justify-content-sm-between {\n    justify-content: space-between !important;\n  }\n  .jhtml .justify-content-sm-around , .jhtml.justify-content-sm-around {\n    justify-content: space-around !important;\n  }\n  .jhtml .justify-content-sm-evenly , .jhtml.justify-content-sm-evenly {\n    justify-content: space-evenly !important;\n  }\n  .jhtml .align-items-sm-start , .jhtml.align-items-sm-start {\n    align-items: flex-start !important;\n  }\n  .jhtml .align-items-sm-end , .jhtml.align-items-sm-end {\n    align-items: flex-end !important;\n  }\n  .jhtml .align-items-sm-center , .jhtml.align-items-sm-center {\n    align-items: center !important;\n  }\n  .jhtml .align-items-sm-baseline , .jhtml.align-items-sm-baseline {\n    align-items: baseline !important;\n  }\n  .jhtml .align-items-sm-stretch , .jhtml.align-items-sm-stretch {\n    align-items: stretch !important;\n  }\n  .jhtml .align-content-sm-start , .jhtml.align-content-sm-start {\n    align-content: flex-start !important;\n  }\n  .jhtml .align-content-sm-end , .jhtml.align-content-sm-end {\n    align-content: flex-end !important;\n  }\n  .jhtml .align-content-sm-center , .jhtml.align-content-sm-center {\n    align-content: center !important;\n  }\n  .jhtml .align-content-sm-between , .jhtml.align-content-sm-between {\n    align-content: space-between !important;\n  }\n  .jhtml .align-content-sm-around , .jhtml.align-content-sm-around {\n    align-content: space-around !important;\n  }\n  .jhtml .align-content-sm-stretch , .jhtml.align-content-sm-stretch {\n    align-content: stretch !important;\n  }\n  .jhtml .align-self-sm-auto , .jhtml.align-self-sm-auto {\n    align-self: auto !important;\n  }\n  .jhtml .align-self-sm-start , .jhtml.align-self-sm-start {\n    align-self: flex-start !important;\n  }\n  .jhtml .align-self-sm-end , .jhtml.align-self-sm-end {\n    align-self: flex-end !important;\n  }\n  .jhtml .align-self-sm-center , .jhtml.align-self-sm-center {\n    align-self: center !important;\n  }\n  .jhtml .align-self-sm-baseline , .jhtml.align-self-sm-baseline {\n    align-self: baseline !important;\n  }\n  .jhtml .align-self-sm-stretch , .jhtml.align-self-sm-stretch {\n    align-self: stretch !important;\n  }\n  .jhtml .order-sm-first , .jhtml.order-sm-first {\n    order: -1 !important;\n  }\n  .jhtml .order-sm-0 , .jhtml.order-sm-0 {\n    order: 0 !important;\n  }\n  .jhtml .order-sm-1 , .jhtml.order-sm-1 {\n    order: 1 !important;\n  }\n  .jhtml .order-sm-2 , .jhtml.order-sm-2 {\n    order: 2 !important;\n  }\n  .jhtml .order-sm-3 , .jhtml.order-sm-3 {\n    order: 3 !important;\n  }\n  .jhtml .order-sm-4 , .jhtml.order-sm-4 {\n    order: 4 !important;\n  }\n  .jhtml .order-sm-5 , .jhtml.order-sm-5 {\n    order: 5 !important;\n  }\n  .jhtml .order-sm-last , .jhtml.order-sm-last {\n    order: 6 !important;\n  }\n  .jhtml .m-sm-0 , .jhtml.m-sm-0 {\n    margin: 0 !important;\n  }\n  .jhtml .m-sm-1 , .jhtml.m-sm-1 {\n    margin: 0.25rem !important;\n  }\n  .jhtml .m-sm-2 , .jhtml.m-sm-2 {\n    margin: 0.5rem !important;\n  }\n  .jhtml .m-sm-3 , .jhtml.m-sm-3 {\n    margin: 1rem !important;\n  }\n  .jhtml .m-sm-4 , .jhtml.m-sm-4 {\n    margin: 1.5rem !important;\n  }\n  .jhtml .m-sm-5 , .jhtml.m-sm-5 {\n    margin: 3rem !important;\n  }\n  .jhtml .m-sm-auto , .jhtml.m-sm-auto {\n    margin: auto !important;\n  }\n  .jhtml .mx-sm-0 , .jhtml.mx-sm-0 {\n    margin-right: 0 !important;\n    margin-left: 0 !important;\n  }\n  .jhtml .mx-sm-1 , .jhtml.mx-sm-1 {\n    margin-right: 0.25rem !important;\n    margin-left: 0.25rem !important;\n  }\n  .jhtml .mx-sm-2 , .jhtml.mx-sm-2 {\n    margin-right: 0.5rem !important;\n    margin-left: 0.5rem !important;\n  }\n  .jhtml .mx-sm-3 , .jhtml.mx-sm-3 {\n    margin-right: 1rem !important;\n    margin-left: 1rem !important;\n  }\n  .jhtml .mx-sm-4 , .jhtml.mx-sm-4 {\n    margin-right: 1.5rem !important;\n    margin-left: 1.5rem !important;\n  }\n  .jhtml .mx-sm-5 , .jhtml.mx-sm-5 {\n    margin-right: 3rem !important;\n    margin-left: 3rem !important;\n  }\n  .jhtml .mx-sm-auto , .jhtml.mx-sm-auto {\n    margin-right: auto !important;\n    margin-left: auto !important;\n  }\n  .jhtml .my-sm-0 , .jhtml.my-sm-0 {\n    margin-top: 0 !important;\n    margin-bottom: 0 !important;\n  }\n  .jhtml .my-sm-1 , .jhtml.my-sm-1 {\n    margin-top: 0.25rem !important;\n    margin-bottom: 0.25rem !important;\n  }\n  .jhtml .my-sm-2 , .jhtml.my-sm-2 {\n    margin-top: 0.5rem !important;\n    margin-bottom: 0.5rem !important;\n  }\n  .jhtml .my-sm-3 , .jhtml.my-sm-3 {\n    margin-top: 1rem !important;\n    margin-bottom: 1rem !important;\n  }\n  .jhtml .my-sm-4 , .jhtml.my-sm-4 {\n    margin-top: 1.5rem !important;\n    margin-bottom: 1.5rem !important;\n  }\n  .jhtml .my-sm-5 , .jhtml.my-sm-5 {\n    margin-top: 3rem !important;\n    margin-bottom: 3rem !important;\n  }\n  .jhtml .my-sm-auto , .jhtml.my-sm-auto {\n    margin-top: auto !important;\n    margin-bottom: auto !important;\n  }\n  .jhtml .mt-sm-0 , .jhtml.mt-sm-0 {\n    margin-top: 0 !important;\n  }\n  .jhtml .mt-sm-1 , .jhtml.mt-sm-1 {\n    margin-top: 0.25rem !important;\n  }\n  .jhtml .mt-sm-2 , .jhtml.mt-sm-2 {\n    margin-top: 0.5rem !important;\n  }\n  .jhtml .mt-sm-3 , .jhtml.mt-sm-3 {\n    margin-top: 1rem !important;\n  }\n  .jhtml .mt-sm-4 , .jhtml.mt-sm-4 {\n    margin-top: 1.5rem !important;\n  }\n  .jhtml .mt-sm-5 , .jhtml.mt-sm-5 {\n    margin-top: 3rem !important;\n  }\n  .jhtml .mt-sm-auto , .jhtml.mt-sm-auto {\n    margin-top: auto !important;\n  }\n  .jhtml .me-sm-0 , .jhtml.me-sm-0 {\n    margin-right: 0 !important;\n  }\n  .jhtml .me-sm-1 , .jhtml.me-sm-1 {\n    margin-right: 0.25rem !important;\n  }\n  .jhtml .me-sm-2 , .jhtml.me-sm-2 {\n    margin-right: 0.5rem !important;\n  }\n  .jhtml .me-sm-3 , .jhtml.me-sm-3 {\n    margin-right: 1rem !important;\n  }\n  .jhtml .me-sm-4 , .jhtml.me-sm-4 {\n    margin-right: 1.5rem !important;\n  }\n  .jhtml .me-sm-5 , .jhtml.me-sm-5 {\n    margin-right: 3rem !important;\n  }\n  .jhtml .me-sm-auto , .jhtml.me-sm-auto {\n    margin-right: auto !important;\n  }\n  .jhtml .mb-sm-0 , .jhtml.mb-sm-0 {\n    margin-bottom: 0 !important;\n  }\n  .jhtml .mb-sm-1 , .jhtml.mb-sm-1 {\n    margin-bottom: 0.25rem !important;\n  }\n  .jhtml .mb-sm-2 , .jhtml.mb-sm-2 {\n    margin-bottom: 0.5rem !important;\n  }\n  .jhtml .mb-sm-3 , .jhtml.mb-sm-3 {\n    margin-bottom: 1rem !important;\n  }\n  .jhtml .mb-sm-4 , .jhtml.mb-sm-4 {\n    margin-bottom: 1.5rem !important;\n  }\n  .jhtml .mb-sm-5 , .jhtml.mb-sm-5 {\n    margin-bottom: 3rem !important;\n  }\n  .jhtml .mb-sm-auto , .jhtml.mb-sm-auto {\n    margin-bottom: auto !important;\n  }\n  .jhtml .ms-sm-0 , .jhtml.ms-sm-0 {\n    margin-left: 0 !important;\n  }\n  .jhtml .ms-sm-1 , .jhtml.ms-sm-1 {\n    margin-left: 0.25rem !important;\n  }\n  .jhtml .ms-sm-2 , .jhtml.ms-sm-2 {\n    margin-left: 0.5rem !important;\n  }\n  .jhtml .ms-sm-3 , .jhtml.ms-sm-3 {\n    margin-left: 1rem !important;\n  }\n  .jhtml .ms-sm-4 , .jhtml.ms-sm-4 {\n    margin-left: 1.5rem !important;\n  }\n  .jhtml .ms-sm-5 , .jhtml.ms-sm-5 {\n    margin-left: 3rem !important;\n  }\n  .jhtml .ms-sm-auto , .jhtml.ms-sm-auto {\n    margin-left: auto !important;\n  }\n  .jhtml .p-sm-0 , .jhtml.p-sm-0 {\n    padding: 0 !important;\n  }\n  .jhtml .p-sm-1 , .jhtml.p-sm-1 {\n    padding: 0.25rem !important;\n  }\n  .jhtml .p-sm-2 , .jhtml.p-sm-2 {\n    padding: 0.5rem !important;\n  }\n  .jhtml .p-sm-3 , .jhtml.p-sm-3 {\n    padding: 1rem !important;\n  }\n  .jhtml .p-sm-4 , .jhtml.p-sm-4 {\n    padding: 1.5rem !important;\n  }\n  .jhtml .p-sm-5 , .jhtml.p-sm-5 {\n    padding: 3rem !important;\n  }\n  .jhtml .px-sm-0 , .jhtml.px-sm-0 {\n    padding-right: 0 !important;\n    padding-left: 0 !important;\n  }\n  .jhtml .px-sm-1 , .jhtml.px-sm-1 {\n    padding-right: 0.25rem !important;\n    padding-left: 0.25rem !important;\n  }\n  .jhtml .px-sm-2 , .jhtml.px-sm-2 {\n    padding-right: 0.5rem !important;\n    padding-left: 0.5rem !important;\n  }\n  .jhtml .px-sm-3 , .jhtml.px-sm-3 {\n    padding-right: 1rem !important;\n    padding-left: 1rem !important;\n  }\n  .jhtml .px-sm-4 , .jhtml.px-sm-4 {\n    padding-right: 1.5rem !important;\n    padding-left: 1.5rem !important;\n  }\n  .jhtml .px-sm-5 , .jhtml.px-sm-5 {\n    padding-right: 3rem !important;\n    padding-left: 3rem !important;\n  }\n  .jhtml .py-sm-0 , .jhtml.py-sm-0 {\n    padding-top: 0 !important;\n    padding-bottom: 0 !important;\n  }\n  .jhtml .py-sm-1 , .jhtml.py-sm-1 {\n    padding-top: 0.25rem !important;\n    padding-bottom: 0.25rem !important;\n  }\n  .jhtml .py-sm-2 , .jhtml.py-sm-2 {\n    padding-top: 0.5rem !important;\n    padding-bottom: 0.5rem !important;\n  }\n  .jhtml .py-sm-3 , .jhtml.py-sm-3 {\n    padding-top: 1rem !important;\n    padding-bottom: 1rem !important;\n  }\n  .jhtml .py-sm-4 , .jhtml.py-sm-4 {\n    padding-top: 1.5rem !important;\n    padding-bottom: 1.5rem !important;\n  }\n  .jhtml .py-sm-5 , .jhtml.py-sm-5 {\n    padding-top: 3rem !important;\n    padding-bottom: 3rem !important;\n  }\n  .jhtml .pt-sm-0 , .jhtml.pt-sm-0 {\n    padding-top: 0 !important;\n  }\n  .jhtml .pt-sm-1 , .jhtml.pt-sm-1 {\n    padding-top: 0.25rem !important;\n  }\n  .jhtml .pt-sm-2 , .jhtml.pt-sm-2 {\n    padding-top: 0.5rem !important;\n  }\n  .jhtml .pt-sm-3 , .jhtml.pt-sm-3 {\n    padding-top: 1rem !important;\n  }\n  .jhtml .pt-sm-4 , .jhtml.pt-sm-4 {\n    padding-top: 1.5rem !important;\n  }\n  .jhtml .pt-sm-5 , .jhtml.pt-sm-5 {\n    padding-top: 3rem !important;\n  }\n  .jhtml .pe-sm-0 , .jhtml.pe-sm-0 {\n    padding-right: 0 !important;\n  }\n  .jhtml .pe-sm-1 , .jhtml.pe-sm-1 {\n    padding-right: 0.25rem !important;\n  }\n  .jhtml .pe-sm-2 , .jhtml.pe-sm-2 {\n    padding-right: 0.5rem !important;\n  }\n  .jhtml .pe-sm-3 , .jhtml.pe-sm-3 {\n    padding-right: 1rem !important;\n  }\n  .jhtml .pe-sm-4 , .jhtml.pe-sm-4 {\n    padding-right: 1.5rem !important;\n  }\n  .jhtml .pe-sm-5 , .jhtml.pe-sm-5 {\n    padding-right: 3rem !important;\n  }\n  .jhtml .pb-sm-0 , .jhtml.pb-sm-0 {\n    padding-bottom: 0 !important;\n  }\n  .jhtml .pb-sm-1 , .jhtml.pb-sm-1 {\n    padding-bottom: 0.25rem !important;\n  }\n  .jhtml .pb-sm-2 , .jhtml.pb-sm-2 {\n    padding-bottom: 0.5rem !important;\n  }\n  .jhtml .pb-sm-3 , .jhtml.pb-sm-3 {\n    padding-bottom: 1rem !important;\n  }\n  .jhtml .pb-sm-4 , .jhtml.pb-sm-4 {\n    padding-bottom: 1.5rem !important;\n  }\n  .jhtml .pb-sm-5 , .jhtml.pb-sm-5 {\n    padding-bottom: 3rem !important;\n  }\n  .jhtml .ps-sm-0 , .jhtml.ps-sm-0 {\n    padding-left: 0 !important;\n  }\n  .jhtml .ps-sm-1 , .jhtml.ps-sm-1 {\n    padding-left: 0.25rem !important;\n  }\n  .jhtml .ps-sm-2 , .jhtml.ps-sm-2 {\n    padding-left: 0.5rem !important;\n  }\n  .jhtml .ps-sm-3 , .jhtml.ps-sm-3 {\n    padding-left: 1rem !important;\n  }\n  .jhtml .ps-sm-4 , .jhtml.ps-sm-4 {\n    padding-left: 1.5rem !important;\n  }\n  .jhtml .ps-sm-5 , .jhtml.ps-sm-5 {\n    padding-left: 3rem !important;\n  }\n  .jhtml .text-sm-start , .jhtml.text-sm-start {\n    text-align: left !important;\n  }\n  .jhtml .text-sm-end , .jhtml.text-sm-end {\n    text-align: right !important;\n  }\n  .jhtml .text-sm-center , .jhtml.text-sm-center {\n    text-align: center !important;\n  }\n}\n@media (min-width: 768px) {\n  .jhtml .float-md-start , .jhtml.float-md-start {\n    float: left !important;\n  }\n  .jhtml .float-md-end , .jhtml.float-md-end {\n    float: right !important;\n  }\n  .jhtml .float-md-none , .jhtml.float-md-none {\n    float: none !important;\n  }\n  .jhtml .d-md-inline , .jhtml.d-md-inline {\n    display: inline !important;\n  }\n  .jhtml .d-md-inline-block , .jhtml.d-md-inline-block {\n    display: inline-block !important;\n  }\n  .jhtml .d-md-block , .jhtml.d-md-block {\n    display: block !important;\n  }\n  .jhtml .d-md-grid , .jhtml.d-md-grid {\n    display: grid !important;\n  }\n  .jhtml .d-md-table , .jhtml.d-md-table {\n    display: table !important;\n  }\n  .jhtml .d-md-table-row , .jhtml.d-md-table-row {\n    display: table-row !important;\n  }\n  .jhtml .d-md-table-cell , .jhtml.d-md-table-cell {\n    display: table-cell !important;\n  }\n  .jhtml .d-md-flex , .jhtml.d-md-flex {\n    display: flex !important;\n  }\n  .jhtml .d-md-inline-flex , .jhtml.d-md-inline-flex {\n    display: inline-flex !important;\n  }\n  .jhtml .d-md-none , .jhtml.d-md-none {\n    display: none !important;\n  }\n  .jhtml .flex-md-fill , .jhtml.flex-md-fill {\n    flex: 1 1 auto !important;\n  }\n  .jhtml .flex-md-row , .jhtml.flex-md-row {\n    flex-direction: row !important;\n  }\n  .jhtml .flex-md-column , .jhtml.flex-md-column {\n    flex-direction: column !important;\n  }\n  .jhtml .flex-md-row-reverse , .jhtml.flex-md-row-reverse {\n    flex-direction: row-reverse !important;\n  }\n  .jhtml .flex-md-column-reverse , .jhtml.flex-md-column-reverse {\n    flex-direction: column-reverse !important;\n  }\n  .jhtml .flex-md-grow-0 , .jhtml.flex-md-grow-0 {\n    flex-grow: 0 !important;\n  }\n  .jhtml .flex-md-grow-1 , .jhtml.flex-md-grow-1 {\n    flex-grow: 1 !important;\n  }\n  .jhtml .flex-md-shrink-0 , .jhtml.flex-md-shrink-0 {\n    flex-shrink: 0 !important;\n  }\n  .jhtml .flex-md-shrink-1 , .jhtml.flex-md-shrink-1 {\n    flex-shrink: 1 !important;\n  }\n  .jhtml .flex-md-wrap , .jhtml.flex-md-wrap {\n    flex-wrap: wrap !important;\n  }\n  .jhtml .flex-md-nowrap , .jhtml.flex-md-nowrap {\n    flex-wrap: nowrap !important;\n  }\n  .jhtml .flex-md-wrap-reverse , .jhtml.flex-md-wrap-reverse {\n    flex-wrap: wrap-reverse !important;\n  }\n  .jhtml .gap-md-0 , .jhtml.gap-md-0 {\n    gap: 0 !important;\n  }\n  .jhtml .gap-md-1 , .jhtml.gap-md-1 {\n    gap: 0.25rem !important;\n  }\n  .jhtml .gap-md-2 , .jhtml.gap-md-2 {\n    gap: 0.5rem !important;\n  }\n  .jhtml .gap-md-3 , .jhtml.gap-md-3 {\n    gap: 1rem !important;\n  }\n  .jhtml .gap-md-4 , .jhtml.gap-md-4 {\n    gap: 1.5rem !important;\n  }\n  .jhtml .gap-md-5 , .jhtml.gap-md-5 {\n    gap: 3rem !important;\n  }\n  .jhtml .justify-content-md-start , .jhtml.justify-content-md-start {\n    justify-content: flex-start !important;\n  }\n  .jhtml .justify-content-md-end , .jhtml.justify-content-md-end {\n    justify-content: flex-end !important;\n  }\n  .jhtml .justify-content-md-center , .jhtml.justify-content-md-center {\n    justify-content: center !important;\n  }\n  .jhtml .justify-content-md-between , .jhtml.justify-content-md-between {\n    justify-content: space-between !important;\n  }\n  .jhtml .justify-content-md-around , .jhtml.justify-content-md-around {\n    justify-content: space-around !important;\n  }\n  .jhtml .justify-content-md-evenly , .jhtml.justify-content-md-evenly {\n    justify-content: space-evenly !important;\n  }\n  .jhtml .align-items-md-start , .jhtml.align-items-md-start {\n    align-items: flex-start !important;\n  }\n  .jhtml .align-items-md-end , .jhtml.align-items-md-end {\n    align-items: flex-end !important;\n  }\n  .jhtml .align-items-md-center , .jhtml.align-items-md-center {\n    align-items: center !important;\n  }\n  .jhtml .align-items-md-baseline , .jhtml.align-items-md-baseline {\n    align-items: baseline !important;\n  }\n  .jhtml .align-items-md-stretch , .jhtml.align-items-md-stretch {\n    align-items: stretch !important;\n  }\n  .jhtml .align-content-md-start , .jhtml.align-content-md-start {\n    align-content: flex-start !important;\n  }\n  .jhtml .align-content-md-end , .jhtml.align-content-md-end {\n    align-content: flex-end !important;\n  }\n  .jhtml .align-content-md-center , .jhtml.align-content-md-center {\n    align-content: center !important;\n  }\n  .jhtml .align-content-md-between , .jhtml.align-content-md-between {\n    align-content: space-between !important;\n  }\n  .jhtml .align-content-md-around , .jhtml.align-content-md-around {\n    align-content: space-around !important;\n  }\n  .jhtml .align-content-md-stretch , .jhtml.align-content-md-stretch {\n    align-content: stretch !important;\n  }\n  .jhtml .align-self-md-auto , .jhtml.align-self-md-auto {\n    align-self: auto !important;\n  }\n  .jhtml .align-self-md-start , .jhtml.align-self-md-start {\n    align-self: flex-start !important;\n  }\n  .jhtml .align-self-md-end , .jhtml.align-self-md-end {\n    align-self: flex-end !important;\n  }\n  .jhtml .align-self-md-center , .jhtml.align-self-md-center {\n    align-self: center !important;\n  }\n  .jhtml .align-self-md-baseline , .jhtml.align-self-md-baseline {\n    align-self: baseline !important;\n  }\n  .jhtml .align-self-md-stretch , .jhtml.align-self-md-stretch {\n    align-self: stretch !important;\n  }\n  .jhtml .order-md-first , .jhtml.order-md-first {\n    order: -1 !important;\n  }\n  .jhtml .order-md-0 , .jhtml.order-md-0 {\n    order: 0 !important;\n  }\n  .jhtml .order-md-1 , .jhtml.order-md-1 {\n    order: 1 !important;\n  }\n  .jhtml .order-md-2 , .jhtml.order-md-2 {\n    order: 2 !important;\n  }\n  .jhtml .order-md-3 , .jhtml.order-md-3 {\n    order: 3 !important;\n  }\n  .jhtml .order-md-4 , .jhtml.order-md-4 {\n    order: 4 !important;\n  }\n  .jhtml .order-md-5 , .jhtml.order-md-5 {\n    order: 5 !important;\n  }\n  .jhtml .order-md-last , .jhtml.order-md-last {\n    order: 6 !important;\n  }\n  .jhtml .m-md-0 , .jhtml.m-md-0 {\n    margin: 0 !important;\n  }\n  .jhtml .m-md-1 , .jhtml.m-md-1 {\n    margin: 0.25rem !important;\n  }\n  .jhtml .m-md-2 , .jhtml.m-md-2 {\n    margin: 0.5rem !important;\n  }\n  .jhtml .m-md-3 , .jhtml.m-md-3 {\n    margin: 1rem !important;\n  }\n  .jhtml .m-md-4 , .jhtml.m-md-4 {\n    margin: 1.5rem !important;\n  }\n  .jhtml .m-md-5 , .jhtml.m-md-5 {\n    margin: 3rem !important;\n  }\n  .jhtml .m-md-auto , .jhtml.m-md-auto {\n    margin: auto !important;\n  }\n  .jhtml .mx-md-0 , .jhtml.mx-md-0 {\n    margin-right: 0 !important;\n    margin-left: 0 !important;\n  }\n  .jhtml .mx-md-1 , .jhtml.mx-md-1 {\n    margin-right: 0.25rem !important;\n    margin-left: 0.25rem !important;\n  }\n  .jhtml .mx-md-2 , .jhtml.mx-md-2 {\n    margin-right: 0.5rem !important;\n    margin-left: 0.5rem !important;\n  }\n  .jhtml .mx-md-3 , .jhtml.mx-md-3 {\n    margin-right: 1rem !important;\n    margin-left: 1rem !important;\n  }\n  .jhtml .mx-md-4 , .jhtml.mx-md-4 {\n    margin-right: 1.5rem !important;\n    margin-left: 1.5rem !important;\n  }\n  .jhtml .mx-md-5 , .jhtml.mx-md-5 {\n    margin-right: 3rem !important;\n    margin-left: 3rem !important;\n  }\n  .jhtml .mx-md-auto , .jhtml.mx-md-auto {\n    margin-right: auto !important;\n    margin-left: auto !important;\n  }\n  .jhtml .my-md-0 , .jhtml.my-md-0 {\n    margin-top: 0 !important;\n    margin-bottom: 0 !important;\n  }\n  .jhtml .my-md-1 , .jhtml.my-md-1 {\n    margin-top: 0.25rem !important;\n    margin-bottom: 0.25rem !important;\n  }\n  .jhtml .my-md-2 , .jhtml.my-md-2 {\n    margin-top: 0.5rem !important;\n    margin-bottom: 0.5rem !important;\n  }\n  .jhtml .my-md-3 , .jhtml.my-md-3 {\n    margin-top: 1rem !important;\n    margin-bottom: 1rem !important;\n  }\n  .jhtml .my-md-4 , .jhtml.my-md-4 {\n    margin-top: 1.5rem !important;\n    margin-bottom: 1.5rem !important;\n  }\n  .jhtml .my-md-5 , .jhtml.my-md-5 {\n    margin-top: 3rem !important;\n    margin-bottom: 3rem !important;\n  }\n  .jhtml .my-md-auto , .jhtml.my-md-auto {\n    margin-top: auto !important;\n    margin-bottom: auto !important;\n  }\n  .jhtml .mt-md-0 , .jhtml.mt-md-0 {\n    margin-top: 0 !important;\n  }\n  .jhtml .mt-md-1 , .jhtml.mt-md-1 {\n    margin-top: 0.25rem !important;\n  }\n  .jhtml .mt-md-2 , .jhtml.mt-md-2 {\n    margin-top: 0.5rem !important;\n  }\n  .jhtml .mt-md-3 , .jhtml.mt-md-3 {\n    margin-top: 1rem !important;\n  }\n  .jhtml .mt-md-4 , .jhtml.mt-md-4 {\n    margin-top: 1.5rem !important;\n  }\n  .jhtml .mt-md-5 , .jhtml.mt-md-5 {\n    margin-top: 3rem !important;\n  }\n  .jhtml .mt-md-auto , .jhtml.mt-md-auto {\n    margin-top: auto !important;\n  }\n  .jhtml .me-md-0 , .jhtml.me-md-0 {\n    margin-right: 0 !important;\n  }\n  .jhtml .me-md-1 , .jhtml.me-md-1 {\n    margin-right: 0.25rem !important;\n  }\n  .jhtml .me-md-2 , .jhtml.me-md-2 {\n    margin-right: 0.5rem !important;\n  }\n  .jhtml .me-md-3 , .jhtml.me-md-3 {\n    margin-right: 1rem !important;\n  }\n  .jhtml .me-md-4 , .jhtml.me-md-4 {\n    margin-right: 1.5rem !important;\n  }\n  .jhtml .me-md-5 , .jhtml.me-md-5 {\n    margin-right: 3rem !important;\n  }\n  .jhtml .me-md-auto , .jhtml.me-md-auto {\n    margin-right: auto !important;\n  }\n  .jhtml .mb-md-0 , .jhtml.mb-md-0 {\n    margin-bottom: 0 !important;\n  }\n  .jhtml .mb-md-1 , .jhtml.mb-md-1 {\n    margin-bottom: 0.25rem !important;\n  }\n  .jhtml .mb-md-2 , .jhtml.mb-md-2 {\n    margin-bottom: 0.5rem !important;\n  }\n  .jhtml .mb-md-3 , .jhtml.mb-md-3 {\n    margin-bottom: 1rem !important;\n  }\n  .jhtml .mb-md-4 , .jhtml.mb-md-4 {\n    margin-bottom: 1.5rem !important;\n  }\n  .jhtml .mb-md-5 , .jhtml.mb-md-5 {\n    margin-bottom: 3rem !important;\n  }\n  .jhtml .mb-md-auto , .jhtml.mb-md-auto {\n    margin-bottom: auto !important;\n  }\n  .jhtml .ms-md-0 , .jhtml.ms-md-0 {\n    margin-left: 0 !important;\n  }\n  .jhtml .ms-md-1 , .jhtml.ms-md-1 {\n    margin-left: 0.25rem !important;\n  }\n  .jhtml .ms-md-2 , .jhtml.ms-md-2 {\n    margin-left: 0.5rem !important;\n  }\n  .jhtml .ms-md-3 , .jhtml.ms-md-3 {\n    margin-left: 1rem !important;\n  }\n  .jhtml .ms-md-4 , .jhtml.ms-md-4 {\n    margin-left: 1.5rem !important;\n  }\n  .jhtml .ms-md-5 , .jhtml.ms-md-5 {\n    margin-left: 3rem !important;\n  }\n  .jhtml .ms-md-auto , .jhtml.ms-md-auto {\n    margin-left: auto !important;\n  }\n  .jhtml .p-md-0 , .jhtml.p-md-0 {\n    padding: 0 !important;\n  }\n  .jhtml .p-md-1 , .jhtml.p-md-1 {\n    padding: 0.25rem !important;\n  }\n  .jhtml .p-md-2 , .jhtml.p-md-2 {\n    padding: 0.5rem !important;\n  }\n  .jhtml .p-md-3 , .jhtml.p-md-3 {\n    padding: 1rem !important;\n  }\n  .jhtml .p-md-4 , .jhtml.p-md-4 {\n    padding: 1.5rem !important;\n  }\n  .jhtml .p-md-5 , .jhtml.p-md-5 {\n    padding: 3rem !important;\n  }\n  .jhtml .px-md-0 , .jhtml.px-md-0 {\n    padding-right: 0 !important;\n    padding-left: 0 !important;\n  }\n  .jhtml .px-md-1 , .jhtml.px-md-1 {\n    padding-right: 0.25rem !important;\n    padding-left: 0.25rem !important;\n  }\n  .jhtml .px-md-2 , .jhtml.px-md-2 {\n    padding-right: 0.5rem !important;\n    padding-left: 0.5rem !important;\n  }\n  .jhtml .px-md-3 , .jhtml.px-md-3 {\n    padding-right: 1rem !important;\n    padding-left: 1rem !important;\n  }\n  .jhtml .px-md-4 , .jhtml.px-md-4 {\n    padding-right: 1.5rem !important;\n    padding-left: 1.5rem !important;\n  }\n  .jhtml .px-md-5 , .jhtml.px-md-5 {\n    padding-right: 3rem !important;\n    padding-left: 3rem !important;\n  }\n  .jhtml .py-md-0 , .jhtml.py-md-0 {\n    padding-top: 0 !important;\n    padding-bottom: 0 !important;\n  }\n  .jhtml .py-md-1 , .jhtml.py-md-1 {\n    padding-top: 0.25rem !important;\n    padding-bottom: 0.25rem !important;\n  }\n  .jhtml .py-md-2 , .jhtml.py-md-2 {\n    padding-top: 0.5rem !important;\n    padding-bottom: 0.5rem !important;\n  }\n  .jhtml .py-md-3 , .jhtml.py-md-3 {\n    padding-top: 1rem !important;\n    padding-bottom: 1rem !important;\n  }\n  .jhtml .py-md-4 , .jhtml.py-md-4 {\n    padding-top: 1.5rem !important;\n    padding-bottom: 1.5rem !important;\n  }\n  .jhtml .py-md-5 , .jhtml.py-md-5 {\n    padding-top: 3rem !important;\n    padding-bottom: 3rem !important;\n  }\n  .jhtml .pt-md-0 , .jhtml.pt-md-0 {\n    padding-top: 0 !important;\n  }\n  .jhtml .pt-md-1 , .jhtml.pt-md-1 {\n    padding-top: 0.25rem !important;\n  }\n  .jhtml .pt-md-2 , .jhtml.pt-md-2 {\n    padding-top: 0.5rem !important;\n  }\n  .jhtml .pt-md-3 , .jhtml.pt-md-3 {\n    padding-top: 1rem !important;\n  }\n  .jhtml .pt-md-4 , .jhtml.pt-md-4 {\n    padding-top: 1.5rem !important;\n  }\n  .jhtml .pt-md-5 , .jhtml.pt-md-5 {\n    padding-top: 3rem !important;\n  }\n  .jhtml .pe-md-0 , .jhtml.pe-md-0 {\n    padding-right: 0 !important;\n  }\n  .jhtml .pe-md-1 , .jhtml.pe-md-1 {\n    padding-right: 0.25rem !important;\n  }\n  .jhtml .pe-md-2 , .jhtml.pe-md-2 {\n    padding-right: 0.5rem !important;\n  }\n  .jhtml .pe-md-3 , .jhtml.pe-md-3 {\n    padding-right: 1rem !important;\n  }\n  .jhtml .pe-md-4 , .jhtml.pe-md-4 {\n    padding-right: 1.5rem !important;\n  }\n  .jhtml .pe-md-5 , .jhtml.pe-md-5 {\n    padding-right: 3rem !important;\n  }\n  .jhtml .pb-md-0 , .jhtml.pb-md-0 {\n    padding-bottom: 0 !important;\n  }\n  .jhtml .pb-md-1 , .jhtml.pb-md-1 {\n    padding-bottom: 0.25rem !important;\n  }\n  .jhtml .pb-md-2 , .jhtml.pb-md-2 {\n    padding-bottom: 0.5rem !important;\n  }\n  .jhtml .pb-md-3 , .jhtml.pb-md-3 {\n    padding-bottom: 1rem !important;\n  }\n  .jhtml .pb-md-4 , .jhtml.pb-md-4 {\n    padding-bottom: 1.5rem !important;\n  }\n  .jhtml .pb-md-5 , .jhtml.pb-md-5 {\n    padding-bottom: 3rem !important;\n  }\n  .jhtml .ps-md-0 , .jhtml.ps-md-0 {\n    padding-left: 0 !important;\n  }\n  .jhtml .ps-md-1 , .jhtml.ps-md-1 {\n    padding-left: 0.25rem !important;\n  }\n  .jhtml .ps-md-2 , .jhtml.ps-md-2 {\n    padding-left: 0.5rem !important;\n  }\n  .jhtml .ps-md-3 , .jhtml.ps-md-3 {\n    padding-left: 1rem !important;\n  }\n  .jhtml .ps-md-4 , .jhtml.ps-md-4 {\n    padding-left: 1.5rem !important;\n  }\n  .jhtml .ps-md-5 , .jhtml.ps-md-5 {\n    padding-left: 3rem !important;\n  }\n  .jhtml .text-md-start , .jhtml.text-md-start {\n    text-align: left !important;\n  }\n  .jhtml .text-md-end , .jhtml.text-md-end {\n    text-align: right !important;\n  }\n  .jhtml .text-md-center , .jhtml.text-md-center {\n    text-align: center !important;\n  }\n}\n@media (min-width: 992px) {\n  .jhtml .float-lg-start , .jhtml.float-lg-start {\n    float: left !important;\n  }\n  .jhtml .float-lg-end , .jhtml.float-lg-end {\n    float: right !important;\n  }\n  .jhtml .float-lg-none , .jhtml.float-lg-none {\n    float: none !important;\n  }\n  .jhtml .d-lg-inline , .jhtml.d-lg-inline {\n    display: inline !important;\n  }\n  .jhtml .d-lg-inline-block , .jhtml.d-lg-inline-block {\n    display: inline-block !important;\n  }\n  .jhtml .d-lg-block , .jhtml.d-lg-block {\n    display: block !important;\n  }\n  .jhtml .d-lg-grid , .jhtml.d-lg-grid {\n    display: grid !important;\n  }\n  .jhtml .d-lg-table , .jhtml.d-lg-table {\n    display: table !important;\n  }\n  .jhtml .d-lg-table-row , .jhtml.d-lg-table-row {\n    display: table-row !important;\n  }\n  .jhtml .d-lg-table-cell , .jhtml.d-lg-table-cell {\n    display: table-cell !important;\n  }\n  .jhtml .d-lg-flex , .jhtml.d-lg-flex {\n    display: flex !important;\n  }\n  .jhtml .d-lg-inline-flex , .jhtml.d-lg-inline-flex {\n    display: inline-flex !important;\n  }\n  .jhtml .d-lg-none , .jhtml.d-lg-none {\n    display: none !important;\n  }\n  .jhtml .flex-lg-fill , .jhtml.flex-lg-fill {\n    flex: 1 1 auto !important;\n  }\n  .jhtml .flex-lg-row , .jhtml.flex-lg-row {\n    flex-direction: row !important;\n  }\n  .jhtml .flex-lg-column , .jhtml.flex-lg-column {\n    flex-direction: column !important;\n  }\n  .jhtml .flex-lg-row-reverse , .jhtml.flex-lg-row-reverse {\n    flex-direction: row-reverse !important;\n  }\n  .jhtml .flex-lg-column-reverse , .jhtml.flex-lg-column-reverse {\n    flex-direction: column-reverse !important;\n  }\n  .jhtml .flex-lg-grow-0 , .jhtml.flex-lg-grow-0 {\n    flex-grow: 0 !important;\n  }\n  .jhtml .flex-lg-grow-1 , .jhtml.flex-lg-grow-1 {\n    flex-grow: 1 !important;\n  }\n  .jhtml .flex-lg-shrink-0 , .jhtml.flex-lg-shrink-0 {\n    flex-shrink: 0 !important;\n  }\n  .jhtml .flex-lg-shrink-1 , .jhtml.flex-lg-shrink-1 {\n    flex-shrink: 1 !important;\n  }\n  .jhtml .flex-lg-wrap , .jhtml.flex-lg-wrap {\n    flex-wrap: wrap !important;\n  }\n  .jhtml .flex-lg-nowrap , .jhtml.flex-lg-nowrap {\n    flex-wrap: nowrap !important;\n  }\n  .jhtml .flex-lg-wrap-reverse , .jhtml.flex-lg-wrap-reverse {\n    flex-wrap: wrap-reverse !important;\n  }\n  .jhtml .gap-lg-0 , .jhtml.gap-lg-0 {\n    gap: 0 !important;\n  }\n  .jhtml .gap-lg-1 , .jhtml.gap-lg-1 {\n    gap: 0.25rem !important;\n  }\n  .jhtml .gap-lg-2 , .jhtml.gap-lg-2 {\n    gap: 0.5rem !important;\n  }\n  .jhtml .gap-lg-3 , .jhtml.gap-lg-3 {\n    gap: 1rem !important;\n  }\n  .jhtml .gap-lg-4 , .jhtml.gap-lg-4 {\n    gap: 1.5rem !important;\n  }\n  .jhtml .gap-lg-5 , .jhtml.gap-lg-5 {\n    gap: 3rem !important;\n  }\n  .jhtml .justify-content-lg-start , .jhtml.justify-content-lg-start {\n    justify-content: flex-start !important;\n  }\n  .jhtml .justify-content-lg-end , .jhtml.justify-content-lg-end {\n    justify-content: flex-end !important;\n  }\n  .jhtml .justify-content-lg-center , .jhtml.justify-content-lg-center {\n    justify-content: center !important;\n  }\n  .jhtml .justify-content-lg-between , .jhtml.justify-content-lg-between {\n    justify-content: space-between !important;\n  }\n  .jhtml .justify-content-lg-around , .jhtml.justify-content-lg-around {\n    justify-content: space-around !important;\n  }\n  .jhtml .justify-content-lg-evenly , .jhtml.justify-content-lg-evenly {\n    justify-content: space-evenly !important;\n  }\n  .jhtml .align-items-lg-start , .jhtml.align-items-lg-start {\n    align-items: flex-start !important;\n  }\n  .jhtml .align-items-lg-end , .jhtml.align-items-lg-end {\n    align-items: flex-end !important;\n  }\n  .jhtml .align-items-lg-center , .jhtml.align-items-lg-center {\n    align-items: center !important;\n  }\n  .jhtml .align-items-lg-baseline , .jhtml.align-items-lg-baseline {\n    align-items: baseline !important;\n  }\n  .jhtml .align-items-lg-stretch , .jhtml.align-items-lg-stretch {\n    align-items: stretch !important;\n  }\n  .jhtml .align-content-lg-start , .jhtml.align-content-lg-start {\n    align-content: flex-start !important;\n  }\n  .jhtml .align-content-lg-end , .jhtml.align-content-lg-end {\n    align-content: flex-end !important;\n  }\n  .jhtml .align-content-lg-center , .jhtml.align-content-lg-center {\n    align-content: center !important;\n  }\n  .jhtml .align-content-lg-between , .jhtml.align-content-lg-between {\n    align-content: space-between !important;\n  }\n  .jhtml .align-content-lg-around , .jhtml.align-content-lg-around {\n    align-content: space-around !important;\n  }\n  .jhtml .align-content-lg-stretch , .jhtml.align-content-lg-stretch {\n    align-content: stretch !important;\n  }\n  .jhtml .align-self-lg-auto , .jhtml.align-self-lg-auto {\n    align-self: auto !important;\n  }\n  .jhtml .align-self-lg-start , .jhtml.align-self-lg-start {\n    align-self: flex-start !important;\n  }\n  .jhtml .align-self-lg-end , .jhtml.align-self-lg-end {\n    align-self: flex-end !important;\n  }\n  .jhtml .align-self-lg-center , .jhtml.align-self-lg-center {\n    align-self: center !important;\n  }\n  .jhtml .align-self-lg-baseline , .jhtml.align-self-lg-baseline {\n    align-self: baseline !important;\n  }\n  .jhtml .align-self-lg-stretch , .jhtml.align-self-lg-stretch {\n    align-self: stretch !important;\n  }\n  .jhtml .order-lg-first , .jhtml.order-lg-first {\n    order: -1 !important;\n  }\n  .jhtml .order-lg-0 , .jhtml.order-lg-0 {\n    order: 0 !important;\n  }\n  .jhtml .order-lg-1 , .jhtml.order-lg-1 {\n    order: 1 !important;\n  }\n  .jhtml .order-lg-2 , .jhtml.order-lg-2 {\n    order: 2 !important;\n  }\n  .jhtml .order-lg-3 , .jhtml.order-lg-3 {\n    order: 3 !important;\n  }\n  .jhtml .order-lg-4 , .jhtml.order-lg-4 {\n    order: 4 !important;\n  }\n  .jhtml .order-lg-5 , .jhtml.order-lg-5 {\n    order: 5 !important;\n  }\n  .jhtml .order-lg-last , .jhtml.order-lg-last {\n    order: 6 !important;\n  }\n  .jhtml .m-lg-0 , .jhtml.m-lg-0 {\n    margin: 0 !important;\n  }\n  .jhtml .m-lg-1 , .jhtml.m-lg-1 {\n    margin: 0.25rem !important;\n  }\n  .jhtml .m-lg-2 , .jhtml.m-lg-2 {\n    margin: 0.5rem !important;\n  }\n  .jhtml .m-lg-3 , .jhtml.m-lg-3 {\n    margin: 1rem !important;\n  }\n  .jhtml .m-lg-4 , .jhtml.m-lg-4 {\n    margin: 1.5rem !important;\n  }\n  .jhtml .m-lg-5 , .jhtml.m-lg-5 {\n    margin: 3rem !important;\n  }\n  .jhtml .m-lg-auto , .jhtml.m-lg-auto {\n    margin: auto !important;\n  }\n  .jhtml .mx-lg-0 , .jhtml.mx-lg-0 {\n    margin-right: 0 !important;\n    margin-left: 0 !important;\n  }\n  .jhtml .mx-lg-1 , .jhtml.mx-lg-1 {\n    margin-right: 0.25rem !important;\n    margin-left: 0.25rem !important;\n  }\n  .jhtml .mx-lg-2 , .jhtml.mx-lg-2 {\n    margin-right: 0.5rem !important;\n    margin-left: 0.5rem !important;\n  }\n  .jhtml .mx-lg-3 , .jhtml.mx-lg-3 {\n    margin-right: 1rem !important;\n    margin-left: 1rem !important;\n  }\n  .jhtml .mx-lg-4 , .jhtml.mx-lg-4 {\n    margin-right: 1.5rem !important;\n    margin-left: 1.5rem !important;\n  }\n  .jhtml .mx-lg-5 , .jhtml.mx-lg-5 {\n    margin-right: 3rem !important;\n    margin-left: 3rem !important;\n  }\n  .jhtml .mx-lg-auto , .jhtml.mx-lg-auto {\n    margin-right: auto !important;\n    margin-left: auto !important;\n  }\n  .jhtml .my-lg-0 , .jhtml.my-lg-0 {\n    margin-top: 0 !important;\n    margin-bottom: 0 !important;\n  }\n  .jhtml .my-lg-1 , .jhtml.my-lg-1 {\n    margin-top: 0.25rem !important;\n    margin-bottom: 0.25rem !important;\n  }\n  .jhtml .my-lg-2 , .jhtml.my-lg-2 {\n    margin-top: 0.5rem !important;\n    margin-bottom: 0.5rem !important;\n  }\n  .jhtml .my-lg-3 , .jhtml.my-lg-3 {\n    margin-top: 1rem !important;\n    margin-bottom: 1rem !important;\n  }\n  .jhtml .my-lg-4 , .jhtml.my-lg-4 {\n    margin-top: 1.5rem !important;\n    margin-bottom: 1.5rem !important;\n  }\n  .jhtml .my-lg-5 , .jhtml.my-lg-5 {\n    margin-top: 3rem !important;\n    margin-bottom: 3rem !important;\n  }\n  .jhtml .my-lg-auto , .jhtml.my-lg-auto {\n    margin-top: auto !important;\n    margin-bottom: auto !important;\n  }\n  .jhtml .mt-lg-0 , .jhtml.mt-lg-0 {\n    margin-top: 0 !important;\n  }\n  .jhtml .mt-lg-1 , .jhtml.mt-lg-1 {\n    margin-top: 0.25rem !important;\n  }\n  .jhtml .mt-lg-2 , .jhtml.mt-lg-2 {\n    margin-top: 0.5rem !important;\n  }\n  .jhtml .mt-lg-3 , .jhtml.mt-lg-3 {\n    margin-top: 1rem !important;\n  }\n  .jhtml .mt-lg-4 , .jhtml.mt-lg-4 {\n    margin-top: 1.5rem !important;\n  }\n  .jhtml .mt-lg-5 , .jhtml.mt-lg-5 {\n    margin-top: 3rem !important;\n  }\n  .jhtml .mt-lg-auto , .jhtml.mt-lg-auto {\n    margin-top: auto !important;\n  }\n  .jhtml .me-lg-0 , .jhtml.me-lg-0 {\n    margin-right: 0 !important;\n  }\n  .jhtml .me-lg-1 , .jhtml.me-lg-1 {\n    margin-right: 0.25rem !important;\n  }\n  .jhtml .me-lg-2 , .jhtml.me-lg-2 {\n    margin-right: 0.5rem !important;\n  }\n  .jhtml .me-lg-3 , .jhtml.me-lg-3 {\n    margin-right: 1rem !important;\n  }\n  .jhtml .me-lg-4 , .jhtml.me-lg-4 {\n    margin-right: 1.5rem !important;\n  }\n  .jhtml .me-lg-5 , .jhtml.me-lg-5 {\n    margin-right: 3rem !important;\n  }\n  .jhtml .me-lg-auto , .jhtml.me-lg-auto {\n    margin-right: auto !important;\n  }\n  .jhtml .mb-lg-0 , .jhtml.mb-lg-0 {\n    margin-bottom: 0 !important;\n  }\n  .jhtml .mb-lg-1 , .jhtml.mb-lg-1 {\n    margin-bottom: 0.25rem !important;\n  }\n  .jhtml .mb-lg-2 , .jhtml.mb-lg-2 {\n    margin-bottom: 0.5rem !important;\n  }\n  .jhtml .mb-lg-3 , .jhtml.mb-lg-3 {\n    margin-bottom: 1rem !important;\n  }\n  .jhtml .mb-lg-4 , .jhtml.mb-lg-4 {\n    margin-bottom: 1.5rem !important;\n  }\n  .jhtml .mb-lg-5 , .jhtml.mb-lg-5 {\n    margin-bottom: 3rem !important;\n  }\n  .jhtml .mb-lg-auto , .jhtml.mb-lg-auto {\n    margin-bottom: auto !important;\n  }\n  .jhtml .ms-lg-0 , .jhtml.ms-lg-0 {\n    margin-left: 0 !important;\n  }\n  .jhtml .ms-lg-1 , .jhtml.ms-lg-1 {\n    margin-left: 0.25rem !important;\n  }\n  .jhtml .ms-lg-2 , .jhtml.ms-lg-2 {\n    margin-left: 0.5rem !important;\n  }\n  .jhtml .ms-lg-3 , .jhtml.ms-lg-3 {\n    margin-left: 1rem !important;\n  }\n  .jhtml .ms-lg-4 , .jhtml.ms-lg-4 {\n    margin-left: 1.5rem !important;\n  }\n  .jhtml .ms-lg-5 , .jhtml.ms-lg-5 {\n    margin-left: 3rem !important;\n  }\n  .jhtml .ms-lg-auto , .jhtml.ms-lg-auto {\n    margin-left: auto !important;\n  }\n  .jhtml .p-lg-0 , .jhtml.p-lg-0 {\n    padding: 0 !important;\n  }\n  .jhtml .p-lg-1 , .jhtml.p-lg-1 {\n    padding: 0.25rem !important;\n  }\n  .jhtml .p-lg-2 , .jhtml.p-lg-2 {\n    padding: 0.5rem !important;\n  }\n  .jhtml .p-lg-3 , .jhtml.p-lg-3 {\n    padding: 1rem !important;\n  }\n  .jhtml .p-lg-4 , .jhtml.p-lg-4 {\n    padding: 1.5rem !important;\n  }\n  .jhtml .p-lg-5 , .jhtml.p-lg-5 {\n    padding: 3rem !important;\n  }\n  .jhtml .px-lg-0 , .jhtml.px-lg-0 {\n    padding-right: 0 !important;\n    padding-left: 0 !important;\n  }\n  .jhtml .px-lg-1 , .jhtml.px-lg-1 {\n    padding-right: 0.25rem !important;\n    padding-left: 0.25rem !important;\n  }\n  .jhtml .px-lg-2 , .jhtml.px-lg-2 {\n    padding-right: 0.5rem !important;\n    padding-left: 0.5rem !important;\n  }\n  .jhtml .px-lg-3 , .jhtml.px-lg-3 {\n    padding-right: 1rem !important;\n    padding-left: 1rem !important;\n  }\n  .jhtml .px-lg-4 , .jhtml.px-lg-4 {\n    padding-right: 1.5rem !important;\n    padding-left: 1.5rem !important;\n  }\n  .jhtml .px-lg-5 , .jhtml.px-lg-5 {\n    padding-right: 3rem !important;\n    padding-left: 3rem !important;\n  }\n  .jhtml .py-lg-0 , .jhtml.py-lg-0 {\n    padding-top: 0 !important;\n    padding-bottom: 0 !important;\n  }\n  .jhtml .py-lg-1 , .jhtml.py-lg-1 {\n    padding-top: 0.25rem !important;\n    padding-bottom: 0.25rem !important;\n  }\n  .jhtml .py-lg-2 , .jhtml.py-lg-2 {\n    padding-top: 0.5rem !important;\n    padding-bottom: 0.5rem !important;\n  }\n  .jhtml .py-lg-3 , .jhtml.py-lg-3 {\n    padding-top: 1rem !important;\n    padding-bottom: 1rem !important;\n  }\n  .jhtml .py-lg-4 , .jhtml.py-lg-4 {\n    padding-top: 1.5rem !important;\n    padding-bottom: 1.5rem !important;\n  }\n  .jhtml .py-lg-5 , .jhtml.py-lg-5 {\n    padding-top: 3rem !important;\n    padding-bottom: 3rem !important;\n  }\n  .jhtml .pt-lg-0 , .jhtml.pt-lg-0 {\n    padding-top: 0 !important;\n  }\n  .jhtml .pt-lg-1 , .jhtml.pt-lg-1 {\n    padding-top: 0.25rem !important;\n  }\n  .jhtml .pt-lg-2 , .jhtml.pt-lg-2 {\n    padding-top: 0.5rem !important;\n  }\n  .jhtml .pt-lg-3 , .jhtml.pt-lg-3 {\n    padding-top: 1rem !important;\n  }\n  .jhtml .pt-lg-4 , .jhtml.pt-lg-4 {\n    padding-top: 1.5rem !important;\n  }\n  .jhtml .pt-lg-5 , .jhtml.pt-lg-5 {\n    padding-top: 3rem !important;\n  }\n  .jhtml .pe-lg-0 , .jhtml.pe-lg-0 {\n    padding-right: 0 !important;\n  }\n  .jhtml .pe-lg-1 , .jhtml.pe-lg-1 {\n    padding-right: 0.25rem !important;\n  }\n  .jhtml .pe-lg-2 , .jhtml.pe-lg-2 {\n    padding-right: 0.5rem !important;\n  }\n  .jhtml .pe-lg-3 , .jhtml.pe-lg-3 {\n    padding-right: 1rem !important;\n  }\n  .jhtml .pe-lg-4 , .jhtml.pe-lg-4 {\n    padding-right: 1.5rem !important;\n  }\n  .jhtml .pe-lg-5 , .jhtml.pe-lg-5 {\n    padding-right: 3rem !important;\n  }\n  .jhtml .pb-lg-0 , .jhtml.pb-lg-0 {\n    padding-bottom: 0 !important;\n  }\n  .jhtml .pb-lg-1 , .jhtml.pb-lg-1 {\n    padding-bottom: 0.25rem !important;\n  }\n  .jhtml .pb-lg-2 , .jhtml.pb-lg-2 {\n    padding-bottom: 0.5rem !important;\n  }\n  .jhtml .pb-lg-3 , .jhtml.pb-lg-3 {\n    padding-bottom: 1rem !important;\n  }\n  .jhtml .pb-lg-4 , .jhtml.pb-lg-4 {\n    padding-bottom: 1.5rem !important;\n  }\n  .jhtml .pb-lg-5 , .jhtml.pb-lg-5 {\n    padding-bottom: 3rem !important;\n  }\n  .jhtml .ps-lg-0 , .jhtml.ps-lg-0 {\n    padding-left: 0 !important;\n  }\n  .jhtml .ps-lg-1 , .jhtml.ps-lg-1 {\n    padding-left: 0.25rem !important;\n  }\n  .jhtml .ps-lg-2 , .jhtml.ps-lg-2 {\n    padding-left: 0.5rem !important;\n  }\n  .jhtml .ps-lg-3 , .jhtml.ps-lg-3 {\n    padding-left: 1rem !important;\n  }\n  .jhtml .ps-lg-4 , .jhtml.ps-lg-4 {\n    padding-left: 1.5rem !important;\n  }\n  .jhtml .ps-lg-5 , .jhtml.ps-lg-5 {\n    padding-left: 3rem !important;\n  }\n  .jhtml .text-lg-start , .jhtml.text-lg-start {\n    text-align: left !important;\n  }\n  .jhtml .text-lg-end , .jhtml.text-lg-end {\n    text-align: right !important;\n  }\n  .jhtml .text-lg-center , .jhtml.text-lg-center {\n    text-align: center !important;\n  }\n}\n@media (min-width: 1200px) {\n  .jhtml .float-xl-start , .jhtml.float-xl-start {\n    float: left !important;\n  }\n  .jhtml .float-xl-end , .jhtml.float-xl-end {\n    float: right !important;\n  }\n  .jhtml .float-xl-none , .jhtml.float-xl-none {\n    float: none !important;\n  }\n  .jhtml .d-xl-inline , .jhtml.d-xl-inline {\n    display: inline !important;\n  }\n  .jhtml .d-xl-inline-block , .jhtml.d-xl-inline-block {\n    display: inline-block !important;\n  }\n  .jhtml .d-xl-block , .jhtml.d-xl-block {\n    display: block !important;\n  }\n  .jhtml .d-xl-grid , .jhtml.d-xl-grid {\n    display: grid !important;\n  }\n  .jhtml .d-xl-table , .jhtml.d-xl-table {\n    display: table !important;\n  }\n  .jhtml .d-xl-table-row , .jhtml.d-xl-table-row {\n    display: table-row !important;\n  }\n  .jhtml .d-xl-table-cell , .jhtml.d-xl-table-cell {\n    display: table-cell !important;\n  }\n  .jhtml .d-xl-flex , .jhtml.d-xl-flex {\n    display: flex !important;\n  }\n  .jhtml .d-xl-inline-flex , .jhtml.d-xl-inline-flex {\n    display: inline-flex !important;\n  }\n  .jhtml .d-xl-none , .jhtml.d-xl-none {\n    display: none !important;\n  }\n  .jhtml .flex-xl-fill , .jhtml.flex-xl-fill {\n    flex: 1 1 auto !important;\n  }\n  .jhtml .flex-xl-row , .jhtml.flex-xl-row {\n    flex-direction: row !important;\n  }\n  .jhtml .flex-xl-column , .jhtml.flex-xl-column {\n    flex-direction: column !important;\n  }\n  .jhtml .flex-xl-row-reverse , .jhtml.flex-xl-row-reverse {\n    flex-direction: row-reverse !important;\n  }\n  .jhtml .flex-xl-column-reverse , .jhtml.flex-xl-column-reverse {\n    flex-direction: column-reverse !important;\n  }\n  .jhtml .flex-xl-grow-0 , .jhtml.flex-xl-grow-0 {\n    flex-grow: 0 !important;\n  }\n  .jhtml .flex-xl-grow-1 , .jhtml.flex-xl-grow-1 {\n    flex-grow: 1 !important;\n  }\n  .jhtml .flex-xl-shrink-0 , .jhtml.flex-xl-shrink-0 {\n    flex-shrink: 0 !important;\n  }\n  .jhtml .flex-xl-shrink-1 , .jhtml.flex-xl-shrink-1 {\n    flex-shrink: 1 !important;\n  }\n  .jhtml .flex-xl-wrap , .jhtml.flex-xl-wrap {\n    flex-wrap: wrap !important;\n  }\n  .jhtml .flex-xl-nowrap , .jhtml.flex-xl-nowrap {\n    flex-wrap: nowrap !important;\n  }\n  .jhtml .flex-xl-wrap-reverse , .jhtml.flex-xl-wrap-reverse {\n    flex-wrap: wrap-reverse !important;\n  }\n  .jhtml .gap-xl-0 , .jhtml.gap-xl-0 {\n    gap: 0 !important;\n  }\n  .jhtml .gap-xl-1 , .jhtml.gap-xl-1 {\n    gap: 0.25rem !important;\n  }\n  .jhtml .gap-xl-2 , .jhtml.gap-xl-2 {\n    gap: 0.5rem !important;\n  }\n  .jhtml .gap-xl-3 , .jhtml.gap-xl-3 {\n    gap: 1rem !important;\n  }\n  .jhtml .gap-xl-4 , .jhtml.gap-xl-4 {\n    gap: 1.5rem !important;\n  }\n  .jhtml .gap-xl-5 , .jhtml.gap-xl-5 {\n    gap: 3rem !important;\n  }\n  .jhtml .justify-content-xl-start , .jhtml.justify-content-xl-start {\n    justify-content: flex-start !important;\n  }\n  .jhtml .justify-content-xl-end , .jhtml.justify-content-xl-end {\n    justify-content: flex-end !important;\n  }\n  .jhtml .justify-content-xl-center , .jhtml.justify-content-xl-center {\n    justify-content: center !important;\n  }\n  .jhtml .justify-content-xl-between , .jhtml.justify-content-xl-between {\n    justify-content: space-between !important;\n  }\n  .jhtml .justify-content-xl-around , .jhtml.justify-content-xl-around {\n    justify-content: space-around !important;\n  }\n  .jhtml .justify-content-xl-evenly , .jhtml.justify-content-xl-evenly {\n    justify-content: space-evenly !important;\n  }\n  .jhtml .align-items-xl-start , .jhtml.align-items-xl-start {\n    align-items: flex-start !important;\n  }\n  .jhtml .align-items-xl-end , .jhtml.align-items-xl-end {\n    align-items: flex-end !important;\n  }\n  .jhtml .align-items-xl-center , .jhtml.align-items-xl-center {\n    align-items: center !important;\n  }\n  .jhtml .align-items-xl-baseline , .jhtml.align-items-xl-baseline {\n    align-items: baseline !important;\n  }\n  .jhtml .align-items-xl-stretch , .jhtml.align-items-xl-stretch {\n    align-items: stretch !important;\n  }\n  .jhtml .align-content-xl-start , .jhtml.align-content-xl-start {\n    align-content: flex-start !important;\n  }\n  .jhtml .align-content-xl-end , .jhtml.align-content-xl-end {\n    align-content: flex-end !important;\n  }\n  .jhtml .align-content-xl-center , .jhtml.align-content-xl-center {\n    align-content: center !important;\n  }\n  .jhtml .align-content-xl-between , .jhtml.align-content-xl-between {\n    align-content: space-between !important;\n  }\n  .jhtml .align-content-xl-around , .jhtml.align-content-xl-around {\n    align-content: space-around !important;\n  }\n  .jhtml .align-content-xl-stretch , .jhtml.align-content-xl-stretch {\n    align-content: stretch !important;\n  }\n  .jhtml .align-self-xl-auto , .jhtml.align-self-xl-auto {\n    align-self: auto !important;\n  }\n  .jhtml .align-self-xl-start , .jhtml.align-self-xl-start {\n    align-self: flex-start !important;\n  }\n  .jhtml .align-self-xl-end , .jhtml.align-self-xl-end {\n    align-self: flex-end !important;\n  }\n  .jhtml .align-self-xl-center , .jhtml.align-self-xl-center {\n    align-self: center !important;\n  }\n  .jhtml .align-self-xl-baseline , .jhtml.align-self-xl-baseline {\n    align-self: baseline !important;\n  }\n  .jhtml .align-self-xl-stretch , .jhtml.align-self-xl-stretch {\n    align-self: stretch !important;\n  }\n  .jhtml .order-xl-first , .jhtml.order-xl-first {\n    order: -1 !important;\n  }\n  .jhtml .order-xl-0 , .jhtml.order-xl-0 {\n    order: 0 !important;\n  }\n  .jhtml .order-xl-1 , .jhtml.order-xl-1 {\n    order: 1 !important;\n  }\n  .jhtml .order-xl-2 , .jhtml.order-xl-2 {\n    order: 2 !important;\n  }\n  .jhtml .order-xl-3 , .jhtml.order-xl-3 {\n    order: 3 !important;\n  }\n  .jhtml .order-xl-4 , .jhtml.order-xl-4 {\n    order: 4 !important;\n  }\n  .jhtml .order-xl-5 , .jhtml.order-xl-5 {\n    order: 5 !important;\n  }\n  .jhtml .order-xl-last , .jhtml.order-xl-last {\n    order: 6 !important;\n  }\n  .jhtml .m-xl-0 , .jhtml.m-xl-0 {\n    margin: 0 !important;\n  }\n  .jhtml .m-xl-1 , .jhtml.m-xl-1 {\n    margin: 0.25rem !important;\n  }\n  .jhtml .m-xl-2 , .jhtml.m-xl-2 {\n    margin: 0.5rem !important;\n  }\n  .jhtml .m-xl-3 , .jhtml.m-xl-3 {\n    margin: 1rem !important;\n  }\n  .jhtml .m-xl-4 , .jhtml.m-xl-4 {\n    margin: 1.5rem !important;\n  }\n  .jhtml .m-xl-5 , .jhtml.m-xl-5 {\n    margin: 3rem !important;\n  }\n  .jhtml .m-xl-auto , .jhtml.m-xl-auto {\n    margin: auto !important;\n  }\n  .jhtml .mx-xl-0 , .jhtml.mx-xl-0 {\n    margin-right: 0 !important;\n    margin-left: 0 !important;\n  }\n  .jhtml .mx-xl-1 , .jhtml.mx-xl-1 {\n    margin-right: 0.25rem !important;\n    margin-left: 0.25rem !important;\n  }\n  .jhtml .mx-xl-2 , .jhtml.mx-xl-2 {\n    margin-right: 0.5rem !important;\n    margin-left: 0.5rem !important;\n  }\n  .jhtml .mx-xl-3 , .jhtml.mx-xl-3 {\n    margin-right: 1rem !important;\n    margin-left: 1rem !important;\n  }\n  .jhtml .mx-xl-4 , .jhtml.mx-xl-4 {\n    margin-right: 1.5rem !important;\n    margin-left: 1.5rem !important;\n  }\n  .jhtml .mx-xl-5 , .jhtml.mx-xl-5 {\n    margin-right: 3rem !important;\n    margin-left: 3rem !important;\n  }\n  .jhtml .mx-xl-auto , .jhtml.mx-xl-auto {\n    margin-right: auto !important;\n    margin-left: auto !important;\n  }\n  .jhtml .my-xl-0 , .jhtml.my-xl-0 {\n    margin-top: 0 !important;\n    margin-bottom: 0 !important;\n  }\n  .jhtml .my-xl-1 , .jhtml.my-xl-1 {\n    margin-top: 0.25rem !important;\n    margin-bottom: 0.25rem !important;\n  }\n  .jhtml .my-xl-2 , .jhtml.my-xl-2 {\n    margin-top: 0.5rem !important;\n    margin-bottom: 0.5rem !important;\n  }\n  .jhtml .my-xl-3 , .jhtml.my-xl-3 {\n    margin-top: 1rem !important;\n    margin-bottom: 1rem !important;\n  }\n  .jhtml .my-xl-4 , .jhtml.my-xl-4 {\n    margin-top: 1.5rem !important;\n    margin-bottom: 1.5rem !important;\n  }\n  .jhtml .my-xl-5 , .jhtml.my-xl-5 {\n    margin-top: 3rem !important;\n    margin-bottom: 3rem !important;\n  }\n  .jhtml .my-xl-auto , .jhtml.my-xl-auto {\n    margin-top: auto !important;\n    margin-bottom: auto !important;\n  }\n  .jhtml .mt-xl-0 , .jhtml.mt-xl-0 {\n    margin-top: 0 !important;\n  }\n  .jhtml .mt-xl-1 , .jhtml.mt-xl-1 {\n    margin-top: 0.25rem !important;\n  }\n  .jhtml .mt-xl-2 , .jhtml.mt-xl-2 {\n    margin-top: 0.5rem !important;\n  }\n  .jhtml .mt-xl-3 , .jhtml.mt-xl-3 {\n    margin-top: 1rem !important;\n  }\n  .jhtml .mt-xl-4 , .jhtml.mt-xl-4 {\n    margin-top: 1.5rem !important;\n  }\n  .jhtml .mt-xl-5 , .jhtml.mt-xl-5 {\n    margin-top: 3rem !important;\n  }\n  .jhtml .mt-xl-auto , .jhtml.mt-xl-auto {\n    margin-top: auto !important;\n  }\n  .jhtml .me-xl-0 , .jhtml.me-xl-0 {\n    margin-right: 0 !important;\n  }\n  .jhtml .me-xl-1 , .jhtml.me-xl-1 {\n    margin-right: 0.25rem !important;\n  }\n  .jhtml .me-xl-2 , .jhtml.me-xl-2 {\n    margin-right: 0.5rem !important;\n  }\n  .jhtml .me-xl-3 , .jhtml.me-xl-3 {\n    margin-right: 1rem !important;\n  }\n  .jhtml .me-xl-4 , .jhtml.me-xl-4 {\n    margin-right: 1.5rem !important;\n  }\n  .jhtml .me-xl-5 , .jhtml.me-xl-5 {\n    margin-right: 3rem !important;\n  }\n  .jhtml .me-xl-auto , .jhtml.me-xl-auto {\n    margin-right: auto !important;\n  }\n  .jhtml .mb-xl-0 , .jhtml.mb-xl-0 {\n    margin-bottom: 0 !important;\n  }\n  .jhtml .mb-xl-1 , .jhtml.mb-xl-1 {\n    margin-bottom: 0.25rem !important;\n  }\n  .jhtml .mb-xl-2 , .jhtml.mb-xl-2 {\n    margin-bottom: 0.5rem !important;\n  }\n  .jhtml .mb-xl-3 , .jhtml.mb-xl-3 {\n    margin-bottom: 1rem !important;\n  }\n  .jhtml .mb-xl-4 , .jhtml.mb-xl-4 {\n    margin-bottom: 1.5rem !important;\n  }\n  .jhtml .mb-xl-5 , .jhtml.mb-xl-5 {\n    margin-bottom: 3rem !important;\n  }\n  .jhtml .mb-xl-auto , .jhtml.mb-xl-auto {\n    margin-bottom: auto !important;\n  }\n  .jhtml .ms-xl-0 , .jhtml.ms-xl-0 {\n    margin-left: 0 !important;\n  }\n  .jhtml .ms-xl-1 , .jhtml.ms-xl-1 {\n    margin-left: 0.25rem !important;\n  }\n  .jhtml .ms-xl-2 , .jhtml.ms-xl-2 {\n    margin-left: 0.5rem !important;\n  }\n  .jhtml .ms-xl-3 , .jhtml.ms-xl-3 {\n    margin-left: 1rem !important;\n  }\n  .jhtml .ms-xl-4 , .jhtml.ms-xl-4 {\n    margin-left: 1.5rem !important;\n  }\n  .jhtml .ms-xl-5 , .jhtml.ms-xl-5 {\n    margin-left: 3rem !important;\n  }\n  .jhtml .ms-xl-auto , .jhtml.ms-xl-auto {\n    margin-left: auto !important;\n  }\n  .jhtml .p-xl-0 , .jhtml.p-xl-0 {\n    padding: 0 !important;\n  }\n  .jhtml .p-xl-1 , .jhtml.p-xl-1 {\n    padding: 0.25rem !important;\n  }\n  .jhtml .p-xl-2 , .jhtml.p-xl-2 {\n    padding: 0.5rem !important;\n  }\n  .jhtml .p-xl-3 , .jhtml.p-xl-3 {\n    padding: 1rem !important;\n  }\n  .jhtml .p-xl-4 , .jhtml.p-xl-4 {\n    padding: 1.5rem !important;\n  }\n  .jhtml .p-xl-5 , .jhtml.p-xl-5 {\n    padding: 3rem !important;\n  }\n  .jhtml .px-xl-0 , .jhtml.px-xl-0 {\n    padding-right: 0 !important;\n    padding-left: 0 !important;\n  }\n  .jhtml .px-xl-1 , .jhtml.px-xl-1 {\n    padding-right: 0.25rem !important;\n    padding-left: 0.25rem !important;\n  }\n  .jhtml .px-xl-2 , .jhtml.px-xl-2 {\n    padding-right: 0.5rem !important;\n    padding-left: 0.5rem !important;\n  }\n  .jhtml .px-xl-3 , .jhtml.px-xl-3 {\n    padding-right: 1rem !important;\n    padding-left: 1rem !important;\n  }\n  .jhtml .px-xl-4 , .jhtml.px-xl-4 {\n    padding-right: 1.5rem !important;\n    padding-left: 1.5rem !important;\n  }\n  .jhtml .px-xl-5 , .jhtml.px-xl-5 {\n    padding-right: 3rem !important;\n    padding-left: 3rem !important;\n  }\n  .jhtml .py-xl-0 , .jhtml.py-xl-0 {\n    padding-top: 0 !important;\n    padding-bottom: 0 !important;\n  }\n  .jhtml .py-xl-1 , .jhtml.py-xl-1 {\n    padding-top: 0.25rem !important;\n    padding-bottom: 0.25rem !important;\n  }\n  .jhtml .py-xl-2 , .jhtml.py-xl-2 {\n    padding-top: 0.5rem !important;\n    padding-bottom: 0.5rem !important;\n  }\n  .jhtml .py-xl-3 , .jhtml.py-xl-3 {\n    padding-top: 1rem !important;\n    padding-bottom: 1rem !important;\n  }\n  .jhtml .py-xl-4 , .jhtml.py-xl-4 {\n    padding-top: 1.5rem !important;\n    padding-bottom: 1.5rem !important;\n  }\n  .jhtml .py-xl-5 , .jhtml.py-xl-5 {\n    padding-top: 3rem !important;\n    padding-bottom: 3rem !important;\n  }\n  .jhtml .pt-xl-0 , .jhtml.pt-xl-0 {\n    padding-top: 0 !important;\n  }\n  .jhtml .pt-xl-1 , .jhtml.pt-xl-1 {\n    padding-top: 0.25rem !important;\n  }\n  .jhtml .pt-xl-2 , .jhtml.pt-xl-2 {\n    padding-top: 0.5rem !important;\n  }\n  .jhtml .pt-xl-3 , .jhtml.pt-xl-3 {\n    padding-top: 1rem !important;\n  }\n  .jhtml .pt-xl-4 , .jhtml.pt-xl-4 {\n    padding-top: 1.5rem !important;\n  }\n  .jhtml .pt-xl-5 , .jhtml.pt-xl-5 {\n    padding-top: 3rem !important;\n  }\n  .jhtml .pe-xl-0 , .jhtml.pe-xl-0 {\n    padding-right: 0 !important;\n  }\n  .jhtml .pe-xl-1 , .jhtml.pe-xl-1 {\n    padding-right: 0.25rem !important;\n  }\n  .jhtml .pe-xl-2 , .jhtml.pe-xl-2 {\n    padding-right: 0.5rem !important;\n  }\n  .jhtml .pe-xl-3 , .jhtml.pe-xl-3 {\n    padding-right: 1rem !important;\n  }\n  .jhtml .pe-xl-4 , .jhtml.pe-xl-4 {\n    padding-right: 1.5rem !important;\n  }\n  .jhtml .pe-xl-5 , .jhtml.pe-xl-5 {\n    padding-right: 3rem !important;\n  }\n  .jhtml .pb-xl-0 , .jhtml.pb-xl-0 {\n    padding-bottom: 0 !important;\n  }\n  .jhtml .pb-xl-1 , .jhtml.pb-xl-1 {\n    padding-bottom: 0.25rem !important;\n  }\n  .jhtml .pb-xl-2 , .jhtml.pb-xl-2 {\n    padding-bottom: 0.5rem !important;\n  }\n  .jhtml .pb-xl-3 , .jhtml.pb-xl-3 {\n    padding-bottom: 1rem !important;\n  }\n  .jhtml .pb-xl-4 , .jhtml.pb-xl-4 {\n    padding-bottom: 1.5rem !important;\n  }\n  .jhtml .pb-xl-5 , .jhtml.pb-xl-5 {\n    padding-bottom: 3rem !important;\n  }\n  .jhtml .ps-xl-0 , .jhtml.ps-xl-0 {\n    padding-left: 0 !important;\n  }\n  .jhtml .ps-xl-1 , .jhtml.ps-xl-1 {\n    padding-left: 0.25rem !important;\n  }\n  .jhtml .ps-xl-2 , .jhtml.ps-xl-2 {\n    padding-left: 0.5rem !important;\n  }\n  .jhtml .ps-xl-3 , .jhtml.ps-xl-3 {\n    padding-left: 1rem !important;\n  }\n  .jhtml .ps-xl-4 , .jhtml.ps-xl-4 {\n    padding-left: 1.5rem !important;\n  }\n  .jhtml .ps-xl-5 , .jhtml.ps-xl-5 {\n    padding-left: 3rem !important;\n  }\n  .jhtml .text-xl-start , .jhtml.text-xl-start {\n    text-align: left !important;\n  }\n  .jhtml .text-xl-end , .jhtml.text-xl-end {\n    text-align: right !important;\n  }\n  .jhtml .text-xl-center , .jhtml.text-xl-center {\n    text-align: center !important;\n  }\n}\n@media (min-width: 1400px) {\n  .jhtml .float-xxl-start , .jhtml.float-xxl-start {\n    float: left !important;\n  }\n  .jhtml .float-xxl-end , .jhtml.float-xxl-end {\n    float: right !important;\n  }\n  .jhtml .float-xxl-none , .jhtml.float-xxl-none {\n    float: none !important;\n  }\n  .jhtml .d-xxl-inline , .jhtml.d-xxl-inline {\n    display: inline !important;\n  }\n  .jhtml .d-xxl-inline-block , .jhtml.d-xxl-inline-block {\n    display: inline-block !important;\n  }\n  .jhtml .d-xxl-block , .jhtml.d-xxl-block {\n    display: block !important;\n  }\n  .jhtml .d-xxl-grid , .jhtml.d-xxl-grid {\n    display: grid !important;\n  }\n  .jhtml .d-xxl-table , .jhtml.d-xxl-table {\n    display: table !important;\n  }\n  .jhtml .d-xxl-table-row , .jhtml.d-xxl-table-row {\n    display: table-row !important;\n  }\n  .jhtml .d-xxl-table-cell , .jhtml.d-xxl-table-cell {\n    display: table-cell !important;\n  }\n  .jhtml .d-xxl-flex , .jhtml.d-xxl-flex {\n    display: flex !important;\n  }\n  .jhtml .d-xxl-inline-flex , .jhtml.d-xxl-inline-flex {\n    display: inline-flex !important;\n  }\n  .jhtml .d-xxl-none , .jhtml.d-xxl-none {\n    display: none !important;\n  }\n  .jhtml .flex-xxl-fill , .jhtml.flex-xxl-fill {\n    flex: 1 1 auto !important;\n  }\n  .jhtml .flex-xxl-row , .jhtml.flex-xxl-row {\n    flex-direction: row !important;\n  }\n  .jhtml .flex-xxl-column , .jhtml.flex-xxl-column {\n    flex-direction: column !important;\n  }\n  .jhtml .flex-xxl-row-reverse , .jhtml.flex-xxl-row-reverse {\n    flex-direction: row-reverse !important;\n  }\n  .jhtml .flex-xxl-column-reverse , .jhtml.flex-xxl-column-reverse {\n    flex-direction: column-reverse !important;\n  }\n  .jhtml .flex-xxl-grow-0 , .jhtml.flex-xxl-grow-0 {\n    flex-grow: 0 !important;\n  }\n  .jhtml .flex-xxl-grow-1 , .jhtml.flex-xxl-grow-1 {\n    flex-grow: 1 !important;\n  }\n  .jhtml .flex-xxl-shrink-0 , .jhtml.flex-xxl-shrink-0 {\n    flex-shrink: 0 !important;\n  }\n  .jhtml .flex-xxl-shrink-1 , .jhtml.flex-xxl-shrink-1 {\n    flex-shrink: 1 !important;\n  }\n  .jhtml .flex-xxl-wrap , .jhtml.flex-xxl-wrap {\n    flex-wrap: wrap !important;\n  }\n  .jhtml .flex-xxl-nowrap , .jhtml.flex-xxl-nowrap {\n    flex-wrap: nowrap !important;\n  }\n  .jhtml .flex-xxl-wrap-reverse , .jhtml.flex-xxl-wrap-reverse {\n    flex-wrap: wrap-reverse !important;\n  }\n  .jhtml .gap-xxl-0 , .jhtml.gap-xxl-0 {\n    gap: 0 !important;\n  }\n  .jhtml .gap-xxl-1 , .jhtml.gap-xxl-1 {\n    gap: 0.25rem !important;\n  }\n  .jhtml .gap-xxl-2 , .jhtml.gap-xxl-2 {\n    gap: 0.5rem !important;\n  }\n  .jhtml .gap-xxl-3 , .jhtml.gap-xxl-3 {\n    gap: 1rem !important;\n  }\n  .jhtml .gap-xxl-4 , .jhtml.gap-xxl-4 {\n    gap: 1.5rem !important;\n  }\n  .jhtml .gap-xxl-5 , .jhtml.gap-xxl-5 {\n    gap: 3rem !important;\n  }\n  .jhtml .justify-content-xxl-start , .jhtml.justify-content-xxl-start {\n    justify-content: flex-start !important;\n  }\n  .jhtml .justify-content-xxl-end , .jhtml.justify-content-xxl-end {\n    justify-content: flex-end !important;\n  }\n  .jhtml .justify-content-xxl-center , .jhtml.justify-content-xxl-center {\n    justify-content: center !important;\n  }\n  .jhtml .justify-content-xxl-between , .jhtml.justify-content-xxl-between {\n    justify-content: space-between !important;\n  }\n  .jhtml .justify-content-xxl-around , .jhtml.justify-content-xxl-around {\n    justify-content: space-around !important;\n  }\n  .jhtml .justify-content-xxl-evenly , .jhtml.justify-content-xxl-evenly {\n    justify-content: space-evenly !important;\n  }\n  .jhtml .align-items-xxl-start , .jhtml.align-items-xxl-start {\n    align-items: flex-start !important;\n  }\n  .jhtml .align-items-xxl-end , .jhtml.align-items-xxl-end {\n    align-items: flex-end !important;\n  }\n  .jhtml .align-items-xxl-center , .jhtml.align-items-xxl-center {\n    align-items: center !important;\n  }\n  .jhtml .align-items-xxl-baseline , .jhtml.align-items-xxl-baseline {\n    align-items: baseline !important;\n  }\n  .jhtml .align-items-xxl-stretch , .jhtml.align-items-xxl-stretch {\n    align-items: stretch !important;\n  }\n  .jhtml .align-content-xxl-start , .jhtml.align-content-xxl-start {\n    align-content: flex-start !important;\n  }\n  .jhtml .align-content-xxl-end , .jhtml.align-content-xxl-end {\n    align-content: flex-end !important;\n  }\n  .jhtml .align-content-xxl-center , .jhtml.align-content-xxl-center {\n    align-content: center !important;\n  }\n  .jhtml .align-content-xxl-between , .jhtml.align-content-xxl-between {\n    align-content: space-between !important;\n  }\n  .jhtml .align-content-xxl-around , .jhtml.align-content-xxl-around {\n    align-content: space-around !important;\n  }\n  .jhtml .align-content-xxl-stretch , .jhtml.align-content-xxl-stretch {\n    align-content: stretch !important;\n  }\n  .jhtml .align-self-xxl-auto , .jhtml.align-self-xxl-auto {\n    align-self: auto !important;\n  }\n  .jhtml .align-self-xxl-start , .jhtml.align-self-xxl-start {\n    align-self: flex-start !important;\n  }\n  .jhtml .align-self-xxl-end , .jhtml.align-self-xxl-end {\n    align-self: flex-end !important;\n  }\n  .jhtml .align-self-xxl-center , .jhtml.align-self-xxl-center {\n    align-self: center !important;\n  }\n  .jhtml .align-self-xxl-baseline , .jhtml.align-self-xxl-baseline {\n    align-self: baseline !important;\n  }\n  .jhtml .align-self-xxl-stretch , .jhtml.align-self-xxl-stretch {\n    align-self: stretch !important;\n  }\n  .jhtml .order-xxl-first , .jhtml.order-xxl-first {\n    order: -1 !important;\n  }\n  .jhtml .order-xxl-0 , .jhtml.order-xxl-0 {\n    order: 0 !important;\n  }\n  .jhtml .order-xxl-1 , .jhtml.order-xxl-1 {\n    order: 1 !important;\n  }\n  .jhtml .order-xxl-2 , .jhtml.order-xxl-2 {\n    order: 2 !important;\n  }\n  .jhtml .order-xxl-3 , .jhtml.order-xxl-3 {\n    order: 3 !important;\n  }\n  .jhtml .order-xxl-4 , .jhtml.order-xxl-4 {\n    order: 4 !important;\n  }\n  .jhtml .order-xxl-5 , .jhtml.order-xxl-5 {\n    order: 5 !important;\n  }\n  .jhtml .order-xxl-last , .jhtml.order-xxl-last {\n    order: 6 !important;\n  }\n  .jhtml .m-xxl-0 , .jhtml.m-xxl-0 {\n    margin: 0 !important;\n  }\n  .jhtml .m-xxl-1 , .jhtml.m-xxl-1 {\n    margin: 0.25rem !important;\n  }\n  .jhtml .m-xxl-2 , .jhtml.m-xxl-2 {\n    margin: 0.5rem !important;\n  }\n  .jhtml .m-xxl-3 , .jhtml.m-xxl-3 {\n    margin: 1rem !important;\n  }\n  .jhtml .m-xxl-4 , .jhtml.m-xxl-4 {\n    margin: 1.5rem !important;\n  }\n  .jhtml .m-xxl-5 , .jhtml.m-xxl-5 {\n    margin: 3rem !important;\n  }\n  .jhtml .m-xxl-auto , .jhtml.m-xxl-auto {\n    margin: auto !important;\n  }\n  .jhtml .mx-xxl-0 , .jhtml.mx-xxl-0 {\n    margin-right: 0 !important;\n    margin-left: 0 !important;\n  }\n  .jhtml .mx-xxl-1 , .jhtml.mx-xxl-1 {\n    margin-right: 0.25rem !important;\n    margin-left: 0.25rem !important;\n  }\n  .jhtml .mx-xxl-2 , .jhtml.mx-xxl-2 {\n    margin-right: 0.5rem !important;\n    margin-left: 0.5rem !important;\n  }\n  .jhtml .mx-xxl-3 , .jhtml.mx-xxl-3 {\n    margin-right: 1rem !important;\n    margin-left: 1rem !important;\n  }\n  .jhtml .mx-xxl-4 , .jhtml.mx-xxl-4 {\n    margin-right: 1.5rem !important;\n    margin-left: 1.5rem !important;\n  }\n  .jhtml .mx-xxl-5 , .jhtml.mx-xxl-5 {\n    margin-right: 3rem !important;\n    margin-left: 3rem !important;\n  }\n  .jhtml .mx-xxl-auto , .jhtml.mx-xxl-auto {\n    margin-right: auto !important;\n    margin-left: auto !important;\n  }\n  .jhtml .my-xxl-0 , .jhtml.my-xxl-0 {\n    margin-top: 0 !important;\n    margin-bottom: 0 !important;\n  }\n  .jhtml .my-xxl-1 , .jhtml.my-xxl-1 {\n    margin-top: 0.25rem !important;\n    margin-bottom: 0.25rem !important;\n  }\n  .jhtml .my-xxl-2 , .jhtml.my-xxl-2 {\n    margin-top: 0.5rem !important;\n    margin-bottom: 0.5rem !important;\n  }\n  .jhtml .my-xxl-3 , .jhtml.my-xxl-3 {\n    margin-top: 1rem !important;\n    margin-bottom: 1rem !important;\n  }\n  .jhtml .my-xxl-4 , .jhtml.my-xxl-4 {\n    margin-top: 1.5rem !important;\n    margin-bottom: 1.5rem !important;\n  }\n  .jhtml .my-xxl-5 , .jhtml.my-xxl-5 {\n    margin-top: 3rem !important;\n    margin-bottom: 3rem !important;\n  }\n  .jhtml .my-xxl-auto , .jhtml.my-xxl-auto {\n    margin-top: auto !important;\n    margin-bottom: auto !important;\n  }\n  .jhtml .mt-xxl-0 , .jhtml.mt-xxl-0 {\n    margin-top: 0 !important;\n  }\n  .jhtml .mt-xxl-1 , .jhtml.mt-xxl-1 {\n    margin-top: 0.25rem !important;\n  }\n  .jhtml .mt-xxl-2 , .jhtml.mt-xxl-2 {\n    margin-top: 0.5rem !important;\n  }\n  .jhtml .mt-xxl-3 , .jhtml.mt-xxl-3 {\n    margin-top: 1rem !important;\n  }\n  .jhtml .mt-xxl-4 , .jhtml.mt-xxl-4 {\n    margin-top: 1.5rem !important;\n  }\n  .jhtml .mt-xxl-5 , .jhtml.mt-xxl-5 {\n    margin-top: 3rem !important;\n  }\n  .jhtml .mt-xxl-auto , .jhtml.mt-xxl-auto {\n    margin-top: auto !important;\n  }\n  .jhtml .me-xxl-0 , .jhtml.me-xxl-0 {\n    margin-right: 0 !important;\n  }\n  .jhtml .me-xxl-1 , .jhtml.me-xxl-1 {\n    margin-right: 0.25rem !important;\n  }\n  .jhtml .me-xxl-2 , .jhtml.me-xxl-2 {\n    margin-right: 0.5rem !important;\n  }\n  .jhtml .me-xxl-3 , .jhtml.me-xxl-3 {\n    margin-right: 1rem !important;\n  }\n  .jhtml .me-xxl-4 , .jhtml.me-xxl-4 {\n    margin-right: 1.5rem !important;\n  }\n  .jhtml .me-xxl-5 , .jhtml.me-xxl-5 {\n    margin-right: 3rem !important;\n  }\n  .jhtml .me-xxl-auto , .jhtml.me-xxl-auto {\n    margin-right: auto !important;\n  }\n  .jhtml .mb-xxl-0 , .jhtml.mb-xxl-0 {\n    margin-bottom: 0 !important;\n  }\n  .jhtml .mb-xxl-1 , .jhtml.mb-xxl-1 {\n    margin-bottom: 0.25rem !important;\n  }\n  .jhtml .mb-xxl-2 , .jhtml.mb-xxl-2 {\n    margin-bottom: 0.5rem !important;\n  }\n  .jhtml .mb-xxl-3 , .jhtml.mb-xxl-3 {\n    margin-bottom: 1rem !important;\n  }\n  .jhtml .mb-xxl-4 , .jhtml.mb-xxl-4 {\n    margin-bottom: 1.5rem !important;\n  }\n  .jhtml .mb-xxl-5 , .jhtml.mb-xxl-5 {\n    margin-bottom: 3rem !important;\n  }\n  .jhtml .mb-xxl-auto , .jhtml.mb-xxl-auto {\n    margin-bottom: auto !important;\n  }\n  .jhtml .ms-xxl-0 , .jhtml.ms-xxl-0 {\n    margin-left: 0 !important;\n  }\n  .jhtml .ms-xxl-1 , .jhtml.ms-xxl-1 {\n    margin-left: 0.25rem !important;\n  }\n  .jhtml .ms-xxl-2 , .jhtml.ms-xxl-2 {\n    margin-left: 0.5rem !important;\n  }\n  .jhtml .ms-xxl-3 , .jhtml.ms-xxl-3 {\n    margin-left: 1rem !important;\n  }\n  .jhtml .ms-xxl-4 , .jhtml.ms-xxl-4 {\n    margin-left: 1.5rem !important;\n  }\n  .jhtml .ms-xxl-5 , .jhtml.ms-xxl-5 {\n    margin-left: 3rem !important;\n  }\n  .jhtml .ms-xxl-auto , .jhtml.ms-xxl-auto {\n    margin-left: auto !important;\n  }\n  .jhtml .p-xxl-0 , .jhtml.p-xxl-0 {\n    padding: 0 !important;\n  }\n  .jhtml .p-xxl-1 , .jhtml.p-xxl-1 {\n    padding: 0.25rem !important;\n  }\n  .jhtml .p-xxl-2 , .jhtml.p-xxl-2 {\n    padding: 0.5rem !important;\n  }\n  .jhtml .p-xxl-3 , .jhtml.p-xxl-3 {\n    padding: 1rem !important;\n  }\n  .jhtml .p-xxl-4 , .jhtml.p-xxl-4 {\n    padding: 1.5rem !important;\n  }\n  .jhtml .p-xxl-5 , .jhtml.p-xxl-5 {\n    padding: 3rem !important;\n  }\n  .jhtml .px-xxl-0 , .jhtml.px-xxl-0 {\n    padding-right: 0 !important;\n    padding-left: 0 !important;\n  }\n  .jhtml .px-xxl-1 , .jhtml.px-xxl-1 {\n    padding-right: 0.25rem !important;\n    padding-left: 0.25rem !important;\n  }\n  .jhtml .px-xxl-2 , .jhtml.px-xxl-2 {\n    padding-right: 0.5rem !important;\n    padding-left: 0.5rem !important;\n  }\n  .jhtml .px-xxl-3 , .jhtml.px-xxl-3 {\n    padding-right: 1rem !important;\n    padding-left: 1rem !important;\n  }\n  .jhtml .px-xxl-4 , .jhtml.px-xxl-4 {\n    padding-right: 1.5rem !important;\n    padding-left: 1.5rem !important;\n  }\n  .jhtml .px-xxl-5 , .jhtml.px-xxl-5 {\n    padding-right: 3rem !important;\n    padding-left: 3rem !important;\n  }\n  .jhtml .py-xxl-0 , .jhtml.py-xxl-0 {\n    padding-top: 0 !important;\n    padding-bottom: 0 !important;\n  }\n  .jhtml .py-xxl-1 , .jhtml.py-xxl-1 {\n    padding-top: 0.25rem !important;\n    padding-bottom: 0.25rem !important;\n  }\n  .jhtml .py-xxl-2 , .jhtml.py-xxl-2 {\n    padding-top: 0.5rem !important;\n    padding-bottom: 0.5rem !important;\n  }\n  .jhtml .py-xxl-3 , .jhtml.py-xxl-3 {\n    padding-top: 1rem !important;\n    padding-bottom: 1rem !important;\n  }\n  .jhtml .py-xxl-4 , .jhtml.py-xxl-4 {\n    padding-top: 1.5rem !important;\n    padding-bottom: 1.5rem !important;\n  }\n  .jhtml .py-xxl-5 , .jhtml.py-xxl-5 {\n    padding-top: 3rem !important;\n    padding-bottom: 3rem !important;\n  }\n  .jhtml .pt-xxl-0 , .jhtml.pt-xxl-0 {\n    padding-top: 0 !important;\n  }\n  .jhtml .pt-xxl-1 , .jhtml.pt-xxl-1 {\n    padding-top: 0.25rem !important;\n  }\n  .jhtml .pt-xxl-2 , .jhtml.pt-xxl-2 {\n    padding-top: 0.5rem !important;\n  }\n  .jhtml .pt-xxl-3 , .jhtml.pt-xxl-3 {\n    padding-top: 1rem !important;\n  }\n  .jhtml .pt-xxl-4 , .jhtml.pt-xxl-4 {\n    padding-top: 1.5rem !important;\n  }\n  .jhtml .pt-xxl-5 , .jhtml.pt-xxl-5 {\n    padding-top: 3rem !important;\n  }\n  .jhtml .pe-xxl-0 , .jhtml.pe-xxl-0 {\n    padding-right: 0 !important;\n  }\n  .jhtml .pe-xxl-1 , .jhtml.pe-xxl-1 {\n    padding-right: 0.25rem !important;\n  }\n  .jhtml .pe-xxl-2 , .jhtml.pe-xxl-2 {\n    padding-right: 0.5rem !important;\n  }\n  .jhtml .pe-xxl-3 , .jhtml.pe-xxl-3 {\n    padding-right: 1rem !important;\n  }\n  .jhtml .pe-xxl-4 , .jhtml.pe-xxl-4 {\n    padding-right: 1.5rem !important;\n  }\n  .jhtml .pe-xxl-5 , .jhtml.pe-xxl-5 {\n    padding-right: 3rem !important;\n  }\n  .jhtml .pb-xxl-0 , .jhtml.pb-xxl-0 {\n    padding-bottom: 0 !important;\n  }\n  .jhtml .pb-xxl-1 , .jhtml.pb-xxl-1 {\n    padding-bottom: 0.25rem !important;\n  }\n  .jhtml .pb-xxl-2 , .jhtml.pb-xxl-2 {\n    padding-bottom: 0.5rem !important;\n  }\n  .jhtml .pb-xxl-3 , .jhtml.pb-xxl-3 {\n    padding-bottom: 1rem !important;\n  }\n  .jhtml .pb-xxl-4 , .jhtml.pb-xxl-4 {\n    padding-bottom: 1.5rem !important;\n  }\n  .jhtml .pb-xxl-5 , .jhtml.pb-xxl-5 {\n    padding-bottom: 3rem !important;\n  }\n  .jhtml .ps-xxl-0 , .jhtml.ps-xxl-0 {\n    padding-left: 0 !important;\n  }\n  .jhtml .ps-xxl-1 , .jhtml.ps-xxl-1 {\n    padding-left: 0.25rem !important;\n  }\n  .jhtml .ps-xxl-2 , .jhtml.ps-xxl-2 {\n    padding-left: 0.5rem !important;\n  }\n  .jhtml .ps-xxl-3 , .jhtml.ps-xxl-3 {\n    padding-left: 1rem !important;\n  }\n  .jhtml .ps-xxl-4 , .jhtml.ps-xxl-4 {\n    padding-left: 1.5rem !important;\n  }\n  .jhtml .ps-xxl-5 , .jhtml.ps-xxl-5 {\n    padding-left: 3rem !important;\n  }\n  .jhtml .text-xxl-start , .jhtml.text-xxl-start {\n    text-align: left !important;\n  }\n  .jhtml .text-xxl-end , .jhtml.text-xxl-end {\n    text-align: right !important;\n  }\n  .jhtml .text-xxl-center , .jhtml.text-xxl-center {\n    text-align: center !important;\n  }\n}\n@media (min-width: 1200px) {\n  .jhtml .fs-1 , .jhtml.fs-1 {\n    font-size: 2.5rem !important;\n  }\n  .jhtml .fs-2 , .jhtml.fs-2 {\n    font-size: 2rem !important;\n  }\n  .jhtml .fs-3 , .jhtml.fs-3 {\n    font-size: 1.75rem !important;\n  }\n  .jhtml .fs-4 , .jhtml.fs-4 {\n    font-size: 1.5rem !important;\n  }\n}\n@media print {\n  .jhtml .d-print-inline , .jhtml.d-print-inline {\n    display: inline !important;\n  }\n  .jhtml .d-print-inline-block , .jhtml.d-print-inline-block {\n    display: inline-block !important;\n  }\n  .jhtml .d-print-block , .jhtml.d-print-block {\n    display: block !important;\n  }\n  .jhtml .d-print-grid , .jhtml.d-print-grid {\n    display: grid !important;\n  }\n  .jhtml .d-print-table , .jhtml.d-print-table {\n    display: table !important;\n  }\n  .jhtml .d-print-table-row , .jhtml.d-print-table-row {\n    display: table-row !important;\n  }\n  .jhtml .d-print-table-cell , .jhtml.d-print-table-cell {\n    display: table-cell !important;\n  }\n  .jhtml .d-print-flex , .jhtml.d-print-flex {\n    display: flex !important;\n  }\n  .jhtml .d-print-inline-flex , .jhtml.d-print-inline-flex {\n    display: inline-flex !important;\n  }\n  .jhtml .d-print-none , .jhtml.d-print-none {\n    display: none !important;\n  }\n}\n.jhtml .form-check-input , .jhtml.form-check-input {\n  -webkit-appearance: none;\n  -moz-appearance: none;\n}\n.jhtml .form-select , .jhtml.form-select {\n  -webkit-appearance: none;\n  -moz-appearance: none;\n}\n.jhtml .form-range , .jhtml.form-range {\n  -webkit-appearance: none;\n  -moz-appearance: none;\n}\n", ""]);
// Exports
module.exports = exports;


/***/ }),

/***/ "./node_modules/css-loader/dist/cjs.js!./css/index.css":
/*!*************************************************************!*\
  !*** ./node_modules/css-loader/dist/cjs.js!./css/index.css ***!
  \*************************************************************/
/***/ ((module, exports, __webpack_require__) => {

// Imports
var ___CSS_LOADER_API_IMPORT___ = __webpack_require__(/*! ../node_modules/css-loader/dist/runtime/api.js */ "./node_modules/css-loader/dist/runtime/api.js");
var ___CSS_LOADER_AT_RULE_IMPORT_0___ = __webpack_require__(/*! -!../node_modules/css-loader/dist/cjs.js!./bootstrap.css */ "./node_modules/css-loader/dist/cjs.js!./css/bootstrap.css");
exports = ___CSS_LOADER_API_IMPORT___(false);
exports.i(___CSS_LOADER_AT_RULE_IMPORT_0___);
// Module
exports.push([module.id, "", ""]);
// Exports
module.exports = exports;


/***/ }),

/***/ "./node_modules/css-loader/dist/runtime/api.js":
/*!*****************************************************!*\
  !*** ./node_modules/css-loader/dist/runtime/api.js ***!
  \*****************************************************/
/***/ ((module) => {

"use strict";


/*
  MIT License http://www.opensource.org/licenses/mit-license.php
  Author Tobias Koppers @sokra
*/
// css base code, injected by the css-loader
// eslint-disable-next-line func-names
module.exports = function (useSourceMap) {
  var list = []; // return the list of modules as css string

  list.toString = function toString() {
    return this.map(function (item) {
      var content = cssWithMappingToString(item, useSourceMap);

      if (item[2]) {
        return "@media ".concat(item[2], " {").concat(content, "}");
      }

      return content;
    }).join('');
  }; // import a list of modules into the list
  // eslint-disable-next-line func-names


  list.i = function (modules, mediaQuery, dedupe) {
    if (typeof modules === 'string') {
      // eslint-disable-next-line no-param-reassign
      modules = [[null, modules, '']];
    }

    var alreadyImportedModules = {};

    if (dedupe) {
      for (var i = 0; i < this.length; i++) {
        // eslint-disable-next-line prefer-destructuring
        var id = this[i][0];

        if (id != null) {
          alreadyImportedModules[id] = true;
        }
      }
    }

    for (var _i = 0; _i < modules.length; _i++) {
      var item = [].concat(modules[_i]);

      if (dedupe && alreadyImportedModules[item[0]]) {
        // eslint-disable-next-line no-continue
        continue;
      }

      if (mediaQuery) {
        if (!item[2]) {
          item[2] = mediaQuery;
        } else {
          item[2] = "".concat(mediaQuery, " and ").concat(item[2]);
        }
      }

      list.push(item);
    }
  };

  return list;
};

function cssWithMappingToString(item, useSourceMap) {
  var content = item[1] || ''; // eslint-disable-next-line prefer-destructuring

  var cssMapping = item[3];

  if (!cssMapping) {
    return content;
  }

  if (useSourceMap && typeof btoa === 'function') {
    var sourceMapping = toComment(cssMapping);
    var sourceURLs = cssMapping.sources.map(function (source) {
      return "/*# sourceURL=".concat(cssMapping.sourceRoot || '').concat(source, " */");
    });
    return [content].concat(sourceURLs).concat([sourceMapping]).join('\n');
  }

  return [content].join('\n');
} // Adapted from convert-source-map (MIT)


function toComment(sourceMap) {
  // eslint-disable-next-line no-undef
  var base64 = btoa(unescape(encodeURIComponent(JSON.stringify(sourceMap))));
  var data = "sourceMappingURL=data:application/json;charset=utf-8;base64,".concat(base64);
  return "/*# ".concat(data, " */");
}

/***/ }),

/***/ "./css/index.css":
/*!***********************!*\
  !*** ./css/index.css ***!
  \***********************/
/***/ ((module, __unused_webpack_exports, __webpack_require__) => {

var api = __webpack_require__(/*! !../node_modules/style-loader/dist/runtime/injectStylesIntoStyleTag.js */ "./node_modules/style-loader/dist/runtime/injectStylesIntoStyleTag.js");
            var content = __webpack_require__(/*! !!../node_modules/css-loader/dist/cjs.js!./index.css */ "./node_modules/css-loader/dist/cjs.js!./css/index.css");

            content = content.__esModule ? content.default : content;

            if (typeof content === 'string') {
              content = [[module.id, content, '']];
            }

var options = {};

options.insert = "head";
options.singleton = false;

var update = api(content, options);



module.exports = content.locals || {};

/***/ }),

/***/ "./node_modules/style-loader/dist/runtime/injectStylesIntoStyleTag.js":
/*!****************************************************************************!*\
  !*** ./node_modules/style-loader/dist/runtime/injectStylesIntoStyleTag.js ***!
  \****************************************************************************/
/***/ ((module, __unused_webpack_exports, __webpack_require__) => {

"use strict";


var isOldIE = function isOldIE() {
  var memo;
  return function memorize() {
    if (typeof memo === 'undefined') {
      // Test for IE <= 9 as proposed by Browserhacks
      // @see http://browserhacks.com/#hack-e71d8692f65334173fee715c222cb805
      // Tests for existence of standard globals is to allow style-loader
      // to operate correctly into non-standard environments
      // @see https://github.com/webpack-contrib/style-loader/issues/177
      memo = Boolean(window && document && document.all && !window.atob);
    }

    return memo;
  };
}();

var getTarget = function getTarget() {
  var memo = {};
  return function memorize(target) {
    if (typeof memo[target] === 'undefined') {
      var styleTarget = document.querySelector(target); // Special case to return head of iframe instead of iframe itself

      if (window.HTMLIFrameElement && styleTarget instanceof window.HTMLIFrameElement) {
        try {
          // This will throw an exception if access to iframe is blocked
          // due to cross-origin restrictions
          styleTarget = styleTarget.contentDocument.head;
        } catch (e) {
          // istanbul ignore next
          styleTarget = null;
        }
      }

      memo[target] = styleTarget;
    }

    return memo[target];
  };
}();

var stylesInDom = [];

function getIndexByIdentifier(identifier) {
  var result = -1;

  for (var i = 0; i < stylesInDom.length; i++) {
    if (stylesInDom[i].identifier === identifier) {
      result = i;
      break;
    }
  }

  return result;
}

function modulesToDom(list, options) {
  var idCountMap = {};
  var identifiers = [];

  for (var i = 0; i < list.length; i++) {
    var item = list[i];
    var id = options.base ? item[0] + options.base : item[0];
    var count = idCountMap[id] || 0;
    var identifier = "".concat(id, " ").concat(count);
    idCountMap[id] = count + 1;
    var index = getIndexByIdentifier(identifier);
    var obj = {
      css: item[1],
      media: item[2],
      sourceMap: item[3]
    };

    if (index !== -1) {
      stylesInDom[index].references++;
      stylesInDom[index].updater(obj);
    } else {
      stylesInDom.push({
        identifier: identifier,
        updater: addStyle(obj, options),
        references: 1
      });
    }

    identifiers.push(identifier);
  }

  return identifiers;
}

function insertStyleElement(options) {
  var style = document.createElement('style');
  var attributes = options.attributes || {};

  if (typeof attributes.nonce === 'undefined') {
    var nonce =  true ? __webpack_require__.nc : 0;

    if (nonce) {
      attributes.nonce = nonce;
    }
  }

  Object.keys(attributes).forEach(function (key) {
    style.setAttribute(key, attributes[key]);
  });

  if (typeof options.insert === 'function') {
    options.insert(style);
  } else {
    var target = getTarget(options.insert || 'head');

    if (!target) {
      throw new Error("Couldn't find a style target. This probably means that the value for the 'insert' parameter is invalid.");
    }

    target.appendChild(style);
  }

  return style;
}

function removeStyleElement(style) {
  // istanbul ignore if
  if (style.parentNode === null) {
    return false;
  }

  style.parentNode.removeChild(style);
}
/* istanbul ignore next  */


var replaceText = function replaceText() {
  var textStore = [];
  return function replace(index, replacement) {
    textStore[index] = replacement;
    return textStore.filter(Boolean).join('\n');
  };
}();

function applyToSingletonTag(style, index, remove, obj) {
  var css = remove ? '' : obj.media ? "@media ".concat(obj.media, " {").concat(obj.css, "}") : obj.css; // For old IE

  /* istanbul ignore if  */

  if (style.styleSheet) {
    style.styleSheet.cssText = replaceText(index, css);
  } else {
    var cssNode = document.createTextNode(css);
    var childNodes = style.childNodes;

    if (childNodes[index]) {
      style.removeChild(childNodes[index]);
    }

    if (childNodes.length) {
      style.insertBefore(cssNode, childNodes[index]);
    } else {
      style.appendChild(cssNode);
    }
  }
}

function applyToTag(style, options, obj) {
  var css = obj.css;
  var media = obj.media;
  var sourceMap = obj.sourceMap;

  if (media) {
    style.setAttribute('media', media);
  } else {
    style.removeAttribute('media');
  }

  if (sourceMap && typeof btoa !== 'undefined') {
    css += "\n/*# sourceMappingURL=data:application/json;base64,".concat(btoa(unescape(encodeURIComponent(JSON.stringify(sourceMap)))), " */");
  } // For old IE

  /* istanbul ignore if  */


  if (style.styleSheet) {
    style.styleSheet.cssText = css;
  } else {
    while (style.firstChild) {
      style.removeChild(style.firstChild);
    }

    style.appendChild(document.createTextNode(css));
  }
}

var singleton = null;
var singletonCounter = 0;

function addStyle(obj, options) {
  var style;
  var update;
  var remove;

  if (options.singleton) {
    var styleIndex = singletonCounter++;
    style = singleton || (singleton = insertStyleElement(options));
    update = applyToSingletonTag.bind(null, style, styleIndex, false);
    remove = applyToSingletonTag.bind(null, style, styleIndex, true);
  } else {
    style = insertStyleElement(options);
    update = applyToTag.bind(null, style, options);

    remove = function remove() {
      removeStyleElement(style);
    };
  }

  update(obj);
  return function updateStyle(newObj) {
    if (newObj) {
      if (newObj.css === obj.css && newObj.media === obj.media && newObj.sourceMap === obj.sourceMap) {
        return;
      }

      update(obj = newObj);
    } else {
      remove();
    }
  };
}

module.exports = function (list, options) {
  options = options || {}; // Force single-tag solution on IE6-9, which has a hard limit on the # of <style>
  // tags it will allow on a page

  if (!options.singleton && typeof options.singleton !== 'boolean') {
    options.singleton = isOldIE();
  }

  list = list || [];
  var lastIdentifiers = modulesToDom(list, options);
  return function update(newList) {
    newList = newList || [];

    if (Object.prototype.toString.call(newList) !== '[object Array]') {
      return;
    }

    for (var i = 0; i < lastIdentifiers.length; i++) {
      var identifier = lastIdentifiers[i];
      var index = getIndexByIdentifier(identifier);
      stylesInDom[index].references--;
    }

    var newLastIdentifiers = modulesToDom(newList, options);

    for (var _i = 0; _i < lastIdentifiers.length; _i++) {
      var _identifier = lastIdentifiers[_i];

      var _index = getIndexByIdentifier(_identifier);

      if (stylesInDom[_index].references === 0) {
        stylesInDom[_index].updater();

        stylesInDom.splice(_index, 1);
      }
    }

    lastIdentifiers = newLastIdentifiers;
  };
};

/***/ })

}]);
//# sourceMappingURL=css_index_css.0a349ad0ce82fa2494b2.js.map